package datavisualizer;

import javax.swing.JInternalFrame;
import javawrapper.ModelException;

/**
 * This class is used by the DetailedOutputFileManager class to manage open
 * requests for data.  Each DataRequest object represents the data needed to
 * make one chart.  This base class rejects all data in its non-abstract data
 * members. Override the ones necessary to get and process required
 * information.
 * <p>Copyright: Copyright (c) Charles D. Canham 2003</p>
 * <p>Company: Institute of Ecosystem Studies</p>
 * @author Lora E. Murphy
 * @version 1.0
 *
 * <br>Edit history:
 * <br>------------------
 * <br>April 28, 2004: Submitted in beta version (LEM)
 * <br>December 10, 2004: Added the writing of data behind charts (LEM)
 */

public abstract class DataRequest
    implements java.awt.event.ActionListener {

  /**Managing object*/
  DetailedOutputFileManager m_oManager;

  /**
   * Constructor.
   * @param sChartName Name of chart to display.
   * @param oManager Managing file manager.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public DataRequest(String sChartName, DetailedOutputFileManager oManager) {
    m_sChartName = sChartName;
    m_oChartFrame = null;
    m_oManager = oManager;
  }

  /**
   * If passed the order to write chart data, does it.
   * @param oEvent ActionEvent Event to process.
   */
  public void actionPerformed(java.awt.event.ActionEvent oEvent) {
    if (oEvent.getActionCommand().equals("WriteChartData")) {
      SaveChartDataToFile();
    }
  }

  /**
   * Shell for saving a chart's data to a file.  This takes care of
   * getting the filename from the user, then passes off the writing of the
   * file to WriteChartDataToFile().
   *
   * <br>Edit history:
   * <br>------------------
   * <br>December 10, 2004: Created (LEM)
   */
  public void SaveChartDataToFile() {
    String sFileName = "";
    try {
      //Allow the user to save a text file
      javax.swing.JFileChooser jChooser = new javax.swing.JFileChooser();
      jChooser.setFileFilter(new javawrapper.TextFileFilter());

      int iReturnVal = jChooser.showSaveDialog(m_oChartFrame);
      if (iReturnVal == javax.swing.JFileChooser.APPROVE_OPTION) {
        //User chose a file - trigger the save
        java.io.File oFile = jChooser.getSelectedFile();
        if (oFile.exists()) {
          iReturnVal = javax.swing.JOptionPane.showConfirmDialog(m_oChartFrame,
              "Do you wish to overwrite the existing file?",
              "Model",
              javax.swing.JOptionPane.YES_NO_OPTION);
        }
        if (!oFile.exists() || iReturnVal == javax.swing.JOptionPane.YES_OPTION) {
          sFileName = oFile.getAbsolutePath();
          if (sFileName.indexOf(".") == -1) {
            sFileName += ".txt";
          }

          WriteChartDataToFile(sFileName);

          //Tell the user
          javax.swing.JOptionPane.showMessageDialog(m_oChartFrame,
              "File has been saved.");
        }
      }
    }
    catch (java.io.IOException oErr) {
      javawrapper.ModelException oExp = new javawrapper.ModelException();
      oExp.sMessage = "There was a problem writing the file " + sFileName + ".";
      oExp.sCallingFunction = "JAVA";
      javawrapper.ErrorGUI oHandler = new javawrapper.ErrorGUI(m_oChartFrame);
      oHandler.WriteErrorMessage(oExp);
    }
    catch (javawrapper.ModelException oErr) {
      javawrapper.ErrorGUI oHandler = new javawrapper.ErrorGUI(m_oChartFrame);
      oHandler.WriteErrorMessage(oErr);
    }
  }

  /**
   * Child classes override this to write the data behind their chart to
   * tab-delimited text.
   * @param sFileName String Filename to write to.
   * @throws java.io.IOException if there is a problem writing the file.
   * @throws ModelException if there is a problem assembling the chart data.
   */
  protected abstract void WriteChartDataToFile(String sFileName) throws java.io.
      IOException, javawrapper.ModelException;

  /**
   * Get the name of the chart.
   * @return Chart name.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public String GetChartName() {
    return m_sChartName;
  }

  /**
   * Get the chart frame.
   * @return Chart frame.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public JInternalFrame GetChart() {
    return m_oChartFrame;
  }

  /**
   * Whether or not this object wants any tree float data members.
   * @return True if float data members are desired, false if not.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public boolean WantAnyTreeFloats() {
    return false;
  }

  /**
   * Whether or not this object wants any tree int data members.
   * @return True if int data members are desired, false if not.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public boolean WantAnyTreeInts() {
    return false;
  }

  /**
   * Whether or not this object wants any tree char data members.
   * @return True if char data members are desired, false if not.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public boolean WantAnyTreeChars() {
    return false;
  }

  /**
   * Whether or not this object wants any tree bool data members.
   * @return True if bool data members are desired, false if not.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public boolean WantAnyTreeBools() {
    return false;
  }

  /**
   * Whether or not this object wants any grid float data members.
   * @return True if float data members are desired, false if not.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public boolean WantAnyGridFloats() {
    return false;
  }

  /**
   * Whether or not this object wants any grid int data members.
   * @return True if int data members are desired, false if not.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public boolean WantAnyGridInts() {
    return false;
  }

  /**
   * Whether or not this object wants any grid char data members.
   * @return True if char data members are desired, false if not.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public boolean WantAnyGridChars() {
    return false;
  }

  /**
   * Whether or not this object wants any grid bool data members.
   * @return True if bool data members are desired, false if not.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public boolean WantAnyGridBools() {
    return false;
  }

  /**
   * Accepts a piece of tree float data from the parser.  If this particular
   * piece is not wanted, do nothing.
   * @param iSpecies Species of the tree from which this value came.
   * @param iType Type of the tree from which this value came.
   * @param iCode Data member code of this value.
   * @param fVal Value.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void AddTreeFloatData(int iSpecies, int iType, int iCode, float fVal) {}

  /**
   * Accepts a piece of tree int data from the parser.  If this particular
   * piece is not wanted, do nothing.
   * @param iSpecies Species of the tree from which this value came.
   * @param iType Type of the tree from which this value came.
   * @param iCode Data member code of this value.
   * @param iVal Value.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void AddTreeIntData(int iSpecies, int iType, int iCode, int iVal) {}

  /**
   * Accepts a piece of tree char data from the parser.  If this particular
   * piece is not wanted, do nothing.
   * @param iSpecies Species of the tree from which this value came.
   * @param iType Type of the tree from which this value came.
   * @param iCode Data member code of this value.
   * @param sVal Value.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void AddTreeCharData(int iSpecies, int iType, int iCode, String sVal) {}

  /**
   * Accepts a piece of tree bool data from the parser.  If this particular
   * piece is not wanted, do nothing.
   * @param iSpecies Species of the tree from which this value came.
   * @param iType Type of the tree from which this value came.
   * @param iCode Data member code of this value.
   * @param bVal Value.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void AddTreeBoolData(int iSpecies, int iType, int iCode, boolean bVal) {}

  /**
   * Accepts a piece of grid float data from the parser.  If this particular
   * piece is not wanted, do nothing.
   * @param sGridName Name of the grid for this data
   * @param iX X number of the cell from which this value came
   * @param iY Y number of the cell from which this value came
   * @param iCode Data member code of this value.
   * @param fVal Value.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void AddGridFloatData(String sGridName, int iX, int iY, int iCode,
                               float fVal) {}

  /**
   * Accepts a piece of grid int data from the parser.  If this particular
   * piece is not wanted, do nothing.
   * @param sGridName Name of the grid for this data
   * @param iX X number of the cell from which this value came
   * @param iY Y number of the cell from which this value came
   * @param iCode Data member code of this value.
   * @param iVal Value.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void AddGridIntData(String sGridName, int iX, int iY, int iCode,
                             int iVal) {}

  /**
   * Accepts a piece of grid char data from the parser.  If this particular
   * piece is not wanted, do nothing.
   * @param sGridName Name of the grid for this data
   * @param iX X number of the cell from which this value came
   * @param iY Y number of the cell from which this value came
   * @param iCode Data member code of this value.
   * @param sVal Value.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void AddGridCharData(String sGridName, int iX, int iY, int iCode,
                              String sVal) {}

  /**
   * Accepts a piece of grid bool data from the parser.  If this particular
   * piece is not wanted, do nothing.
   * @param sGridName Name of the grid for this data
   * @param iX X number of the cell from which this value came
   * @param iY Y number of the cell from which this value came
   * @param iCode Data member code of this value.
   * @param bVal Value.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void AddGridBoolData(String sGridName, int iX, int iY, int iCode,
                              boolean bVal) {}

  /**
   * Accepts a tree float data member code for future reference when passed
   * float data members.
   * @param iSpecies The species for which this is a data member.
   * @param iType The tree type for which this is a data member.
   * @param sLabel The label of the data member.
   * @param iCode The data member code.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void AddTreeFloatDataMemberCode(int iSpecies, int iType, String sLabel,
                                         int iCode) {}

  /**
   * Accepts a tree int data member code for future reference when passed int
   * data members.
   * @param iSpecies The species for which this is a data member.
   * @param iType The tree type for which this is a data member.
   * @param sLabel The label of the data member.
   * @param iCode The data member code.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void AddTreeIntDataMemberCode(int iSpecies, int iType, String sLabel,
                                       int iCode) {}

  /**
   * Accepts a tree char data member code for future reference when passed char
   * data members.
   * @param iSpecies The species for which this is a data member.
   * @param iType The tree type for which this is a data member.
   * @param sLabel The label of the data member.
   * @param iCode The data member code.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void AddTreeCharDataMemberCode(int iSpecies, int iType, String sLabel,
                                        int iCode) {}

  /**
   * Accepts a tree bool data member code for future reference when passed bool
   * data members.
   * @param iSpecies The species for which this is a data member.
   * @param iType The tree type for which this is a data member.
   * @param sLabel The label of the data member.
   * @param iCode The data member code.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void AddTreeBoolDataMemberCode(int iSpecies, int iType, String sLabel,
                                        int iCode) {}

  /**
   * Accepts a grid float data member code for future reference when passed
   * float data members.
   * @param sGridName Name of the grid
   * @param sLabel The label of the data member.
   * @param iCode The data member code.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void AddGridFloatDataMemberCode(String sGridName, String sLabel,
                                         int iCode) {}

  /**
   * Accepts a grid int data member code for future reference when passed int
   * data members.
   * @param sGridName Name of the grid
   * @param sLabel The label of the data member.
   * @param iCode The data member code.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void AddGridIntDataMemberCode(String sGridName, String sLabel,
                                       int iCode) {}

  /**
   * Accepts a grid char data member code for future reference when passed char
   * data members.
   * @param sGridName Name of the grid
   * @param sLabel The label of the data member.
   * @param iCode The data member code.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void AddGridCharDataMemberCode(String sGridName, String sLabel,
                                        int iCode) {}

  /**
   * Accepts a grid bool data member code for future reference when passed bool
   * data members.
   * @param sGridName Name of the grid
   * @param sLabel The label of the data member.
   * @param iCode The data member code.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void AddGridBoolDataMemberCode(String sGridName, String sLabel,
                                        int iCode) {}

  /**
   * Clears out data for garbage collection in preparation for new data being
   * read in.
   * @throws ModelException if there are problems.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  abstract void ClearData() throws ModelException;

  /**
   * Draws the chart for this data request according to its individual settings
   * and data.
   * @param oLegend The legend for the chart.
   * @param sChartTitle The title for the chart.
   * @return The chart.
   * @throws ModelException if anything goes wrong with the chart drawing.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  abstract JInternalFrame DrawChart(Legend oLegend, String sChartTitle) throws
      ModelException;

  /**
   * Redraws the chart using the existing dataset.
   * @param oLegend The legend for this chart.
   * @throws ModelException if anything goes wrong with the chart drawing.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  abstract void UpdateChart(Legend oLegend) throws ModelException;

  protected JInternalFrame m_oChartFrame;

  /**<Chart for this request.  This is not automatically set.*/
  protected String m_sChartName; /**<Name string of chart.*/
}
