package datavisualizer;

import java.awt.event.ActionEvent;
import javax.swing.*;

import java.util.Vector;

import javawrapper.ModelException;

/**
 * This controls the drawing of a histogram for a single piece of information.
 * The label of the data is passed in the constructor.  The histogram is
 * defaulted to 10 divisions equally divided between their minimum and maximum
 * values.
 * <p>Copyright: Copyright (c) Charles D. Canham 2003</p>
 * <p>Company: Institute of Ecosystem Studies</p>
 * @author Lora E. Murphy
 * @version 1.0
 *
 * <br>Edit history:
 * <br>------------------
 * <br>April 28, 2004: Submitted in beta version (LEM)
 * <br>July 19, 2004:  Gave user bin control (LEM)
 * <br>December 6, 2004:  Added total bar (LEM)
 */

public class HistogramDataRequest
    extends DataRequest
    implements java.awt.event.ActionListener {

  /**Data member's label, as it appears in the detailed output file*/
  private String m_sLabel;

  /**Whether our data member being graphed is a float or integer*/
  private boolean m_bIsFloat, m_bIsInt;

  /**The dataset that will be graphed*/
  protected ModelHistogramDataset m_oDataset;

  /**Field displaying the number of bins in the histogram*/
  private JTextField m_jNumBins = new JTextField("");

  /**Field displaying the bin size in the histogram*/
  private JTextField m_jBinSize = new JTextField("");

  /**The data member codes for this data member - one per species / type*/
  private int[][] mp_iDataMemberCodes;

  /**The data - kept in an array of vectors - one per species*/
  private Vector[] mp_oData;

  /**Plot area in hectares*/
  private float m_fPlotAreaInHectares,
      /**Size of the histogram bins*/
      m_fBinSize;

  /**Number of bins in the histogram*/
  private int m_iNumBins;

  /**Whether or not to use a logarithmic axis on the Y*/
  private boolean m_bUseLogarithmicAxis;

  /**Whether or not we have to recalculate the bins on chart update (i.e. we
     are charting an unknown data type without defaults)*/
  private boolean m_bRecalcBinsOnUpdate;

  /**Whether or not to display a total bar in each bin*/
  private boolean m_bShowTotal = true;

  /**
   * Constructor.
   * @param sLabel Label of the piece of data, as it would be passed as a data
   * member code.
   * @param bIsFloat Whether this is float data.  If this is set to false,
   * integer data is assumed.
   * @param oManager The parent detailed output file manager.
   * @param sGraphName Name of graph string.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public HistogramDataRequest(String sLabel, boolean bIsFloat,
                              DetailedOutputFileManager oManager,
                              String sGraphName) {
    super(sGraphName, oManager);

    //Default use of logarithmic axis to true
    m_bUseLogarithmicAxis = true;

    //Calculate the plot area in hectares
    float fXLength = oManager.GetXPlotLength(),
        fYLength = oManager.GetYPlotLength();

    m_fPlotAreaInHectares = fXLength * fYLength / 10000;

    //Set up flags
    m_sLabel = sLabel;
    if (bIsFloat) {
      m_bIsFloat = true;
      m_bIsInt = false;
    }
    else {
      m_bIsFloat = false;
      m_bIsInt = true;
    }
    m_oManager = oManager;
    int iNumSpecies = oManager.GetNumberOfSpecies(),
        iNumTypes = oManager.GetNumberOfTypes(),
        i, j;

    //Declare our data member codes array and initialize our codes to -1
    mp_iDataMemberCodes = new int[iNumSpecies][];
    for (i = 0; i < iNumSpecies; i++) {
      mp_iDataMemberCodes[i] = new int[iNumTypes];
      for (j = 0; j < iNumTypes; j++) {
        mp_iDataMemberCodes[i][j] = -1;
      }
    }

    //Declare our data vector array
    mp_oData = new Vector[iNumSpecies];
    for (i = 0; i < iNumSpecies; i++) {
      mp_oData[i] = new Vector(0);
    }

    //Be somewhat smart about our bin size and number of bins - if we
    //recognize the value being histogrammed...
    if (sLabel.equalsIgnoreCase("dbh")) {
      m_iNumBins = 20;
      m_fBinSize = 10;
      m_bRecalcBinsOnUpdate = false;
    }
    else if (sLabel.equalsIgnoreCase("height")) {
      m_iNumBins = 50;
      m_fBinSize = 1;
      m_bRecalcBinsOnUpdate = false;
    }
    else {
      m_iNumBins = 10;
      m_fBinSize = 0;
      m_bRecalcBinsOnUpdate = true;
    }
  }

  /**
   * Writes the histogram's data to tab-delimited text.
   * @param sFileName String Filename to write to.
   * @throws java.io.IOException if there's a problem writing the file.
   */
  protected void WriteChartDataToFile(String sFileName) throws java.io.
      IOException {
    java.io.FileWriter jOut = null;
    try {

      float fValue;
      int i, j, iNumSeries, iNumItems;

      //If there are no series in the dataset, add them now (in case the
      //file was re-parsed while this chart was open)
      if (m_oDataset.getSeriesCount() == 0) {
        AddSeriesToDataset();
        //Adjust values to amounts per hectare
        iNumSeries = m_oDataset.getSeriesCount();
        for (i = 0; i < iNumSeries; i++) {
          iNumItems = m_oDataset.getItemCount(i);
          for (j = 0; j < iNumItems; j++) {
            fValue = m_oDataset.getYValue(i, j).intValue();
            fValue /= m_fPlotAreaInHectares;
            m_oDataset.setYValue(i, j, fValue);
          }
        }
      }

      jOut = new java.io.FileWriter(sFileName);

      //Write the chart header
      jOut.write(m_oChartFrame.getTitle() + "\n");
      jOut.write("Timestep: " +
                 String.valueOf( ( (DetailedOutputLegend) m_oManager.GetLegend()).
                                GetCurrentTimestep()) + "\n");
      jOut.write("Values are in number per hectare.\n");

      //Get the number of series - this will be the number of species, with
      //a possible totals series
      iNumSeries = m_oDataset.getSeriesCount();
      if (iNumSeries <= 0) {
        jOut.write("This histogram has no data.");
        return;
      }

      //Write out the header columns
      jOut.write("Species");
      for (i = 1; i <= m_iNumBins; i++) {
        jOut.write("\t" + String.valueOf( (i - 1) * m_fBinSize) + " - " +
                   String.valueOf(i * m_fBinSize));
      }
      jOut.write("\n");

      //Write out the series
      for (i = 0; i < iNumSeries; i++) {
        jOut.write(m_oDataset.getSeriesName(i));
        iNumItems = m_oDataset.getItemCount(i);
        for (j = 0; j < iNumItems; j++) {
          jOut.write("\t" + String.valueOf(m_oDataset.getYValue(i, j)));
        }
        jOut.write("\n");
      }

    }
    catch (ModelException oErr) {
      ; //do nothing
    }
    finally {
      if (jOut != null) {
        jOut.close();
      }
    }
  }

  /**
   * Performs actions for the controls in the Histogram window.
   * @param oEvent Event triggered.
   */
  public void actionPerformed(ActionEvent oEvent) {
    try {
      if (oEvent.getSource() instanceof JCheckBox) {
        JCheckBox jBox = (JCheckBox) oEvent.getSource();
        if (jBox.getName().equals("log_checkbox")) {
          if (jBox.isSelected() != m_bUseLogarithmicAxis) {
            m_bUseLogarithmicAxis = jBox.isSelected();
            UpdateChart(m_oManager.GetLegend());
          }
        }
        else if (jBox.getName().equals("total_checkbox")) {
          if (jBox.isSelected() != m_bShowTotal) {
            m_bShowTotal = jBox.isSelected();

            //If the dataset already exists, simply calling an update will
            //have no effect; so either remove or add the series, as
            //requested
            if (m_oDataset != null && m_oDataset.getSeriesCount() > 0) {
              if (m_bShowTotal == true) {
                m_oDataset.AddTotalSeries();
              }
              else {
                for (int i = 0; i < m_oDataset.getSeriesCount(); i++) {
                  if (m_oDataset.getSeriesName(i).equalsIgnoreCase("total")) {
                    m_oDataset.RemoveSeries(i);
                    break;
                  }
                }
              }
            }
            UpdateChart(m_oManager.GetLegend());
          }
        }
      }
      else if (oEvent.getActionCommand().equals("UpdateBins")) {

        //Make sure the values in the bins are recognizable, greater-than-zero numbers
        int iNumBins = 0;
        float fBinSize = 0;
        try {
          Integer oNumBins = new Integer(m_jNumBins.getText());
          Float oBinSize = new Float(m_jBinSize.getText());
          iNumBins = oNumBins.intValue();
          fBinSize = oBinSize.floatValue();
        }
        catch (java.lang.NumberFormatException oErr) {
          JOptionPane.showMessageDialog(m_oChartFrame,
                                        "The value for number of bins or bin size is not a recognized number.");
          m_jNumBins.setText(String.valueOf(m_iNumBins));
          m_jBinSize.setText(String.valueOf(m_fBinSize));
          return;
        }

        if (iNumBins <= 0 || fBinSize <= 0) {
          JOptionPane.showMessageDialog(m_oChartFrame,
                                        "The values for number of bins and bin size must be greater than zero.");
          m_jNumBins.setText(String.valueOf(m_iNumBins));
          m_jBinSize.setText(String.valueOf(m_fBinSize));
          return;
        }

        if (iNumBins != m_iNumBins || fBinSize != m_fBinSize) {
          m_iNumBins = iNumBins;
          m_fBinSize = fBinSize;

          //Clear the dataset to force a reconstruction
          m_oDataset = null;
          m_oDataset = new ModelHistogramDataset();

          m_bRecalcBinsOnUpdate = false;
          UpdateChart(m_oManager.GetLegend());
        }
      }
      super.actionPerformed(oEvent);
    }
    catch (javawrapper.ModelException oErr) {
      javawrapper.ErrorGUI oHandler = new javawrapper.ErrorGUI(m_oChartFrame);
      oHandler.WriteErrorMessage(oErr);
    }
  }

  /**
   * Gets the bin size for this data request's histogram.
   * @return The bin size.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public float GetBinSize() {
    return m_fBinSize;
  }

  /**
   * Gets whether or not to use a logarithmic axis.
   * @return True if a logarithmic axis should be used, or false if a linear
   * axis should be used.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public boolean GetUseLogarithmicAxis() {
    return m_bUseLogarithmicAxis;
  }

  /**
   * If the label matches the one given in the constructor, and the data being
   * histogrammed is a float, this captures the values passed.
   * @param iSpecies Species for this data member
   * @param iType Type for this data member
   * @param sLabel Data member label.
   * @param iCode Data member code.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void AddTreeFloatDataMemberCode(int iSpecies, int iType, String sLabel,
                                         int iCode) {
    if (m_bIsFloat) {
      if (m_sLabel.equals(sLabel)) {
        mp_iDataMemberCodes[iSpecies][iType] = iCode;
      }
    }
  }

  /**
   * Creates the histogram using data accumulated thus far.
   * @param oLegend Legend for this chart.
   * @param sChartTitle Title for this chart.
   * @return The new chart window.
   * @throws datavisualizer.ModelException Passing through underlying
   * exceptions.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   * <br>June 28, 2004:  Made the histogram bars display floats (LEM)
   */
  JInternalFrame DrawChart(Legend oLegend, String sChartTitle) throws
      ModelException {
    float fValue;
    int i, j, iNumSeries, iNumItems;

    //Create our new dataset
    m_oDataset = new ModelHistogramDataset();

    AddSeriesToDataset();

    ModelHistogramDataset oAdjustedDataset = UpdateForVisible(oLegend);

    //Adjust values to amounts per hectare
    iNumSeries = oAdjustedDataset.getSeriesCount();
    for (i = 0; i < iNumSeries; i++) {
      iNumItems = oAdjustedDataset.getItemCount(i);
      for (j = 0; j < iNumItems; j++) {
        fValue = oAdjustedDataset.getYValue(i, j).intValue();
        fValue /= m_fPlotAreaInHectares;
        oAdjustedDataset.setYValue(i, j, fValue);
      }
    }

    //Make the chart
    m_oChartFrame = DataGrapher.DrawHistogram(oAdjustedDataset, m_sLabel,
                                              "Number per hectare",
                                              sChartTitle, oLegend, this);

    //Add the extra controls to the top of the histogram window
    JPanel jControls = new JPanel(new java.awt.FlowLayout(java.awt.FlowLayout.
        LEFT));

    //Bins boxes
    JLabel jTemp = new JLabel("Number of bins:");
    jTemp.setFont(new javawrapper.ModelFont());
    jControls.add(jTemp);
    m_jNumBins.setFont(new javawrapper.ModelFont());
    m_jNumBins.setText(String.valueOf(m_iNumBins));
    m_jNumBins.setPreferredSize(new java.awt.Dimension(50,
        (int) m_jNumBins.getPreferredSize().getHeight()));
    jControls.add(m_jNumBins);

    jTemp = new JLabel("Bin size:");
    jTemp.setFont(new javawrapper.ModelFont());
    jControls.add(jTemp);
    m_jBinSize.setFont(new javawrapper.ModelFont());
    java.text.NumberFormat oFormat = java.text.NumberFormat.getInstance();
    oFormat.setMaximumFractionDigits(2);
    m_jBinSize.setText(oFormat.format(m_fBinSize));
    m_jBinSize.setPreferredSize(new java.awt.Dimension(50,
        (int) m_jBinSize.getPreferredSize().getHeight()));
    jControls.add(m_jBinSize);

    //Button to change
    JButton jButton = new JButton("Change");
    jButton.setFont(new javawrapper.ModelFont());
    jButton.setActionCommand("UpdateBins");
    jButton.addActionListener(this);
    jControls.add(jButton);

    //Checkbox for logarithmic axis
    JCheckBox jUseLogarithmic = new JCheckBox("Use Logarithmic Axis",
                                              GetUseLogarithmicAxis());
    jUseLogarithmic.setFont(new javawrapper.ModelFont());
    jUseLogarithmic.setName("log_checkbox");
    jUseLogarithmic.addActionListener(this);
    jControls.add(jUseLogarithmic);

    //Checkbox for showing total bars
    JCheckBox jUseTotals = new JCheckBox("Show Totals", m_bShowTotal);
    jUseTotals.setFont(new javawrapper.ModelFont());
    jUseTotals.setName("total_checkbox");
    jUseTotals.addActionListener(this);
    jControls.add(jUseTotals);

    //Remove the chart, which is the content panel
    org.jfree.chart.ChartPanel jChart = null;
    for (i = 0; i < m_oChartFrame.getContentPane().getComponentCount(); i++) {
      if (m_oChartFrame.getContentPane().getComponent(i) instanceof org.jfree.
          chart.ChartPanel) {
        jChart = (org.jfree.chart.ChartPanel) m_oChartFrame.getContentPane().
            getComponent(i);
      }
    }

    //Recreate the content pane with the controls and the chart
    JPanel jContentPanel = new JPanel(new java.awt.BorderLayout());
    jContentPanel.setLayout(new java.awt.BorderLayout());
    jContentPanel.add(jControls, java.awt.BorderLayout.NORTH);
    jContentPanel.add(jChart, java.awt.BorderLayout.CENTER);

    m_oChartFrame.setContentPane(jContentPanel);

    return m_oChartFrame;

  }

  /**
   * Copies our dataset and includes only species marked as visible in the
   * legend.
   * @param oLegend Legend for this chart.
   * @return The copied dataset.
   * @throws ModelException Passing through an underlying exception.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  protected ModelHistogramDataset UpdateForVisible(Legend oLegend) throws
      ModelException {
    ModelHistogramDataset oNewDataset = (ModelHistogramDataset) m_oDataset.
        clone();
    String sName;
    int i;

    for (i = 0; i < oNewDataset.getSeriesCount(); i++) {
      sName = oNewDataset.getSeriesName(i);

      if (sName.equalsIgnoreCase("total") == false) {

        //If it's enabled, add the series to the copy
        if (!oLegend.GetIsSpeciesSelected(m_oManager.GetSpeciesCodeFromName(
            sName))) {
          oNewDataset.RemoveSeries(i);
          i--;
        }
      }
    }

    return oNewDataset;
  }

  /**
   * Updates the histogram.
   * @param oLegend Legend for this chart.
   * @throws datavisualizer.ModelException Passing through an underlying
   * exception.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  void UpdateChart(Legend oLegend) throws ModelException {
    float fValue;
    int i, j, iNumSeries, iNumItems;

    //If there are no series in the dataset, add them now (in case the
    //file was re-parsed while this chart was open)
    if (m_oDataset.getSeriesCount() == 0) {
      AddSeriesToDataset();
      //Adjust values to amounts per hectare
      iNumSeries = m_oDataset.getSeriesCount();
      for (i = 0; i < iNumSeries; i++) {
        iNumItems = m_oDataset.getItemCount(i);
        for (j = 0; j < iNumItems; j++) {
          fValue = m_oDataset.getYValue(i, j).intValue();
          fValue /= m_fPlotAreaInHectares;
          m_oDataset.setYValue(i, j, fValue);
        }
      }
    }

    ModelHistogramDataset oAdjustedDataset = UpdateForVisible(oLegend);

    DataGrapher.UpdateHistogram(oAdjustedDataset, m_oChartFrame, oLegend,
                                this);

    //Update bin information, if we're updating
    if (m_bRecalcBinsOnUpdate) {
      m_jNumBins.setText(String.valueOf(m_iNumBins));
      java.text.NumberFormat oFormat = java.text.NumberFormat.getInstance();
      oFormat.setMaximumFractionDigits(2);
      m_jBinSize.setText(oFormat.format(m_fBinSize));
    }
  }

  /**
   * This will add all series to the dataset.
   * @throws ModelException Passing through from underlying
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   * <br>December 6, 2004:  Added a series for total (LEM)
   */
  protected void AddSeriesToDataset() throws ModelException {
    double[] p_oNewSeries;
    float fMax = 0;
    int i, j;

    //If our binsize is 0, figure it out from the data.  Get the maximum value
    //in the dataset.  We'll want to refigure this at every update because we
    //don't know the final limits of whatever we're graphing.
    if (m_bRecalcBinsOnUpdate) {
      //Get the maximum value of all series
      for (i = 0; i < mp_oData.length; i++) {
        for (j = 0; j < mp_oData[i].size(); j++) {
          Number oNumber = (Number) mp_oData[i].elementAt(j);
          if (oNumber.floatValue() > fMax) {
            fMax = oNumber.floatValue();
          }
        }
      }

      m_fBinSize = (fMax + 1) / m_iNumBins;
    }

    //Add all accumulated data to the dataset
    for (i = 0; i < mp_oData.length; i++) {
      p_oNewSeries = new double[mp_oData[i].size()];

      for (j = 0; j < p_oNewSeries.length; j++) {
        if (m_bIsFloat) {
          Float fVal = (Float) mp_oData[i].elementAt(j);
          p_oNewSeries[j] = fVal.doubleValue();
        }
        else {
          Integer iVal = (Integer) mp_oData[i].elementAt(j);
          p_oNewSeries[j] = iVal.doubleValue();
        }
      }

      if (p_oNewSeries.length > 0) {
        m_oDataset.addSeries(m_oManager.GetSpeciesNameFromCode(i), p_oNewSeries,
                             m_iNumBins, m_fBinSize);
      }
    }

    //If totals are desired, add them
    if (m_bShowTotal) {
      m_oDataset.AddTotalSeries();
    }

  }

  /**
   * This will add the integer data, if the data for this chart is an int that
   * matches the data member code.
   * @param iSpecies Species for this data.
   * @param iType Type for this data.  We don't care, other than for matching codes.
   * @param iCode Data member code.
   * @param iVal Value.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void AddTreeIntData(int iSpecies, int iType, int iCode, int iVal) {
    if (m_bIsInt && mp_iDataMemberCodes[iSpecies][iType] == iCode) {

      //Add this piece of data to our dataset
      mp_oData[iSpecies].add(new Integer(iVal));
    }
  }

  /**
   * If the label matches the one given in the constructor, and the data being
   * histogrammed is an int, this captures the values passed.
   * @param iSpecies Species for this data member
   * @param iType Type for this data member
   * @param sLabel Data member label.
   * @param iCode Data member code.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void AddTreeIntDataMemberCode(int iSpecies, int iType, String sLabel,
                                       int iCode) {
    if (m_bIsInt) {
      if (m_sLabel.equals(sLabel)) {
        mp_iDataMemberCodes[iSpecies][iType] = iCode;
      }
    }
  }

  /**
   * If our data piece that's being graphed is an integer, this returns true.
   * Otherwise, false.
   * @return Whether this object's data piece is an integer.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public boolean WantAnyTreeInts() {
    return m_bIsInt;
  }

  /**
   * If our data piece that's being graphed is a float, this returns true.
   * Otherwise, false.
   * @return Whether this object's data piece is a float.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public boolean WantAnyTreeFloats() {
    return m_bIsFloat;
  }

  /**
   * This nulls our dataset and recreates it, and clears all of our data
   * vectors.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  void ClearData() {

    m_oDataset = null;
    m_oDataset = new ModelHistogramDataset();

    for (int i = 0; i < mp_oData.length; i++) {
      mp_oData[i].clear();
    }
  }

  /**
   * This will add the float data, if the data for this chart is a float that
   * matches the data member code.
   * @param iSpecies Species for this data.
   * @param iType Type for this data.  We don't care, other than for matching codes.
   * @param iCode Data member code.
   * @param fVal Value.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void AddTreeFloatData(int iSpecies, int iType, int iCode, float fVal) {
    if (m_bIsFloat && mp_iDataMemberCodes[iSpecies][iType] == iCode) {

      //Add this piece of data to our dataset
      mp_oData[iSpecies].add(new Float(fVal));
    }
  }
}
