package javawrapper;

import java.io.BufferedWriter;
import java.util.Vector;
import org.xml.sax.Attributes;

/**
 * This class functions as an organizer for model-level behaviors.  Each object
 * of this class manages a group of similar behavior objects.
 * <p>
 * This is the behavior level of which the GUIManager is aware.  It keeps a
 * known set of these objects around, which can respond to a common set of
 * requests, and relies on them to manage the complexity of the individual
 * behaviors which they own.
 * <p>Copyright: Copyright (c) 2003 Charles D. Canham</p>
 * <p>Company: Institute of Ecosystem Studies</p>
 * @author Lora E. Murphy
 * @version 1.0
 *
 * <br>Edit history:
 * <br>------------------
 * <br>April 28, 2004: Submitted in beta version (LEM)
 */

abstract class BehaviorTypeBase
    extends WorkerBase {

  /**
   * Constructor.
   * @param oManager GUIManager object.
   * @param sName Name of this object to display to the user.
   * @param sXMLTag Parent XML tag for all of this object's data.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public BehaviorTypeBase(GUIManager oManager, String sName, String sXMLTag) {
    super(oManager, sName, sXMLTag);
    m_iGridParseIndex = -1;
  }

  /**
   * Gets which species should have values for a given ModelVector.  This polls
   * all behaviors; if a behavior is enabled, and this ModelVector is on its
   * required data list, all species for which it is enabled are set to true.
   * @param p_oData The ModelVector to check.
   * @param oPop TreePopulation object.
   * @return A list of flags, one per species, that indicate for each whether
   * p_oData should have a value for that species.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public boolean[] GetWhichSpeciesUsed(ModelVector p_oData, TreePopulation oPop) {
    ModelData oCompare;
    boolean[] p_bReturnFlags = new boolean[oPop.GetNumberOfSpecies()];
    String sName = p_oData.GetDescriptor();
    int i, j;
    for (i = 0; i < p_bReturnFlags.length; i++) {
      p_bReturnFlags[i] = false;
    }

    //Now go through each behavior
    if (mp_oChildBehaviors == null) {
      return p_bReturnFlags;
    }

    for (i = 0; i < mp_oChildBehaviors.length; i++) {
      if (mp_oChildBehaviors[i].m_bIsEnabled) {
        for (j = 0; j < mp_oChildBehaviors[i].mp_oRequiredData.size(); j++) {
          if (mp_oChildBehaviors[i].mp_oRequiredData.elementAt(j) instanceof
              ModelData) {
            oCompare = (ModelData) mp_oChildBehaviors[i].mp_oRequiredData.
                elementAt(j);
            if (oCompare.GetDescriptor().equals(sName)) {
              p_bReturnFlags = MergeBooleans(p_bReturnFlags,
                                             mp_oChildBehaviors[i].
                                             GetWhichSpeciesUsed(oPop));
              break;
            }
          }
        }
      }
    }
    return p_bReturnFlags;
  }

  /**
   * Determines if any of the child behaviors are currently enabled.
   * @return True if there are any child behaviors enabled.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public boolean AnyBehaviorsEnabled() {
    int i;
    boolean bAnybodyEnabled = false;

    if (mp_oChildBehaviors != null) {
      for (i = 0; i < mp_oChildBehaviors.length; i++) {
        if (mp_oChildBehaviors[i].m_bIsEnabled) {
          bAnybodyEnabled = true;
          break;
        }
      }
    }

    return bAnybodyEnabled;
  }

  /**
   * Finds a behavior by its XML tag.
   * @param sXMLTag String The XML tag for which to find a behavior.
   * @return Behavior Behavior for the XML tag, or NULL if none of the
   * behaviors has that tag.
   */
  public Behavior GetBehaviorByXMLTag(String sXMLTag) {
    int j;
    if (mp_oChildBehaviors != null) {

      for (j = 0; j < mp_oChildBehaviors.length; j++) {

        if (mp_oChildBehaviors[j].GetParameterFileBehaviorName().equals(sXMLTag)) {
          return mp_oChildBehaviors[j];
        }
      }
    }
    return null;
  }

  /**
   * List of child behaviors a behavior group is managing.  Notice that this
   * array is not created in this class.  Also note that you can't count on the
   * order remaining constant, so to work with behaviors, use a
   * GetBehaviorBy* method.
   */
  protected Behavior[] mp_oChildBehaviors;

  /**
   * List of grid objects associated with these behaviors - notice this
   * array is not created in this class
   */
  protected Grid[] mp_oGrids;

  protected int
      /**When parsing grid maps, this is the index of the current map (or -1
       * if the current map is not owned by this behavior).*/
      m_iGridParseIndex,
      /**When parsing packages, this is the index of the current package (or
       * -1 if the current map is not owned by this behavior).*/
      m_iPackageIndex,
      /**When parsing grid maps, the current X grid cell*/
      m_iCurrentXCell,
      /**When parsing grid maps, the current Y grid cell*/
      m_iCurrentYCell;
  /**
   * Gets the list of grid objects which the currently enabled behaviors would
   * be expected to create - and thus would be available to output, etc.
   * @return The list of grids enabled, or null if none are enabled.
   * @throws ModelException passing through from called methods.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public Grid[] GetEnabledGridObjects() throws ModelException {
    if (mp_oChildBehaviors == null) {
      return null;
    }

    Vector oGrids = new Vector(0);
    Grid oExistingGrid, oNewGrid;
    int i, j, k;
    boolean bAlreadyOnList;

    for (i = 0; i < mp_oChildBehaviors.length; i++) {
      if (mp_oChildBehaviors[i].m_bIsEnabled) {
        for (j = 0; j < mp_oChildBehaviors[i].GetNumberOfGrids(); j++) {
          oNewGrid = mp_oChildBehaviors[i].GetGrid(j);
          //Make sure each grid is not already on the list
          bAlreadyOnList = false;
          for (k = 0; k < oGrids.size(); k++) {
            oExistingGrid = (Grid) oGrids.elementAt(k);
            if (oExistingGrid.GetName().equals(oNewGrid.GetName())) {
              bAlreadyOnList = true;
              break;
            }
          }
          if (!bAlreadyOnList) {
            oGrids.add(oNewGrid);
          }
        }
      }
    }

    if (oGrids.size() == 0) {
      return null;
    }

    Grid[] p_oGrids = new Grid[oGrids.size()];
    for (i = 0; i < p_oGrids.length; i++) {
      p_oGrids[i] = (Grid) oGrids.elementAt(i);
    }

    return p_oGrids;
  }

  /**
   * Returns all grid objects, whether they are used or not.
   * @return All grid objects for a behavior grouping, or null if there are
   * none.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public Grid[] GetAllGridObjects() {
    return mp_oGrids;
  }

  /**
   * Get all Behavior objects for this group.
   * @return The array of Behavior objects.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public Behavior[] GetAllBehaviors() {
    return mp_oChildBehaviors;
  }

  /**
   * Writes behavior tags to the parameter file.  For all enabled behaviors,
   * writes them in order along with their appropriate applyTo tags.
   * @param out File stream to write to.
   * @param oPop Tree population object.
   * @throws ModelException if there is a problem writing the file.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  void WriteBehaviorNodes(BufferedWriter out, TreePopulation oPop) throws
      ModelException {
    try {
      int i, j; //loop counters
      if (null != mp_oChildBehaviors) {
        for (i = 0; i < mp_oChildBehaviors.length; i++) {
          if (mp_oChildBehaviors[i].m_bIsEnabled) {

            //Write out the behavior tag
            out.write("<behavior><behaviorName>");
            out.write(mp_oChildBehaviors[i].GetParameterFileBehaviorName());
            out.write("</behaviorName><version>");
            out.write(String.valueOf(mp_oChildBehaviors[i].GetVersion()));
            out.write("</version>");

            //Now write the appliesTo tags
            for (j = 0; j < mp_oChildBehaviors[i].GetNumberOfCombos(); j++) {
              //When writing the species tags, replace spaces with underscores;
              //the XML parser can't handle spaces in attributes
              out.write("<applyTo species=\"");
              out.write(oPop.GetSpeciesNameFromCode(mp_oChildBehaviors[i].
                  GetSpeciesTypeCombo(j).
                  GetSpecies()).replace(' ', '_'));
              out.write("\" type=\"");
              out.write(oPop.GetTypeNameFromCode(mp_oChildBehaviors[i].
                                                 GetSpeciesTypeCombo(j).
                                                 GetType()));
              out.write("\"/>");
            }

            //Now closing behavior tag
            out.write("</behavior>");
          }
        }
      }
    }
    catch (java.io.IOException e) {
      ModelException oErr = new ModelException();
      oErr.iErrorCode = ErrorGUI.BAD_FILE;
      oErr.sCallingFunction = "JAVA";
      oErr.sMessage = "There was a problem writing the parameter file.";
      throw (oErr);
    }
  }

  /**
   * Writes grid maps for any grid that has values.  If there are no grids,
   * or if there are no grid map values in any grid, nothing is written.
   * @param jOut File to write to.
   * @throws ModelException if there is a problem writing to the file.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void WriteGridXML(BufferedWriter jOut) throws ModelException {

    if (mp_oGrids != null) {
      Plot oPlot = m_oManager.GetPlot();
      int i;
      for (i = 0; i < mp_oGrids.length; i++) {
        mp_oGrids[i].WriteXML(jOut, oPlot);
      }
    }
  }

  /**
   * Clears all grid map values for this behavior's grids.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void ClearGridMapValues() {
    if (mp_oGrids != null) {
      int i;
      for (i = 0; i < mp_oGrids.length; i++) {
        mp_oGrids[i].ClearMapValues();
      }
    }
  }

  /**
   * Sets which species for a given type use a particular behavior.
   * <p>
   * This method would normally not be used, as this object manages its own
   * data and would make the decision of which behaviors are enabled and which
   * are not.  However, in the case of input of a tab-delimited parameter file,
   * the information must be fed to this class.
   * @param p_iVals Array of use flags, one per species.  A 0 value means that
   * the given species doesn't use the behavior, a non-zero value means they
   * do.
   * @param iType Tree type to which the species list applies.
   * @param oPop Tree population object for species/type validation.
   * @throws ModelException if the type number is invalid, or if not all
   * species are represented.
   * @param sKey Key string of behavior to be set.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  protected void SetUsesBehavior(Integer[] p_iVals, int iType,
                                 TreePopulation oPop, String sKey) throws
      ModelException {
    //Make sure the size of the array matches the number of species
    if (p_iVals.length != oPop.GetNumberOfSpecies()) {
      ModelException oErr = new ModelException();
      oErr.iErrorCode = ErrorGUI.BAD_DATA;
      oErr.sMessage = "Data not provided for all species.";
      oErr.sCallingFunction = "JAVA";
      throw (oErr);
    }

    //Find behavior
    Behavior oBeh = GetBehaviorByKey(sKey);
    SpeciesTypeCombo oCombo;
    int i, j;
    boolean bUsed;
    if (null != oBeh) {

      //Clear all combos with this type in it
      for (j = 0; j < oBeh.GetNumberOfCombos(); j++) {

        if (oBeh.GetSpeciesTypeCombo(j).GetType() == iType) {
          oBeh.DeleteSpeciesTypeCombo(j);
          j = -1; //We changed the size of the combo list - start over
        }
      }

      //Verify that one of the use flags is non-zero - otherwise there's
      //nothing do do here
      bUsed = false;
      for (j = 0; j < p_iVals.length; j++) {
        if (p_iVals[j].compareTo(new Integer(0)) > 0) {
          bUsed = true;
        }
      }
      if (bUsed) {
        oBeh.m_bIsEnabled = true;
        for (j = 0; j < p_iVals.length; j++) {
          if (p_iVals[j].compareTo(new Integer(0)) > 0) {
            oBeh.AddSpeciesTypeCombo(new SpeciesTypeCombo(j, iType, oPop));
          }
        }
      }
      else {
        oBeh.m_bIsEnabled = false;
      }
    }
  }

  /**
   * Convenience method for getting a behavior.
   * @param sKey The short key name of the behavior.
   * @return The behavior if found, null if not.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public Behavior GetBehaviorByKey(String sKey) {
    Behavior oBeh;
    int i;
    if (mp_oChildBehaviors == null) {
      return null;
    }
    for (i = 0; i < mp_oChildBehaviors.length; i++) {
      if (mp_oChildBehaviors[i].GetKey() == sKey) {
        return mp_oChildBehaviors[i];
      }
    }
    return null;
  }

  /**
   * Convenience method for getting a behavior.
   * @param sDescriptor The descriptive name of the behavior.
   * @return The behavior if found, null if not.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public Behavior GetBehaviorByDisplayName(String sDescriptor) {
    Behavior oBeh;
    int i;
    if (mp_oChildBehaviors == null) {
      return null;
    }
    for (i = 0; i < mp_oChildBehaviors.length; i++) {
      if (mp_oChildBehaviors[i].GetDescriptor() == sDescriptor) {
        return mp_oChildBehaviors[i];
      }
    }
    return null;
  }

  /**
   * Formats data for display in a set of JTables.  For all required behaviors,
   * their required data objects are placed in a table.
   * Each element in the vector returned is one table to display, stored in a
   * vector.  Each single table vector's first element is an Object array with
   * the header row for the table, and the second element is an Object[][]
   * with the table's data.  Data labels should be placed in the first array
   * bucket.
   * <p>
   * This function operates by looking at each behavior.  For each behavior
   * which is enabled, its required data objects are added to a data vector.
   * The vector is a vector of vectors where the data objects are grouped
   * together by number of columns they require.
   * <p>
   * Once the sorted lists are in place, they are transformed to the table
   * format and returned.
   * @param oPop TreePopulation object.
   * @return Vector of vectors suitable for creating a set of JTables, or null
   * if there is no data to display.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  /*  public Vector FormatDataForDisplay(TreePopulation oPop) {
      Vector p_oSingles = new Vector(), //single data objects
   p_oSpeciesSpecific = new Vector(); //SpeciesSpecific objects - vector of vectors
      //grouped by what species they apply to
      ModelData p_oDataObject;

      int i, j;
      boolean[] p_bAppliesTo = null;
      boolean bAny = false;

      //First check to see if any child behaviors are enabled - if not, return
      //null
      for (i = 0; i < mp_oChildBehaviors.length; i++) {
        if (mp_oChildBehaviors[i].m_bIsEnabled) {
          bAny = true;
        }
      }
      if (!bAny) {
        return null;
      }

      //Go through the child behaviors again and start adding and sorting them
      for (i = 0; i < mp_oChildBehaviors.length; i++) {
        if (mp_oChildBehaviors[i].m_bIsEnabled) {
          p_bAppliesTo = mp_oChildBehaviors[i].GetWhichSpeciesUsed(oPop);
   for (j = 0; j < mp_oChildBehaviors[i].GetNumberOfRequiredDataObjects();
               j++) {

   p_oDataObject = (ModelData) mp_oChildBehaviors[i].GetRequiredData(j);
            AddDataObjectToDisplayArrays(p_oDataObject, p_oSingles,
                                         p_oSpeciesSpecific, p_bAppliesTo);
          }
        }
      }

      return FormatTable(p_oSingles, p_oSpeciesSpecific, oPop);
    } */

  /**
   * Accepts an XML parent tag (empty, no data) from the parser.  This is
   * looking for grid data.
   * @param sXMLTag The XML tag.
   * @param oAttributes The attributes of this object.
   * @throws ModelException if there is a problem reading this data.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void ReadXMLParentTag(String sXMLTag, Attributes oAttributes) throws
      ModelException {

    if (sXMLTag.equals("ma_v")) {

      //Get the X and Y
      if (m_iGridParseIndex > -1) {
        m_iCurrentXCell = new Integer(oAttributes.getValue("x")).intValue();
        m_iCurrentYCell = new Integer(oAttributes.getValue("y")).intValue();
        m_iPackageIndex = -1;
      }
    }
    else if (sXMLTag.equals("pkg")) {
      if (m_iGridParseIndex > -1) {
        m_iPackageIndex++;
      }
    }
    else if (sXMLTag.equals("grid")) {

      //This is a grid - does it belong to this object?
      String sGridName = oAttributes.getValue("gridName");
      m_iGridParseIndex = -1;
      m_iPackageIndex = -1;

      if (mp_oGrids != null) {
        for (int i = 0; i < mp_oGrids.length; i++) {
          if (mp_oGrids[i].GetName().equals(sGridName)) {
            m_iGridParseIndex = i;
            break;
          }
        }
      }

    }
    else if (sXMLTag.equals("ma_floatCodes") ||
             sXMLTag.equals("ma_intCodes") ||
             sXMLTag.equals("ma_boolCodes") ||
             sXMLTag.equals("ma_charCodes")) {
      m_iPackageIndex = -1;

    }
    else if (sXMLTag.equals("ma_packageFloatCodes") ||
             sXMLTag.equals("ma_packageIntCodes") ||
             sXMLTag.equals("ma_packageBoolCodes") ||
             sXMLTag.equals("ma_packageCharCodes")) {
      //Set this flag so we can tell that the next data member codes are package
      m_iPackageIndex = 0;
    }
    else {

      //If this is not a recognized grid tag, then grids are not currently
      //being parsed; reset the grid index
      m_iGridParseIndex = -1;
      m_iPackageIndex = -1;
    }
  }

  /**
   * This method is looking for ma_floatCode, ma_intCode, ma_charCode,
   * ma_boolCode, fl, int, ch, and bl.  For the last four, they are only set
   * if m_iCurrentSpecies and m_iCurrentTreeType are greater than -1.  Thus
   * grid values with these tags will be ignored.
   * @param sXMLTag XML tag of data object whose value is to be set.
   * @param sXMLParentTag The immediate parent tag that sXMLTag is within.
   * @param oAttributes Attributes of the object.  Ignored, but may be needed
   * by overriding objects.
   * @param oData Data value appropriate to the data type
   * @return true if the value was set successfully; false if the value could
   * not be found.  (This would not be an error, because I need a way to
   * cycle through the objects until one of the objects comes up with a
   * match.)
   * @throws ModelException if the value could not be assigned to the data
   * object.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public boolean SetSingleValueByXMLTag(String sXMLTag, String sXMLParentTag,
                                        Attributes oAttributes,
                                        Object oData) throws
      ModelException {
    if (sXMLTag != null) {
      if (sXMLTag.equals("fl")) {
        if (m_iGridParseIndex > -1) {

          //Get the code
          int iCode = new Integer(oAttributes.getValue("c")).intValue();
          Float fValue = new Float(String.valueOf(oData));

          //Set it for the grid
          mp_oGrids[m_iGridParseIndex].SetGridValue(m_iCurrentXCell,
              m_iCurrentYCell, iCode, fValue, m_oManager.GetPlot());
          return true;
        }
      }
      else if (sXMLTag.equals("int")) {
        if (m_iGridParseIndex > -1) {

          //Get the code
          int iCode = new Integer(oAttributes.getValue("c")).intValue();
          Integer iValue = new Integer(String.valueOf(oData));

          //Set it for the grid
          mp_oGrids[m_iGridParseIndex].SetGridValue(m_iCurrentXCell,
              m_iCurrentYCell, iCode, iValue, m_oManager.GetPlot());

          return true;
        }
      }
      else if (sXMLTag.equals("ch")) {
        if (m_iGridParseIndex > -1) {

          //Get the code
          int iCode = new Integer(oAttributes.getValue("c")).intValue();

          //Set it for the grid
          mp_oGrids[m_iGridParseIndex].SetGridValue(m_iCurrentXCell,
              m_iCurrentYCell, iCode, String.valueOf(oData), m_oManager.GetPlot());
          return true;
        }
      }
      else if (sXMLTag.equals("bl")) {
        if (m_iGridParseIndex > -1) {

          //Get the code
          int iCode = new Integer(oAttributes.getValue("c")).intValue();
          Boolean bValue = new Boolean(String.valueOf(oData));

          //Set it for the grid
          mp_oGrids[m_iGridParseIndex].SetGridValue(m_iCurrentXCell,
              m_iCurrentYCell, iCode, bValue, m_oManager.GetPlot());
          return true;
        }
      }
      if (sXMLTag.equals("pfl")) {
        if (m_iGridParseIndex > -1) {

          //Get the code
          int iCode = new Integer(oAttributes.getValue("c")).intValue();
          Float fValue = new Float(String.valueOf(oData));

          //Set it for the grid
          mp_oGrids[m_iGridParseIndex].SetGridPackageValue(m_iCurrentXCell,
              m_iCurrentYCell, m_iPackageIndex, iCode, fValue,
              m_oManager.GetPlot());
          return true;
        }
      }
      else if (sXMLTag.equals("pint")) {
        if (m_iGridParseIndex > -1) {

          //Get the code
          int iCode = new Integer(oAttributes.getValue("c")).intValue();
          Integer iValue = new Integer(String.valueOf(oData));

          //Set it for the grid
          mp_oGrids[m_iGridParseIndex].SetGridPackageValue(m_iCurrentXCell,
              m_iCurrentYCell, m_iPackageIndex, iCode, iValue,
              m_oManager.GetPlot());

          return true;
        }
      }
      else if (sXMLTag.equals("pch")) {
        if (m_iGridParseIndex > -1) {

          //Get the code
          int iCode = new Integer(oAttributes.getValue("c")).intValue();

          //Set it for the grid
          mp_oGrids[m_iGridParseIndex].SetGridPackageValue(m_iCurrentXCell,
              m_iCurrentYCell, m_iPackageIndex, iCode, String.valueOf(oData),
              m_oManager.GetPlot());
          return true;
        }
      }
      else if (sXMLTag.equals("pbl")) {
        if (m_iGridParseIndex > -1) {

          //Get the code
          int iCode = new Integer(oAttributes.getValue("c")).intValue();
          Boolean bValue = new Boolean(String.valueOf(oData));

          //Set it for the grid
          mp_oGrids[m_iGridParseIndex].SetGridPackageValue(m_iCurrentXCell,
              m_iCurrentYCell, m_iPackageIndex, iCode, bValue,
              m_oManager.GetPlot());
          return true;
        }
      }
      else if (sXMLTag.equals("ma_floatCode")) {
        if (m_iGridParseIndex > -1) {
          String sLabel = oAttributes.getValue("label");
          int iIndex = new Integer( (String) oData).intValue();
          if (m_iPackageIndex > -1) {
            mp_oGrids[m_iGridParseIndex].SetGridPackageFloatCode(sLabel, iIndex);
            return true;
          }
          else {
            mp_oGrids[m_iGridParseIndex].SetGridFloatCode(sLabel, iIndex);
            return true;
          }
        }
      }
      else if (sXMLTag.equals("ma_intCode")) {
        if (m_iGridParseIndex > -1) {
          String sLabel = oAttributes.getValue("label");
          int iIndex = new Integer( (String) oData).intValue();
          if (m_iPackageIndex > -1) {
            mp_oGrids[m_iGridParseIndex].SetGridPackageIntCode(sLabel, iIndex);
            return true;
          }
          else {
            mp_oGrids[m_iGridParseIndex].SetGridIntCode(sLabel, iIndex);
            return true;
          }
        }
      }
      else if (sXMLTag.equals("ma_boolCode")) {
        if (m_iGridParseIndex > -1) {
          String sLabel = oAttributes.getValue("label");
          int iIndex = new Integer( (String) oData).intValue();
          if (m_iPackageIndex > -1) {
            mp_oGrids[m_iGridParseIndex].SetGridPackageBoolCode(sLabel, iIndex);
            return true;
          }
          else {
            mp_oGrids[m_iGridParseIndex].SetGridBoolCode(sLabel, iIndex);
            return true;
          }
        }
      }
      else if (sXMLTag.equals("ma_charCode")) {
        if (m_iGridParseIndex > -1) {
          String sLabel = oAttributes.getValue("label");
          int iIndex = new Integer( (String) oData).intValue();
          if (m_iPackageIndex > -1) {
            mp_oGrids[m_iGridParseIndex].SetGridPackageCharCode(sLabel, iIndex);
            return true;
          }
          else {
            mp_oGrids[m_iGridParseIndex].SetGridCharCode(sLabel, iIndex);
            return true;
          }
        }
      }
      else if (sXMLTag.equals("ma_lengthXCells")) {
        if (m_iGridParseIndex > -1) {

          float fLength = new Float( (String) oData).floatValue();
          mp_oGrids[m_iGridParseIndex].SetXCellLength(fLength);
          return true;
        }
      }
      else if (sXMLTag.equals("ma_lengthYCells")) {
        if (m_iGridParseIndex > -1) {

          float fLength = new Float( (String) oData).floatValue();
          mp_oGrids[m_iGridParseIndex].SetYCellLength(fLength);
          return true;
        }
      }
      else if (sXMLTag.equals("ma_plotLenX")) {
        if (m_iGridParseIndex > -1) {
          if (java.lang.Math.abs(m_oManager.GetPlot().GetPlotXLength() -
                                 new Float( (String) oData).floatValue()) >
              0.001) {
            ModelException oErr = new ModelException();
            oErr.sMessage = "In the grid map for " +
                mp_oGrids[m_iGridParseIndex].GetName() +
                ", there is an invalid " +
                "X plot length of \"" + String.valueOf(oData) + "\".";
            throw (oErr);
          }
          return true;
        }
      }
      else if (sXMLTag.equals("ma_plotLenY")) {
        if (m_iGridParseIndex > -1) {
          if (java.lang.Math.abs(m_oManager.GetPlot().GetPlotYLength() -
                                 new Float( (String) oData).floatValue()) >
              0.001) {
            ModelException oErr = new ModelException();
            oErr.sMessage = "In the grid map for " +
                mp_oGrids[m_iGridParseIndex].GetName() +
                ", there is an invalid " +
                "Y plot length of \"" + String.valueOf(oData) + "\".";
            throw (oErr);
          }
          return true;
        }
      }
    }
    return super.SetSingleValueByXMLTag(sXMLTag, sXMLParentTag,
                                        oAttributes, oData);
  }

  /**
   * Override this function in order to be able to handle grid maps.
   * @param sXMLTag Parent XML tag of data vector whose value is to be set.
   * @param sXMLParentTag The immediate parent tag that sXMLTag is within.
   * @param p_oData Vector of data values appropriate to the data type
   * @param p_sChildXMLTags The XML tags of the child elements
   * @param p_bAppliesTo Array of booleans saying which of the vector values
   * should be set.  This is important in the case of species-specifics - the
   * vector index is the species number but not all species are set.
   * @param oParentAttributes Attributes of parent tag.  May be useful when
   * overridding this for unusual tags.
   * @param p_oAttributes Attributes passed from parser.  This may be needed
   * when overriding this function.  Basic species-specific values are
   * already handled by this function.
   * @return true if the value was set successfully; false if the value could
   * not be found.  (This would not be an error, because I need a way to
   * cycle through the objects until one of the objects comes up with a
   * match.)  If a match to a data object is made via XML tag, but the found
   * object is not a ModelVector, this returns false.
   * @throws ModelException if the value could not be assigned to the data
   * object.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public boolean SetVectorValueByXMLTag(String sXMLTag, String sXMLParentTag,
                                        Vector p_oData,
                                        String[] p_sChildXMLTags,
                                        boolean[] p_bAppliesTo,
                                        Attributes oParentAttributes,
                                        Attributes[] p_oAttributes) throws
      ModelException {

    //These may be tree map codes rolled up into a vector - handle them
    //like individuals
    if (m_iGridParseIndex > -1) {
      int i;
      boolean bReturn = false;
      for (i = 0; i < p_sChildXMLTags.length; i++) {
        if (SetSingleValueByXMLTag(p_sChildXMLTags[i], sXMLParentTag,
                                   p_oAttributes[i],
                                   p_oData.elementAt(i))) {
          bReturn = true;
        }
      }

      if (bReturn) {
        return bReturn;
      }
    }
    return super.SetVectorValueByXMLTag(sXMLTag, sXMLParentTag, p_oData,
                                        p_sChildXMLTags,
                                        p_bAppliesTo, oParentAttributes,
                                        p_oAttributes);
  }

  /**
   * Writes a set of species-specific float values to a parameter file.  The
   * data is only written if it is required for a behavior that is enabled; and
   * then only for the species required by any behavior for which it is
   * required.
   * @param out The file to write to.
   * @param p_oData The vector of data pieces.
   * @param oPop Tree population object to get species names
   * @throws java.io.IOException Passes on exceptions from FileWriter
   * @throws ModelException Passes on from TreePopulation
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void WriteSpeciesSpecificValue(BufferedWriter out, ModelVector p_oData,
                                        TreePopulation oPop) throws
      java.io.IOException, ModelException {

    if (p_oData == null || p_oData.GetValue().size() == 0) {
      return;
    }

    //Determine which species for which this is enabled
    boolean[] p_bUse;
    if (p_oData.GetMustApplyToAllSpecies()) {
      p_bUse = new boolean[oPop.GetNumberOfSpecies()];
      for (int i = 0; i < p_bUse.length; i++) {
        p_bUse[i] = true;
      }
    }
    else {
      p_bUse = GetWhichSpeciesUsed(p_oData, oPop);
    }

    super.WriteSpeciesSpecificValue(out, p_oData, oPop, p_bUse);
  }

  /**
   * Writes the XML data to a parameter file for the behaviors owned by this
   * object.  Overridden so that only the data enabled by behaviors is written.
   * @param jOut File stream to write to.
   * @param oPop TreePopulation object.
   * @throws ModelException if there is a problem writing the file or
   * validating the data.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void WriteXML(BufferedWriter jOut, TreePopulation oPop) throws
      ModelException {

    //Make sure a behavior is enabled - if not, return
    if (!AnyBehaviorsEnabled()) {
      return;
    }

    super.WriteXML(jOut, oPop);
  }

  /**
   * Determines whether a piece of data is required for any enabled behaviors
   * in the mp_oChildBehaviors list.
   * @param oData Piece of data to check.
   * @return True if an enabled behavior requires it; false otherwise.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public boolean IsThisDataUsed(ModelData oData) {
    ModelData oCompare;
    String sName = oData.GetDescriptor();
    int i, j;
    boolean bIsUsed = false;

    for (i = 0; i < mp_oChildBehaviors.length; i++) {
      if (mp_oChildBehaviors[i].m_bIsEnabled) {
        for (j = 0; j < mp_oChildBehaviors[i].mp_oRequiredData.size(); j++) {
          if (mp_oChildBehaviors[i].mp_oRequiredData.elementAt(j) instanceof
              ModelData) {
            oCompare = (ModelData) mp_oChildBehaviors[i].mp_oRequiredData.
                elementAt(j);
            if (oCompare.GetDescriptor().equals(sName)) {
              bIsUsed = true;
              break;
            }
          }
          if (bIsUsed) {
            break;
          }
        }
      }
    }

    return bIsUsed;
  }

  /**
   * Writes a piece of data to an XML file.
   * @param jOut The file to write to.
   * @param oData The data being written.
   * @throws java.io.IOException Passes on exceptions from FileWriter
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void WriteData(BufferedWriter jOut, ModelData oData) throws java.io.
      IOException {
    if (IsThisDataUsed(oData)) {
      super.WriteData(jOut, oData);
    }
  }

}
