package javawrapper;

import javax.swing.*;
import java.awt.*;
import java.awt.event.*;

import java.util.Vector;

/**
* Window the user uses to set up detailed output options for grids.  A combo
* box displays a list of all grids set up by BehaviorTypeBase-descended
* objects. When the user chooses one of these grids, a list of all its data
* members and package data members appears so the user can choose what to save.
* <p>
* We have to have a way to tell the regular data members from the package
* data members.  They are all displayed together, not in separate boxes; so
* we append a "(p)" flag to the beginning of package data member display
* names.  A custom renderer splits it back off for display so the user doesn't
* see it; and it will also be split off before notifying the
* DetailedGridSettings object that it's been picked.
* <p>Copyright: Copyright (c) Charles D. Canham 2003</p>
* <p>Company: Institute of Ecosystem Studies</p>
* @author Lora E. Murphy
* @version 1.0
*
* <br>Edit history:
* <br>------------------
* <br>April 28, 2004: Submitted in beta version (LEM)
*/
public class DetailedOutputGridSetup
    extends JDialog
    implements ActionListener {

  /**OutputBehaviors object to exchange data with*/
  protected OutputBehaviors m_oOutputBehaviors;

  /**For detailed output - collection of DetailedGridSettings objects representing the
   * temp settings that changes will be made to*/
  protected Vector mp_oDetailedGridSaveSettings = new Vector(0);

  JLabel jLabel1 = new JLabel();
  JComboBox m_jGridListCombo = new JComboBox();
  JLabel jLabel4 = new JLabel();
  JLabel jLabel2 = new JLabel();
  JButton jAddButton = new JButton();
  DefaultListModel jDataMemberListModel = new DefaultListModel(),
      jSaveListModel = new DefaultListModel();
  JTextField jTimestepsEdit = new JTextField();
  JLabel jLabel5 = new JLabel();
  JScrollPane jScrollPane1 = new JScrollPane();
  JScrollPane jScrollPane2 = new JScrollPane();
  JList jSavedList = new JList(jSaveListModel);
  GridBagLayout gridBagLayout1 = new GridBagLayout();
  JPanel jComponentPanel = new JPanel();

  /**List of grids to display*/
  protected Grid[] mp_oGridList;
  JButton jRemoveButton = new JButton();
  JList jDataMemberList = new JList(jDataMemberListModel);

  private String m_sHelpID = "windows.detailed_output_grid_setup";

  /**
   * Constructor.  Builds the GUI.
   * @param oParent Parent dialog in which to display this dialog.
   * @param oOutput OutputBehaviors object, to draw data from to display and
   * to send changes to
   * @throws ModelException passed through from underlying methods
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public DetailedOutputGridSetup(JDialog oParent,
                          OutputBehaviors oOutput) throws ModelException {
    super(oParent, "Set Up Grid Save Options", true);
    try {

      m_oOutputBehaviors = oOutput;

      Vector p_oGrids = new Vector(0);
      BehaviorTypeBase[] p_oBehaviors = oOutput.m_oManager.GetAllBehaviorGroups();
      Grid[] oSetGrids;
      int i, j;

      jbInit();

      //Help ID
      oOutput.m_oManager.GetHelpBroker().enableHelpKey(this.getRootPane(),
          m_sHelpID, null);

      //Default save frequency
      jTimestepsEdit.setText("1");

      //Retrieve the list of enabled grid objects
      if (null != p_oBehaviors) {
        for (i = 0; i < p_oBehaviors.length; i++) {
          oSetGrids = p_oBehaviors[i].GetEnabledGridObjects();
          if (null != oSetGrids) {
            for (j = 0; j < oSetGrids.length; j++) {
              p_oGrids.add(oSetGrids[j]);
            }
          }
        }
      }

      if (p_oGrids.size() == 0) {
        m_jGridListCombo.addItem("---No grids available---");
        mp_oGridList = null;
      }
      else {

        m_jGridListCombo.addItem("---Please select grid---");
        mp_oGridList = new Grid[p_oGrids.size()];
        for (i = 0; i < mp_oGridList.length; i++) {
          mp_oGridList[i] = (Grid) p_oGrids.elementAt(i);
          m_jGridListCombo.addItem(mp_oGridList[i].GetName());
        }
      }

      //Setup settings
      LoadSettings();
      UpdateSaves();
    }
    catch (Exception e) {
      e.printStackTrace();
    }
  }

  /**
   * Causes all possible grids and grid members to be saved for the run.  This
   * can be called without the window's being visible.  Its results are
   * immediately applied to the output behavior - so this is not subject
   * to a Cancel button.
   * @throws ModelException passed through from called methods.
   */
  public void SaveAll() {
    DetailedGridSettings oSettings;
    DataMember[] p_oMembers;
    int i, j;

    m_oOutputBehaviors.ClearDetailedGridSettings();

    for (i = 0; i < mp_oGridList.length; i++) {
      oSettings = new DetailedGridSettings(mp_oGridList[i].GetName());
      oSettings.SetSaveFrequency(1);
      p_oMembers = mp_oGridList[i].GetDataMembers();
      if (p_oMembers != null) {
        for (j = 0; j < p_oMembers.length; j++) {

          if (p_oMembers[j].GetType() == DataMember.FLOAT) {
            oSettings.AddFloat(p_oMembers[j].GetCodeName());
          }
          else if (p_oMembers[j].GetType() == DataMember.INTEGER) {
            oSettings.AddInt(p_oMembers[j].GetCodeName());
          }
          else if (p_oMembers[j].GetType() == DataMember.BOOLEAN) {
            oSettings.AddBool(p_oMembers[j].GetCodeName());
          }
          else if (p_oMembers[j].GetType() == DataMember.CHAR) {
            oSettings.AddChar(p_oMembers[j].GetCodeName());
          }
        }
      }

      p_oMembers = mp_oGridList[i].GetPackageDataMembers();
      if (p_oMembers != null) {
        for (j = 0; j < p_oMembers.length; j++) {

          if (p_oMembers[j].GetType() == DataMember.FLOAT) {
            oSettings.AddPackageFloat(p_oMembers[j].GetCodeName());
          }
          else if (p_oMembers[j].GetType() == DataMember.INTEGER) {
            oSettings.AddPackageInt(p_oMembers[j].GetCodeName());
          }
          else if (p_oMembers[j].GetType() == DataMember.BOOLEAN) {
            oSettings.AddPackageBool(p_oMembers[j].GetCodeName());
          }
          else if (p_oMembers[j].GetType() == DataMember.CHAR) {
            oSettings.AddPackageChar(p_oMembers[j].GetCodeName());
          }
        }
      }

      m_oOutputBehaviors.AddDetailedGridSettings(oSettings);
    }
  }

  /**
   * Responds to button clicks.
   * @param oEvent The event telling us what button was clicked.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void actionPerformed(ActionEvent oEvent) {
    if (oEvent.getActionCommand().equals("OK")) {
      DetailedGridSettings oTempSettings, oPermSettings;
      int i, j;

      m_oOutputBehaviors.ClearDetailedGridSettings();

      //Copy over the data we've accumulated to the output behaviors
      for (i = 0; i < mp_oDetailedGridSaveSettings.size(); i++) {
        oTempSettings = (DetailedGridSettings) mp_oDetailedGridSaveSettings.
            elementAt(i);
        oPermSettings = (DetailedGridSettings) oTempSettings.clone();

        m_oOutputBehaviors.AddDetailedGridSettings(oPermSettings);
      }

      mp_oDetailedGridSaveSettings.clear();

      //Close window
      this.setVisible(false);
      this.dispose();
    }
    else if (oEvent.getActionCommand().equals("Cancel")) {
      //Close window
      this.setVisible(false);
      this.dispose();
    }
    else if (oEvent.getActionCommand().equals("Add")) {
      DoAdd();
    }
    else if (oEvent.getActionCommand().equals("Remove")) {
      DoRemove();
    }
    else if (oEvent.getSource().equals(m_jGridListCombo)) {
      //Get the value that was selected
      String sGrid = (String) m_jGridListCombo.getSelectedItem();
      if (sGrid.length() == 0 || sGrid.startsWith("---")) {
        jDataMemberListModel.clear();
        return;
      }

      //Get the grid that was selected
      DataMember[] p_oDataMembers;
      int i, j;
      for (i = 0; i < mp_oGridList.length; i++) {
        if (sGrid.equals(mp_oGridList[i].GetName())) {

          //Get the list of data members for this grid
          jDataMemberListModel.clear();
          p_oDataMembers = mp_oGridList[i].GetDataMembers();
          for (j = 0; j < p_oDataMembers.length; j++) {
            jDataMemberListModel.addElement(p_oDataMembers[j].GetDisplayName());
          }
          p_oDataMembers = mp_oGridList[i].GetPackageDataMembers();
          if (p_oDataMembers != null) {
            for (j = 0; j < p_oDataMembers.length; j++) {
              jDataMemberListModel.addElement("(p)" +
                                              p_oDataMembers[j].GetDisplayName());
            }
          }
        }
      }
    }
  }

  /**
   * Removes the data members for a grid from the list of settings.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  protected void DoRemove() {
    DetailedGridSettings oSettings;
    Object[] p_oSelected;
    String sData = "", sGridName;
    int iSpecies, i, j;

    //Get the list of selected strings
    p_oSelected = jSavedList.getSelectedValues();
    //Nothing selected?  Exit.
    if (p_oSelected.length == 0) {
      return;
    }

    //Get each save string
    for (i = 0; i < p_oSelected.length; i++) {

      sData = (String) p_oSelected[i];

      //Get the grid name
      sGridName = sData.substring(0, sData.indexOf(" - "));

      //Delete this setting
      for (j = 0; j < mp_oDetailedGridSaveSettings.size(); j++) {
        oSettings = (DetailedGridSettings) mp_oDetailedGridSaveSettings.
            elementAt(j);
        if (sGridName.equals(oSettings.GetName())) {
          mp_oDetailedGridSaveSettings.remove(j);
          j--;
        }
      }
    }

    UpdateSaves();
  }

  /**
   * Adds the data members for a grid to the list of settings.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  protected void DoAdd() {
    Object[] p_oSelected; //selected options from list boxes
    String[] p_oSelectedDataMembers; //string version of selected data members
    DataMember[] p_oGridDataMembers, p_oGridPackageDataMembers;
    Grid oGrid = null;
    DetailedGridSettings oSettings;
    String sGridName, sDataMember;
    int iSaveFreq = 0, i, j;

    //Get the save frequency
    try {
      iSaveFreq = new Integer(jTimestepsEdit.getText()).intValue();
    }
    catch (java.lang.NumberFormatException oErr) {
      //This means that they typed something other than a number
      JOptionPane jPane = new JOptionPane();
      jPane.showMessageDialog(this,
                              "Timestep save frequency must be a number.",
                              "Error", JOptionPane.ERROR_MESSAGE);
      return;
    }

    //Get the selected grid
    sGridName = (String) m_jGridListCombo.getSelectedItem();
    if (sGridName.length() == 0 || sGridName.startsWith("---")) {
      return;
    }

    for (i = 0; i < mp_oGridList.length; i++) {
      if (sGridName == mp_oGridList[i].GetName()) {
        oGrid = mp_oGridList[i];
      }
    }

    if (oGrid == null) {
      return;
    }
    p_oGridDataMembers = oGrid.GetDataMembers();
    p_oGridPackageDataMembers = oGrid.GetPackageDataMembers();

    //Get the selected data members
    p_oSelected = jDataMemberList.getSelectedValues();
    //Nothing selected?  Exit.
    if (p_oSelected.length == 0) {
      return;
    }

    p_oSelectedDataMembers = new String[p_oSelected.length];
    for (i = 0; i < p_oSelected.length; i++) {
      p_oSelectedDataMembers[i] = (String) p_oSelected[i];
    }

    //Now put what's to save in our data vector
    oSettings = new DetailedGridSettings(sGridName);
    oSettings.SetSaveFrequency(iSaveFreq);
    for (i = 0; i < p_oSelectedDataMembers.length; i++) {

      //Find the corresponding data member in the grid
      sDataMember = (String) p_oSelectedDataMembers[i];

      if (sDataMember.startsWith("(p)")) {
        //This is a package data member - strip off the "(p)"
        sDataMember = sDataMember.substring(3);

        for (j = 0; j < p_oGridPackageDataMembers.length; j++) {
          if (sDataMember.equals(p_oGridPackageDataMembers[j].GetDisplayName())) {
            if (p_oGridPackageDataMembers[j].GetType() == DataMember.FLOAT) {
              oSettings.AddPackageFloat(p_oGridPackageDataMembers[j].
                                        GetCodeName());
            }
            else if (p_oGridPackageDataMembers[j].GetType() ==
                     DataMember.INTEGER) {
              oSettings.AddPackageInt(p_oGridPackageDataMembers[j].GetCodeName());
            }
            else if (p_oGridPackageDataMembers[j].GetType() ==
                     DataMember.BOOLEAN) {
              oSettings.AddPackageBool(p_oGridPackageDataMembers[j].GetCodeName());
            }
            else if (p_oGridPackageDataMembers[j].GetType() == DataMember.CHAR) {
              oSettings.AddPackageChar(p_oGridPackageDataMembers[j].GetCodeName());
            }
          }
        }
      }
      else {

        for (j = 0; j < p_oGridDataMembers.length; j++) {
          if (sDataMember.equals(p_oGridDataMembers[j].GetDisplayName())) {
            if (p_oGridDataMembers[j].GetType() == DataMember.FLOAT) {
              oSettings.AddFloat(p_oGridDataMembers[j].GetCodeName());
            }
            else if (p_oGridDataMembers[j].GetType() == DataMember.INTEGER) {
              oSettings.AddInt(p_oGridDataMembers[j].GetCodeName());
            }
            else if (p_oGridDataMembers[j].GetType() == DataMember.BOOLEAN) {
              oSettings.AddBool(p_oGridDataMembers[j].GetCodeName());
            }
            else if (p_oGridDataMembers[j].GetType() == DataMember.CHAR) {
              oSettings.AddChar(p_oGridDataMembers[j].GetCodeName());
            }
          }
        }
      }
    }

    //If this grid is already represented, overwrite it
    for (j = 0; j < mp_oDetailedGridSaveSettings.size(); j++) {
      DetailedGridSettings oTempSettings = (DetailedGridSettings)
          mp_oDetailedGridSaveSettings.elementAt(j);
      if (oTempSettings.GetName() == oSettings.GetName()) {
        mp_oDetailedGridSaveSettings.remove(j);
        j--;
      }
    }

    mp_oDetailedGridSaveSettings.add(oSettings);

    UpdateSaves();
  }

  /**
   * Creates the window.  Auto-generated except for tweaks by me.
   * @throws java.lang.Exception if something goes wrong with the window
   * building.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  protected void jbInit() throws Exception {

    jComponentPanel.setLayout(gridBagLayout1);
    this.getContentPane().setLayout(new BorderLayout());
    jLabel1.setFont(new ModelFont());
    jLabel1.setText("Select a grid:");
    jLabel4.setFont(new ModelFont());
    jLabel4.setText("For this grid save:");
    jLabel2.setFont(new ModelFont());
    jLabel2.setText("Save every X timestep(s)");

    jAddButton.setFont(new ModelFont());
    jAddButton.setText("Add");
    jAddButton.setActionCommand("Add");
    jAddButton.addActionListener(this);

    jTimestepsEdit.setText("");
    jTimestepsEdit.setFont(new ModelFont());
    jLabel5.setFont(new ModelFont());
    jLabel5.setText("What\'s being saved:");
    m_jGridListCombo.addActionListener(this);
    m_jGridListCombo.setFont(new ModelFont());
    jSavedList.setFont(new ModelFont());
    jSavedList.setPreferredSize(new Dimension(0, 0));
    jSavedList.setVisibleRowCount(8);
    jScrollPane1.setOpaque(true);
    jScrollPane1.setPreferredSize(new Dimension(260, 132));
    jRemoveButton.setFont(new ModelFont());
    jRemoveButton.setActionCommand("Remove");
    jRemoveButton.setText("Remove");
    jRemoveButton.addActionListener(this);
    jScrollPane2.setOpaque(true);
    jScrollPane2.setPreferredSize(new Dimension(260, 132));
    jDataMemberList.setCellRenderer(new DetailedOutputGridRenderer());
    jComponentPanel.add(jLabel1,
                        new GridBagConstraints(0, 0, 2, 1, 0.0, 0.0
                                               , GridBagConstraints.WEST,
                                               GridBagConstraints.NONE,
                                               new Insets(0, 0, 0, 74), 0, 0));
    jComponentPanel.add(m_jGridListCombo,
                        new GridBagConstraints(0, 1, 3, 1, 1.0, 0.0
                                               , GridBagConstraints.CENTER,
                                               GridBagConstraints.HORIZONTAL,
                                               new Insets(5, 0, 0, 0), 144, 0));
    jComponentPanel.add(jLabel4,
                        new GridBagConstraints(0, 3, 2, 1, 0.0, 0.0
                                               , GridBagConstraints.WEST,
                                               GridBagConstraints.NONE,
                                               new Insets(0, 0, 5, 0), 0, 0));
    jComponentPanel.add(jLabel2,
                        new GridBagConstraints(0, 2, 2, 1, 0.0, 0.0
                                               , GridBagConstraints.WEST,
                                               GridBagConstraints.NONE,
                                               new Insets(9, 0, 7, 0), 0, 0));
    jComponentPanel.add(jScrollPane1,
                        new GridBagConstraints(0, 4, 4, 1, 1.0, 1.0
                                               , GridBagConstraints.CENTER,
                                               GridBagConstraints.BOTH,
                                               new Insets(0, 0, 0, 0), 0, 0));
    jScrollPane1.getViewport().add(jDataMemberList, null);
    jComponentPanel.add(jScrollPane2,
                        new GridBagConstraints(0, 8, 5, 1, 1.0, 1.0
                                               , GridBagConstraints.WEST,
                                               GridBagConstraints.BOTH,
                                               new Insets(0, 0, 0, 0), 0, 0));
    jComponentPanel.add(jLabel5,
                        new GridBagConstraints(0, 7, 2, 1, 0.0, 0.0
                                               , GridBagConstraints.WEST,
                                               GridBagConstraints.NONE,
                                               new Insets(16, 0, 5, 0), 0, 0));
    jComponentPanel.add(jAddButton,
                        new GridBagConstraints(0, 6, 1, 1, 0.0, 0.0
                                               , GridBagConstraints.WEST,
                                               GridBagConstraints.NONE,
                                               new Insets(13, 0, 0, 0), 0, 0));
    jComponentPanel.add(jTimestepsEdit,
                        new GridBagConstraints(2, 2, 1, 1, 1.0, 0.0
                                               , GridBagConstraints.CENTER,
                                               GridBagConstraints.HORIZONTAL,
                                               new Insets(13, 5, 7, 100), 0, 0));
    jComponentPanel.add(jRemoveButton,
                        new GridBagConstraints(0, 9, 1, 1, 0.0, 0.0
                                               , GridBagConstraints.WEST,
                                               GridBagConstraints.NONE,
                                               new Insets(10, 0, 0, 0), 0, 0));
    jScrollPane2.getViewport().add(jSavedList, null);

    jComponentPanel.setBorder(BorderFactory.createEmptyBorder(10, 10, 10, 10));
    this.getContentPane().add(jComponentPanel, BorderLayout.CENTER);

    this.getContentPane().add(new OKCancelButtonPanel(this,
        m_oOutputBehaviors.m_oManager.GetHelpBroker(), m_sHelpID),
                              BorderLayout.SOUTH);
  }

  /**
   * Loads the settings from output into this form's temp array for display.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  protected void LoadSettings() {
    DetailedGridSettings oPermSettings, oTempSettings;
    int i;

    mp_oDetailedGridSaveSettings.clear();

    for (i = 0; i < m_oOutputBehaviors.GetNumberOfDetailedGridSettings(); i++) {
      oPermSettings = (DetailedGridSettings) m_oOutputBehaviors.
          GetDetailedGridSetting(i);
      oTempSettings = (DetailedGridSettings) oPermSettings.clone();
      mp_oDetailedGridSaveSettings.add(oTempSettings);

    }
  }

  /**
   * Displays what's being saved in each save list based on what's in
   * mp_oDetailedTreeSaveSettings
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  protected void UpdateSaves() {
    TreePopulation oPop = m_oOutputBehaviors.GetTreePopulation();
    DetailedGridSettings oSettings;
    String sSave = "";
    int i, j;

    //Clear the existing contents of the save list
    jSaveListModel.clear();

    //Go through our settings and list each one in the appropriate box
    for (i = 0; i < mp_oDetailedGridSaveSettings.size(); i++) {
      sSave = "";
      oSettings = (DetailedGridSettings) mp_oDetailedGridSaveSettings.elementAt(
          i);

      for (j = 0; j < oSettings.GetNumberOfBools(); j++) {
        if (sSave.length() == 0) {
          sSave = oSettings.GetBool(j).GetDisplayName();
        }
        else {
          sSave = sSave + ", " + oSettings.GetBool(j).GetDisplayName();
        }
      }

      for (j = 0; j < oSettings.GetNumberOfChars(); j++) {
        if (sSave.length() == 0) {
          sSave = oSettings.GetChar(j).GetDisplayName();
        }
        else {
          sSave = sSave + ", " + oSettings.GetChar(j).GetDisplayName();
        }
      }

      for (j = 0; j < oSettings.GetNumberOfInts(); j++) {
        if (sSave.length() == 0) {
          sSave = oSettings.GetInt(j).GetDisplayName();
        }
        else {
          sSave = sSave + ", " + oSettings.GetInt(j).GetDisplayName();
        }
      }

      for (j = 0; j < oSettings.GetNumberOfFloats(); j++) {
        if (sSave.length() == 0) {
          sSave = oSettings.GetFloat(j).GetDisplayName();
        }
        else {
          sSave = sSave + ", " + oSettings.GetFloat(j).GetDisplayName();
        }
      }
      for (j = 0; j < oSettings.GetNumberOfPackageBools(); j++) {
        if (sSave.length() == 0) {
          sSave = oSettings.GetPackageBool(j).GetDisplayName();
        }
        else {
          sSave = sSave + ", " + oSettings.GetPackageBool(j).GetDisplayName();
        }
      }

      for (j = 0; j < oSettings.GetNumberOfPackageChars(); j++) {
        if (sSave.length() == 0) {
          sSave = oSettings.GetPackageChar(j).GetDisplayName();
        }
        else {
          sSave = sSave + ", " + oSettings.GetPackageChar(j).GetDisplayName();
        }
      }

      for (j = 0; j < oSettings.GetNumberOfPackageInts(); j++) {
        if (sSave.length() == 0) {
          sSave = oSettings.GetPackageInt(j).GetDisplayName();
        }
        else {
          sSave = sSave + ", " + oSettings.GetPackageInt(j).GetDisplayName();
        }
      }

      for (j = 0; j < oSettings.GetNumberOfPackageFloats(); j++) {
        if (sSave.length() == 0) {
          sSave = oSettings.GetPackageFloat(j).GetDisplayName();
        }
        else {
          sSave = sSave + ", " + oSettings.GetPackageFloat(j).GetDisplayName();
        }
      }

      sSave = oSettings.GetName() + " - " +
          sSave + " every " + String.valueOf(oSettings.GetSaveFrequency()) +
          " timestep(s)";

      jSaveListModel.addElement(sSave);
    }
  }
}

/**
 * This renders choices in a JList so that, if they have a prepended "(p)"
 * string, it is not shown to the user.
 * <p>Copyright: Copyright (c) Charles D. Canham 2003</p>
 * <p>Company: Institute of Ecosystem Studies</p>
 * @author Lora E. Murphy
 * @version 1.0
 *
 * <br>Edit history:
 * <br>------------------
 * <br>April 28, 2004: Submitted in beta version (LEM)
 */
class DetailedOutputGridRenderer
    extends JLabel
    implements ListCellRenderer {
  public DetailedOutputGridRenderer() {
    setOpaque(true);
    setHorizontalAlignment(LEFT);
    setVerticalAlignment(CENTER);
    setFont(new ModelFont());
  }

  /*
   * This method displays the file choice, with a max of 50 chars.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public Component getListCellRendererComponent(
      JList list,
      Object value,
      int index,
      boolean isSelected,
      boolean cellHasFocus) {

    if (isSelected) {
      setBackground(list.getSelectionBackground());
      setForeground(list.getSelectionForeground());
    }
    else {
      setBackground(list.getBackground());
      setForeground(list.getForeground());
    }

    //Get the string to display
    String sText = (String) value;

    //If the string starts with "(p)", trim that off
    if (sText.startsWith("(p)")) {
      sText = sText.substring(3);
    }

    setText(sText);

    return this;
  }
}
