package javawrapper;

/**
 * This is the organizer class for all disperse behaviors.
 * <p>Copyright: Copyright (c) Charles D. Canham 2003</p>
 * <p>Company: Institute of Ecosystem Studies</p>
 * @author Lora E. Murphy
 * @version 1.0
 *
 * <br>Edit history:
 * <br>------------------
 * <br>April 28, 2004: Submitted in beta version (LEM)
 */

public class DisperseBehaviors
    extends BehaviorTypeBase {

  //Data
  public final static int WEIBULL = 0;
  public final static int LOGNORMAL = 1;
  public final static int CANOPY = 0;
  public final static int GAP = 1;

  public final static int NUMBER_OF_DISPERSE_FUNCTIONS = 2;
  public final static int NUMBER_OF_FOREST_COVERS = 2;

  /**
   * STR for disperse function.  Array is 3D - first index is which disperse
   * function is used - weibull or lognormal.  The second index is cover -
   * canopy or gap.  The third index is species.
   */
  private ModelVector[][] mp_fSTR,
      /**
       * Beta for disperse function.  Array is 3D - first index is which disperse
       * function is used - weibull or lognormal.  The second index is cover -
       * canopy or gap.  The third index is species.
       */
      mp_fBeta,
      /**
       * Theta (if weibull) or Xb (if lognormal) for disperse function.
       * Array is 3D - first index is which disperse
       * function is used - weibull or lognormal.  The second index is cover -
       * canopy or gap.  The third index is species.
       */
      mp_fThetaOrXb,
      /**
       * Dispersal (if weibull) or X0 (if lognormal) for disperse function.
       * Array is 3D - first index is which disperse
       * function is used - weibull or lognormal.  The second index is cover -
       * canopy or gap.  The third index is species.
       */
      mp_fDispOrX0;

  private ModelVector[]
      /**Which disperse function to use under each forest cover - valid values
       * are WEIBULL and LOGNORMAL - this is a vector of ModelEnums*/
      mp_iWhichFunctionUsed;

  private ModelVector
      /**Slope of lambda for non spatial dispersal for each species*/
      mp_fSlopeOfLambda = new ModelVector(
      "Slope Mean Non-Spatial Seed Rain, seeds/m2/ha of BA/yr",
      "di_nonSpatialSlopeOfLambda",
      "di_nssolVal", 0,
      ModelVector.FLOAT);

  /**Intercept of lambda for non spatial dispersal for each species*/
  private ModelVector mp_fInterceptOfLambda = new ModelVector(
      "Intercept of Mean Non-Spatial Seed Rain, seeds/m2/yr",
      "di_nonSpatialInterceptOfLambda",
      "di_nsiolVal", 0, ModelVector.FLOAT);

  /**Minimum DBH for reproduction for each species*/
  private ModelVector mp_fMinDbhForReproduction;

  /**STR for stump dispersal for each species*/
  private ModelVector mp_fStumpSTR;

  /**Beta for stump dispersal for each species*/
  private ModelVector mp_fStumpBeta;

  /**Azimuth direction of maximum dispersal distance, in radians*/
  private ModelVector mp_fDirectionMaxDispersal = new ModelVector(
      "Azimuth Direction of Max Dispersal Distance, in rad",
      "di_directionMaxDispersal", "di_dmdVal", 0, ModelVector.FLOAT);

  /**Amplitude of anisotropic effect*/
  private ModelVector mp_fAnisotropicAmplitude = new ModelVector(
      "Amplitude of Anisotropic Effect", "di_anisotropicAmplitude", "di_aaVal",
      0, ModelVector.FLOAT);

  /**Standard deviation if seed distribution method is normal or lognormal*/
  private ModelVector mp_fStandardDeviation = new ModelVector(
      "Seed Dist. Std. Deviation (Normal or Lognormal)",
      "di_standardDeviation", "di_sdVal", 0, ModelVector.FLOAT);

  /**Clumping parameter if seed distribution is negative binomial*/
  private ModelVector mp_fClumpingParameter = new ModelVector(
      "Seed Dist. Clumping Parameter (Neg. Binomial)",
      "di_clumpingParameter", "di_cpVal", 0, ModelVector.FLOAT);

  /**Seed distribution*/
  private ModelEnum m_iSeedDistributionMethod =
      new ModelEnum(new int[] {0, 1, 2, 3, 4}
                    ,
                    new String[] {"Deterministic", "Poisson", "Lognormal",
                    "Normal", "Negative binomial"}
                    , "Seed Distribution", "di_seedDistributionMethod");

  private ModelFloat
      /**Maximum search radius, in meters, for neighbors for isotropic and
       * anisotropic disperse*/
      m_fMaxSearchRadius = new ModelFloat(0,
                                          "Maximum Search Distance for Neighbor Parents, in m",
                                          "di_maxSearchRadius");

  private ModelInt
      /**Max number of parent trees that can be in a grid cell for it to still be
       * marked as gap */
      m_iMaxGapDensity = new ModelInt(0,
                                      "Maximum Parent Trees Allowed in Gap Cell",
                                      "di_maxGapDensity");

  /**
   * Constructor
   * @param oManager GUIManager object.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public DisperseBehaviors(GUIManager oManager) {
    super(oManager, "Disperse", "disperse");

    m_sHelpID = "data.disperse_behaviors";

    int i, j;

    //Initialize all our arrays and prep our vectors

    mp_fSTR = new ModelVector[NUMBER_OF_DISPERSE_FUNCTIONS][];
    mp_fBeta = new ModelVector[NUMBER_OF_DISPERSE_FUNCTIONS][];
    mp_fThetaOrXb = new ModelVector[NUMBER_OF_DISPERSE_FUNCTIONS][];
    mp_fDispOrX0 = new ModelVector[NUMBER_OF_DISPERSE_FUNCTIONS][];

    for (i = 0; i < NUMBER_OF_DISPERSE_FUNCTIONS; i++) {
      mp_fSTR[i] = new ModelVector[NUMBER_OF_FOREST_COVERS];
      mp_fBeta[i] = new ModelVector[NUMBER_OF_FOREST_COVERS];
      mp_fThetaOrXb[i] = new ModelVector[NUMBER_OF_FOREST_COVERS];
      mp_fDispOrX0[i] = new ModelVector[NUMBER_OF_FOREST_COVERS];

      for (j = 0; j < NUMBER_OF_FOREST_COVERS; j++) {
        if (WEIBULL == i && GAP == j) {
          mp_fSTR[i][j] = new ModelVector("Weibull Gap Annual STR",
                                          "di_weibullGapSTR",
                                          "di_wgsVal", 0, ModelVector.FLOAT);
          mp_fBeta[i][j] = new ModelVector("Weibull Gap Beta",
                                           "di_weibullGapBeta", "di_wgbVal", 0,
                                           ModelVector.FLOAT);
          mp_fThetaOrXb[i][j] = new ModelVector("Weibull Gap Theta",
                                                "di_weibullGapTheta",
                                                "di_wgtVal", 0,
                                                ModelVector.FLOAT);
          mp_fDispOrX0[i][j] = new ModelVector("Weibull Gap Dispersal",
                                               "di_weibullGapDispersal",
                                               "di_wgdVal", 0,
                                               ModelVector.FLOAT);
        }
        else if (WEIBULL == i && CANOPY == j) {
          mp_fSTR[i][j] = new ModelVector("Weibull Canopy Annual STR",
                                          "di_weibullCanopySTR", "di_wcsVal", 0,
                                          ModelVector.FLOAT);
          mp_fBeta[i][j] = new ModelVector("Weibull Canopy Beta",
                                           "di_weibullCanopyBeta", "di_wcbVal",
                                           0, ModelVector.FLOAT);
          mp_fThetaOrXb[i][j] = new ModelVector("Weibull Canopy Theta",
                                                "di_weibullCanopyTheta",
                                                "di_wctVal", 0,
                                                ModelVector.FLOAT);
          mp_fDispOrX0[i][j] = new ModelVector("Weibull Canopy Dispersal",
                                               "di_weibullCanopyDispersal",
                                               "di_wcdVal", 0,
                                               ModelVector.FLOAT);
        }
        else if (LOGNORMAL == i && GAP == j) {
          mp_fSTR[i][j] = new ModelVector("Lognormal Gap Annual STR",
                                          "di_lognormalGapSTR", "di_lgsVal", 0,
                                          ModelVector.FLOAT);
          mp_fBeta[i][j] = new ModelVector("Lognormal Gap Beta",
                                           "di_lognormalGapBeta", "di_lgbVal",
                                           0, ModelVector.FLOAT);
          mp_fThetaOrXb[i][j] = new ModelVector("Lognormal Gap Xb",
                                                "di_lognormalGapXb",
                                                "di_lgxbVal", 0,
                                                ModelVector.FLOAT);
          mp_fDispOrX0[i][j] = new ModelVector("Lognormal Gap X0",
                                               "di_lognormalGapX0",
                                               "di_lgx0Val", 0,
                                               ModelVector.FLOAT);
        }
        else {
          mp_fSTR[i][j] = new ModelVector("Lognormal Canopy Annual STR",
                                          "di_lognormalCanopySTR", "di_lcsVal",
                                          0, ModelVector.FLOAT);
          mp_fBeta[i][j] = new ModelVector("Lognormal Canopy Beta",
                                           "di_lognormalCanopyBeta",
                                           "di_lcbVal", 0, ModelVector.FLOAT);
          mp_fThetaOrXb[i][j] = new ModelVector("Lognormal Canopy Xb",
                                                "di_lognormalCanopyXb",
                                                "di_lcxbVal", 0,
                                                ModelVector.FLOAT);
          mp_fDispOrX0[i][j] = new ModelVector("Lognormal Canopy X0",
                                               "di_lognormalCanopyX0",
                                               "di_lcx0Val", 0,
                                               ModelVector.FLOAT);
        }
      }
    }

    mp_iWhichFunctionUsed = new ModelVector[NUMBER_OF_FOREST_COVERS];

    for (i = 0; i < NUMBER_OF_FOREST_COVERS; i++) {
      if (CANOPY == i) {
        mp_iWhichFunctionUsed[i] = new ModelVector("Canopy Function Used",
            "di_canopyFunction", "di_cfVal", 0, ModelVector.MODEL_ENUM);
      }
      else {
        mp_iWhichFunctionUsed[i] = new ModelVector("Gap Function Used",
            "di_gapFunction", "di_gfVal", 0, ModelVector.MODEL_ENUM);
      }
    }

    mp_fMinDbhForReproduction = new ModelVector(
        "Minimum DBH for Reproduction, in cm", "di_minDbhForReproduction",
        "di_mdfrVal", 0, ModelVector.FLOAT);
    mp_fStumpSTR = new ModelVector("STR for Stumps", "di_suckerSTR", "di_ssVal",
                                   0, ModelVector.FLOAT);
    mp_fStumpBeta = new ModelVector("Beta for Stumps", "di_suckerBeta",
                                    "di_sbVal", 0, ModelVector.FLOAT);

    //Set up behavior list
    mp_oChildBehaviors = new Behavior[4];

    //Add a behavior for non spatial dispersal
    mp_oChildBehaviors[0] = new Behavior("non spatial disperse",
                                         "Non-spatial disperse",
                                         "non spatial disperse", (float) 1);
    mp_oChildBehaviors[0].AddRequiredData(mp_fMinDbhForReproduction);
    mp_oChildBehaviors[0].AddRequiredData(mp_fSlopeOfLambda);
    mp_oChildBehaviors[0].AddRequiredData(mp_fInterceptOfLambda);
    mp_oChildBehaviors[0].AddRequiredData(m_iSeedDistributionMethod);
    mp_oChildBehaviors[0].AddRequiredData(mp_fStandardDeviation);
    mp_oChildBehaviors[0].AddRequiredData(mp_fClumpingParameter);
    //Disallow for seedlings
    mp_oChildBehaviors[0].SetCanApplyTo(TreePopulation.SEEDLING, false);

    //Add a behavior for non-gap disperse
    mp_oChildBehaviors[1] = new Behavior("non-gap disperse",
                                         "Non-gap spatial disperse",
                                         "non-gap disperse",
                                         (float) 1);
    mp_oChildBehaviors[1].AddRequiredData(mp_fMinDbhForReproduction);
    mp_oChildBehaviors[1].AddRequiredData(mp_iWhichFunctionUsed[CANOPY]);
    for (i = 0; i < NUMBER_OF_DISPERSE_FUNCTIONS; i++) {
      mp_oChildBehaviors[1].AddRequiredData(mp_fSTR[i][CANOPY]);
      mp_oChildBehaviors[1].AddRequiredData(mp_fBeta[i][CANOPY]);
      mp_oChildBehaviors[1].AddRequiredData(mp_fThetaOrXb[i][CANOPY]);
      mp_oChildBehaviors[1].AddRequiredData(mp_fDispOrX0[i][CANOPY]);
    }
    mp_oChildBehaviors[1].AddRequiredData(mp_fStumpSTR);
    mp_oChildBehaviors[1].AddRequiredData(mp_fStumpBeta);
    mp_oChildBehaviors[1].AddRequiredData(m_iSeedDistributionMethod);
    mp_oChildBehaviors[1].AddRequiredData(mp_fStandardDeviation);
    mp_oChildBehaviors[1].AddRequiredData(mp_fClumpingParameter);
    //Disallow for seedlings
    mp_oChildBehaviors[1].SetCanApplyTo(TreePopulation.SEEDLING, false);
    //Allow for stumps
    mp_oChildBehaviors[1].SetCanApplyTo(TreePopulation.STUMP, true);

    //Add a behavior for gap disperse
    mp_oChildBehaviors[2] = new Behavior("gap disperse", "Gap spatial disperse",
                                         "gap disperse", (float) 1);
    //Disallow for seedlings
    mp_oChildBehaviors[2].SetCanApplyTo(TreePopulation.SEEDLING, false);
    //Allow for stumps
    mp_oChildBehaviors[2].SetCanApplyTo(TreePopulation.STUMP, true);

    mp_oChildBehaviors[2].AddRequiredData(mp_fMinDbhForReproduction);
    mp_oChildBehaviors[2].AddRequiredData(m_iMaxGapDensity);
    mp_oChildBehaviors[2].AddRequiredData(m_iSeedDistributionMethod);
    mp_oChildBehaviors[2].AddRequiredData(mp_fStandardDeviation);
    mp_oChildBehaviors[2].AddRequiredData(mp_fClumpingParameter);
    for (i = 0; i < NUMBER_OF_FOREST_COVERS; i++) {
      mp_oChildBehaviors[2].AddRequiredData(mp_iWhichFunctionUsed[i]);
    }
    for (i = 0; i < NUMBER_OF_DISPERSE_FUNCTIONS; i++) {
      for (j = 0; j < NUMBER_OF_FOREST_COVERS; j++) {
        mp_oChildBehaviors[2].AddRequiredData(mp_fSTR[i][j]);
        mp_oChildBehaviors[2].AddRequiredData(mp_fBeta[i][j]);
        mp_oChildBehaviors[2].AddRequiredData(mp_fThetaOrXb[i][j]);
        mp_oChildBehaviors[2].AddRequiredData(mp_fDispOrX0[i][j]);
      }
    }

    mp_oChildBehaviors[2].AddRequiredData(mp_fStumpSTR);
    mp_oChildBehaviors[2].AddRequiredData(mp_fStumpBeta);

    //Add a behavior for isotropic disperse
    /*    mp_oChildBehaviors[3] = new Behavior("isotropic disperse",
                                             "Isotropic spatial disperse",
                                             "isotropic disperse", 1);
        //Disallow for seedlings
        mp_oChildBehaviors[3].SetCanApplyTo(TreePopulation.SEEDLING, false);
        mp_oChildBehaviors[3].AddRequiredData(m_fMaxSearchRadius);
        mp_oChildBehaviors[3].AddRequiredData(mp_fMinDbhForReproduction);
        mp_oChildBehaviors[3].AddRequiredData(mp_iWhichFunctionUsed[CANOPY]);
        mp_oChildBehaviors[3].AddRequiredData(m_iSeedDistributionMethod);
        mp_oChildBehaviors[3].AddRequiredData(mp_fStandardDeviation);
        mp_oChildBehaviors[3].AddRequiredData(mp_fClumpingParameter);
        for (i = 0; i < NUMBER_OF_DISPERSE_FUNCTIONS; i++) {
          mp_oChildBehaviors[3].AddRequiredData(mp_fSTR[i][CANOPY]);
          mp_oChildBehaviors[3].AddRequiredData(mp_fBeta[i][CANOPY]);
          mp_oChildBehaviors[3].AddRequiredData(mp_fThetaOrXb[i][CANOPY]);
          mp_oChildBehaviors[3].AddRequiredData(mp_fDispOrX0[i][CANOPY]);
        } */

    //Add a behavior for anisotropic disperse
    mp_oChildBehaviors[3] = new Behavior("anisotropic disperse",
                                         "Anisotropic spatial disperse",
                                         "anisotropic disperse", (float) 1);
    //Disallow for seedlings
    mp_oChildBehaviors[3].SetCanApplyTo(TreePopulation.SEEDLING, false);
    mp_oChildBehaviors[3].AddRequiredData(m_fMaxSearchRadius);
    mp_oChildBehaviors[3].AddRequiredData(mp_fMinDbhForReproduction);
    mp_oChildBehaviors[3].AddRequiredData(mp_iWhichFunctionUsed[CANOPY]);
    mp_oChildBehaviors[3].AddRequiredData(m_iSeedDistributionMethod);
    mp_oChildBehaviors[3].AddRequiredData(mp_fStandardDeviation);
    mp_oChildBehaviors[3].AddRequiredData(mp_fClumpingParameter);
    for (i = 0; i < NUMBER_OF_DISPERSE_FUNCTIONS; i++) {
      mp_oChildBehaviors[3].AddRequiredData(mp_fSTR[i][CANOPY]);
      mp_oChildBehaviors[3].AddRequiredData(mp_fBeta[i][CANOPY]);
      mp_oChildBehaviors[3].AddRequiredData(mp_fThetaOrXb[i][CANOPY]);
      mp_oChildBehaviors[3].AddRequiredData(mp_fDispOrX0[i][CANOPY]);
    }
    mp_oChildBehaviors[3].AddRequiredData(mp_fDirectionMaxDispersal);
    mp_oChildBehaviors[3].AddRequiredData(mp_fAnisotropicAmplitude);

    //All data - in the order it will be written in XML
    mp_oAllData.add(mp_fMinDbhForReproduction);
    //Weibull canopy parameters
    mp_oAllData.add(mp_fSTR[WEIBULL][CANOPY]);
    mp_oAllData.add(mp_fBeta[WEIBULL][CANOPY]);
    mp_oAllData.add(mp_fDispOrX0[WEIBULL][CANOPY]);
    mp_oAllData.add(mp_fThetaOrXb[WEIBULL][CANOPY]);
    //Weibull gap parameters
    mp_oAllData.add(mp_fSTR[WEIBULL][GAP]);
    mp_oAllData.add(mp_fBeta[WEIBULL][GAP]);
    mp_oAllData.add(mp_fDispOrX0[WEIBULL][GAP]);
    mp_oAllData.add(mp_fThetaOrXb[WEIBULL][GAP]);
    //Lognormal canopy parameters
    mp_oAllData.add(mp_fSTR[LOGNORMAL][CANOPY]);
    mp_oAllData.add(mp_fBeta[LOGNORMAL][CANOPY]);
    mp_oAllData.add(mp_fDispOrX0[LOGNORMAL][CANOPY]);
    mp_oAllData.add(mp_fThetaOrXb[LOGNORMAL][CANOPY]);
    //Lognormal gap parameters
    mp_oAllData.add(mp_fSTR[LOGNORMAL][GAP]);
    mp_oAllData.add(mp_fBeta[LOGNORMAL][GAP]);
    mp_oAllData.add(mp_fDispOrX0[LOGNORMAL][GAP]);
    mp_oAllData.add(mp_fThetaOrXb[LOGNORMAL][GAP]);
    //Stump parameters
    mp_oAllData.add(mp_fStumpSTR);
    mp_oAllData.add(mp_fStumpBeta);

    mp_oAllData.add(m_iMaxGapDensity);

    mp_oAllData.add(mp_iWhichFunctionUsed[GAP]);
    mp_oAllData.add(mp_iWhichFunctionUsed[CANOPY]);
    //Bath recruitment parameters
    mp_oAllData.add(mp_fSlopeOfLambda);
    mp_oAllData.add(mp_fInterceptOfLambda);
    //Anisotropic disperse parameters
    mp_oAllData.add(m_fMaxSearchRadius);
    mp_oAllData.add(mp_fDirectionMaxDispersal);
    mp_oAllData.add(mp_fAnisotropicAmplitude);
    //Seed distribution parameters
    mp_oAllData.add(m_iSeedDistributionMethod);
    mp_oAllData.add(mp_fStandardDeviation);
    mp_oAllData.add(mp_fClumpingParameter);
  }

  /**
   * Does setup.  Sets up the substrate favorability grid.
   * @param oPop TreePopulation object.
   * @throws ModelException if there's a problem setting behavior use data.
   * @todo Something for stump removal?
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void DoSetup(TreePopulation oPop) throws ModelException {
    int iNumSpecies = oPop.GetNumberOfSpecies(), i, j, k;

    //Set up the "which function" vector
    for (i = 0; i < NUMBER_OF_FOREST_COVERS; i++) {
      mp_iWhichFunctionUsed[i].GetValue().clear();
      for (j = 0; j < iNumSpecies; j++) {
        mp_iWhichFunctionUsed[i].GetValue().add(new ModelEnum(new int[] {
            WEIBULL, LOGNORMAL}
            ,
            new String[] {"Weibull", "Lognormal"}
            , "Function used", ""));
      }
    }

    //Set up the seeds grid
    DataMember[] p_oDataMembers = new DataMember[iNumSpecies + 2];
    mp_oGrids = new Grid[1];

    String sGridName = "Dispersed Seeds";

    //The accessible data members are a seed count for each species and
    //gap status
    for (i = 0; i < iNumSpecies; i++) {

      p_oDataMembers[i] = new DataMember("Number of seeds for " +
                                         oPop.GetSpeciesNameFromCode(i).
                                         replace('_', ' '),
                                         "seeds_" + i,
                                         DataMember.FLOAT);
    }

    p_oDataMembers[iNumSpecies] = new DataMember("Gap status", "Is Gap",
                                                 DataMember.BOOLEAN);

    p_oDataMembers[iNumSpecies +
        1] = new DataMember("Adult tree count", "count",
                            DataMember.INTEGER);

    //Create our grid
    mp_oGrids[0] = new Grid(sGridName, p_oDataMembers, null, 8, 8);

    //Assign the grid to all behaviors
    for (i = 0; i < mp_oChildBehaviors.length; i++) {
      mp_oChildBehaviors[i].AddGrid(mp_oGrids[0]);
    }

    //Initialize our arrays with zeroes if they haven't been set
    for (i = 0; i < NUMBER_OF_DISPERSE_FUNCTIONS; i++) {
      for (j = 0; j < NUMBER_OF_FOREST_COVERS; j++) {
        if (mp_fSTR[i][j].GetValue().size() == 0) {
          for (k = 0; k < iNumSpecies; k++) {
            mp_fSTR[i][j].GetValue().add(new Float(0));
          }
        }
        if (mp_fBeta[i][j].GetValue().size() == 0) {
          for (k = 0; k < iNumSpecies; k++) {
            mp_fBeta[i][j].GetValue().add(new Float(0));
          }
        }
        if (mp_fThetaOrXb[i][j].GetValue().size() == 0) {
          for (k = 0; k < iNumSpecies; k++) {
            mp_fThetaOrXb[i][j].GetValue().add(new Float(0));
          }
        }
        if (mp_fDispOrX0[i][j].GetValue().size() == 0) {
          for (k = 0; k < iNumSpecies; k++) {
            mp_fDispOrX0[i][j].GetValue().add(new Float(0));
          }
        }
      }
    }
    if (mp_fStumpSTR.GetValue().size() == 0) {
      for (k = 0; k < iNumSpecies; k++) {
        mp_fStumpSTR.GetValue().add(new Float(0));
      }
    }
    if (mp_fStumpBeta.GetValue().size() == 0) {
      for (k = 0; k < iNumSpecies; k++) {
        mp_fStumpBeta.GetValue().add(new Float(0));
      }
    }

  }

  /**
   * Sets the value for STR for suckering stumps.
   * @param p_fVals One stump STR for each species.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void SetStumpSTR(Float[] p_fVals) {
    SetVectorValues(mp_fStumpSTR, p_fVals);
  }

  /**
   * Sets the value for beta for suckering stumps.
   * @param p_fVals One stump beta for each species.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void SetStumpBeta(Float[] p_fVals) {
    SetVectorValues(mp_fStumpBeta, p_fVals);
  }

  /**
   * Sets the minimum dbh for reproduction for all species.
   * @param p_fVals Minimum dbh for reproduction, one for each species.
   * @throws ModelException if any value is less than 0.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void SetMinimumDbhForReproduction(Float[] p_fVals) throws
      ModelException {
    MakeSureAllPositive(p_fVals, mp_fMinDbhForReproduction.GetDescriptor());
    SetVectorValues(mp_fMinDbhForReproduction, p_fVals);
  }

  /**
   * Sets the slope of lambda values for all species.
   * @param p_fVals Slope of lambda values.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void SetSlopeOfLambda(Float[] p_fVals) {
    SetVectorValues(mp_fSlopeOfLambda, p_fVals);
  }

  /**
   * Sets the intercept of lambda values for all species.
   * @param p_fVals Intercept of lambda values.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void SetInterceptOfLambda(Float[] p_fVals) {
    SetVectorValues(mp_fInterceptOfLambda, p_fVals);
  }

  /**
   * Sets the maximum gap density.
   * @param i Value to set.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void SetMaximumGapDensity(int i) {
    m_iMaxGapDensity.SetValue(i);
  }

  /**
   * Sets the seed distribution method.
   * @param i Seed distribution method.
   * @throws ModelException if the value is invalid.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void SetSeedDistribution(int i) throws ModelException {
    m_iSeedDistributionMethod.SetValue(i);
  }

  /**
   * Sets the STR values for all species for a particular disperse function
   * and forest cover.
   * @param iDispFunc Disperse function number.
   * @param iForestCov Forest cover number.
   * @param p_fVals Array of Floats containing the values.
   * @throws ModelException if iDispFunc or iForestCov are not valid.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void SetSTR(int iDispFunc, int iForestCov, Float[] p_fVals) throws
      ModelException {
    if (iDispFunc < 0 || iDispFunc >= NUMBER_OF_DISPERSE_FUNCTIONS ||
        iForestCov < 0 || iForestCov >= NUMBER_OF_FOREST_COVERS) {
      ModelException exp = new ModelException();
      exp.iErrorCode = ErrorGUI.ILLEGAL_OP;
      exp.sMessage = "Invalid index passed to STR disperse setting function";
      exp.sCallingFunction = "JAVA";
      throw (exp);
    }
    SetVectorValues(mp_fSTR[iDispFunc][iForestCov], p_fVals);
  }

  /**
   * Sets the Dispersal (if for the Weibull function) or X0 (if for the
   * lognormal function) values for all species for a particular disperse
   * function and forest cover.
   * @param iDispFunc Disperse function number.
   * @param iForestCov Forest cover number.
   * @param p_fVals Array of Floats containing the values.
   * @throws ModelException if iDispFunc or iForestCov are not valid.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void SetDispersalOrX0(int iDispFunc, int iForestCov, Float[] p_fVals) throws
      ModelException {
    if (iDispFunc < 0 || iDispFunc >= NUMBER_OF_DISPERSE_FUNCTIONS ||
        iForestCov < 0 || iForestCov >= NUMBER_OF_FOREST_COVERS) {
      ModelException exp = new ModelException();
      exp.iErrorCode = ErrorGUI.ILLEGAL_OP;
      exp.sMessage =
          "Invalid index passed to Dispersal/X0 disperse setting function";
      exp.sCallingFunction = "JAVA";
      throw (exp);
    }
    SetVectorValues(mp_fDispOrX0[iDispFunc][iForestCov], p_fVals);
  }

  /**
   * Sets the Theta (if for the Weibull function) or Xb (if for the
   * lognormal function) values for all species for a particular disperse
   * function and forest cover.
   * @param iDispFunc Disperse function number.
   * @param iForestCov Forest cover number.
   * @param p_fVals Array of Floats containing the STR values.
   * @throws ModelException if iDispFunc or iForestCov are not valid.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void SetThetaOrXb(int iDispFunc, int iForestCov, Float[] p_fVals) throws
      ModelException {
    if (iDispFunc < 0 || iDispFunc >= NUMBER_OF_DISPERSE_FUNCTIONS ||
        iForestCov < 0 || iForestCov >= NUMBER_OF_FOREST_COVERS) {
      ModelException exp = new ModelException();
      exp.iErrorCode = ErrorGUI.ILLEGAL_OP;
      exp.sMessage =
          "Invalid index passed to Theta/Xb disperse setting function";
      exp.sCallingFunction = "JAVA";
      throw (exp);
    }
    SetVectorValues(mp_fThetaOrXb[iDispFunc][iForestCov], p_fVals);
  }

  /**
   * Sets the beta values for all species for a particular disperse function
   * and forest cover.
   * @param iDispFunc Disperse function number.
   * @param iForestCov Forest cover number.
   * @param p_fVals Array of Floats containing the values.
   * @throws ModelException if iDispFunc or iForestCov are not valid.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void SetBeta(int iDispFunc, int iForestCov, Float[] p_fVals) throws
      ModelException {
    if (iDispFunc < 0 || iDispFunc >= NUMBER_OF_DISPERSE_FUNCTIONS ||
        iForestCov < 0 || iForestCov >= NUMBER_OF_FOREST_COVERS) {
      ModelException exp = new ModelException();
      exp.iErrorCode = ErrorGUI.ILLEGAL_OP;
      exp.sMessage = "Invalid index passed to beta disperse setting function";
      exp.sCallingFunction = "JAVA";
      throw (exp);
    }
    SetVectorValues(mp_fBeta[iDispFunc][iForestCov], p_fVals);
  }

  /**
   * Sets the disperse function values for all species for a particular
   * forest cover.
   * @param iForestCov Forest cover number.
   * @param p_iVals Function flags for each species - valid values are WEIBULL
   *  and LOGNORMAL.
   * @throws ModelException if the forest cover isn't valid or if the function
   *  values are invalid.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void SetDisperseFunctions(int iForestCov, Integer[] p_iVals) throws
      ModelException {
    if (iForestCov < 0 || iForestCov >= NUMBER_OF_FOREST_COVERS) {
      ModelException exp = new ModelException();
      exp.iErrorCode = ErrorGUI.ILLEGAL_OP;
      exp.sMessage =
          "Invalid forest cover passed to disperse function setter.";
      exp.sCallingFunction = "JAVA";
      throw (exp);
    }
    ModelEnum oData;
    int i;

    for (i = 0; i < p_iVals.length; i++) {
      oData = (ModelEnum) mp_iWhichFunctionUsed[iForestCov].GetValue().
          elementAt(i);
      oData.SetValue(p_iVals[i].intValue());
    }
  }

  /**
   * Validates the data in preparation for parameter file writing or some such.
   * @param oPop TreePopulation object.
   * @throws ModelException if:
   * <ul>
   * <li>Any value in mp_fMinDbhForReproduction is 0</li>
   * <li>There aren't values for all species for any disperse function used</li>
   * <li>Max search radius for anisotropic disperse is not a positive number</li>
   * and less than half the shortest plot distance</li>
   * <li>The direction of max dispersal is not between 0 and 2PI</li>
   * <li>Spatial disperse is used and a value for weibull theta is greater than
   * or equal to 50</li>
   * <li>Spatial disperse is used and a value for beta is greater than 25</li>
   * </ul>
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void ValidateData(TreePopulation oPop) throws ModelException {
    Behavior oNoGap = GetBehaviorByKey("non-gap disperse"),
        oGap = GetBehaviorByKey("gap disperse");
    int iNumSpecies = oPop.GetNumberOfSpecies(),
        i, j, k;

    //If there are no behaviors enabled, exit
    if (!AnyBehaviorsEnabled()) {
      return;
    }

    boolean[] p_bApplies = null;
    for (i = 0; i < mp_oChildBehaviors.length; i++) {
      if (mp_oChildBehaviors[i].m_bIsEnabled) {
        if (p_bApplies == null) {
          p_bApplies = mp_oChildBehaviors[i].GetWhichSpeciesUsed(oPop);
        }
        else {
          p_bApplies = MergeBooleans(p_bApplies,
                                     mp_oChildBehaviors[i].
                                     GetWhichSpeciesUsed(oPop));
        }
      }
    }
    MakeSureAllPositive(mp_fMinDbhForReproduction, p_bApplies);

    if (oGap.m_bIsEnabled) {
      for (i = 0; i < NUMBER_OF_DISPERSE_FUNCTIONS; i++) {
        for (j = 0; j < NUMBER_OF_FOREST_COVERS; j++) {
          MakeSureRightSize(mp_fSTR[i][j], iNumSpecies);
          MakeSureRightSize(mp_fBeta[i][j], iNumSpecies);
          MakeSureRightSize(mp_fThetaOrXb[i][j], iNumSpecies);
          MakeSureRightSize(mp_fDispOrX0[i][j], iNumSpecies);
        }
      }
    }
    else if (oNoGap.m_bIsEnabled) {
      for (i = 0; i < NUMBER_OF_DISPERSE_FUNCTIONS; i++) {
        MakeSureRightSize(mp_fSTR[i][CANOPY], iNumSpecies);
        MakeSureRightSize(mp_fBeta[i][CANOPY], iNumSpecies);
        MakeSureRightSize(mp_fThetaOrXb[i][CANOPY], iNumSpecies);
        MakeSureRightSize(mp_fDispOrX0[i][CANOPY], iNumSpecies);
      }
    }

    //Make sure all values for weibull theta are less than 50
    if (oGap.m_bIsEnabled) {
      p_bApplies = oGap.GetWhichSpeciesUsed(oPop);
      for (i = 0; i < NUMBER_OF_DISPERSE_FUNCTIONS; i++) {
        if (i == WEIBULL) {
          for (j = 0; j < NUMBER_OF_FOREST_COVERS; j++) {
            for (k = 0; k < mp_fThetaOrXb[i][j].GetValue().size(); k++) {
              if (p_bApplies[k]) {
                float fNumber = ( (Float) mp_fThetaOrXb[i][j].GetValue().
                                 elementAt(k)).floatValue();
                if (fNumber >= 50) {
                  throw (new ModelException(ErrorGUI.BAD_DATA, "JAVA",
                                            "The values in " +
                                            mp_fThetaOrXb[i][j].GetDescriptor() +
                                            " must be less than 50 to avoid math errors."));
                }
              }
            }
          }
        }
      }
    }
    else if (oNoGap.m_bIsEnabled) {
      p_bApplies = oNoGap.GetWhichSpeciesUsed(oPop);
      for (i = 0; i < NUMBER_OF_DISPERSE_FUNCTIONS; i++) {
        if (i == WEIBULL) {
          for (k = 0; k < mp_fThetaOrXb[i][CANOPY].GetValue().size(); k++) {
            if (p_bApplies[k]) {
              float fNumber = ( (Float) mp_fThetaOrXb[i][CANOPY].GetValue().
                               elementAt(k)).floatValue();
              if (fNumber >= 50) {
                throw (new ModelException(ErrorGUI.BAD_DATA, "JAVA",
                                          "The values in " +
                                          mp_fThetaOrXb[i][CANOPY].
                                          GetDescriptor() +
                                          " must be less than 50 to avoid math errors."));
              }
            }
          }
        }
      }
    }

    //Make sure all values for beta are less than 25
    if (oGap.m_bIsEnabled || oNoGap.m_bIsEnabled) {
      p_bApplies = oGap.GetWhichSpeciesUsed(oPop);
      p_bApplies = MergeBooleans(p_bApplies, oNoGap.GetWhichSpeciesUsed(oPop));
      for (i = 0; i < NUMBER_OF_DISPERSE_FUNCTIONS; i++) {
        for (k = 0; k < mp_fBeta[i][CANOPY].GetValue().size(); k++) {
          if (p_bApplies[k]) {
            float fNumber = ( (Float) mp_fBeta[i][CANOPY].GetValue().
                             elementAt(k)).floatValue();
            if (fNumber > 25) {
              throw (new ModelException(ErrorGUI.BAD_DATA, "JAVA",
                                        "The values in " +
                                        mp_fBeta[i][CANOPY].
                                        GetDescriptor() +
                                        " must be less than 25 to avoid math errors."));
            }
          }
        }
      }
    }
    if (oGap.m_bIsEnabled) {
      p_bApplies = oGap.GetWhichSpeciesUsed(oPop);
      for (i = 0; i < NUMBER_OF_DISPERSE_FUNCTIONS; i++) {
        for (k = 0; k < mp_fBeta[i][GAP].GetValue().size(); k++) {
          if (p_bApplies[k]) {
            float fNumber = ( (Float) mp_fBeta[i][GAP].GetValue().
                             elementAt(k)).floatValue();
            if (fNumber > 25) {
              throw (new ModelException(ErrorGUI.BAD_DATA, "JAVA",
                                        "The values in " +
                                        mp_fBeta[i][GAP].
                                        GetDescriptor() +
                                        " must be less than 25 to avoid math errors."));
            }
          }
        }
      }
    }

    if ( //GetBehaviorByKey("isotropic disperse").m_bIsEnabled ||
        GetBehaviorByKey("anisotropic disperse").m_bIsEnabled) {
      MakeSurePositive(m_fMaxSearchRadius);
      //Find out what half the shortest plot length is
      Plot oPlot = m_oManager.GetPlot();
      float fMax = java.lang.Math.min(oPlot.GetPlotXLength(),
                                      oPlot.GetPlotYLength()) / 2;
      if (m_fMaxSearchRadius.GetValue() > fMax) {
        ModelException oErr = new ModelException();
        oErr.iErrorCode = ErrorGUI.BAD_DATA;
        oErr.sMessage = m_sName + " - " + m_fMaxSearchRadius.GetDescriptor() +
            " must not be greater than half the shortest plot length.";
        oErr.sCallingFunction = "JAVA";
        throw (oErr);
      }
    }

    if (GetBehaviorByKey("anisotropic disperse").m_bIsEnabled) {
      for (i = 0; i < mp_fDirectionMaxDispersal.GetValue().size(); i++) {
        if (mp_fDirectionMaxDispersal.GetValue().elementAt(i) != null) {
          Float fAngle = (Float) mp_fDirectionMaxDispersal.GetValue().elementAt(
              i);
          if (fAngle.floatValue() < 0 ||
              fAngle.floatValue() >= (2 * java.lang.Math.PI)) {
            ModelException oErr = new ModelException();
            oErr.iErrorCode = ErrorGUI.BAD_DATA;
            oErr.sMessage = m_sName + " - " +
                mp_fDirectionMaxDispersal.GetDescriptor() +
                " values must be between 0 and 2PI.";
            oErr.sCallingFunction = "JAVA";
            throw (oErr);
          }
        }
      }
    }
  }
}
