package javawrapper;

import javawrapper.*;
import java.io.*;
import java.util.Vector;
import org.xml.sax.Attributes;

/**
 * This class manages data for disturbance behaviors.  Planting is not included.
 * <p>Copyright: Copyright (c) Charles D. Canham 2003</p>
 * <p>Company: Institute of Ecosystem Studies</p>
 * @author Lora E. Murphy
 * @version 1.0
 *
 * <br>Edit history:
 * <br>------------------
 * <br>April 28, 2004: Submitted in beta version (LEM)
 * <br>May 4, 2004:  Extended this to interact with the HarvestWindow
 * class (LEM)
 * <br>May 12, 2004:  Added support for parsing harvest data from a parameter
 * file (LEM)
 * <br>May 13, 2004:  Added the disease behavior (LEM)
 * <br>September 9, 2004:  Added storm behavior (LEM)
 * <br>September 29, 2004: Added storm damage applier behavior (LEM)
 * <br>December 7, 2004:  Changed natural disturbance to episodic mortality (LEM)
 * <br>April 4, 2005:  Added more parameters to storm (LEM)
 * <br>May 2, 2005:  Added more parameters to storm (LEM)
 * <br>May 12, 2005:  Added selection harvest (LEM)
 */

public class DisturbanceBehaviors
    extends BehaviorTypeBase {

  /**Partial cut.  Matches the old code's enum value - that's important!.*/
  public static final int PARTIAL_CUT = 1;

  /**Gap cut.  Matches the old code's enum value - that's important!.*/
  public static final int GAP_CUT = 2;

  /**Clear cut.  Matches the old code's enum value - that's important!.*/
  public static final int CLEAR_CUT = 3;

  /**Percentage basal area.  Matches the old code's enum value - that's important!.*/
  public static final int PERCENTAGE_BASAL_AREA = 2;

  /**Absolute basal area.  Matches the old code's enum value - that's important!.*/
  public static final int ABSOLUTE_BASAL_AREA = 3;

  /**Percentage density.  Matches the old code's enum value - that's important!.*/
  public static final int PERCENTAGE_DENSITY = 0;

  /**Absolute density.  Matches the old code's enum value - that's important!.*/
  public static final int ABSOLUTE_DENSITY = 1;

  /**Our list of harvest cuts, held as HarvestData objects*/
  protected Vector mp_oHarvestCuts = new Vector(0);

  /**List of episodic mortality events, held as HarvestData objects (ignoring the cut type)*/
  protected Vector mp_oMortEpisodes = new Vector(0);

  /**Minimum DBH to which storm damage applies for each species*/
  protected ModelVector mp_fMinDBH = new ModelVector(
      "Minimum DBH for Storm Damage, in cm",
      "st_minDBH", "st_mdVal", 0, ModelVector.FLOAT);

  /**Storm damage intercept for medium damage (a) for each species*/
  protected ModelVector mp_fStmDmgInterceptMed = new ModelVector(
      "Storm Damage Intercept (a) for Medium Damage",
      "st_stmDmgInterceptMed", "st_sdimVal", 0, ModelVector.FLOAT);

  /**Storm damage intercept for complete damage (a) for each species*/
  protected ModelVector mp_fStmDmgInterceptFull = new ModelVector(
      "Storm Damage Intercept (a) for Heavy Damage",
      "st_stmDmgInterceptFull", "st_sdifVal", 0, ModelVector.FLOAT);

  /**Storm intensity coefficient (b) for each species*/
  protected ModelVector mp_fStmIntensityCoeff = new ModelVector(
      "Storm Intensity Coefficient (b)",
      "st_stmIntensityCoeff", "st_sicVal", 0, ModelVector.FLOAT);

  /**Storm DBH coefficient (d) for each species*/
  protected ModelVector mp_fStmDBHCoeff = new ModelVector(
      "Storm DBH Coefficient (d)",
      "st_stmDBHCoeff", "st_sdcVal", 0, ModelVector.FLOAT);

  /**Storm - proportion of trees with medium damage that are immediately
   * killed - 04/04/05 LEM*/
  protected ModelVector mp_fStmPropMedDmgDie = new ModelVector(
      "Storm - Proportion of Trees with Medium Damage that Die",
      "st_stmPropDieMedDmg", "st_spdmdVal", 0, ModelVector.FLOAT);

  /**Storm - proportion of trees with total damage that are immediately
   * killed - 04/04/05 LEM*/
  protected ModelVector mp_fStmPropFullDmgDie = new ModelVector(
      "Storm - Proportion of Trees with Heavy Damage that Die",
      "st_stmPropDieFullDmg", "st_spdfdVal", 0, ModelVector.FLOAT);

  /**Storm - proportion of trees with total damage that die that tip up -
   * 05/01/05 LEM*/
  protected ModelVector mp_fStmPropFullTipUp = new ModelVector(
      "Storm - Prop. Heavy Damage Dead Trees that Tip Up",
      "st_stmPropTipUpFullDmg", "st_sptufdVal", 0, ModelVector.FLOAT);

  /**Storm return interval for 0-0.1 severity storm*/
  protected ModelFloat m_fClass1RtrnInt = new ModelFloat(0,
      "Return Interval for Severity Storm Class 0 - 0.1", "st_s1ReturnInterval");

  /**Storm return interval for 0.1-0.2 severity storm*/
  protected ModelFloat m_fClass2RtrnInt = new ModelFloat(0,
      "Return Interval for Severity Storm Class 0.1 - 0.2",
      "st_s2ReturnInterval");

  /**Storm return interval for 0.2-0.3 severity storm*/
  protected ModelFloat m_fClass3RtrnInt = new ModelFloat(0,
      "Return Interval for Severity Storm Class 0.2 - 0.3",
      "st_s3ReturnInterval");

  /**Storm return interval for 0.3-0.4 severity storm*/
  protected ModelFloat m_fClass4RtrnInt = new ModelFloat(0,
      "Return Interval for Severity Storm Class 0.3 - 0.4",
      "st_s4ReturnInterval");

  /**Storm return interval for 0.4-0.5 severity storm*/
  protected ModelFloat m_fClass5RtrnInt = new ModelFloat(0,
      "Return Interval for Severity Storm Class 0.4 - 0.5",
      "st_s5ReturnInterval");

  /**Storm return interval for 0.5-0.6 severity storm*/
  protected ModelFloat m_fClass6RtrnInt = new ModelFloat(0,
      "Return Interval for Severity Storm Class 0.5 - 0.6",
      "st_s6ReturnInterval");

  /**Storm return interval for 0.6-0.7 severity storm*/
  protected ModelFloat m_fClass7RtrnInt = new ModelFloat(0,
      "Return Interval for Severity Storm Class 0.6 - 0.7",
      "st_s7ReturnInterval");

  /**Storm return interval for 0.7-0.8 severity storm*/
  protected ModelFloat m_fClass8RtrnInt = new ModelFloat(0,
      "Return Interval for Severity Storm Class 0.7 - 0.8",
      "st_s8ReturnInterval");

  /**Storm return interval for 0.8-0.9 severity storm*/
  protected ModelFloat m_fClass9RtrnInt = new ModelFloat(0,
      "Return Interval for Severity Storm Class 0.8 - 0.9",
      "st_s9ReturnInterval");

  /**Storm return interval for 0.9-1.0 severity storm*/
  protected ModelFloat m_fClass10RtrnInt = new ModelFloat(0,
      "Return Interval for Severity Storm Class 0.9 - 1.0",
      "st_s10ReturnInterval");

  /**Standard deviation for normal and lognormal stochastic storm damage functions*/
  protected ModelFloat m_fStdDev = new ModelFloat(0,
                                                  "Standard Deviation (lognormal or normal)",
                                                  "st_standardDeviation");

  /**Clumping parameter for negative binomial stochastic storm damage functions*/
  protected ModelFloat m_fClump = new ModelFloat(0,
                                                 "Clumping Parameter (neg. binomial)",
                                                 "st_clumpingParameter");

  /**Selection harvest cut range 1 lower DBH - LEM 05-12-05*/
  protected ModelFloat m_fCut1LoDBH = new ModelFloat(0,
      "Selection Harvest Cut Range 1 Lower DBH (cm)", "sha_loDBH");

  /**Selection harvest cut range 1 upper DBH - LEM 05-12-05*/
  protected ModelFloat m_fCut1HiDBH = new ModelFloat(0,
      "Selection Harvest Cut Range 1 Upper DBH (cm)", "sha_hiDBH");

  /**Selection harvest cut range 1 target BA - LEM 05-12-05*/
  protected ModelFloat m_fCut1BA = new ModelFloat(0,
      "Selection Harvest Cut Range 1 Target Basal Area (m2/ha)",
                                                  "sha_target_BA");

  /**Selection harvest cut range 2 lower DBH - LEM 05-12-05*/
  protected ModelFloat m_fCut2LoDBH = new ModelFloat(0,
      "Selection Harvest Cut Range 2 Lower DBH (cm)", "sha_loDBH");

  /**Selection harvest cut range 2 upper DBH - LEM 05-12-05*/
  protected ModelFloat m_fCut2HiDBH = new ModelFloat(0,
      "Selection Harvest Cut Range 2 Upper DBH (cm)", "sha_hiDBH");

  /**Selection harvest cut range 2 target BA - LEM 05-12-05*/
  protected ModelFloat m_fCut2BA = new ModelFloat(0,
      "Selection Harvest Cut Range 2 Target Basal Area (m2/ha)",
                                                  "sha_target_BA");

  /**Selection harvest cut range 3 lower DBH - LEM 05-12-05*/
  protected ModelFloat m_fCut3LoDBH = new ModelFloat(0,
      "Selection Harvest Cut Range 3 Lower DBH (cm)", "sha_loDBH");

  /**Selection harvest cut range 3 upper DBH - LEM 05-12-05*/
  protected ModelFloat m_fCut3HiDBH = new ModelFloat(0,
      "Selection Harvest Cut Range 3 Upper DBH (cm)", "sha_hiDBH");

  /**Selection harvest cut range 3 target BA - LEM 05-12-05*/
  protected ModelFloat m_fCut3BA = new ModelFloat(0,
      "Selection Harvest Cut Range 3 Target Basal Area (m2/ha)",
                                                  "sha_target_BA");

  /**Selection harvest cut range 4 lower DBH - LEM 05-12-05*/
  protected ModelFloat m_fCut4LoDBH = new ModelFloat(0,
      "Selection Harvest Cut Range 4 Lower DBH (cm)", "sha_loDBH");

  /**Selection harvest cut range 4 upper DBH - LEM 05-12-05*/
  protected ModelFloat m_fCut4HiDBH = new ModelFloat(0,
      "Selection Harvest Cut Range 4 Upper DBH (cm)", "sha_hiDBH");

  /**Selection harvest cut range 4 target BA - LEM 05-12-05*/
  protected ModelFloat m_fCut4BA = new ModelFloat(0,
      "Selection Harvest Cut Range 4 Target Basal Area (m2/ha)",
                                                  "sha_target_BA");

  /**Selection harvest initial age - LEM 05-12-05*/
  protected ModelInt m_iInitialAge = new ModelInt(0,
      "Selection Harvest Initial Age", "sha_InitialAge");

  /**Number of years damaged trees stay damaged*/
  protected ModelInt m_iNumYearsToHeal = new ModelInt(0,
      "Number of Years Damaged Trees Take to Heal",
      "st_numYearsToHeal");

  /**Number of years storm-damaged snags last before being removed -
   * 05/01/05 LEM*/
  protected ModelInt m_iNumSnagYears = new ModelInt(0,
      "Number of Years Storm-Damaged Snags Last",
      "st_numYearsStormSnags");

  /**Storm susceptibility pattern*/
  protected ModelEnum m_iSusceptibility =
      new ModelEnum(new int[] {0, 1}
                    ,
                    new String[] {"Mapped", "Uniform"}
                    , "Plot Storm Susceptibility Pattern", "st_susceptibility");

  /**Storm stochasticity pattern*/
  protected ModelEnum m_iStochasticity =
      new ModelEnum(new int[] {0, 1}
                    ,
                    new String[] {"Deterministic", "Stochastic"}
                    , "Storm Damage Application", "st_stochasticity");

  /**Probability distribution function*/
  protected ModelEnum m_iProbDistFunc =
      new ModelEnum(new int[] {0, 1, 2}
                    ,
                    new String[] {"Lognormal", "Normal", "Negative Binomial"}
                    , "Stochastic Pattern Damage Distribution",
                    "st_probFunction");

  /**Placeholder for reading cut ranges from an XML file.  This one is
   * for DBH range lower bound*/
  protected float m_iCutRangeLowerBound = -1;

  /**Placeholder for reading cut ranges from an XML file.  This one is
   * for DBH range upper bound*/
  protected float m_iCutRangeUpperBound = -1;

  /**Placeholder for reading cut ranges from an XML file.  This one is
   * for range cut amount*/
  protected float m_iCutRangeAmount = -1;

  /**Placeholder for reading selection harvest cut ranges from an XML file.*/
  protected int m_iSelectionHarvestCutRange = 0;

  /**
   * Constructor.
   * @param oManager GUIManager object.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public DisturbanceBehaviors(GUIManager oManager) {
    super(oManager, "Disturbance", "");

    //Set up disturbance behaviors
    mp_oChildBehaviors = new Behavior[5];

    mp_oChildBehaviors[0] = new Behavior("Selection Harvest",
                                         "Selection Harvest",
                                         "SelectionHarvest",
                                         (float) 1.0);
    mp_oChildBehaviors[0].m_bMustHaveTrees = false;
    mp_oChildBehaviors[0].AddRequiredData(m_iInitialAge);
    mp_oChildBehaviors[0].AddRequiredData(m_fCut1LoDBH);
    mp_oChildBehaviors[0].AddRequiredData(m_fCut1HiDBH);
    mp_oChildBehaviors[0].AddRequiredData(m_fCut1BA);
    mp_oChildBehaviors[0].AddRequiredData(m_fCut2LoDBH);
    mp_oChildBehaviors[0].AddRequiredData(m_fCut2HiDBH);
    mp_oChildBehaviors[0].AddRequiredData(m_fCut2BA);
    mp_oChildBehaviors[0].AddRequiredData(m_fCut3LoDBH);
    mp_oChildBehaviors[0].AddRequiredData(m_fCut3HiDBH);
    mp_oChildBehaviors[0].AddRequiredData(m_fCut3BA);
    mp_oChildBehaviors[0].AddRequiredData(m_fCut4LoDBH);
    mp_oChildBehaviors[0].AddRequiredData(m_fCut4HiDBH);
    mp_oChildBehaviors[0].AddRequiredData(m_fCut4BA);

    //Harvest behavior
    mp_oChildBehaviors[1] = new Behavior("Harvest",
                                         "Harvest", "harvest", (float) 1);
    mp_oChildBehaviors[1].m_bCanEditAutomatically = false;
    mp_oChildBehaviors[1].m_bMustHaveTrees = false;

    //Episodic mortality behavior
    mp_oChildBehaviors[2] = new Behavior("Episodic Mortality",
                                         "Episodic Mortality",
                                         "episodic mortality",
                                         (float) 1);
    mp_oChildBehaviors[2].m_bCanEditAutomatically = false;
    mp_oChildBehaviors[2].m_bMustHaveTrees = false;

    //Storm behavior
    mp_oChildBehaviors[3] = new Behavior("Storms", "Storms", "storm", (float) 1);
    mp_oChildBehaviors[3].m_bMustHaveTrees = false;
    mp_oChildBehaviors[3].AddRequiredData(m_fClass1RtrnInt);
    mp_oChildBehaviors[3].AddRequiredData(m_fClass2RtrnInt);
    mp_oChildBehaviors[3].AddRequiredData(m_fClass3RtrnInt);
    mp_oChildBehaviors[3].AddRequiredData(m_fClass4RtrnInt);
    mp_oChildBehaviors[3].AddRequiredData(m_fClass5RtrnInt);
    mp_oChildBehaviors[3].AddRequiredData(m_fClass6RtrnInt);
    mp_oChildBehaviors[3].AddRequiredData(m_fClass7RtrnInt);
    mp_oChildBehaviors[3].AddRequiredData(m_fClass8RtrnInt);
    mp_oChildBehaviors[3].AddRequiredData(m_fClass9RtrnInt);
    mp_oChildBehaviors[3].AddRequiredData(m_fClass10RtrnInt);
    mp_oChildBehaviors[3].AddRequiredData(m_iSusceptibility);
    mp_oChildBehaviors[3].AddRequiredData(m_iStochasticity);
    mp_oChildBehaviors[3].AddRequiredData(m_iProbDistFunc);
    mp_oChildBehaviors[3].AddRequiredData(m_fStdDev);
    mp_oChildBehaviors[3].AddRequiredData(m_fClump);

    mp_oChildBehaviors[4] = new Behavior("Storm Damage Applier",
                                         "Storm Damage Applier",
                                         "storm damage applier",
                                         (float) 2.0);
    mp_oChildBehaviors[4].m_bMustHaveTrees = true;
    mp_oChildBehaviors[4].SetCanApplyTo(TreePopulation.SEEDLING, false);
    mp_oChildBehaviors[4].AddRequiredData(mp_fMinDBH);
    mp_oChildBehaviors[4].AddRequiredData(mp_fStmDmgInterceptMed);
    mp_oChildBehaviors[4].AddRequiredData(mp_fStmDmgInterceptFull);
    mp_oChildBehaviors[4].AddRequiredData(mp_fStmIntensityCoeff);
    mp_oChildBehaviors[4].AddRequiredData(mp_fStmDBHCoeff);
    mp_oChildBehaviors[4].AddRequiredData(mp_fStmPropMedDmgDie);
    mp_oChildBehaviors[4].AddRequiredData(mp_fStmPropFullDmgDie);
    mp_oChildBehaviors[4].AddRequiredData(mp_fStmPropFullTipUp);
    mp_oChildBehaviors[4].AddRequiredData(m_iNumYearsToHeal);
    mp_oChildBehaviors[4].AddRequiredData(m_iNumSnagYears);

    mp_oAllData.add(m_fClass1RtrnInt);
    mp_oAllData.add(m_fClass2RtrnInt);
    mp_oAllData.add(m_fClass3RtrnInt);
    mp_oAllData.add(m_fClass4RtrnInt);
    mp_oAllData.add(m_fClass5RtrnInt);
    mp_oAllData.add(m_fClass6RtrnInt);
    mp_oAllData.add(m_fClass7RtrnInt);
    mp_oAllData.add(m_fClass8RtrnInt);
    mp_oAllData.add(m_fClass9RtrnInt);
    mp_oAllData.add(m_fClass10RtrnInt);
    mp_oAllData.add(m_iSusceptibility);
    mp_oAllData.add(m_iStochasticity);
    mp_oAllData.add(m_iProbDistFunc);
    mp_oAllData.add(m_fStdDev);
    mp_oAllData.add(m_fClump);
    mp_oAllData.add(mp_fMinDBH);
    mp_oAllData.add(mp_fStmDmgInterceptMed);
    mp_oAllData.add(mp_fStmDmgInterceptFull);
    mp_oAllData.add(mp_fStmIntensityCoeff);
    mp_oAllData.add(mp_fStmDBHCoeff);
    mp_oAllData.add(mp_fStmPropMedDmgDie);
    mp_oAllData.add(mp_fStmPropFullDmgDie);
    mp_oAllData.add(m_iNumYearsToHeal);
    mp_oAllData.add(mp_fStmPropFullTipUp);
    mp_oAllData.add(m_iNumSnagYears);
    mp_oAllData.add(m_iInitialAge);
    mp_oAllData.add(m_fCut1LoDBH);
    mp_oAllData.add(m_fCut1HiDBH);
    mp_oAllData.add(m_fCut1BA);
    mp_oAllData.add(m_fCut2LoDBH);
    mp_oAllData.add(m_fCut2HiDBH);
    mp_oAllData.add(m_fCut2BA);
    mp_oAllData.add(m_fCut3LoDBH);
    mp_oAllData.add(m_fCut3HiDBH);
    mp_oAllData.add(m_fCut3BA);
    mp_oAllData.add(m_fCut4LoDBH);
    mp_oAllData.add(m_fCut4HiDBH);
    mp_oAllData.add(m_fCut4BA);

    //We have to wait to set up the grids
  }

  /**
   * Reads in an old tab-delimited harvest regime file.  Any existing harvest
   * data will be replaced with what's in the file.
   * @param sFileName Filename of harvest regime file.
   * @param oPlot Plot object.
   * @param oPop TreePopulation object.
   * @throws ModelException Wrapping an IO exception, or if the harvest
   * parameters do not match the currently loaded parameter file.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void ReadOldHarvestRegimeFile(String sFileName, Plot oPlot,
                                       TreePopulation oPop) throws
      ModelException {
    FileReader oIn = null;
    try {
      oIn = new FileReader(sFileName);
      Cell[] p_oCells;
      HarvestData oHarvest = null;
      Vector oLine, oHarvestList = new Vector(0);
      String sVal;
      Integer iVal;
      float fLowDbh, fHighDbh, fAmountToCut,
          fOldXCellLength, fOldYCellLength; //old X and Y cell lengths
      int iNumOldGridCells,
          iOldNumXGrids, iOldNumYGrids, //old number of X and Y grids
          iNumSpecies = oPop.GetNumberOfSpecies(),
          iTimestep,
          iX, iY,
          iSp, iGrid, i;

      ModelFileFunctions.SkipLine(oIn);

      //Get the first data line
      oLine = ModelFileFunctions.ReadLine(oIn);

      //Validate the plot lengths - remember they're reversed from the old to the
      //new coordinate system
      sVal = (String) oLine.elementAt(2); //old X cell length
      fOldXCellLength = new Float(sVal).floatValue();
      sVal = (String) oLine.elementAt(0); //old number of X cells
      iOldNumXGrids = new Integer(sVal).intValue();

      sVal = (String) oLine.elementAt(3); //old Y plot length
      fOldYCellLength = new Float(sVal).floatValue();
      sVal = (String) oLine.elementAt(1); //old number of Y cells
      iOldNumYGrids = new Integer(sVal).intValue();

      if (fOldXCellLength * (float) iOldNumXGrids != oPlot.GetPlotYLength() ||
          fOldYCellLength * (float) iOldNumYGrids != oPlot.GetPlotXLength()) {
        ModelException oErr = new ModelException();
        oErr.sCallingFunction = "JAVA";
        oErr.iErrorCode = ErrorGUI.BAD_FILE;
        oErr.sMessage = "The plot dimensions for this harvest regime file " +
            "do not match the plot dimensions for this parameter file.";
        throw oErr;
      }

      //Validate the number of species
      sVal = (String) oLine.elementAt(4);
      if (new Integer(sVal).intValue() != iNumSpecies) {
        ModelException oErr = new ModelException();
        oErr.sCallingFunction = "JAVA";
        oErr.iErrorCode = ErrorGUI.BAD_FILE;
        oErr.sMessage = "The number of species for this harvest regime file (" +
            sVal +
            ") does not match the plot length for this parameter file (" +
            String.valueOf(iNumSpecies) + ".";
        throw oErr;
      }

      //Check the what's to load flag - if planting only, skip out
      sVal = (String) oLine.elementAt(5);
      if (new Integer(sVal).intValue() == 2) { //2 = planting only
        return;
      }

      //Calculate the number of grid cells in this file
      sVal = (String) oLine.elementAt(0); //number of X cells
      iNumOldGridCells = new Integer(sVal).intValue();
      sVal = (String) oLine.elementAt(1); //number of Y cells
      iNumOldGridCells *= new Integer(sVal).intValue();

      for (iGrid = 0; iGrid < iNumOldGridCells; iGrid++) {
        for (iSp = 0; iSp < iNumSpecies; iSp++) {

          //Then create event records and read in the data
          oLine = ModelFileFunctions.ReadLine(oIn);

          //When this record is over, it will be designated by a -1 in the
          //first bucket of the line.  Read all the info for this grid and
          //species until we get to that end-of-record designator.
          sVal = (String) oLine.elementAt(0);
          iVal = new Integer(sVal);
          while (iVal.intValue() > -1) {

            //If there are only two values on the line - no values - skip out
            if (oLine.size() > 2) {
              //First value is grid cell - make sure it matches our current grid
              if (iVal.intValue() != iGrid) {
                ModelException oErr = new ModelException();
                oErr.iErrorCode = ErrorGUI.TAB_FILE_PREMATURE_END;
                oErr.sCallingFunction = "ReadOldHarvestRegimeFile";
                oErr.sMessage = sFileName;
              }

              //Second value is species number - make sure it matches our current
              //species
              sVal = (String) oLine.elementAt(1);
              if (new Integer(sVal).intValue() != iSp) {
                ModelException oErr = new ModelException();
                oErr.iErrorCode = ErrorGUI.TAB_FILE_PREMATURE_END;
                oErr.sCallingFunction = "ReadOldHarvestRegimeFile";
                oErr.sMessage = sFileName;
              }

              //Third value is timestep
              sVal = (String) oLine.elementAt(2);
              iTimestep = new Integer(sVal).intValue();

              //Create a new harvest data record
              oHarvest = null;
              oHarvest = new HarvestData();

              //Set timestep
              oHarvest.SetTimestep(iTimestep);

              //Set species
              oHarvest.AddSpecies(iSp);

              //Add cut amount type
              sVal = (String) oLine.elementAt(11);
              oHarvest.SetCutAmountType(new Integer(sVal).intValue());

              //Add cut type
              sVal = (String) oLine.elementAt(16);
              oHarvest.SetCutType(new Integer(sVal).intValue());

              //Add grid cells
              p_oCells = Cell.TransformOldGridCell(iGrid, fOldXCellLength,
                  fOldYCellLength,
                  iOldNumXGrids, iOldNumYGrids, oPlot);
              for (i = 0; i < p_oCells.length; i++) {
                oHarvest.AddCell(p_oCells[i]);
              }

              //Add cut ranges - there are four
              sVal = (String) oLine.elementAt(3); //first range low dbh
              fLowDbh = new Float(sVal).floatValue();
              sVal = (String) oLine.elementAt(7); //first range high dbh
              fHighDbh = new Float(sVal).floatValue();
              sVal = (String) oLine.elementAt(12); //first range low cut amount
              fAmountToCut = new Float(sVal).floatValue();
              if (fAmountToCut > 0) {
                oHarvest.AddCutRange(fLowDbh, fHighDbh, fAmountToCut);
              }

              sVal = (String) oLine.elementAt(4); //second range low dbh
              fLowDbh = new Float(sVal).floatValue();
              sVal = (String) oLine.elementAt(8); //second range high dbh
              fHighDbh = new Float(sVal).floatValue();
              sVal = (String) oLine.elementAt(13); //second range low cut amount
              fAmountToCut = new Float(sVal).floatValue();
              if (fAmountToCut > 0) {
                oHarvest.AddCutRange(fLowDbh, fHighDbh, fAmountToCut);
              }

              sVal = (String) oLine.elementAt(5); //third range low dbh
              fLowDbh = new Float(sVal).floatValue();
              sVal = (String) oLine.elementAt(9); //third range high dbh
              fHighDbh = new Float(sVal).floatValue();
              sVal = (String) oLine.elementAt(14); //third range low cut amount
              fAmountToCut = new Float(sVal).floatValue();
              if (fAmountToCut > 0) {
                oHarvest.AddCutRange(fLowDbh, fHighDbh, fAmountToCut);
              }

              sVal = (String) oLine.elementAt(6); //fourth range low dbh
              fLowDbh = new Float(sVal).floatValue();
              sVal = (String) oLine.elementAt(10); //fourth range high dbh
              fHighDbh = new Float(sVal).floatValue();
              sVal = (String) oLine.elementAt(15); //fourth range low cut amount
              fAmountToCut = new Float(sVal).floatValue();
              if (fAmountToCut > 0) {
                oHarvest.AddCutRange(fLowDbh, fHighDbh, fAmountToCut);
              }

              //Add to our list
              oHarvestList = AddHarvestData(oHarvestList, oHarvest);

            }

            oLine = ModelFileFunctions.ReadLine(oIn);
            sVal = (String) oLine.elementAt(0);
            iVal = new Integer(sVal);
          }
        }
      }

      //Now compact our list and copy it to our existing list
      mp_oHarvestCuts = CompactHarvestList(oHarvestList);

      Behavior oBeh = GetBehaviorByKey("Harvest");
      if (mp_oHarvestCuts.size() > 0) {
        oBeh.m_bIsEnabled = true;
      }
      else {
        oBeh.m_bIsEnabled = false;
      }
    }
    catch (java.io.IOException oErr) {
      ModelException oExp = new ModelException();
      oExp.iErrorCode = ErrorGUI.BAD_FILE;
      oExp.sCallingFunction = "ReadOldHarvestRegimeFile";
      oExp.sMessage = sFileName;
      throw (oExp);
    }
    finally {
      try {
        oIn.close();
      }
      catch (java.io.IOException e) {
        ;
      }
    }
  }

  /**
   * This adds the data contained in a HarvestData object to a list of harvest
   * data.  A HarvestData object is first checked to make sure there are any
   * cut ranges defined; if not, nothing happens.  If there is data to be
   * added, it is compared to existing HarvestData objects on the list.  If
   * they match except for the grid cell, then the new object's grid cell is
   * added to the existing object and the new object thrown away.  If it is
   * truly new data, it is added to the list.
   * <p>
   * IMPORTANT:  It is assumed that the harvest data list is all single-species
   * cuts.  This makes this method not very generalized.
   * @param oList The list of HarvestData objects to add to.
   * @param oNewHarvest The HarvestData object to add.
   * @return The new list with the data added.
   * @throws ModelException Passing through underlying exceptions.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  protected Vector AddHarvestData(Vector oList, HarvestData oNewHarvest) throws
      ModelException {
    HarvestData oExistingHarvest;
    int i, j, k;
    boolean bMatch, bAllMatch;

    if (null == oNewHarvest) {
      return oList;
    }

    //If there are no cut ranges, don't add
    if (oNewHarvest.GetNumberOfCutRanges() == 0) {
      return null;
    }

    //If the list is currently null, create it
    if (null == oList) {
      oList = new Vector(0);
      oList.add(oNewHarvest);
      return oList;
    }

    for (i = 0; i < oList.size(); i++) {

      oExistingHarvest = (HarvestData) oList.elementAt(i);

      //Compare all the data
      if (oExistingHarvest.GetCutAmountType() == oNewHarvest.GetCutAmountType() &&
          oExistingHarvest.GetCutType() == oNewHarvest.GetCutType() &&
          oExistingHarvest.GetTimestep() == oNewHarvest.GetTimestep() &&
          oExistingHarvest.GetNumberOfCutRanges() ==
          oNewHarvest.GetNumberOfCutRanges() &&
          oExistingHarvest.GetSpecies(0) == oNewHarvest.GetSpecies(0)) {

        //Make sure the cut ranges are the same
        bAllMatch = true;
        for (j = 0; j < oNewHarvest.GetNumberOfCutRanges(); j++) {
          bMatch = false;
          for (k = 0; k < oExistingHarvest.GetNumberOfCutRanges(); k++) {
            if (oNewHarvest.GetLowerBound(j) ==
                oExistingHarvest.GetLowerBound(k) &&
                oNewHarvest.GetUpperBound(j) ==
                oExistingHarvest.GetUpperBound(k) &&
                oNewHarvest.GetCutAmount(j) == oExistingHarvest.GetCutAmount(k)) {
              bMatch = true;
              break;
            }
          }
          if (!bMatch) {
            bAllMatch = false;
            break;
          }
        }

        if (bAllMatch) {
          //Add this one's grid cell to the existing list
          for (j = 0; j < oNewHarvest.GetNumberOfCells(); j++) {
            oExistingHarvest.AddCell(oNewHarvest.GetCell(j));
          }
          return oList;
        }
      }
    }

    //If we've gotten to here, then the new object's data is unique and should
    //be added.
    oList.add(oNewHarvest);
    return oList;
  }

  /**
   * Takes a list of HarvestData objects where each is one single species, and
   * replaces it with a list of HarvestData objects for multiple species for
   * any whose data match.
   * @param oList The list to compact.
   * @return The compacted list.
   * @throws ModelException Passing through underlying exceptions
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  protected Vector CompactHarvestList(Vector oList) throws ModelException {

    Vector oTempList = new Vector(0);
    HarvestData oExistingHarvest, oNewHarvest;
    Cell oNewCell, oExistingCell;
    int i, j, k, m;
    boolean bUnique = true, bMatch, bAllMatch;

    //If the list is null, return
    if (null == oList) {
      return null;
    }

    for (i = 0; i < oList.size(); i++) {

      oNewHarvest = (HarvestData) oList.elementAt(i);
      bUnique = true;
      //See if this object's data matches any existing object
      for (j = 0; j < oTempList.size(); j++) {

        oExistingHarvest = (HarvestData) oTempList.elementAt(j);

        if (oExistingHarvest.GetCutAmountType() == oNewHarvest.GetCutAmountType() &&
            oExistingHarvest.GetCutType() == oNewHarvest.GetCutType() &&
            oExistingHarvest.GetTimestep() == oNewHarvest.GetTimestep() &&
            oExistingHarvest.GetNumberOfCutRanges() ==
            oNewHarvest.GetNumberOfCutRanges() &&
            oExistingHarvest.GetNumberOfCells() == oNewHarvest.GetNumberOfCells()) {

          //Test cut ranges
          bAllMatch = true;
          for (k = 0; k < oNewHarvest.GetNumberOfCutRanges(); k++) {
            bMatch = false;
            for (m = 0; m < oExistingHarvest.GetNumberOfCutRanges(); m++) {
              if (oNewHarvest.GetLowerBound(k) ==
                  oExistingHarvest.GetLowerBound(m) &&
                  oNewHarvest.GetUpperBound(k) ==
                  oExistingHarvest.GetUpperBound(m) &&
                  oNewHarvest.GetCutAmount(k) ==
                  oExistingHarvest.GetCutAmount(m)) {
                bMatch = true;
                break;
              }
            }
            if (!bMatch) {
              bAllMatch = false;
              break;
            }
          }

          //If cut ranges checked out, test the cell list
          if (bAllMatch) {

            bAllMatch = true;
            for (k = 0; k < oNewHarvest.GetNumberOfCells(); k++) {
              oNewCell = (Cell) oNewHarvest.GetCell(k);
              bMatch = false;
              for (m = 0; m < oExistingHarvest.GetNumberOfCells(); m++) {
                oExistingCell = (Cell) oExistingHarvest.GetCell(m);
                if (oNewCell.GetX() == oExistingCell.GetX() &&
                    oNewCell.GetY() == oExistingCell.GetY()) {
                  bMatch = true;
                  break;
                }
              }
              if (!bMatch) {
                bAllMatch = false;
                break;
              }
            }
            if (bAllMatch) {
              bUnique = false;
              oExistingHarvest.AddSpecies(oNewHarvest.GetSpecies(0));
            }
          }
        }
      }

      if (bUnique) {
        //This object has unique data - add it to the temp list
        oTempList.add(oNewHarvest);
      }
    }

    //Return the temp list to the existing list
    return oTempList;
  }

  /**
   * Writes the storm data to the parameter file.  Does nothing if the
   * storm behaviors are not enabled.
   * @param jOut FileWriter File to write to.
   * @throws IOException if there is a problem writing the file.
   * @throws ModelException passed through from called functions.
   */
  private void WriteStorm(BufferedWriter jOut) throws java.io.IOException,
      ModelException {

    if (GetBehaviorByKey("Storms").m_bIsEnabled) {
      jOut.write("<storm>");

      WriteData(jOut, m_fClass1RtrnInt);
      WriteData(jOut, m_fClass2RtrnInt);
      WriteData(jOut, m_fClass3RtrnInt);
      WriteData(jOut, m_fClass4RtrnInt);
      WriteData(jOut, m_fClass5RtrnInt);
      WriteData(jOut, m_fClass6RtrnInt);
      WriteData(jOut, m_fClass7RtrnInt);
      WriteData(jOut, m_fClass8RtrnInt);
      WriteData(jOut, m_fClass9RtrnInt);
      WriteData(jOut, m_fClass10RtrnInt);
      WriteData(jOut, m_iSusceptibility);
      WriteData(jOut, m_iStochasticity);
      if (m_iStochasticity.GetStringValue().equals("Stochastic")) {
        WriteData(jOut, m_iProbDistFunc);

        if (m_iProbDistFunc.GetStringValue().equalsIgnoreCase("Lognormal") ||
            m_iProbDistFunc.GetStringValue().equalsIgnoreCase("Normal")) {
          WriteData(jOut, m_fStdDev);
        }
        else if (m_iProbDistFunc.GetStringValue().equalsIgnoreCase(
            "Negative Binomial")) {
          WriteData(jOut, m_fClump);
        }
      }

      if (GetBehaviorByKey("Storm Damage Applier").m_bIsEnabled) {
        TreePopulation oPop = m_oManager.GetTreePopulation();
        boolean[] p_bAppliesTo = GetBehaviorByKey("Storm Damage Applier").
            GetWhichSpeciesUsed(oPop);
        WriteSpeciesSpecificValue(jOut, mp_fMinDBH, oPop, p_bAppliesTo);
        WriteSpeciesSpecificValue(jOut, mp_fStmDmgInterceptMed, oPop,
                                  p_bAppliesTo);
        WriteSpeciesSpecificValue(jOut, mp_fStmDmgInterceptFull, oPop,
                                  p_bAppliesTo);
        WriteSpeciesSpecificValue(jOut, mp_fStmIntensityCoeff, oPop,
                                  p_bAppliesTo);
        WriteSpeciesSpecificValue(jOut, mp_fStmDBHCoeff, oPop, p_bAppliesTo);
        WriteSpeciesSpecificValue(jOut, mp_fStmPropMedDmgDie, oPop,
                                  p_bAppliesTo);
        WriteSpeciesSpecificValue(jOut, mp_fStmPropFullDmgDie, oPop,
                                  p_bAppliesTo);
        WriteSpeciesSpecificValue(jOut, mp_fStmPropFullTipUp, oPop,
                                  p_bAppliesTo);
        WriteData(jOut, m_iNumYearsToHeal);
        WriteData(jOut, m_iNumSnagYears);
      }

      jOut.write("</storm>");
    }
  }

  /**
   * Writes the selection harvest data to the parameter file.  Does nothing if
   * the selection harvest behavior is not enabled.
   * @param jOut FileWriter File to write to.
   * @throws IOException if there is a problem writing the file.
   * @throws ModelException passed through from called functions.
   */
  private void WriteSelectionHarvest(BufferedWriter jOut) throws java.io.
      IOException,
      ModelException {

    if (GetBehaviorByKey("Selection Harvest").m_bIsEnabled) {
      jOut.write("<SelectionHarvest>");

      WriteData(jOut, m_iInitialAge);

      if (m_fCut1BA.GetValue() > 0) {
        jOut.write("<sha_CutRange>");
        WriteData(jOut, m_fCut1LoDBH);
        WriteData(jOut, m_fCut1HiDBH);
        WriteData(jOut, m_fCut1BA);
        jOut.write("</sha_CutRange>");
      }
      if (m_fCut2BA.GetValue() > 0) {
        jOut.write("<sha_CutRange>");
        WriteData(jOut, m_fCut2LoDBH);
        WriteData(jOut, m_fCut2HiDBH);
        WriteData(jOut, m_fCut2BA);
        jOut.write("</sha_CutRange>");
      }
      if (m_fCut3BA.GetValue() > 0) {
        jOut.write("<sha_CutRange>");
        WriteData(jOut, m_fCut3LoDBH);
        WriteData(jOut, m_fCut3HiDBH);
        WriteData(jOut, m_fCut3BA);
        jOut.write("</sha_CutRange>");
      }
      if (m_fCut4BA.GetValue() > 0) {
        jOut.write("<sha_CutRange>");
        WriteData(jOut, m_fCut4LoDBH);
        WriteData(jOut, m_fCut4HiDBH);
        WriteData(jOut, m_fCut4BA);
        jOut.write("</sha_CutRange>");
      }

      jOut.write("</SelectionHarvest>");
    }
  }

  /**
   * Writes the episodic mortality data for the parameter file.  Does nothing
   * if there is no episodic mortality data.
   * @param jOut FileWriter File to write to.
   * @param oPop Tree population object.
   * @throws java.io.IOException if there is a problem writing the file
   * @throws ModelException if there are unrecognized disturbance cut types
   * (should never happen).
   */
  private void WriteMortalityEpisode(BufferedWriter jOut, TreePopulation oPop) throws
      java.io.IOException, ModelException {
    HarvestData oData;
    Cell oCell;
    int i, iNumObjects = mp_oMortEpisodes.size(), j, iTemp;

    if (iNumObjects > 0) {

      jOut.write("<disturbance>");

      //Write each harvest object's data
      for (i = 0; i < iNumObjects; i++) {
        oData = (HarvestData) mp_oMortEpisodes.elementAt(i);

        jOut.write("<ds_deathEvent>");

        //Write each species
        iTemp = oData.GetNumberOfSpecies();
        for (j = 0; j < iTemp; j++) {
          jOut.write("<ds_applyToSpecies species=\"" +
                     oPop.GetSpeciesNameFromCode(oData.GetSpecies(j)) +
                     "\"/>");
        }

        //Write the timestep
        jOut.write("<ds_timestep>" + String.valueOf(oData.GetTimestep()) +
                   "</ds_timestep>");

        //Write the cut amount type
        jOut.write("<ds_cutAmountType>");
        iTemp = oData.GetCutAmountType();
        if (iTemp == ABSOLUTE_BASAL_AREA) {
          jOut.write("absolute basal area");
        }
        else if (iTemp == ABSOLUTE_DENSITY) {
          jOut.write("absolute density");
        }
        else if (iTemp == PERCENTAGE_BASAL_AREA) {
          jOut.write("percent of basal area");
        }
        else if (iTemp == PERCENTAGE_DENSITY) {
          jOut.write("percent of density");
        }
        else {
          ModelException oErr = new ModelException();
          oErr.iErrorCode = ErrorGUI.BAD_DATA;
          oErr.sCallingFunction = "JAVA";
          oErr.sMessage =
              "When writing parameter file, came across unrecognized " +
              "mortality episode cut type \"" + String.valueOf(iTemp) + "\".";
          throw oErr;
        }
        jOut.write("</ds_cutAmountType>");

        //Write the cut ranges
        iTemp = oData.GetNumberOfCutRanges();
        for (j = 0; j < iTemp; j++) {

          jOut.write("<ds_dbhRange>");
          jOut.write("<ds_low>" + String.valueOf(oData.GetLowerBound(j)) +
                     "</ds_low>");
          jOut.write("<ds_high>" + String.valueOf(oData.GetUpperBound(j)) +
                     "</ds_high>");
          jOut.write("<ds_amountToCut>" +
                     String.valueOf(oData.GetCutAmount(j)) +
                     "</ds_amountToCut>");
          jOut.write("</ds_dbhRange>");
        }

        //Write the list of cells
        iTemp = oData.GetNumberOfCells();
        for (j = 0; j < iTemp; j++) {

          oCell = (Cell) oData.GetCell(j);

          jOut.write("<ds_applyToCell x=\"" + String.valueOf(oCell.GetX())
                     + "\" y=\"" + String.valueOf(oCell.GetY()) + "\"/>");
        }

        jOut.write("</ds_deathEvent>");
      }

      jOut.write("</disturbance>");
    }
  }

  /**
   * Writes the harvest data for the parameter file.  Does nothing if there
   * is no harvest data.
   * @param jOut FileWriter File to write to.
   * @param oPop Tree population object.
   * @throws java.io.IOException if there is a problem writing the file
   * @throws ModelException if there are unrecognized harvest cut types (should
   * never happen).
   */
  private void WriteHarvest(BufferedWriter jOut, TreePopulation oPop) throws
      java.
      io.IOException, ModelException {
    HarvestData oData;
    Cell oCell;
    int i, iNumObjects = mp_oHarvestCuts.size(), j, iTemp;

    //************************
     // Harvest
     //************************

      if (iNumObjects > 0) {

        jOut.write("<harvest>");

        //Write each harvest object's data
        for (i = 0; i < iNumObjects; i++) {
          oData = (HarvestData) mp_oHarvestCuts.elementAt(i);

          jOut.write("<ha_cutEvent>");

          //Write each species
          iTemp = oData.GetNumberOfSpecies();
          for (j = 0; j < iTemp; j++) {
            jOut.write("<ha_applyToSpecies species=\"" +
                       oPop.GetSpeciesNameFromCode(oData.GetSpecies(j)) +
                       "\"/>");
          }

          //Write the timestep
          jOut.write("<ha_timestep>" + String.valueOf(oData.GetTimestep()) +
                     "</ha_timestep>");

          //Write the cut type
          jOut.write("<ha_cutType>");
          iTemp = oData.GetCutType();
          if (iTemp == GAP_CUT) {
            jOut.write("gap");
          }
          else if (iTemp == PARTIAL_CUT) {
            jOut.write("partial");
          }
          else if (iTemp == CLEAR_CUT) {
            jOut.write("clear");
          }
          else {
            ModelException oErr = new ModelException();
            oErr.iErrorCode = ErrorGUI.BAD_DATA;
            oErr.sCallingFunction = "JAVA";
            oErr.sMessage =
                "When writing parameter file, came across unrecognized " +
                "harvest cut type \"" + String.valueOf(iTemp) + "\".";
            throw oErr;
          }
          jOut.write("</ha_cutType>");

          //Write the cut amount type
          jOut.write("<ha_cutAmountType>");
          iTemp = oData.GetCutAmountType();
          if (iTemp == ABSOLUTE_BASAL_AREA) {
            jOut.write("absolute basal area");
          }
          else if (iTemp == ABSOLUTE_DENSITY) {
            jOut.write("absolute density");
          }
          else if (iTemp == PERCENTAGE_BASAL_AREA) {
            jOut.write("percent of basal area");
          }
          else if (iTemp == PERCENTAGE_DENSITY) {
            jOut.write("percent of density");
          }
          else {
            ModelException oErr = new ModelException();
            oErr.iErrorCode = ErrorGUI.BAD_DATA;
            oErr.sCallingFunction = "JAVA";
            oErr.sMessage =
                "When writing parameter file, came across unrecognized " +
                "harvest cut type \"" + String.valueOf(iTemp) + "\".";
            throw oErr;
          }
          jOut.write("</ha_cutAmountType>");

          //Write the cut ranges
          iTemp = oData.GetNumberOfCutRanges();
          for (j = 0; j < iTemp; j++) {

            jOut.write("<ha_dbhRange>");
            jOut.write("<ha_low>" + String.valueOf(oData.GetLowerBound(j)) +
                       "</ha_low>");
            jOut.write("<ha_high>" + String.valueOf(oData.GetUpperBound(j)) +
                       "</ha_high>");
            jOut.write("<ha_amountToCut>" +
                       String.valueOf(oData.GetCutAmount(j)) +
                       "</ha_amountToCut>");
            jOut.write("</ha_dbhRange>");
          }

          //Write the list of cells
          iTemp = oData.GetNumberOfCells();
          for (j = 0; j < iTemp; j++) {

            oCell = (Cell) oData.GetCell(j);

            jOut.write("<ha_applyToCell x=\"" + String.valueOf(oCell.GetX())
                       + "\" y=\"" + String.valueOf(oCell.GetY()) + "\"/>");
          }

          jOut.write("</ha_cutEvent>");
        }

        jOut.write("</harvest>");
      }

  }

  /**
   * Validates the data prior to writing it.  What it does:
   * <ul>
   * <li>Causes all HarvestData objects to validate themselves.</li>
   * <li>If the Storms behavior is enabled, this makes sure that all storm
   * return intervals are greater than 0.</li>
   * <li>If the Storms behavior is enabled and the susceptibility is set to
   * "mapped", this makes sure that there are values for the grid object
   * "Storm Susceptibility" and that they are all between 0 and 1.</li>
   * <li>If the Storms behavior is enabled and the damage pattern is set to
   * "mapped", this makes sure that the grid resolution for the "Storm
   * Susceptibility" and "Storm Damage" grids is the same.</li>
   * <li>If the Storm Damage Applier behavior is enabled, this checks to make
   * sure the Storms behavior is enabled.</li>
   * <li>If the Storm Damage Applier behavior is enabled, this checks to make
   * sure the value in m_iNumYearsToHeal is positive.</li>
   * <li>If the Storm Damage Applier behavior is enabled, this checks to make
   * sure the value in m_iNumSnagYears is not negative.</li>
   * <li>If the Storm Damage Applier behavior is enabled, this makes sure that
   * the proportions of trees that die with medium and total damage are all
   * proportions.</li>
   * <li>If the Storm Damage Applier behavior is enabled, this makes sure that
   * the proportion of dead trees that tip up are all proportions.</li>
   * <li>If Selection Harvest is enabled, this makes sure that target basal
   * area is not less than 0 and cut ranges don't overlap</li>
   * </ul>
   * @param oPop TreePopulation object
   * @throws ModelException if one of the data objects is not valid.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   * <br>April 4, 2005:  Added new storm damage applier parameters (LEM)
   * <br>May 2, 2005:  Added new storm damage applier parameters (LEM)
   * <br>May 12, 2005:  Added selection harvest stuff (LEM)
   */
  public void ValidateData(TreePopulation oPop) throws ModelException {
    HarvestData oData;
    Plot oPlot = m_oManager.GetPlot();
    int i, iNumObjects;

    //Disable the Harvest behavior if there are no harvest events
    //No longer - Selection Harvest may enable it
    //if (iNumObjects == 0) {
    //  GetBehaviorByKey("Harvest").m_bIsEnabled = false;
    // }

    //Validate each harvest event
    iNumObjects = mp_oHarvestCuts.size();
    for (i = 0; i < iNumObjects; i++) {
      oData = (HarvestData) mp_oHarvestCuts.elementAt(i);
      oData.ValidateCut(oPop, oPlot);
    }

    //Disable the Episodic Mortality behavior if there are no mortality episodes
    //events
    //if (iNumObjects == 0) {
    //  GetBehaviorByKey("Episodic Mortality").m_bIsEnabled = false;
    //}

    //Validate each Episodic Mortality event
    iNumObjects = mp_oMortEpisodes.size();
    for (i = 0; i < iNumObjects; i++) {
      oData = (HarvestData) mp_oMortEpisodes.elementAt(i);
      oData.ValidateCut(oPop, oPlot);
    }

    //Storms behavior validation
    if (GetBehaviorByKey("Storms").m_bIsEnabled) {

      //Make sure all the return intervals are positive
      MakeSurePositive(m_fClass1RtrnInt);
      MakeSurePositive(m_fClass2RtrnInt);
      MakeSurePositive(m_fClass3RtrnInt);
      MakeSurePositive(m_fClass4RtrnInt);
      MakeSurePositive(m_fClass5RtrnInt);
      MakeSurePositive(m_fClass6RtrnInt);
      MakeSurePositive(m_fClass7RtrnInt);
      MakeSurePositive(m_fClass8RtrnInt);
      MakeSurePositive(m_fClass9RtrnInt);
      MakeSurePositive(m_fClass10RtrnInt);

      //If the damage pattern is mapped, make sure there are values for the
      //"Storm Susceptibility" grid and that they are all between 0 and 1
      if (m_iSusceptibility.GetStringValue().equalsIgnoreCase("mapped")) {
        Grid oSusceptibility = null;
        for (i = 0; i < mp_oGrids.length; i++) {
          if (mp_oGrids[i].GetName().equalsIgnoreCase("Storm Susceptibility")) {
            oSusceptibility = mp_oGrids[i];
            break;
          }
        }
        if (oSusceptibility == null) {
          //"Storm Susceptibility" is missing - shouldn't happen, but a
          //good double-check
          ModelException oErr = new ModelException(ErrorGUI.DATA_MISSING,
              "JAVA",
              "The grid \"Storm Susceptibility\" is required for mapped storm damage patterns.");
          throw (oErr);
        }

        iNumObjects = oSusceptibility.mp_oGridVals.size();
        if (iNumObjects == 0) {
          //Map is missing for "Storm Susceptibility"
          ModelException oErr = new ModelException(ErrorGUI.DATA_MISSING,
              "JAVA",
              "A map of values for the grid \"Storm Susceptibility\" is required for mapped storm damage patterns.");
          throw (oErr);
        }

        for (i = 0; i < iNumObjects; i++) {
          GridValue oVal = (GridValue) oSusceptibility.mp_oGridVals.elementAt(i);
          if (oVal.mp_fFloats == null || oVal.mp_fFloats[0] == null ||
              oVal.mp_fFloats[0].floatValue() < 0 ||
              oVal.mp_fFloats[0].floatValue() > 1) {
            ModelException oErr = new ModelException(ErrorGUI.DATA_MISSING,
                "JAVA",
                "All values in the map for the grid \"Storm Susceptibility\" must be between 0 and 1.");
            throw (oErr);
          }
        }

        //Get the grid "Storm Damage" and make sure it has the same resolution
        //as "Storm Susceptibility"
        Grid oDamage = null;
        for (i = 0; i < mp_oGrids.length; i++) {
          if (mp_oGrids[i].GetName().equalsIgnoreCase("Storm Damage")) {
            oDamage = mp_oGrids[i];
            break;
          }
        }
        if (oDamage != null) {
          if (oDamage.m_bEdited) {
            if (java.lang.Math.abs(oDamage.GetXCellLength() -
                                   oSusceptibility.GetXCellLength()) > 0.01 ||
                java.lang.Math.abs(oDamage.GetYCellLength() -
                                   oSusceptibility.GetYCellLength()) > 0.01) {
              ModelException oErr = new ModelException(ErrorGUI.BAD_DATA,
                  "JAVA",
                  "The grids \"Storm Susceptibility\" and \"Storm Damage\" " +
                  "must have matching grid cell resolutions.");
              throw (oErr);

            }
          }
        }
      }
    }

    if (GetBehaviorByKey("Storm Damage Applier").m_bIsEnabled) {
      if (GetBehaviorByKey("Storms").m_bIsEnabled == false) {
        ModelException oErr = new ModelException(ErrorGUI.BAD_DATA,
                                                 "JAVA",
                                                 "When the behavior \"Storm Damage Applier\" is enabled, " +
                                                 "the behavior \"Storms\" must be enabled as well.");
        throw (oErr);
      }
      boolean[] p_bAppliesTo = GetBehaviorByKey("Storm Damage Applier").
          GetWhichSpeciesUsed(oPop);
      MakeSurePositive(m_iNumYearsToHeal);
      //LEM 04/04/05
      MakeSureAllAreProportions(mp_fStmPropMedDmgDie, p_bAppliesTo);
      MakeSureAllAreProportions(mp_fStmPropFullDmgDie, p_bAppliesTo);
      //LEM 05/02/05
      MakeSureAllAreProportions(mp_fStmPropFullTipUp, p_bAppliesTo);
      MakeSureNonNegative(m_iNumSnagYears);
    }

    //LEM 05-12-05
    if (GetBehaviorByKey("Selection Harvest").m_bIsEnabled) {

      GetBehaviorByKey("Harvest").m_bIsEnabled = true;

      MakeSureNonNegative(m_fCut1BA);
      MakeSureNonNegative(m_fCut2BA);
      MakeSureNonNegative(m_fCut3BA);
      MakeSureNonNegative(m_fCut4BA);

      //Make sure the upper bound of each cut range is larger than the lower
      //bound
      if (m_fCut1HiDBH.GetValue() < m_fCut1LoDBH.GetValue() ||
          m_fCut2HiDBH.GetValue() < m_fCut2LoDBH.GetValue() ||
          m_fCut3HiDBH.GetValue() < m_fCut3LoDBH.GetValue() ||
          m_fCut4HiDBH.GetValue() < m_fCut4LoDBH.GetValue()) {
        throw new ModelException(ErrorGUI.BAD_DATA, "JAVA", "Selection Harvest cut range upper DBHs must be greater than corresponding lower DBHs.");
      }
    }
  }

  /**
   * Overridden because disturbance writing is special.
   * @param jOut File to write to.
   * @param oPop TreePopulation object
   * @throws javawrapper.ModelException if there's a problem writing the file.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   * <br>May 13, 2004:  Added episodic mortality writing.  (LEM)
   * <br>May 12, 2005:  Added selection harvest writing (LEM)
   */
  public void WriteXML(BufferedWriter jOut, TreePopulation oPop) throws
      ModelException {
    try {

      WriteHarvest(jOut, oPop);
      WriteMortalityEpisode(jOut, oPop);
      WriteStorm(jOut);
      WriteSelectionHarvest(jOut); //LEM 05-12-05

    }
    catch (java.io.IOException e) {
      ModelException oErr = new ModelException();
      oErr.iErrorCode = ErrorGUI.BAD_FILE;
      oErr.sCallingFunction = "JAVA";
      oErr.sMessage = "There was a problem writing the parameter file.";
      throw (oErr);

    }
  }

  /**
   * Sets up the grid objects.
   * @param oPop TreePopulation object.
   * @throws javawrapper.ModelException passthrough only.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   * <br>May 13, 2004:  Added episodic mortality grid.  (LEM)
   * <br>April 4, 2005:  Defaulted new storm params (LEM)
   */
  public void DoSetup(TreePopulation oPop) throws javawrapper.ModelException {

    int iNumSpecies = oPop.GetNumberOfSpecies(),
        i, j, k, m;

    mp_oGrids = new Grid[4];

    //********************************
     //Harvest grid
     //********************************

    String sGridName = "Harvest Results";

    //The accessible data members are one for each cut range and species
    String[] p_sBaseMembers = {
        "Cut Density", "Cut Basal Area"};
    DataMember[] p_oDataMembers = new DataMember[p_sBaseMembers.length *
        iNumSpecies *
        HarvestData.NUMBER_ALLOWED_CUT_RANGES + 1];

    p_oDataMembers[0] = new DataMember("Harvest Type", "Harvest Type",
                                       DataMember.INTEGER);

    m = 1;
    for (i = 0; i < p_sBaseMembers.length; i++) {
      if (p_sBaseMembers[i].equals("Cut Density")) {
        for (j = 0; j < iNumSpecies; j++) {
          for (k = 0; k < HarvestData.NUMBER_ALLOWED_CUT_RANGES; k++) {
            p_oDataMembers[m] = new DataMember(p_sBaseMembers[i] + ", " +
                                               oPop.GetSpeciesNameFromCode(j).
                                               replace('_', ' ')
                                               + ", cut range " + k,
                                               p_sBaseMembers[i] + "_" + k +
                                               "_" +
                                               j,
                                               DataMember.INTEGER);
            m++;
          }
        }
      }
      else {
        for (j = 0; j < iNumSpecies; j++) {
          for (k = 0; k < HarvestData.NUMBER_ALLOWED_CUT_RANGES; k++) {
            p_oDataMembers[m] = new DataMember(p_sBaseMembers[i] + ", " +
                                               oPop.GetSpeciesNameFromCode(j).
                                               replace('_', ' ')
                                               + ", cut range " + k,
                                               p_sBaseMembers[i] + "_" + k +
                                               "_" +
                                               j,
                                               DataMember.FLOAT);
            m++;
          }
        }
      }
    }

    //Create our grid
    mp_oGrids[0] = new Grid(sGridName, p_oDataMembers, null, 8, 8);

    //Add it to the harvest behavior
    GetBehaviorByKey("Harvest").AddGrid(mp_oGrids[0]);

    //********************************
     //Episodic Mortality grid
     //********************************

    sGridName = "Mortality Episode Results";

    //The accessible data members are one for each cut range and species
    p_oDataMembers = new DataMember[p_sBaseMembers.length *
        iNumSpecies *
        HarvestData.NUMBER_ALLOWED_CUT_RANGES];

    m = 0;
    for (i = 0; i < p_sBaseMembers.length; i++) {
      if (p_sBaseMembers[i].equals("Cut Density")) {
        for (j = 0; j < iNumSpecies; j++) {
          for (k = 0; k < HarvestData.NUMBER_ALLOWED_CUT_RANGES; k++) {
            p_oDataMembers[m] = new DataMember(p_sBaseMembers[i] + ", " +
                                               oPop.GetSpeciesNameFromCode(j).
                                               replace('_', ' ')
                                               + ", cut range " + k,
                                               p_sBaseMembers[i] + "_" + k +
                                               "_" +
                                               j,
                                               DataMember.INTEGER);
            m++;
          }
        }
      }
      else {
        for (j = 0; j < iNumSpecies; j++) {
          for (k = 0; k < HarvestData.NUMBER_ALLOWED_CUT_RANGES; k++) {
            p_oDataMembers[m] = new DataMember(p_sBaseMembers[i] + ", " +
                                               oPop.GetSpeciesNameFromCode(j).
                                               replace('_', ' ')
                                               + ", cut range " + k,
                                               p_sBaseMembers[i] + "_" + k +
                                               "_" +
                                               j,
                                               DataMember.FLOAT);
            m++;
          }
        }
      }
    }

    //Create our grid
    mp_oGrids[1] = new Grid(sGridName, p_oDataMembers, null, 8, 8);

    //Add it to the episodic mortality behavior
    GetBehaviorByKey("Episodic Mortality").AddGrid(mp_oGrids[1]);

    //********************************
     //Storm Damage grid
     //********************************

    sGridName = "Storm Damage";

    //The accessible data members are one for each cut range and species
    p_oDataMembers = new DataMember[1];
    p_oDataMembers[0] = new DataMember("Damage Index", "dmg_index",
                                       DataMember.FLOAT);

    //Create our grid
    mp_oGrids[2] = new Grid(sGridName, p_oDataMembers, null, 8, 8);

    //Add to the Storms behavior
    GetBehaviorByKey("Storms").AddGrid(mp_oGrids[2]);

    //********************************
     //Storm Susceptibility grid
     //********************************

    sGridName = "Storm Susceptibility";

    //The accessible data members are one for each cut range and species
    p_oDataMembers = new DataMember[1];
    p_oDataMembers[0] = new DataMember("Susceptibility", "index",
                                       DataMember.FLOAT);

    //Create our grid
    mp_oGrids[3] = new Grid(sGridName, p_oDataMembers, null, 8, 8);

    //Add to the Storms behavior
    GetBehaviorByKey("Storms").AddGrid(mp_oGrids[3]);

    //Default the proportions that die to 0
    if (mp_fStmPropMedDmgDie.GetValue().size() == 0) {
      for (i = 0; i < iNumSpecies; i++) {
        mp_fStmPropMedDmgDie.GetValue().add(new Float(0.0));
      }
    }
    if (mp_fStmPropFullDmgDie.GetValue().size() == 0) {
      for (i = 0; i < iNumSpecies; i++) {
        mp_fStmPropFullDmgDie.GetValue().add(new Float(0.0));
      }
    }
  }

  /**
   * Accepts an XML parent tag (empty, no data) from the parser.  This method
   * watches for the following tags:
   * <ul>
   * <li>ha_applyToCell</li>
   * <li>ha_applyToSpecies</li>
   * <li>ha_dbhRange</li>
   * <li>ha_cutEvent</li>
   * <li>ds_applyToCell</li>
   * <li>ds_applyToSpecies</li>
   * <li>ds_dbhRange</li>
   * <li>ds_deathEvent</li>
   * <li>SelectionHarvest</li>
   * <li>sha_CutRange</li>
   * </ul>
   * @param sXMLTag The XML tag.
   * @param oAttributes The attributes of this object.
   * @throws ModelException if there is a problem reading this data.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>May 12, 2004: Created (LEM)
   */
  public void ReadXMLParentTag(String sXMLTag, Attributes oAttributes) throws
      ModelException {

    if (sXMLTag.equals("ds_applyToCell")) {

      //Get the X and Y values
      int iX = new Integer(oAttributes.getValue("x")).intValue(),
          iY = new Integer(oAttributes.getValue("y")).intValue();

      //Assign them to the newest mortality episode
      HarvestData oMortEpisode = (HarvestData) mp_oMortEpisodes.elementAt(
          mp_oMortEpisodes.size() - 1);
      oMortEpisode.AddCell(iX, iY, m_oManager.GetPlot());

    }
    else if (sXMLTag.equals("ds_applyToSpecies")) {

      //Extract species
      TreePopulation oPop = m_oManager.GetTreePopulation();
      int iSpecies = oPop.GetSpeciesCodeFromName(oAttributes.getValue("species"));

      //Get the last harvest event
      if (mp_oMortEpisodes.size() == 0) {
        ModelException oErr = new ModelException();
        oErr.iErrorCode = ErrorGUI.BAD_DATA;
        oErr.sCallingFunction = "JAVA";
        oErr.sMessage = "Mortality episode parameters are incorrectly specified in the parameter file.  Found tag \"ds_applyToSpecies\" outside of disturbance event tag.";
        throw (oErr);
      }
      HarvestData oMortEpisode = (HarvestData) mp_oMortEpisodes.elementAt(
          mp_oMortEpisodes.size() - 1);

      //Add this species to that event
      oMortEpisode.AddSpecies(iSpecies);

    }
    else if (sXMLTag.equals("ds_dbhRange")) {

      //Reset our placeholders for cut range
      m_iCutRangeAmount = -1;
      m_iCutRangeLowerBound = -1;
      m_iCutRangeUpperBound = -1;
    }
    else if (sXMLTag.equals("ha_applyToCell")) {

      //Get the X and Y values
      int iX = new Integer(oAttributes.getValue("x")).intValue(),
          iY = new Integer(oAttributes.getValue("y")).intValue();

      //Assign them to the newest harvest event
      HarvestData oHarvest = (HarvestData) mp_oHarvestCuts.elementAt(
          mp_oHarvestCuts.size() - 1);
      oHarvest.AddCell(iX, iY, m_oManager.GetPlot());

    }
    else if (sXMLTag.equals("ha_applyToSpecies")) {

      //Extract species
      TreePopulation oPop = m_oManager.GetTreePopulation();
      int iSpecies = oPop.GetSpeciesCodeFromName(oAttributes.getValue("species"));

      //Get the last harvest event
      HarvestData oHarvest = (HarvestData) mp_oHarvestCuts.elementAt(
          mp_oHarvestCuts.size() - 1);

      //Add this species to that event
      oHarvest.AddSpecies(iSpecies);

    }
    else if (sXMLTag.equals("ha_dbhRange")) {

      //Reset our placeholders for cut range
      m_iCutRangeAmount = -1;
      m_iCutRangeLowerBound = -1;
      m_iCutRangeUpperBound = -1;
    }
    else if (sXMLTag.equals("ds_deathEvent")) {

      //Create a new mortality episode
      HarvestData oMortEpisode = new HarvestData();
      mp_oMortEpisodes.add(oMortEpisode);

      //Set it to partial cut - what all mortality episodes get
      oMortEpisode.SetCutType(PARTIAL_CUT);
    }
    else if (sXMLTag.equals("ha_cutEvent")) {

      //Create a new harvest event
      HarvestData oHarvest = new HarvestData();
      mp_oHarvestCuts.add(oHarvest);
    }
    else if (sXMLTag.equals("sha_CutRange")) { //LEM 05-12-05
      m_iSelectionHarvestCutRange++;
    }
    else if (sXMLTag.equals("harvest")) {

      //Enable the harvest behavior
      GetBehaviorByKey("Harvest").m_bIsEnabled = true;
    }
    else if (sXMLTag.equals("disturbance")) {

      //Enable the episodic mortality behavior
      GetBehaviorByKey("Episodic Mortality").m_bIsEnabled = true;
    }
    else if (sXMLTag.equals("SelectionHarvest")) { //LEM 05-12-05

      //Enable the harvest behavior and selection harvest behavior
      GetBehaviorByKey("Harvest").m_bIsEnabled = true;
      GetBehaviorByKey("Selection Harvest").m_bIsEnabled = true;
      m_iSelectionHarvestCutRange = 0;
    }
    super.ReadXMLParentTag(sXMLTag, oAttributes);
  }

  /**
   * This method looks for the following tags:
   * <ul>
   * <li>ds_timestep
   * <li>ds_cutAmountType
   * <li>ds_low
   * <li>ds_high
   * <li>ds_amountToCut
   * <li>ha_timestep
   * <li>ha_cutType
   * <li>ha_cutAmountType
   * <li>ha_low
   * <li>ha_high
   * <li>ha_amountToCut
   * </ul>
   * @param sXMLTag XML tag of data object whose value is to be set.
   * @param sXMLParentTag The immediate parent tag that sXMLTag is within.
   * @param oAttributes Attributes of the object.  Ignored, but may be needed
   * by overriding objects.
   * @param oData Data value appropriate to the data type
   * @return true if the value was set successfully; false if the value could
   * not be found.  (This would not be an error, because I need a way to
   * cycle through the objects until one of the objects comes up with a
   * match.)
   * @throws ModelException if the value could not be assigned to the data
   * object, or if the cut type or cut type amount values are unrecognized.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public boolean SetSingleValueByXMLTag(String sXMLTag, String sXMLParentTag,
                                        Attributes oAttributes,
                                        Object oData) throws
      ModelException {
    if (sXMLTag != null) {
      if (sXMLTag.equals("ds_timestep")) {

        //Get the timestep value
        int iTimestep = new Integer(String.valueOf(oData)).intValue();

        //Get the last harvest event and set this as the timestep
        HarvestData oMortEpisode = (HarvestData) mp_oMortEpisodes.elementAt(
            mp_oMortEpisodes.size() - 1);
        oMortEpisode.SetTimestep(iTimestep);

        return true;
      }
      else if (sXMLTag.equals("ds_cutAmountType")) {

        //Get the cut amount type string
        String sCut = String.valueOf(oData);

        //Get the last harvest event
        HarvestData oMortEpisode = (HarvestData) mp_oMortEpisodes.elementAt(
            mp_oMortEpisodes.size() - 1);

        if (sCut.equalsIgnoreCase("percent of basal area")) {
          oMortEpisode.SetCutAmountType(PERCENTAGE_BASAL_AREA);
        }
        else if (sCut.equalsIgnoreCase("percent of density")) {
          oMortEpisode.SetCutAmountType(PERCENTAGE_DENSITY);
        }
        else if (sCut.equalsIgnoreCase("absolute basal area")) {
          oMortEpisode.SetCutAmountType(ABSOLUTE_BASAL_AREA);
        }
        else if (sCut.equalsIgnoreCase("absolute density")) {
          oMortEpisode.SetCutAmountType(ABSOLUTE_DENSITY);
        }
        else {
          //Unrecognized - error
          ModelException oErr = new ModelException();
          oErr.iErrorCode = ErrorGUI.BAD_DATA;
          oErr.sCallingFunction = "JAVA";
          oErr.sMessage =
              "Episodic mortality event in file has an unrecognized cut amount type - " +
              sCut + ".";
          throw (oErr);
        }

        return true;
      }
      else if (sXMLTag.equals("ds_low")) {

        //Extract the value
        float fVal = new Float(String.valueOf(oData)).floatValue();

        //Check to see whether the high and amount values have already been
        //read
        if (m_iCutRangeAmount > -1 && m_iCutRangeUpperBound > -1) {

          //Assign them
          HarvestData oMortEpisode = (HarvestData) mp_oMortEpisodes.elementAt(
              mp_oMortEpisodes.size() - 1);

          oMortEpisode.AddCutRange(fVal, m_iCutRangeUpperBound,
                                   m_iCutRangeAmount);
          m_iCutRangeAmount = -1;
          m_iCutRangeUpperBound = -1;
          m_iCutRangeLowerBound = -1;
        }

        //Not time to assign - stash this value
        m_iCutRangeLowerBound = fVal;

        return true;
      }
      else if (sXMLTag.equals("ds_high")) {
        //Extract the value
        float fVal = new Float(String.valueOf(oData)).floatValue();

        //Check to see whether the low and amount values have already been
        //read
        if (m_iCutRangeAmount > -1 && m_iCutRangeLowerBound > -1) {

          //Assign them
          HarvestData oMortEpisode = (HarvestData) mp_oMortEpisodes.elementAt(
              mp_oMortEpisodes.size() - 1);

          oMortEpisode.AddCutRange(m_iCutRangeLowerBound, fVal,
                                   m_iCutRangeAmount);
          m_iCutRangeAmount = -1;
          m_iCutRangeUpperBound = -1;
          m_iCutRangeLowerBound = -1;
        }

        //Not time to assign - stash this value
        m_iCutRangeUpperBound = fVal;

        return true;
      }
      else if (sXMLTag.equals("ds_amountToCut")) {
        //Extract the value
        float fVal = new Float(String.valueOf(oData)).floatValue();

        //Check to see whether the low and high bounds have already been
        //read
        if (m_iCutRangeUpperBound > -1 && m_iCutRangeLowerBound > -1) {

          //Assign them
          HarvestData oMortEpisode = (HarvestData) mp_oMortEpisodes.elementAt(
              mp_oMortEpisodes.size() - 1);

          oMortEpisode.AddCutRange(m_iCutRangeLowerBound, m_iCutRangeUpperBound,
                                   fVal);
          m_iCutRangeAmount = -1;
          m_iCutRangeUpperBound = -1;
          m_iCutRangeLowerBound = -1;
        }

        //Not time to assign - stash this value
        m_iCutRangeAmount = fVal;

        return true;
      }
      else if (sXMLTag.equals("ha_timestep")) {

        //Get the timestep value
        int iTimestep = new Integer(String.valueOf(oData)).intValue();

        //Get the last harvest event and set this as the timestep
        HarvestData oHarvest = (HarvestData) mp_oHarvestCuts.elementAt(
            mp_oHarvestCuts.size() - 1);
        oHarvest.SetTimestep(iTimestep);

        return true;
      }
      else if (sXMLTag.equals("ha_cutType")) {

        //Get the cut type string
        String sCut = String.valueOf(oData);

        //Get the last harvest event
        HarvestData oHarvest = (HarvestData) mp_oHarvestCuts.elementAt(
            mp_oHarvestCuts.size() - 1);

        if (sCut.equalsIgnoreCase("gap")) {
          oHarvest.SetCutType(GAP_CUT);
        }
        else if (sCut.equalsIgnoreCase("partial")) {
          oHarvest.SetCutType(PARTIAL_CUT);
        }
        else if (sCut.equalsIgnoreCase("clear")) {
          oHarvest.SetCutType(CLEAR_CUT);
        }
        else {
          //Unrecognized - error
          ModelException oErr = new ModelException();
          oErr.iErrorCode = ErrorGUI.BAD_DATA;
          oErr.sCallingFunction = "JAVA";
          oErr.sMessage = "Harvest cut in file has an unrecognized cut type - " +
              sCut + ".";
          throw (oErr);
        }

        return true;
      }
      else if (sXMLTag.equals("ha_cutAmountType")) {

        //Get the cut amount type string
        String sCut = String.valueOf(oData);

        //Get the last harvest event
        HarvestData oHarvest = (HarvestData) mp_oHarvestCuts.elementAt(
            mp_oHarvestCuts.size() - 1);

        if (sCut.equalsIgnoreCase("percent of basal area")) {
          oHarvest.SetCutAmountType(PERCENTAGE_BASAL_AREA);
        }
        else if (sCut.equalsIgnoreCase("percent of density")) {
          oHarvest.SetCutAmountType(PERCENTAGE_DENSITY);
        }
        else if (sCut.equalsIgnoreCase("absolute basal area")) {
          oHarvest.SetCutAmountType(ABSOLUTE_BASAL_AREA);
        }
        else if (sCut.equalsIgnoreCase("absolute density")) {
          oHarvest.SetCutAmountType(ABSOLUTE_DENSITY);
        }
        else {
          //Unrecognized - error
          ModelException oErr = new ModelException();
          oErr.iErrorCode = ErrorGUI.BAD_DATA;
          oErr.sCallingFunction = "JAVA";
          oErr.sMessage =
              "Harvest cut in file has an unrecognized cut amount type - " +
              sCut + ".";
          throw (oErr);
        }

        return true;
      }
      else if (sXMLTag.equals("ha_low")) {

        //Extract the value
        float fVal = new Float(String.valueOf(oData)).floatValue();

        //Check to see whether the high and amount values have already been
        //read
        if (m_iCutRangeAmount > -1 && m_iCutRangeUpperBound > -1) {

          //Assign them
          HarvestData oHarvest = (HarvestData) mp_oHarvestCuts.elementAt(
              mp_oHarvestCuts.size() - 1);

          oHarvest.AddCutRange(fVal, m_iCutRangeUpperBound, m_iCutRangeAmount);
          m_iCutRangeAmount = -1;
          m_iCutRangeUpperBound = -1;
          m_iCutRangeLowerBound = -1;
        }

        //Not time to assign - stash this value
        m_iCutRangeLowerBound = fVal;

        return true;
      }
      else if (sXMLTag.equals("ha_high")) {
        //Extract the value
        float fVal = new Float(String.valueOf(oData)).floatValue();

        //Check to see whether the low and amount values have already been
        //read
        if (m_iCutRangeAmount > -1 && m_iCutRangeLowerBound > -1) {

          //Assign them
          HarvestData oHarvest = (HarvestData) mp_oHarvestCuts.elementAt(
              mp_oHarvestCuts.size() - 1);

          oHarvest.AddCutRange(m_iCutRangeLowerBound, fVal, m_iCutRangeAmount);
          m_iCutRangeAmount = -1;
          m_iCutRangeUpperBound = -1;
          m_iCutRangeLowerBound = -1;
        }

        //Not time to assign - stash this value
        m_iCutRangeUpperBound = fVal;

        return true;
      }
      else if (sXMLTag.equals("ha_amountToCut")) {
        //Extract the value
        float fVal = new Float(String.valueOf(oData)).floatValue();

        //Check to see whether the low and high bounds have already been
        //read
        if (m_iCutRangeUpperBound > -1 && m_iCutRangeLowerBound > -1) {

          //Assign them
          HarvestData oHarvest = (HarvestData) mp_oHarvestCuts.elementAt(
              mp_oHarvestCuts.size() - 1);

          oHarvest.AddCutRange(m_iCutRangeLowerBound, m_iCutRangeUpperBound,
                               fVal);
          m_iCutRangeAmount = -1;
          m_iCutRangeUpperBound = -1;
          m_iCutRangeLowerBound = -1;
        }

        //Not time to assign - stash this value
        m_iCutRangeAmount = fVal;

        return true;
      }
      else if (sXMLTag.equals("sha_target_BA")) {
        //Extract the value
        float fVal = new Float(String.valueOf(oData)).floatValue();

        //Assuming we have a valid cut range counter, assign this to the
        //appropriate variable
        if (m_iSelectionHarvestCutRange == 1) {
          m_fCut1BA.SetValue(fVal);
        }
        if (m_iSelectionHarvestCutRange == 2) {
          m_fCut2BA.SetValue(fVal);
        }
        if (m_iSelectionHarvestCutRange == 3) {
          m_fCut3BA.SetValue(fVal);
        }
        if (m_iSelectionHarvestCutRange == 4) {
          m_fCut4BA.SetValue(fVal);

        }
        return true;
      }
      else if (sXMLTag.equals("sha_loDBH")) {
        //Extract the value
        float fVal = new Float(String.valueOf(oData)).floatValue();

        //Assuming we have a valid cut range counter, assign this to the
        //appropriate variable
        if (m_iSelectionHarvestCutRange == 1) {
          m_fCut1LoDBH.SetValue(fVal);
        }
        if (m_iSelectionHarvestCutRange == 2) {
          m_fCut2LoDBH.SetValue(fVal);
        }
        if (m_iSelectionHarvestCutRange == 3) {
          m_fCut3LoDBH.SetValue(fVal);
        }
        if (m_iSelectionHarvestCutRange == 4) {
          m_fCut4LoDBH.SetValue(fVal);

        }
        return true;
      }
      else if (sXMLTag.equals("sha_hiDBH")) {
        //Extract the value
        float fVal = new Float(String.valueOf(oData)).floatValue();

        //Assuming we have a valid cut range counter, assign this to the
        //appropriate variable
        if (m_iSelectionHarvestCutRange == 1) {
          m_fCut1HiDBH.SetValue(fVal);
        }
        if (m_iSelectionHarvestCutRange == 2) {
          m_fCut2HiDBH.SetValue(fVal);
        }
        if (m_iSelectionHarvestCutRange == 3) {
          m_fCut3HiDBH.SetValue(fVal);
        }
        if (m_iSelectionHarvestCutRange == 4) {
          m_fCut4HiDBH.SetValue(fVal);

        }
        return true;
      }
    }
    return super.SetSingleValueByXMLTag(sXMLTag, sXMLParentTag, oAttributes,
                                        oData);

  }
}
