package javawrapper;

/**
 * This is the organizer class for all establishment behaviors.
 * <p>Copyright: Copyright (c) Charles D. Canham 2003</p>
 * <p>Company: Institute of Ecosystem Studies</p>
 * @author Lora E. Murphy
 * @version 2.0
 *
 * <br>Edit history:
 * <br>------------------
 * <br>April 28, 2004: Submitted in beta version (LEM)
 * <br>October 13, 2004:  Added Puerto Rico establishment
 * <br>April 20, 2005:  Created new structure for establishment behaviors
 * <br>June 7, 2005:  Updated substrate favorability to match new
 * substrate (LEM)
 */

public class EstablishmentBehaviors
    extends BehaviorTypeBase {

  public final static int MOUND = 2;
  public final static int GROUND = 3;

  /**Fresh logs favorability for each species*/
  protected ModelVector[] mp_fFreshLogsFavorability;

  /**Decayed logs favorability for each species*/
  protected ModelVector[] mp_fDecayedLogsFavorability;

  /**Scarified soil favorability for each species*/
  protected ModelVector[] mp_fScarifiedSoilFavorability;

  /**Forest floor litter favorability for each species*/
  protected ModelVector[] mp_fForestFloorLitterFavorability;

  /**Forest floor moss favorability for each species*/
  protected ModelVector[] mp_fForestFloorMossFavorability;

  /**Tip-Up favorability for each species*/
  protected ModelVector[] mp_fTipUpFavorability;

  /**Light-dependent seed survival - Optimum GLI at which establishment is not
   * reduced for each species*/
  protected ModelVector mp_fOptimumGLI = new ModelVector(
      "GLI of Optimum Establishment, 0-100", "es_optimumGLI", "es_ogVal", 0,
      ModelVector.FLOAT);

  /**Light-dependent seed survival - Slope of establishment dropoff below the
   * optimum GLI for each species*/
  protected ModelVector mp_fLowSlope = new ModelVector(
      "Slope of Dropoff Below Optimum GLI", "es_lowSlope", "es_lsVal", 0,
      ModelVector.FLOAT);

  /**Light-dependent seed survival - Slope of establishment dropoff above the
   * optimum GLI for each species*/
  protected ModelVector mp_fHighSlope = new ModelVector(
      "Slope of Dropoff Above the Optimum GLI", "es_highSlope", "es_hsVal", 0,
      ModelVector.FLOAT);

  /**Light-dependent seed survival - Light extinction coefficient of
   * undamaged trees*/
  protected ModelVector mp_fLightExtCoeffNoDmg = new ModelVector(
      "Light Extinction Coeff of Undamaged Trees (0-1)",
      "li_lightExtinctionCoefficient", "li_lecVal", 0,
      ModelVector.FLOAT, true);

  /**Light-dependent seed survival - Light extinction coefficient of medium
   * damaged trees*/
  protected ModelVector mp_fLightExtCoeffPartDmg = new ModelVector(
      "Light Extinction Coeff of Medium Damage Trees (0-1)",
      "es_lightExtCoeffPartDmg", "es_lecpdVal", 0,
      ModelVector.FLOAT, true);

  /**Light-dependent seed survival - Light extinction coefficient of complete
   * damaged trees*/
  protected ModelVector mp_fLightExtCoeffFullDmg = new ModelVector(
      "Light Extinction Coeff of Complete Damage Trees (0-1)",
      "es_lightExtCoeffFullDmg", "es_lecfdVal", 0,
      ModelVector.FLOAT, true);

  /**Light-dependent seed survival - snag age class 1 light extinction
   * coefficient*/
  protected ModelVector mp_fSnagClass1LightExtinctionCoefficient = new
      ModelVector("Snag Age Class 1 Light Extinction Coefficient (0-1)",
                  "li_snag1LightExtinctionCoefficient", "li_s1lecVal", 0,
                  ModelVector.FLOAT, true);

  /**Light-dependent seed survival - snag age class 2 light extinction
   * coefficient*/
  protected ModelVector mp_fSnagClass2LightExtinctionCoefficient = new
      ModelVector("Snag Age Class 2 Light Extinction Coefficient (0-1)",
                  "li_snag2LightExtinctionCoefficient", "li_s2lecVal", 0,
                  ModelVector.FLOAT, true);

  /**Light-dependent seed survival - snag age class 3 light extinction
   * coefficient*/
  protected ModelVector mp_fSnagClass3LightExtinctionCoefficient = new
      ModelVector("Snag Age Class 3 Light Extinction Coefficient (0-1)",
                  "li_snag3LightExtinctionCoefficient", "li_s3lecVal", 0,
                  ModelVector.FLOAT, true);

  /**Density-dependent seed survival - Density-dependence slope*/
  protected ModelVector mp_fDensDepSlope = new ModelVector(
      "Slope of Density Dependence", "es_densDepSlope", "es_ddsVal", 0,
      ModelVector.FLOAT);

  /**Density-dependent seed survival - Density-dependence steepness*/
  protected ModelVector mp_fDensDepSteepness = new ModelVector(
      "Steepness of Density Dependence", "es_densDepSteepness",
      "es_ddstVal", 0,
      ModelVector.FLOAT);

  /**Proportional seed survival - Proportion of each species' seeds that
   * survives*/
  ModelVector mp_fProportionGerminating = new ModelVector(
      "Proportion Germinating Between 0 and 1",
      "ge_proportionGerminating", "ge_pgVal", 0, ModelVector.FLOAT);

  /**Light-dependent seed survival - Height at which to calculate light*/
  protected ModelFloat m_fLightHeight = new ModelFloat(0,
      "Height in m At Which to Calculate GLI", "es_lightHeight");

  /**Light-dependent seed survival - Beam fraction of global radiation*/
  protected ModelFloat m_fBeamFractionOfGlobalRadiation = new ModelFloat(0,
      "Beam Fraction of Global Radiation", "li_beamFractGlobalRad");

  /**Light-dependent seed survival - Clear sky transmission coefficient*/
  protected ModelFloat m_fClearSkyTransmissionCoefficient = new ModelFloat(0,
      "Clear Sky Transmission Coefficient", "li_clearSkyTransCoeff");

  /**Light-dependent seed survival - Minimum sun angle in radians for GLI*/
  protected ModelFloat m_fMinSunAngle = new ModelFloat(0,
      "Minimum Solar Angle for GLI Calculations, in rad", "li_minSunAngle");

  /**Proportion of the plot which is mound*/
  protected ModelFloat m_fMoundProportion = new ModelFloat(0,
      "Proportion of Plot Area that is Mound", "es_moundProportion");

  /**Mean mound height, in m*/
  protected ModelFloat m_fMeanMoundHeight = new ModelFloat(0,
      "Mean Height of Mounds, in m", "es_meanMoundHeight");

  /**Standard deviation of mound height*/
  protected ModelFloat m_fMoundStdDev = new ModelFloat(0,
      "Standard Deviation of Mound Height, in m", "es_moundStdDev");

  /**Mean log height, in m*/
  protected ModelFloat m_fMeanFreshLogHeight = new ModelFloat(0,
      "Mean Height of Fresh Log Substrate, in m", "es_meanFreshLogHeight");

  /**Standard deviation of fresh log height, in m*/
  protected ModelFloat m_fFreshLogStdDev = new ModelFloat(0,
      "Standard Deviation of Fresh Log Substrate Height, in m",
      "es_freshLogStdDev");

  /**Number of years of respite from fern shading*/
  protected ModelInt m_iMaxRespite = new ModelInt(0,
                                                  "# Years Respite from Fern Shading for Seeds on Fresh Logs",
                                                  "es_maxRespite");

  /**Light-dependent seed survival - Number of azimuth divisions for GLI*/
  protected ModelInt m_iNumAziDiv = new ModelInt(0,
                                                 "Number of Azimuth Sky Divisions for GLI Light Calculations",
                                                 "li_numAziGrids");

  /**Light-dependent seed survival - Number of altitude divisions for GLI*/
  protected ModelInt m_iNumAltDiv = new ModelInt(0,
                                                 "Number of Altitude Sky Divisions for GLI Light Calculations",
                                                 "li_numAltGrids");

  /**Light-dependent seed survival - Start of growing season as Julian day*/
  protected ModelInt m_iJulianDayGrowthStarts = new ModelInt(0,
      "First Day of Growing Season for GLI Light Calculations",
      "li_julianDayGrowthStarts");

  /**Light-dependent seed survival - End of growing season as Julian day*/
  protected ModelInt m_iJulianDayGrowthEnds = new ModelInt(0,
      "Last Day of Growing Season for GLI Light Calculations",
      "li_julianDayGrowthEnds");

  /**Light-dependent seed survival - Upper age limit of snag size class 1*/
  protected ModelInt m_iSnagAgeClass1 = new ModelInt(0,
      "Upper Age (Yrs) of Snag Light Extinction Class 1", "li_snagAgeClass1");

  /**Light-dependent seed survival - Upper age limit of snag size class 2*/
  protected ModelInt m_iSnagAgeClass2 = new ModelInt(0,
      "Upper Age (Yrs) of Snag Light Extinction Class 2", "li_snagAgeClass2");

  /**@deprecated Remove around 6.10 or so
   * Whether or not to use the seedling efficiency routine*/
  protected ModelEnum m_iUseSeedlingEfficiencyRoutine =
      new ModelEnum(new int[] {0, 1}
                    ,
                    new String[] {"false", "true"}
                    , "Use Seedling Efficiency Routine",
                    "es_useSeedlingEfficiencyRoutine");

  /**
   * Constructor
   * @param oManager GUIManager object.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   * <br>April 20, 2005:  Created new structure for establishment behaviors
   * <br>June 13, 2005:  Added microtopographic establishment
   */
  public EstablishmentBehaviors(GUIManager oManager) {
    super(oManager, "Establishment", "establishment");

    m_sHelpID = "data.establishment_behaviors";

    int i;

    //Initialize all our arrays and prep our vectors
    mp_fFreshLogsFavorability = new ModelVector[DisperseBehaviors.
        NUMBER_OF_FOREST_COVERS + 2];
    mp_fDecayedLogsFavorability = new ModelVector[DisperseBehaviors.
        NUMBER_OF_FOREST_COVERS + 2];
    mp_fScarifiedSoilFavorability = new ModelVector[DisperseBehaviors.
        NUMBER_OF_FOREST_COVERS + 2];
    mp_fForestFloorLitterFavorability = new ModelVector[DisperseBehaviors.
        NUMBER_OF_FOREST_COVERS + 2];
    mp_fForestFloorMossFavorability = new ModelVector[DisperseBehaviors.
        NUMBER_OF_FOREST_COVERS + 2];
    mp_fTipUpFavorability = new ModelVector[DisperseBehaviors.
        NUMBER_OF_FOREST_COVERS + 2];

    for (i = 0; i < DisperseBehaviors.NUMBER_OF_FOREST_COVERS + 2; i++) {
      if (DisperseBehaviors.CANOPY == i) {
        mp_fFreshLogsFavorability[i] = new ModelVector(
            "Fraction Seeds Germinating on Canopy Fresh Logs",
            "es_freshLogCanopyFav",
            "es_flcfVal", 0, ModelVector.FLOAT);
        mp_fDecayedLogsFavorability[i] = new ModelVector(
            "Fraction Seeds Germinating on Canopy Decayed Logs",
            "es_decayedLogCanopyFav",
            "es_dlcfVal", 0, ModelVector.FLOAT);
        mp_fScarifiedSoilFavorability[i] = new ModelVector(
            "Fraction Seeds Germinating on Canopy Scarified Soil",
            "es_scarifiedSoilCanopyFav", "es_sscfVal",
            0, ModelVector.FLOAT);
        mp_fForestFloorLitterFavorability[i] = new ModelVector(
            "Fraction Seeds Germinating on Canopy Forest Floor Litter",
            "es_forestFloorLitterCanopyFav",
            "es_fflcfVal", 0, ModelVector.FLOAT);
        mp_fForestFloorMossFavorability[i] = new ModelVector(
            "Fraction Seeds Germinating on Canopy Forest Floor Moss",
            "es_forestFloorMossCanopyFav",
            "es_ffmcfVal", 0, ModelVector.FLOAT);
        mp_fTipUpFavorability[i] = new ModelVector(
            "Fraction Seeds Germinating on Canopy Tip-Up",
            "es_tipUpCanopyFav", "es_tucfVal",
            0, ModelVector.FLOAT);
      }
      else if (DisperseBehaviors.GAP == i) {
        mp_fFreshLogsFavorability[i] = new ModelVector(
            "Fraction Seeds Germinating on Gap Fresh Logs", "es_freshLogGapFav",
            "es_flgfVal", 0,
            ModelVector.FLOAT);
        mp_fDecayedLogsFavorability[i] = new ModelVector(
            "Fraction Seeds Germinating on Gap Decayed Logs",
            "es_decayedLogGapFav",
            "es_dlgfVal", 0,
            ModelVector.FLOAT);
        mp_fScarifiedSoilFavorability[i] = new ModelVector(
            "Fraction Seeds Germinating on Gap Scarified Soil",
            "es_scarifiedSoilGapFav",
            "es_ssgfVal", 0,
            ModelVector.FLOAT);
        mp_fForestFloorLitterFavorability[i] = new ModelVector(
            "Fraction Seeds Germinating on Gap Forest Floor Litter",
            "es_forestFloorLitterGapFav", "es_fflgfVal",
            0, ModelVector.FLOAT);
        mp_fForestFloorMossFavorability[i] = new ModelVector(
            "Fraction Seeds Germinating on Gap Forest Floor Moss",
            "es_forestFloorMossGapFav", "es_ffmgfVal",
            0, ModelVector.FLOAT);
        mp_fTipUpFavorability[i] = new ModelVector(
            "Fraction Seeds Germinating on Gap Tip-Up", "es_tipUpGapFav",
            "es_tugfVal", 0,
            ModelVector.FLOAT);
      }
      else if (GROUND == i) {
        mp_fFreshLogsFavorability[i] = new ModelVector(
            "Fraction Seeds Germinating on Ground Fresh Logs",
            "es_freshLogCanopyFav",
            "es_flcfVal", 0, ModelVector.FLOAT);
        mp_fDecayedLogsFavorability[i] = new ModelVector(
            "Fraction Seeds Germinating on Ground Decayed Logs",
            "es_decayedLogCanopyFav",
            "es_dlcfVal", 0, ModelVector.FLOAT);
        mp_fScarifiedSoilFavorability[i] = new ModelVector(
            "Fraction Seeds Germinating on Ground Scarified Soil",
            "es_scarifiedSoilCanopyFav", "es_sscfVal",
            0, ModelVector.FLOAT);
        mp_fForestFloorLitterFavorability[i] = new ModelVector(
            "Fraction Seeds Germinating on Ground Forest Floor Litter",
            "es_forestFloorLitterCanopyFav",
            "es_fflcfVal", 0, ModelVector.FLOAT);
        mp_fForestFloorMossFavorability[i] = new ModelVector(
            "Fraction Seeds Germinating on Ground Forest Floor Moss",
            "es_forestFloorMossCanopyFav",
            "es_ffmcfVal", 0, ModelVector.FLOAT);
        mp_fTipUpFavorability[i] = new ModelVector(
            "Fraction Seeds Germinating on Ground Tip-Up",
            "es_tipUpCanopyFav", "es_tucfVal",
            0, ModelVector.FLOAT);
      }
      else if (MOUND == i) {
        mp_fFreshLogsFavorability[i] = new ModelVector(
            "Fraction Seeds Germinating on Mound Fresh Logs",
            "es_freshLogMoundFav",
            "es_flmfVal", 0, ModelVector.FLOAT);
        mp_fDecayedLogsFavorability[i] = new ModelVector(
            "Fraction Seeds Germinating on Mound Decayed Logs",
            "es_decayedLogMoundFav",
            "es_dlmfVal", 0, ModelVector.FLOAT);
        mp_fScarifiedSoilFavorability[i] = new ModelVector(
            "Fraction Seeds Germinating on Mound Scarified Soil",
            "es_scarifiedSoilMoundFav", "es_ssmfVal",
            0, ModelVector.FLOAT);
        mp_fForestFloorLitterFavorability[i] = new ModelVector(
            "Fraction Seeds Germinating on Mound Forest Floor Litter",
            "es_forestFloorLitterMoundFav",
            "es_fflmfVal", 0, ModelVector.FLOAT);
        mp_fForestFloorMossFavorability[i] = new ModelVector(
            "Fraction Seeds Germinating on Mound Forest Floor Moss",
            "es_forestFloorMossMoundFav",
            "es_ffmmfVal", 0, ModelVector.FLOAT);
        mp_fTipUpFavorability[i] = new ModelVector(
            "Fraction Seeds Germinating on Mound Tip-Up",
            "es_tipUpMoundFav", "es_tumfVal",
            0, ModelVector.FLOAT);
      }
    }

    //Set up behavior list vector
    mp_oChildBehaviors = new Behavior[8];

    //No gap substrate-dependent seed survival
    mp_oChildBehaviors[0] = new Behavior("no gap substrate seed survival",
                                         "Substrate Dependent Seed Survival No Gap Status",
                                         "No Gap Substrate Seed Survival",
                                         (float) 1.1);
    mp_oChildBehaviors[0].AddRequiredData(mp_fFreshLogsFavorability[
                                          DisperseBehaviors.CANOPY]);
    mp_oChildBehaviors[0].AddRequiredData(mp_fDecayedLogsFavorability[
                                          DisperseBehaviors.CANOPY]);
    mp_oChildBehaviors[0].AddRequiredData(mp_fTipUpFavorability[
                                          DisperseBehaviors.CANOPY]);
    mp_oChildBehaviors[0].AddRequiredData(mp_fScarifiedSoilFavorability[
                                          DisperseBehaviors.CANOPY]);
    mp_oChildBehaviors[0].AddRequiredData(mp_fForestFloorLitterFavorability[
                                          DisperseBehaviors.CANOPY]);
    mp_oChildBehaviors[0].AddRequiredData(mp_fForestFloorMossFavorability[
                                          DisperseBehaviors.CANOPY]);
    //Allow for seeds, disallow for all others
    mp_oChildBehaviors[0].SetCanApplyTo(TreePopulation.SEED, true);
    mp_oChildBehaviors[0].SetCanApplyTo(TreePopulation.SEEDLING, false);
    mp_oChildBehaviors[0].SetCanApplyTo(TreePopulation.SAPLING, false);
    mp_oChildBehaviors[0].SetCanApplyTo(TreePopulation.ADULT, false);

    //Gap substrate-dependent seed survival
    mp_oChildBehaviors[1] = new Behavior("gap substrate seed survival",
                                         "Substrate Dependent Seed Survival With Gap Status",
                                         "Gap Substrate Seed Survival",
                                         (float) 1.1);
    for (i = 0; i < DisperseBehaviors.NUMBER_OF_FOREST_COVERS; i++) {
      mp_oChildBehaviors[1].AddRequiredData(mp_fFreshLogsFavorability[i]);
      mp_oChildBehaviors[1].AddRequiredData(mp_fDecayedLogsFavorability[i]);
      mp_oChildBehaviors[1].AddRequiredData(mp_fTipUpFavorability[i]);
      mp_oChildBehaviors[1].AddRequiredData(mp_fScarifiedSoilFavorability[i]);
      mp_oChildBehaviors[1].AddRequiredData(mp_fForestFloorLitterFavorability[i]);
      mp_oChildBehaviors[1].AddRequiredData(mp_fForestFloorMossFavorability[i]);
    }
    //Allow for seeds, disallow for all others
    mp_oChildBehaviors[1].SetCanApplyTo(TreePopulation.SEED, true);
    mp_oChildBehaviors[1].SetCanApplyTo(TreePopulation.SEEDLING, false);
    mp_oChildBehaviors[1].SetCanApplyTo(TreePopulation.SAPLING, false);
    mp_oChildBehaviors[1].SetCanApplyTo(TreePopulation.ADULT, false);

    //Microtopographic substrate seed survival
    mp_oChildBehaviors[2] = new Behavior("micro substrate seed survival",
                                         "Substrate Based Seed Survival With Microtopography",
                                         "Microtopographic Substrate Seed Survival",
                                         (float) 1.1);
    mp_oChildBehaviors[2].AddRequiredData(mp_fFreshLogsFavorability[GROUND]);
    mp_oChildBehaviors[2].AddRequiredData(mp_fDecayedLogsFavorability[GROUND]);
    mp_oChildBehaviors[2].AddRequiredData(mp_fTipUpFavorability[GROUND]);
    mp_oChildBehaviors[2].AddRequiredData(mp_fScarifiedSoilFavorability[GROUND]);
    mp_oChildBehaviors[2].AddRequiredData(mp_fForestFloorLitterFavorability[
                                          GROUND]);
    mp_oChildBehaviors[2].AddRequiredData(mp_fForestFloorMossFavorability[
                                          GROUND]);

    mp_oChildBehaviors[2].AddRequiredData(mp_fFreshLogsFavorability[MOUND]);
    mp_oChildBehaviors[2].AddRequiredData(mp_fDecayedLogsFavorability[MOUND]);
    mp_oChildBehaviors[2].AddRequiredData(mp_fTipUpFavorability[MOUND]);
    mp_oChildBehaviors[2].AddRequiredData(mp_fScarifiedSoilFavorability[MOUND]);
    mp_oChildBehaviors[2].AddRequiredData(mp_fForestFloorLitterFavorability[
                                          MOUND]);
    mp_oChildBehaviors[2].AddRequiredData(mp_fForestFloorMossFavorability[MOUND]);

    mp_oChildBehaviors[2].AddRequiredData(m_fMoundProportion);

    //Allow for seeds, disallow for all others
    mp_oChildBehaviors[2].SetCanApplyTo(TreePopulation.SEED, true);
    mp_oChildBehaviors[2].SetCanApplyTo(TreePopulation.SEEDLING, false);
    mp_oChildBehaviors[2].SetCanApplyTo(TreePopulation.SAPLING, false);
    mp_oChildBehaviors[2].SetCanApplyTo(TreePopulation.ADULT, false);

    //Light-dependent seed survival
    mp_oChildBehaviors[3] = new Behavior("light seed survival",
                                         "Light Dependent Seed Survival",
                                         "Light Dependent Seed Survival",
                                         (float) 1);
    mp_oChildBehaviors[3].AddRequiredData(mp_fOptimumGLI);
    mp_oChildBehaviors[3].AddRequiredData(mp_fLowSlope);
    mp_oChildBehaviors[3].AddRequiredData(mp_fHighSlope);
    mp_oChildBehaviors[3].AddRequiredData(mp_fLightExtCoeffNoDmg);
    mp_oChildBehaviors[3].AddRequiredData(mp_fLightExtCoeffPartDmg);
    mp_oChildBehaviors[3].AddRequiredData(mp_fLightExtCoeffFullDmg);
    mp_oChildBehaviors[3].AddRequiredData(
        mp_fSnagClass1LightExtinctionCoefficient);
    mp_oChildBehaviors[3].AddRequiredData(
        mp_fSnagClass2LightExtinctionCoefficient);
    mp_oChildBehaviors[3].AddRequiredData(
        mp_fSnagClass3LightExtinctionCoefficient);
    mp_oChildBehaviors[3].AddRequiredData(m_fLightHeight);
    mp_oChildBehaviors[3].AddRequiredData(m_fBeamFractionOfGlobalRadiation);
    mp_oChildBehaviors[3].AddRequiredData(m_fClearSkyTransmissionCoefficient);
    mp_oChildBehaviors[3].AddRequiredData(m_iJulianDayGrowthStarts);
    mp_oChildBehaviors[3].AddRequiredData(m_iJulianDayGrowthEnds);
    mp_oChildBehaviors[3].AddRequiredData(m_iNumAltDiv);
    mp_oChildBehaviors[3].AddRequiredData(m_iNumAziDiv);
    mp_oChildBehaviors[3].AddRequiredData(m_fMinSunAngle);
    mp_oChildBehaviors[3].AddRequiredData(m_iSnagAgeClass1);
    mp_oChildBehaviors[3].AddRequiredData(m_iSnagAgeClass2);
    //Allow for seeds, disallow for all others
    mp_oChildBehaviors[3].SetCanApplyTo(TreePopulation.SEED, true);
    mp_oChildBehaviors[3].SetCanApplyTo(TreePopulation.SEEDLING, false);
    mp_oChildBehaviors[3].SetCanApplyTo(TreePopulation.SAPLING, false);
    mp_oChildBehaviors[3].SetCanApplyTo(TreePopulation.ADULT, false);

    //Density-dependent seed survival
    mp_oChildBehaviors[4] = new Behavior("density seed survival",
                                         "Density Dependent Seed Survival",
                                         "Density Dependent Seed Survival",
                                         (float) 1);
    mp_oChildBehaviors[4].AddRequiredData(mp_fDensDepSlope);
    mp_oChildBehaviors[4].AddRequiredData(mp_fDensDepSteepness);
    //Allow for seeds, disallow for all others
    mp_oChildBehaviors[4].SetCanApplyTo(TreePopulation.SEED, true);
    mp_oChildBehaviors[4].SetCanApplyTo(TreePopulation.SEEDLING, false);
    mp_oChildBehaviors[4].SetCanApplyTo(TreePopulation.SAPLING, false);
    mp_oChildBehaviors[4].SetCanApplyTo(TreePopulation.ADULT, false);

    //Proportional seed survival
    mp_oChildBehaviors[5] = new Behavior("germination",
                                         "Proportional Seed Survival",
                                         "germination", (float) 1);
    //Disallow for all but seeds
    mp_oChildBehaviors[5].SetCanApplyTo(TreePopulation.SEED, true);
    mp_oChildBehaviors[5].SetCanApplyTo(TreePopulation.SEEDLING, false);
    mp_oChildBehaviors[5].SetCanApplyTo(TreePopulation.SAPLING, false);
    mp_oChildBehaviors[5].SetCanApplyTo(TreePopulation.ADULT, false);

    //Add the germination proportion
    mp_oChildBehaviors[5].AddRequiredData(mp_fProportionGerminating);

    //Establishment
    mp_oChildBehaviors[6] = new Behavior("establishment",
                                         "Seed Establishment",
                                         "Establishment", (float) 1);
    //Disallow for all but seeds
    mp_oChildBehaviors[6].SetCanApplyTo(TreePopulation.SEED, true);
    mp_oChildBehaviors[6].SetCanApplyTo(TreePopulation.SEEDLING, false);
    mp_oChildBehaviors[6].SetCanApplyTo(TreePopulation.SAPLING, false);
    mp_oChildBehaviors[6].SetCanApplyTo(TreePopulation.ADULT, false);

    //Establishment with Microtopography
    mp_oChildBehaviors[7] = new Behavior("micro establishment",
                                         "Establishment with Microtopography",
                                         "Micro Establishment", (float) 1);
    mp_oChildBehaviors[7].AddRequiredData(m_fMoundProportion);
    mp_oChildBehaviors[7].AddRequiredData(m_fMeanMoundHeight);
    mp_oChildBehaviors[7].AddRequiredData(m_fMoundStdDev);
    mp_oChildBehaviors[7].AddRequiredData(m_fMeanFreshLogHeight);
    mp_oChildBehaviors[7].AddRequiredData(m_fFreshLogStdDev);
    mp_oChildBehaviors[7].AddRequiredData(m_iMaxRespite);
    //Disallow for all but seeds
    mp_oChildBehaviors[7].SetCanApplyTo(TreePopulation.SEED, true);
    mp_oChildBehaviors[7].SetCanApplyTo(TreePopulation.SEEDLING, false);
    mp_oChildBehaviors[7].SetCanApplyTo(TreePopulation.SAPLING, false);
    mp_oChildBehaviors[7].SetCanApplyTo(TreePopulation.ADULT, false);

    //All data - in the order it will be written in XML
    //Canopy favorabilities
    mp_oAllData.add(mp_fScarifiedSoilFavorability[DisperseBehaviors.CANOPY]);
    mp_oAllData.add(mp_fTipUpFavorability[DisperseBehaviors.CANOPY]);
    mp_oAllData.add(mp_fFreshLogsFavorability[DisperseBehaviors.CANOPY]);
    mp_oAllData.add(mp_fDecayedLogsFavorability[DisperseBehaviors.CANOPY]);
    mp_oAllData.add(mp_fForestFloorLitterFavorability[DisperseBehaviors.CANOPY]);
    mp_oAllData.add(mp_fForestFloorMossFavorability[DisperseBehaviors.CANOPY]);
    //Gap favorabilities
    mp_oAllData.add(mp_fScarifiedSoilFavorability[DisperseBehaviors.GAP]);
    mp_oAllData.add(mp_fTipUpFavorability[DisperseBehaviors.GAP]);
    mp_oAllData.add(mp_fFreshLogsFavorability[DisperseBehaviors.GAP]);
    mp_oAllData.add(mp_fDecayedLogsFavorability[DisperseBehaviors.GAP]);
    mp_oAllData.add(mp_fForestFloorLitterFavorability[DisperseBehaviors.GAP]);
    mp_oAllData.add(mp_fForestFloorMossFavorability[DisperseBehaviors.GAP]);
    //Ground favorabilities
    mp_oAllData.add(mp_fScarifiedSoilFavorability[GROUND]);
    mp_oAllData.add(mp_fTipUpFavorability[GROUND]);
    mp_oAllData.add(mp_fFreshLogsFavorability[GROUND]);
    mp_oAllData.add(mp_fDecayedLogsFavorability[GROUND]);
    mp_oAllData.add(mp_fForestFloorLitterFavorability[GROUND]);
    mp_oAllData.add(mp_fForestFloorMossFavorability[GROUND]);
    //Mound favorabilities
    mp_oAllData.add(mp_fScarifiedSoilFavorability[MOUND]);
    mp_oAllData.add(mp_fTipUpFavorability[MOUND]);
    mp_oAllData.add(mp_fFreshLogsFavorability[MOUND]);
    mp_oAllData.add(mp_fDecayedLogsFavorability[MOUND]);
    mp_oAllData.add(mp_fForestFloorLitterFavorability[MOUND]);
    mp_oAllData.add(mp_fForestFloorMossFavorability[MOUND]);

    mp_oAllData.add(mp_fOptimumGLI);
    mp_oAllData.add(mp_fLowSlope);
    mp_oAllData.add(mp_fHighSlope);
    mp_oAllData.add(mp_fLightExtCoeffNoDmg);
    mp_oAllData.add(mp_fLightExtCoeffPartDmg);
    mp_oAllData.add(mp_fLightExtCoeffFullDmg);
    mp_oAllData.add(mp_fSnagClass1LightExtinctionCoefficient);
    mp_oAllData.add(mp_fSnagClass2LightExtinctionCoefficient);
    mp_oAllData.add(mp_fSnagClass3LightExtinctionCoefficient);
    mp_oAllData.add(m_fLightHeight);
    mp_oAllData.add(m_fBeamFractionOfGlobalRadiation);
    mp_oAllData.add(m_fClearSkyTransmissionCoefficient);
    mp_oAllData.add(m_iJulianDayGrowthStarts);
    mp_oAllData.add(m_iJulianDayGrowthEnds);
    mp_oAllData.add(m_iNumAltDiv);
    mp_oAllData.add(m_iNumAziDiv);
    mp_oAllData.add(m_fMinSunAngle);
    mp_oAllData.add(m_iSnagAgeClass1);
    mp_oAllData.add(m_iSnagAgeClass2);
    mp_oAllData.add(mp_fDensDepSlope);
    mp_oAllData.add(mp_fDensDepSteepness);
    mp_oAllData.add(mp_fProportionGerminating);
    mp_oAllData.add(m_fMoundProportion);
    mp_oAllData.add(m_fMeanMoundHeight);
    mp_oAllData.add(m_fMoundStdDev);
    mp_oAllData.add(m_fMeanFreshLogHeight);
    mp_oAllData.add(m_fFreshLogStdDev);
    mp_oAllData.add(m_iMaxRespite);

    //Deprecated - for backwards compatibility
    mp_oAllData.add(m_iUseSeedlingEfficiencyRoutine);

  }

  /**
   * Does setup.  Sets up the substrate favorability grid.
   * @param oPop TreePopulation object.
   * @throws ModelException if there's a problem setting behavior use data.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   * <br>June 7, 2005:  Added new substrate defaults (LEM)
   */
  public void DoSetup(TreePopulation oPop) throws ModelException {
    int iNumSpecies = oPop.GetNumberOfSpecies(), i, k;

    //Set up the substrate favorability grid
    DataMember[] p_oDataMembers = new DataMember[iNumSpecies];
    mp_oGrids = new Grid[1];

    String sGridName = "Substrate Favorability";

    //The accessible data members are one for each species
    for (i = 0; i < iNumSpecies; i++) {

      p_oDataMembers[i] = new DataMember("Favorability Index - " +
                                         oPop.GetSpeciesNameFromCode(i).
                                         replace('_', ' '),
                                         "Favorability Index" + i,
                                         DataMember.FLOAT);

    }

    //Create our grid
    mp_oGrids[0] = new Grid(sGridName, p_oDataMembers, null, 8, 8);

    GetBehaviorByKey("no gap substrate seed survival").AddGrid(mp_oGrids[0]);
    GetBehaviorByKey("gap substrate seed survival").AddGrid(mp_oGrids[0]);
    GetBehaviorByKey("micro substrate seed survival").AddGrid(mp_oGrids[0]);

    for (i = 0; i < mp_fFreshLogsFavorability.length; i++) {
      if (mp_fFreshLogsFavorability[i].GetValue().size() == 0) {
        for (k = 0; k < iNumSpecies; k++) {
          mp_fFreshLogsFavorability[i].GetValue().add(new Float(0));
        }
      }
      if (mp_fDecayedLogsFavorability[i].GetValue().size() == 0) {
        for (k = 0; k < iNumSpecies; k++) {
          mp_fDecayedLogsFavorability[i].GetValue().add(new Float(0));
        }
      }
      if (mp_fTipUpFavorability[i].GetValue().size() == 0) {
        for (k = 0; k < iNumSpecies; k++) {
          mp_fTipUpFavorability[i].GetValue().add(new Float(0));
        }
      }
      if (mp_fScarifiedSoilFavorability[i].GetValue().size() == 0) {
        for (k = 0; k < iNumSpecies; k++) {
          mp_fScarifiedSoilFavorability[i].GetValue().add(new Float(0));
        }
      }
      if (mp_fForestFloorLitterFavorability[i].GetValue().size() == 0) {
        for (k = 0; k < iNumSpecies; k++) {
          mp_fForestFloorLitterFavorability[i].GetValue().add(new Float(0));
        }
      }
      if (mp_fForestFloorMossFavorability[i].GetValue().size() == 0) {
        for (k = 0; k < iNumSpecies; k++) {
          mp_fForestFloorMossFavorability[i].GetValue().add(new Float(0));
        }
      }
    }
  }

  /**
   * Validates data.
   * @param oPop TreePopulation object.
   * @throws ModelException if:
   * <ul>
   * <li>Light-dependent seed survival is enabled, and any of the light
   * extinction coefficient values are not between 0 and 1.</li>
   * <li>Light-dependent seed survival is enabled, and the value for
   * m_fLightHeight is not zero or greater.</li>
   * <li>Light-dependent seed survival is enabled, and the values for
   * mp_fOptimumGLI are not between 0 and 100.</li>
   * <li>Light-dependent seed survival is enabled, and
   * m_fBeamFractionOfGlobalRadiation is not a valid proportion.</li>
   * <li>Light-dependent seed survival is enabled, and either
   * m_iJulianDayGrowthStarts or m_iJulianDayGrowthEnds is not between 1 and
   * 365 (inclusive)</li>
   * <li>Light-dependent seed survival is enabled, and either m_iNumAltDiv or
   * m_iNumAziDiv is not greater than 0</li>
   * <li>Light-dependent seed survival is enabled, and
   * m_fClearSkyTransmissionCoefficient is 0.</li>
   * <li>if proportional seed survival is enabled and all
   * values in mp_fProportionGerminating are not proportions.</li>
   * <li>Substrate-dependent seed survival is enabled and a favorability value
   * is not a proportion.</li>
   * <li>Microtopographic seed survival is enabled and the proportion of mound
   * value isn't a proportion</li>
   * <li>If the run has snags in it and the values for the snag light
   * extinction coefficients aren't proportions, or the snag age classes aren't
   * positive numbers.</li>
   * <li>Microtopographic establishment is enabled and:  proportion of mound
   * value isn't a proportion, or the Beer's law light filter isn't enabled,
   * or substrate isn't enabled</li>
   * </ul>
   *
   * This will also make sure that establishment is enabled if any of the
   * others are.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   * <br>June 7, 2005:  Added new substrate values (LEM)
   */
  public void ValidateData(TreePopulation oPop) throws ModelException {
    Behavior oBeh;
    boolean[] p_bAppliesTo = new boolean[oPop.GetNumberOfSpecies()];
    java.util.Vector p_oValues;

    if (!AnyBehaviorsEnabled()) {
      return;
    }

    //If neither establishment is enabled, enable basic establishment and
    //assign it all the same species that are used elsewhere
    oBeh = GetBehaviorByKey("establishment");
    if (oBeh.m_bIsEnabled == false && GetBehaviorByKey("micro establishment").m_bIsEnabled == false) {
      int i;
      for (i = 0; i < p_bAppliesTo.length; i++) {
        p_bAppliesTo[i] = false;
      }
      for (i = 0; i < mp_oChildBehaviors.length; i++) {
        if (mp_oChildBehaviors[i].m_sKey.equals("establishment") == false &&
            mp_oChildBehaviors[i].m_bIsEnabled) {
          p_bAppliesTo = MergeBooleans(p_bAppliesTo,
                                       mp_oChildBehaviors[i].
                                       GetWhichSpeciesUsed(oPop));
        }
      }

      oBeh.ClearSpeciesTypeCombos();
      for (i = 0; i < p_bAppliesTo.length; i++) {
        oBeh.AddSpeciesTypeCombo(new SpeciesTypeCombo(i, TreePopulation.SEED,
            oPop));
      }
    }
    //Make sure that the actual establishment behaviors are in the last
    //position
    if (mp_oChildBehaviors[mp_oChildBehaviors.length - 1].equals(oBeh) == false) {
      int i, iIndex = 0;
      for (i = 0; i < mp_oChildBehaviors.length; i++) {
        if (mp_oChildBehaviors[i].equals(oBeh)) {
          iIndex = i;
        }
      }
      for (i = iIndex + 1; i < mp_oChildBehaviors.length; i++) {
        mp_oChildBehaviors[i - 1] = mp_oChildBehaviors[i];
      }
      mp_oChildBehaviors[mp_oChildBehaviors.length - 1] = oBeh;
    }

    oBeh = GetBehaviorByKey("micro establishment");
    if (mp_oChildBehaviors[mp_oChildBehaviors.length - 1].equals(oBeh) == false) {
      int i, iIndex = 0;
      for (i = 0; i < mp_oChildBehaviors.length; i++) {
        if (mp_oChildBehaviors[i].equals(oBeh)) {
          iIndex = i;
        }
      }
      for (i = iIndex + 1; i < mp_oChildBehaviors.length; i++) {
        mp_oChildBehaviors[i - 1] = mp_oChildBehaviors[i];
      }
      mp_oChildBehaviors[mp_oChildBehaviors.length - 1] = oBeh;
    }

    //Light-dependent validation
    oBeh = GetBehaviorByKey("light seed survival");
    if (oBeh.m_bIsEnabled) {
      MakeSureNonNegative(m_fLightHeight);
      p_bAppliesTo = oBeh.GetWhichSpeciesUsed(oPop);
      p_oValues = mp_fOptimumGLI.GetValue();
      int i;
      for (i = 0; i < p_oValues.size(); i++) {
        if (p_bAppliesTo[i]) {
          Float fValue = (Float) p_oValues.elementAt(i);
          if (fValue.floatValue() < 0 || fValue.floatValue() > 100) {
            throw (new ModelException(ErrorGUI.BAD_DATA, "JAVA",
                                      "All values in " +
                                      mp_fOptimumGLI.GetDescriptor() +
                                      " must be between 0 and 100."));
          }
        }
      }
      MakeSureAllAreProportions(mp_fLightExtCoeffNoDmg);
      MakeSureAllAreProportions(mp_fLightExtCoeffPartDmg);
      MakeSureAllAreProportions(mp_fLightExtCoeffFullDmg);
      MakeSurePositive(m_fClearSkyTransmissionCoefficient);
      MakeSureIsProportion(m_fBeamFractionOfGlobalRadiation);
      if (m_iJulianDayGrowthStarts.GetValue() < 1 ||
          m_iJulianDayGrowthStarts.GetValue() > 365 ||
          m_iJulianDayGrowthEnds.GetValue() < 1 ||
          m_iJulianDayGrowthEnds.GetValue() > 365) {
        ModelException oErr = new ModelException();
        oErr.iErrorCode = ErrorGUI.BAD_DATA;
        oErr.sCallingFunction = "JAVA";
        oErr.sMessage = m_iJulianDayGrowthStarts.GetDescriptor() + " and " +
            m_iJulianDayGrowthEnds.GetDescriptor() +
            " must have values between 1 and 365.";
        throw (oErr);
      }
      MakeSurePositive(m_iNumAltDiv);
      MakeSurePositive(m_iNumAziDiv);
      MakeSurePositive(m_fMinSunAngle);

      //If there are any snags enabled anywhere, make sure that the snag data is
      //filled out
      if (m_oManager.GetSnagAwareness()) {
        MakeSureAllAreProportions(mp_fSnagClass1LightExtinctionCoefficient);
        MakeSureAllAreProportions(mp_fSnagClass2LightExtinctionCoefficient);
        MakeSureAllAreProportions(mp_fSnagClass3LightExtinctionCoefficient);
        MakeSurePositive(m_iSnagAgeClass1);
        MakeSurePositive(m_iSnagAgeClass2);
        if (m_iSnagAgeClass2.GetValue() <= m_iSnagAgeClass1.GetValue()) {
          throw new ModelException(ErrorGUI.BAD_DATA, "JAVA",
                                   m_iSnagAgeClass2.GetDescriptor() +
                                   " must be greater than " +
                                   m_iSnagAgeClass1.GetDescriptor());
        }
      }
    }

    //Substrate-dependent
    if (GetBehaviorByKey("no gap substrate seed survival").m_bIsEnabled ||
        GetBehaviorByKey("gap substrate seed survival").m_bIsEnabled) {

      //Canopy favorabilities
      oBeh = GetBehaviorByKey("no gap substrate seed survival");
      int i;
      for (i = 0; i < p_bAppliesTo.length; i++) {
        p_bAppliesTo[i] = false;
      }
      if (oBeh.m_bIsEnabled == true) {
        p_bAppliesTo = oBeh.GetWhichSpeciesUsed(oPop);
      }
      oBeh = GetBehaviorByKey("gap substrate seed survival");
      if (oBeh.m_bIsEnabled == true) {
        p_bAppliesTo = MergeBooleans(p_bAppliesTo,
                                     oBeh.GetWhichSpeciesUsed(oPop));

      }
      MakeSureAllAreProportions(mp_fDecayedLogsFavorability[DisperseBehaviors.
                                CANOPY], p_bAppliesTo);
      MakeSureAllAreProportions(mp_fFreshLogsFavorability[DisperseBehaviors.
                                CANOPY], p_bAppliesTo);
      MakeSureAllAreProportions(mp_fScarifiedSoilFavorability[DisperseBehaviors.
                                CANOPY], p_bAppliesTo);
      MakeSureAllAreProportions(mp_fForestFloorLitterFavorability[
                                DisperseBehaviors.
                                CANOPY], p_bAppliesTo);
      MakeSureAllAreProportions(mp_fForestFloorMossFavorability[
                                DisperseBehaviors.
                                CANOPY], p_bAppliesTo);
      MakeSureAllAreProportions(mp_fTipUpFavorability[DisperseBehaviors.CANOPY],
                                p_bAppliesTo);

      oBeh = GetBehaviorByKey("gap substrate seed survival");
      if (oBeh.m_bIsEnabled == true) {
        p_bAppliesTo = oBeh.GetWhichSpeciesUsed(oPop);

        MakeSureAllAreProportions(mp_fDecayedLogsFavorability[DisperseBehaviors.
                                  GAP], p_bAppliesTo);
        MakeSureAllAreProportions(mp_fFreshLogsFavorability[DisperseBehaviors.
                                  GAP], p_bAppliesTo);
        MakeSureAllAreProportions(mp_fScarifiedSoilFavorability[
                                  DisperseBehaviors.GAP], p_bAppliesTo);
        MakeSureAllAreProportions(mp_fForestFloorLitterFavorability[
                                  DisperseBehaviors.
                                  GAP], p_bAppliesTo);
        MakeSureAllAreProportions(mp_fForestFloorMossFavorability[
                                  DisperseBehaviors.
                                  GAP], p_bAppliesTo);
        MakeSureAllAreProportions(mp_fTipUpFavorability[DisperseBehaviors.
                                  GAP], p_bAppliesTo);
      }
    }

    oBeh = GetBehaviorByKey("micro substrate seed survival");
    if (oBeh.m_bIsEnabled == true) {
      p_bAppliesTo = oBeh.GetWhichSpeciesUsed(oPop);

      MakeSureAllAreProportions(mp_fDecayedLogsFavorability[MOUND],
                                p_bAppliesTo);
      MakeSureAllAreProportions(mp_fFreshLogsFavorability[MOUND], p_bAppliesTo);
      MakeSureAllAreProportions(mp_fScarifiedSoilFavorability[MOUND],
                                p_bAppliesTo);
      MakeSureAllAreProportions(mp_fForestFloorLitterFavorability[MOUND],
                                p_bAppliesTo);
      MakeSureAllAreProportions(mp_fForestFloorMossFavorability[MOUND],
                                p_bAppliesTo);
      MakeSureAllAreProportions(mp_fTipUpFavorability[MOUND], p_bAppliesTo);

      MakeSureAllAreProportions(mp_fDecayedLogsFavorability[GROUND],
                                p_bAppliesTo);
      MakeSureAllAreProportions(mp_fFreshLogsFavorability[GROUND], p_bAppliesTo);
      MakeSureAllAreProportions(mp_fScarifiedSoilFavorability[GROUND],
                                p_bAppliesTo);
      MakeSureAllAreProportions(mp_fForestFloorLitterFavorability[GROUND],
                                p_bAppliesTo);
      MakeSureAllAreProportions(mp_fForestFloorMossFavorability[GROUND],
                                p_bAppliesTo);
      MakeSureAllAreProportions(mp_fTipUpFavorability[GROUND], p_bAppliesTo);

      MakeSureIsProportion(m_fMoundProportion);
    }

    //Proportional seed survival
    oBeh = GetBehaviorByKey("germination");
    if (oBeh.m_bIsEnabled == true) {
      p_bAppliesTo = oBeh.GetWhichSpeciesUsed(oPop);

      MakeSureAllAreProportions(mp_fProportionGerminating, p_bAppliesTo);
    }

    //Microtopographic establishment
    oBeh = GetBehaviorByKey("micro establishment");
    if (oBeh.m_bIsEnabled == true) {
      MakeSureIsProportion(m_fMoundProportion);

      //Make sure that the light filter behavior is enabled
      if (m_oManager.GetLightBehaviors().GetBehaviorByKey("light filter").m_bIsEnabled == false) {
        throw new ModelException(ErrorGUI.CANT_FIND_OBJECT, "JAVA",
                                 "You must use the \""+
                                 m_oManager.GetLightBehaviors().GetBehaviorByKey("light filter").GetDescriptor()+
                                 "\" behavior with the \""+
                                 oBeh.GetDescriptor() + "\" behavior.");
      }

      //Make sure that the substrate behavior is enabled
      if (m_oManager.GetSubstrateBehaviors().GetBehaviorByKey("substrate").m_bIsEnabled == false) {
        throw new ModelException(ErrorGUI.CANT_FIND_OBJECT, "JAVA",
                                 "You must use the \""+
                                 m_oManager.GetSubstrateBehaviors().GetBehaviorByKey("substrate").GetDescriptor()+
                                 "\" behavior with the \""+
                                 oBeh.GetDescriptor() + "\" behavior.");
      }
    }

  }

  /**
   * Traps for the changed behavior names.  This could be removed probably
   * by about 6.10.
   * @param sXMLTag String The XML tag for which to find a behavior.
   * @return Behavior Behavior for the XML tag, or NULL if none of the
   * behaviors has that tag.
   */
  public Behavior GetBehaviorByXMLTag(String sXMLTag) {

    if (sXMLTag.equals("no gap no substrate establishment")) {
      return GetBehaviorByKey("establishment");
    }
    else if (sXMLTag.equals("no gap substrate establishment")) {
      return GetBehaviorByKey("no gap substrate seed survival");
    }
    else if (sXMLTag.equals("gap and substrate establishment")) {
      return GetBehaviorByKey("gap substrate seed survival");
    }
    return super.GetBehaviorByXMLTag(sXMLTag);
  }

  /**
   * Find light values for PR Establishment and reject those for light
   * behaviors.
   * @param sXMLTag XML tag of object to assign.
   * @param sXMLParentTag The immediate parent tag that sXMLTag is within.
   * @param oAttributes Ignored.
   * @param oData Data to assign.
   * @return true if assigned, or false if not.
   * @throws ModelException not thrown here.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public boolean SetSingleValueByXMLTag(String sXMLTag, String sXMLParentTag,
                                        org.xml.sax.Attributes oAttributes,
                                        Object oData) throws
      ModelException {

    if (sXMLParentTag.equals(m_sXMLTag) ||
        sXMLParentTag.equals("ma_v") ||
        sXMLParentTag.equals("pkg") ||
        sXMLParentTag.equals("ma_intCodes") ||
        sXMLParentTag.equals("ma_floatCodes") ||
        sXMLParentTag.equals("ma_charCodes") ||
        sXMLParentTag.equals("ma_boolCodes") ||
        sXMLParentTag.equals("ma_packageIntCodes") ||
        sXMLParentTag.equals("ma_packageFloatCodes") ||
        sXMLParentTag.equals("ma_packageCharCodes") ||
        sXMLParentTag.equals("ma_packageBoolCodes") ||
        sXMLParentTag.equals("grid")) {
      return super.SetSingleValueByXMLTag(sXMLTag, sXMLParentTag,
                                          oAttributes, oData);
    }
    return false;
  }

  /**
   * Overridden for backwards compatibility and to make sure that only light
   * values that apply to light behaviors are set.  There are other behaviors
   * with light values, and this will make sure that no behavior "steals" these
   * parameters.
   * @param sXMLTag Parent XML tag of data vector whose value is to be set.
   * @param sXMLParentTag The immediate parent tag that sXMLTag is within.
   * @param p_oData Vector of data values appropriate to the data type
   * @param p_sChildXMLTags The XML tags of the child elements
   * @param p_bAppliesTo Array of booleans saying which of the vector values
   * should be set.  This is important in the case of species-specifics - the
   * vector index is the species number but not all species are set.
   * @param oParentAttributes Attributes of parent tag.  May be useful when
   * overridding this for unusual tags.
   * @param p_oAttributes Attributes passed from parser.  This may be needed
   * when overriding this function.  Basic species-specific values are
   * already handled by this function.
   * @return true if the value was set successfully; false if the value could
   * not be found.
   * @throws ModelException if the value could not be assigned to the data
   * object.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>January 7, 2005: Created (LEM)
   * <br>April 20, 2005:  Updated for backwards compatibility with substrate
   * (LEM)
   */
  public boolean SetVectorValueByXMLTag(String sXMLTag, String sXMLParentTag,
                                        java.util.Vector p_oData,
                                        String[] p_sChildXMLTags,
                                        boolean[] p_bAppliesTo,
                                        org.xml.sax.Attributes
                                        oParentAttributes,
                                        org.xml.sax.Attributes[] p_oAttributes) throws
      ModelException {

    int i;

    //Capture the changed substrate favorability codes and change them to the
    //new tags - this can be removed by about 6.10
    if (sXMLParentTag.equals(m_sXMLTag)) {
      if (sXMLTag.equals("es_gapMossCanopyFav")) {
        sXMLTag = "es_scarifiedSoilCanopyFav";
        for (i = 0; i < p_sChildXMLTags.length; i++) {
          if (p_sChildXMLTags[i].equals("es_gmcfVal")) {
            p_sChildXMLTags[i] = "es_sscfVal";
          }
        }
        return super.SetVectorValueByXMLTag(sXMLTag, sXMLParentTag, p_oData,
                                            p_sChildXMLTags, p_bAppliesTo,
                                            oParentAttributes, p_oAttributes);
      }
      else if (sXMLTag.equals("es_organicsCanopyFav")) {
        sXMLTag = "es_tipUpCanopyFav";
        for (i = 0; i < p_sChildXMLTags.length; i++) {
          if (p_sChildXMLTags[i].equals("es_ocfVal")) {
            p_sChildXMLTags[i] = "es_flcfVal";
          }
        }
        return super.SetVectorValueByXMLTag(sXMLTag, sXMLParentTag, p_oData,
                                            p_sChildXMLTags, p_bAppliesTo,
                                            oParentAttributes, p_oAttributes);
      }
      else if (sXMLTag.equals("es_mossyLogCanopyFav")) {
        sXMLTag = "es_decayedLogCanopyFav";
        for (i = 0; i < p_sChildXMLTags.length; i++) {
          if (p_sChildXMLTags[i].equals("es_mlcfVal")) {
            p_sChildXMLTags[i] = "es_dlcfVal";
          }
        }
        return super.SetVectorValueByXMLTag(sXMLTag, sXMLParentTag, p_oData,
                                            p_sChildXMLTags, p_bAppliesTo,
                                            oParentAttributes, p_oAttributes);
      }
      else if (sXMLTag.equals("es_canopyMossCanopyFav")) {
        sXMLTag = "es_forestFloorLitterCanopyFav";
        for (i = 0; i < p_sChildXMLTags.length; i++) {
          if (p_sChildXMLTags[i].equals("es_cmcfVal")) {
            p_sChildXMLTags[i] = "es_fflcfVal";
          }
        }
        //Set moss to be the same
        String sTempXMLTag = "es_forestFloorMossCanopyFav";
        String[] p_sTempChildTags = new String[p_sChildXMLTags.length];
        for (i = 0; i < p_sTempChildTags.length; i++)
          p_sTempChildTags[i] = "es_ffmcfVal";
        SetVectorValueByXMLTag(sTempXMLTag, sXMLParentTag, p_oData,
                                            p_sTempChildTags, p_bAppliesTo,
                                            oParentAttributes, p_oAttributes);
        return super.SetVectorValueByXMLTag(sXMLTag, sXMLParentTag, p_oData,
                                            p_sChildXMLTags, p_bAppliesTo,
                                            oParentAttributes, p_oAttributes);
      }
      else if (sXMLTag.equals("es_gapMossGapFav")) {
        sXMLTag = "es_scarifiedSoilGapFav";
        for (i = 0; i < p_sChildXMLTags.length; i++) {
          if (p_sChildXMLTags[i].equals("es_gmgfVal")) {
            p_sChildXMLTags[i] = "es_ssgfVal";
          }
        }
        return super.SetVectorValueByXMLTag(sXMLTag, sXMLParentTag, p_oData,
                                            p_sChildXMLTags, p_bAppliesTo,
                                            oParentAttributes, p_oAttributes);
      }
      else if (sXMLTag.equals("es_organicsGapFav")) {
        sXMLTag = "es_tipUpGapFav";
        for (i = 0; i < p_sChildXMLTags.length; i++) {
          if (p_sChildXMLTags[i].equals("es_ogfVal")) {
            p_sChildXMLTags[i] = "es_flgfVal";
          }
        }
        return super.SetVectorValueByXMLTag(sXMLTag, sXMLParentTag, p_oData,
                                            p_sChildXMLTags, p_bAppliesTo,
                                            oParentAttributes, p_oAttributes);
      }
      else if (sXMLTag.equals("es_mossyLogGapFav")) {
        sXMLTag = "es_decayedLogGapFav";
        for (i = 0; i < p_sChildXMLTags.length; i++) {
          if (p_sChildXMLTags[i].equals("es_mlgfVal")) {
            p_sChildXMLTags[i] = "es_dlgfVal";
          }
        }
        return super.SetVectorValueByXMLTag(sXMLTag, sXMLParentTag, p_oData,
                                            p_sChildXMLTags, p_bAppliesTo,
                                            oParentAttributes, p_oAttributes);
      }
      else if (sXMLTag.equals("es_canopyMossGapFav")) {
        sXMLTag = "es_forestFloorLitterGapFav";
        for (i = 0; i < p_sChildXMLTags.length; i++) {
          if (p_sChildXMLTags[i].equals("es_cmgfVal")) {
            p_sChildXMLTags[i] = "es_fflgfVal";
          }
        }
        //Set moss to be the same
        String sTempXMLTag = "es_forestFloorMossGapFav";
        String[] p_sTempChildTags = new String[p_sChildXMLTags.length];
        for (i = 0; i < p_sTempChildTags.length; i++)
          p_sTempChildTags[i] = "es_ffmgfVal";
        SetVectorValueByXMLTag(sTempXMLTag, sXMLParentTag, p_oData,
                                            p_sTempChildTags, p_bAppliesTo,
                                            oParentAttributes, p_oAttributes);
        return super.SetVectorValueByXMLTag(sXMLTag, sXMLParentTag, p_oData,
                                            p_sChildXMLTags, p_bAppliesTo,
                                            oParentAttributes, p_oAttributes);
      }
      else if (sXMLTag.equals("es_forestFloorCanopyFav")) {
        sXMLTag = "es_forestFloorLitterCanopyFav";
        for (i = 0; i < p_sChildXMLTags.length; i++) {
          if (p_sChildXMLTags[i].equals("es_ffcfVal")) {
            p_sChildXMLTags[i] = "es_ffmcfVal";
          }
        }
        //Set moss to be the same
        String sTempXMLTag = "es_forestFloorMossCanopyFav";
        String[] p_sTempChildTags = new String[p_sChildXMLTags.length];
        for (i = 0; i < p_sTempChildTags.length; i++)
          p_sTempChildTags[i] = "es_ffmcfVal";
        SetVectorValueByXMLTag(sTempXMLTag, sXMLParentTag, p_oData,
                                            p_sTempChildTags, p_bAppliesTo,
                                            oParentAttributes, p_oAttributes);
        return super.SetVectorValueByXMLTag(sXMLTag, sXMLParentTag, p_oData,
                                            p_sChildXMLTags, p_bAppliesTo,
                                            oParentAttributes, p_oAttributes);
      }
      else if (sXMLTag.equals("es_forestFloorGapFav")) {
        sXMLTag = "es_forestFloorLitterGapFav";
        for (i = 0; i < p_sChildXMLTags.length; i++) {
          if (p_sChildXMLTags[i].equals("es_ffgfVal")) {
            p_sChildXMLTags[i] = "es_ffmgfVal";
          }
        }
        //Set moss to be the same
        String sTempXMLTag = "es_forestFloorMossGapFav";
        String[] p_sTempChildTags = new String[p_sChildXMLTags.length];
        for (i = 0; i < p_sTempChildTags.length; i++)
          p_sTempChildTags[i] = "es_ffmgfVal";
        SetVectorValueByXMLTag(sTempXMLTag, sXMLParentTag, p_oData,
                                            p_sTempChildTags, p_bAppliesTo,
                                            oParentAttributes, p_oAttributes);
        return super.SetVectorValueByXMLTag(sXMLTag, sXMLParentTag, p_oData,
                                            p_sChildXMLTags, p_bAppliesTo,
                                            oParentAttributes, p_oAttributes);
      }
    }

    if (sXMLParentTag.equals(m_sXMLTag) ||
        sXMLParentTag.equals("ma_v") ||
        sXMLParentTag.equals("pkg") ||
        sXMLParentTag.equals("ma_intCodes") ||
        sXMLParentTag.equals("ma_floatCodes") ||
        sXMLParentTag.equals("ma_charCodes") ||
        sXMLParentTag.equals("ma_boolCodes") ||
        sXMLParentTag.equals("ma_packageIntCodes") ||
        sXMLParentTag.equals("ma_packageFloatCodes") ||
        sXMLParentTag.equals("ma_packageCharCodes") ||
        sXMLParentTag.equals("ma_packageBoolCodes") ||
        sXMLParentTag.equals("grid")) {
      return super.SetVectorValueByXMLTag(sXMLTag, sXMLParentTag, p_oData,
                                          p_sChildXMLTags, p_bAppliesTo,
                                          oParentAttributes, p_oAttributes);
    }
    return false;
  }
}
