package javawrapper;

import java.util.Vector;
import java.io.FileReader;
import javax.swing.*;
import javax.swing.table.TableModel;
import javax.help.HelpBroker;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.XMLReaderFactory;
import org.xml.sax.InputSource;

/**
 * The GUI manager provides all behind-the-scenes functionality for the Main
 * Window.  It performs a function very similar to the Simulation Manager in
 * the C++ code.
 * <p>Copyright: Copyright (c) Charles D. Canham 2003</p>
 * <p>Company: Institute of Ecosystem Studies </p>
 * @author Lora E. Murphy
 * @version 1.0
 *
 * <br>Edit history:
 * <br>------------------
 * <br>April 28, 2004: Submitted in beta version (LEM)
 */

public class GUIManager {

  /**Path and name of parameter currently loaded, if any.*/
  private String m_sParameterFile;

  /**Flag for whether or not a loaded parameter file was modified*/
  private boolean m_bWasParameterFileModified;

  /**This is an array of behaviors and populations, in the order in which
     they would appear in the parameter file*/
  private WorkerBase[] mp_oManagedObjects;

  /**Pointer to main application window.*/
  private MainWindow m_oMainWindow;

  /**This stands for the parameter file XML file type.  This matches the
   * value in the C++ file "DataTypes.h".*/
  protected final String PARAMETER_FILE = "01";

  /**This stands for the detailed output file XML file type.  This matches the
   * value in the C++ file "DataTypes.h".*/
  protected final String DETAILED_OUTPUT_FILE = "06";

  /**This stands for the detailed output timestep file XML file type.  This
   * matches the value in the C++ file "DataTypes.h".*/
  protected final String DETAILED_OUTPUT_TIMESTEP = "07";

  /**This stands for the batch file XML file type.  This
   * matches the value in the C++ file "DataTypes.h".*/
  protected final String BATCH_FILE = "04";

  /**
   * Constructor.  This creates the objects with no data in them so they can
   * be filled during a new parameter file setup.
   * @param oWindow Main application window.
   * @throws ModelException if something goes wrong with object setup.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public GUIManager(MainWindow oWindow) throws ModelException {
    m_oMainWindow = oWindow;
    ClearCurrentData();
  }

  /**
   * Whether or not the current parameter file has been modified.
   * @return true if modified, false if not.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public boolean HasParameterFileBeenModified() {
    return m_bWasParameterFileModified;
  }

  /**
   * Gets the HelpBroker object which manages help display for the entire
   * application.
   * @return System HelpBroker object.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public HelpBroker GetHelpBroker() {
    return m_oMainWindow.m_oHelpBroker;
  }

  /**
   * This can be used to propose that a new parameter file is being created.
   * If there is currently no parameter file name, this will accept the
   * proposal and place the key "<New>" in the parameter file name.  Then,
   * everything is set as through there is a parameter file entered.  All
   * objects are allowed to DoSetup().
   * <p>
   * If it is not time to create a new parameter file, nothing happens.
   * <p>
   * This is called by the TreeSetup dialog when changes have been made because
   * it's the first step in creating a new file.  It does not know whether it's
   * modifying an existing file or creating a new one, so it calls this so
   * the GUIManager object can decide.
   * @throws ModelException if anything goes wrong with setup.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void ProposeNewParameterFile() throws ModelException {
    if (m_sParameterFile.length() == 0) {
      m_sParameterFile = "<New>";
      m_oMainWindow.SetModelState(MainWindowStateSetter.PAR_FILE_LOADED);
      m_bWasParameterFileModified = true;
      DoSetup();
    }
  }

  /**
   * Gets the parameter file name.
   * @return The parameter file name.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public String GetParameterFileName() {
    return m_sParameterFile;
  }

  /**
   * Display the window which allows the user to set up output options.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void DisplayOutputWindow() {
    //If there's no parameter file loaded, don't do this
    if (m_sParameterFile.length() == 0) {
      JOptionPane jPane = new JOptionPane();
      jPane.showMessageDialog(m_oMainWindow, "Please load a parameter file.",
                              "Error",
                              JOptionPane.ERROR_MESSAGE);
      return;
    }

    GetOutputBehaviors().DisplayWindow(m_oMainWindow);
    m_bWasParameterFileModified = true;
  }

  /**
   * Displays the window which allows the user to set up trees.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void DisplayTreeSetupWindow() {
    TreeSetup oTreeWindow = new TreeSetup(m_oMainWindow, GetTreePopulation());
    oTreeWindow.pack();
    oTreeWindow.setLocationRelativeTo(null);
    oTreeWindow.setVisible(true);
    m_bWasParameterFileModified = true;
  }

  /**
   * Displays the window which allows the user to set up grids.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void DisplayGridSetupWindow() {
    GridSetup oGridWindow = new GridSetup(m_oMainWindow, this);
    oGridWindow.pack();
    oGridWindow.setLocationRelativeTo(null);
    oGridWindow.setVisible(true);
    m_bWasParameterFileModified = true;
  }

  /**
   * This passes a tree map file on to the tree population.
   * @param sFileName Name of file.
   * @throws ModelException if the tree population rejects the file.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void InputTreeMap(String sFileName) throws ModelException {

    GetTreePopulation().AddTabDelimTreeMapFile(m_oMainWindow, sFileName);
    m_bWasParameterFileModified = true;
  }

  /**
   * Accepts an XML file for input.  The file type is determined, and then
   * action is taken depending on type.  If the file type is a parameter file
   * or a detailed output setup file, existing data is cleared and then the
   * file is parsed.  If the file type is detailed output timestep, the user is
   * asked whether they want to replace or add to existing data.  The
   * appropriate step is then taken.
   * @param sFileName File name of file to input.
   * @param oWindow Parent window.
   * @throws ModelException if the file cannot be recognized, cannot be parsed,
   * or is invalid; or if this is a data file and there is no parameter file
   * loaded.
   * @return boolean Whether the file that was read was a data file.  A batch
   * file causes this to be false; all else is true (assuming no errors).
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public boolean InputXMLFile(String sFileName, JFrame oWindow) throws
      ModelException {
    FileReader oIn = null;
    boolean bReadDataFile = false;
    try {
      //Determine the file type
      oIn = new FileReader(sFileName);
      char[] cBuf = new char[100];
      int iRead = 0;
      String sTag = "fileCode=\"",
          sFileType = "";

      //Look for the file code
      while (iRead > -1) {
        iRead = oIn.read(cBuf, 0, 100);
        String sBuf = new String(cBuf);
        int iIndex = sBuf.indexOf(sTag);
        if (iIndex > -1) {
          int iStarter = iIndex + sTag.length() + 4;
          sFileType = sBuf.substring(iStarter, iStarter + 2);
          break;
        }
      }

      if (sFileType.equals(PARAMETER_FILE) ||
          sFileType.equals(DETAILED_OUTPUT_FILE)) {
        InputXMLParameterFile(sFileName);
        bReadDataFile = true;
      }
      else if (sFileType.equals(DETAILED_OUTPUT_TIMESTEP)) {
        //Make sure there's a parameter file loaded
        if (m_oMainWindow.GetModelState() ==
            MainWindowStateSetter.PAR_FILE_LOADED) {
          InputXMLDataFile(sFileName, oWindow);
          bReadDataFile = true;
        }
        else {
          throw (new ModelException(ErrorGUI.BAD_COMMAND, "JAVA",
                                    "You must load a parameter file before loading a file of this type."));
        }
      }
      else if (sFileType.equals(BATCH_FILE)) {
        //Load and display the batch file window

        BatchSetup oBatchWindow = new BatchSetup(m_oMainWindow, this, sFileName);
        oBatchWindow.pack();
        oBatchWindow.setLocationRelativeTo(null);
        oBatchWindow.setVisible(true);
      }
      else {
        ModelException oErr = new ModelException();
        oErr.iErrorCode = ErrorGUI.BAD_FILE_TYPE;
        oErr.sMessage = "The file " + sFileName + " cannot be recognized as a "
            + "SORTIE file.";
        oErr.sCallingFunction = "JAVA";
        throw (oErr);
      }

      return bReadDataFile;

    }
    catch (java.io.IOException oE) {
      ModelException oErr = new ModelException();
      oErr.iErrorCode = ErrorGUI.BAD_FILE;
      oErr.sCallingFunction = "InputXMLFile - JAVA";
      oErr.sMessage = sFileName;
      throw (oErr);
    }
    finally {
      try {
        if (oIn != null) {
          oIn.close();
        }
      }
      catch (java.io.IOException e) {
        ;
      }
    }
  }

  /**
   * Inputs an XML parameter file.
   * @param sFileName XML parameter file name
   * @throws ModelException if the file cannot be parsed or is invalid.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  protected void InputXMLParameterFile(String sFileName) throws ModelException {
    ClearCurrentData();
    java.io.FileInputStream oFileStream = null;

    //Build a parser and parse the file
    try {
      //Create our parser
      //Create the parser this way - don't go through the SAXParserFactory!
      //Parsers created that way throw MalformedURLExceptions when asked to
      //read local files.
      XMLReader oParser = XMLReaderFactory.createXMLReader(
          "org.apache.xerces.parsers.SAXParser");
      //Create the handler that looks for setup data and register it
      DefaultHandler oHandler = new ParameterFileParser(this);
      oParser.setContentHandler(oHandler);
      //Parse the file
      //Create a FileInputStream object to wrap in an InputSource object.
      //This method allows characters with diacritical marks in the filename,
      //whereas simply feeding the filename to InputSource directly does not
      oFileStream = new java.io.FileInputStream(sFileName);
      InputSource oToParse = new InputSource(oFileStream);
      oParser.parse(oToParse);

      m_sParameterFile = sFileName;
      m_bWasParameterFileModified = false;

      //Don't validate here - users should have the chance to load a file and
      //fix it
    }
    catch (SAXException oE) {
      ModelException oErr = new ModelException();
      oErr.sMessage = oE.getMessage();
      oErr.iErrorCode = ErrorGUI.BAD_XML_FILE;
      oErr.sCallingFunction = "JAVA";
      throw (oErr);
    }
    catch (java.io.IOException oE) {
      ModelException oErr = new ModelException();
      oErr.sMessage = sFileName;
      oErr.iErrorCode = ErrorGUI.BAD_FILE;
      oErr.sCallingFunction = "InputXMLParameterFile - JAVA";
      throw (oErr);
    }
    finally {
      try {
        if (oFileStream != null) {
          oFileStream.close();
        }
      }
      catch (java.io.IOException oErr) {
        ; //do nothing
      }
    }
  }

  /**
   * Inputs an XML file which is not a parameter type.  This would probably be
   * a detailed output timestep file, a tree map, or a grid map.  This method
   * does not clear out existing data before parsing the file.
   * @param sFileName XML parameter file name
   * @param oWindow Parent application window
   * @throws ModelException if the file cannot be parsed or is invalid.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  protected void InputXMLDataFile(String sFileName, JFrame oWindow) throws
      ModelException {

    //Determine if the user wants to add to or replace any current data
    Object[] oButtonOptions = {
        "Replace", "Add To", "Cancel"};
    int iChoice = JOptionPane.showOptionDialog(oWindow,
                                               "What would you like to do with existing tree map and grid map data?",
                                               "", JOptionPane.YES_NO_OPTION,
                                               JOptionPane.QUESTION_MESSAGE,
                                               null, //don't use a custom Icon
                                               oButtonOptions, //the titles of buttons
                                               oButtonOptions[0]); //default button title

    if (iChoice == 2) {
      return;
    }
    if (iChoice == 0) {

      //Clean existing data but not settings
      GetTreePopulation().ClearTrees();

      int i;
      for (i = 0; i < mp_oManagedObjects.length; i++) {
        if (mp_oManagedObjects[i] instanceof BehaviorTypeBase) {
          BehaviorTypeBase oBehavior = (BehaviorTypeBase) mp_oManagedObjects[i];
          oBehavior.ClearGridMapValues();
        }
      }
    }

    //Build a parser and parse the file
    java.io.FileInputStream oFileStream = null;
    try {
      //Create our parser
      //Create the parser this way - don't go through the SAXParserFactory!
      //Parsers created that way throw MalformedURLExceptions when asked to
      //read local files.
      XMLReader oParser = XMLReaderFactory.createXMLReader(
          "org.apache.xerces.parsers.SAXParser");
      //Create the handler that looks for setup data and register it
      DefaultHandler oHandler = new ParameterFileParser(this);
      oParser.setContentHandler(oHandler);
      //Parse the file
      //Create a FileInputStream object to wrap in an InputSource object.
      //This method allows characters with diacritical marks in the filename,
      //whereas simply feeding the filename to InputSource directly does not
      oFileStream = new java.io.FileInputStream(sFileName);
      InputSource oToParse = new InputSource(oFileStream);
      oParser.parse(oToParse);

      m_bWasParameterFileModified = true;
    }
    catch (SAXException oE) {
      ModelException oErr = new ModelException();
      oErr.sMessage = oE.getMessage();
      oErr.iErrorCode = ErrorGUI.BAD_XML_FILE;
      oErr.sCallingFunction = "JAVA";
      throw (oErr);
    }
    catch (java.io.IOException oE) {
      ModelException oErr = new ModelException();
      oErr.sMessage = sFileName;
      oErr.iErrorCode = ErrorGUI.BAD_FILE;
      oErr.sCallingFunction = "InputXMLParameterFile - JAVA";
      throw (oErr);
    }
    finally {
      try {
        if (oFileStream != null) {
          oFileStream.close();
        }
      }
      catch (java.io.IOException oErr) {
        ; //do nothing
      }
    }
  }

  /**
   * Gets whether or not the run is snag-aware.  This searches through all
   * behaviors looking for any that are applied to snags.  If none, then the
   * run is not snag-aware.
   * @return True if the run has any snags anywhere; false if it doesn't.
   * @throws ModelException passed through from called methods.  Should never
   * be thrown.
   */
  public boolean GetSnagAwareness() throws ModelException {
    int i, j, k;

    BehaviorTypeBase[] p_oBehaviorGroups = GetAllBehaviorGroups();
    for (i = 0; i < p_oBehaviorGroups.length; i++) {
      Behavior[] p_oBehaviors = p_oBehaviorGroups[i].GetAllBehaviors();
      if (p_oBehaviors != null) {
        for (j = 0; j < p_oBehaviors.length; j++) {
          if (p_oBehaviors[j].m_bIsEnabled) {
            for (k = 0; k < p_oBehaviors[j].GetNumberOfCombos(); k++) {
              if (p_oBehaviors[j].GetSpeciesTypeCombo(k).GetType() ==
                  TreePopulation.SNAG) {
                return true;
              }
            }
          }
        }
      }
    }
    return false;
  }

  /**
   * Reads float values from a tab-delimited file into a Float array.  The
   * values are expected to be on one line separated by tabs.  Once the
   * specified number of values is read, the file stream is advanced to the
   * beginning of the next line, even if there are more values present.
   * <p>
   * There is no provision made for making sure that the array is large enough
   * to hold the number of parameters being read into it.
   *
   * @param in The file stream to read from.
   * @param p_fVarName A Float array in which to read the values from file.
   * @param iNumParams The number of Float values being read.
   *
   * @throws ModelException with error code TAB_PAR_FILE_PREMATURE_END if the
   * end of the file is reached at any point.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */

  protected void ReadParam(FileReader in, Float[] p_fVarName, int iNumParams) throws
      ModelException {

    try {

      Vector p_oLine = ModelFileFunctions.ReadLine(in);
      String sData;
      int i;
      //Make sure there were enough values read
      if (p_oLine.size() < iNumParams + 1) { //+1 for comment
        ModelException exp = new ModelException();
        exp.iErrorCode = ErrorGUI.TAB_FILE_PREMATURE_END;
        throw (exp);
      }

      for (i = 1; i <= iNumParams; i++) {
        sData = (String) p_oLine.elementAt(i);
        p_fVarName[i - 1] = new Float(sData);
      }
    }
    catch (java.io.IOException e) {
      ModelException exp = new ModelException();
      exp.iErrorCode = ErrorGUI.TAB_FILE_PREMATURE_END;
      throw (exp);
    }
  }

  /**
   * Reads integer values from a tab-delimited file into an Integer array.  The
   * values are expected to be on one line separated by tabs.  Once the
   * specified number of values is read, the file stream is advanced to the
   * beginning of the next line, even if there are more values present.
   * <p>
   * There is no provision made for making sure that the array is large enough
   * to hold the number of parameters being read into it.
   *
   * @param in The file stream to read from.
   * @param p_iVarName An Integer array in which to read the values from file.
   * @param iNumParams The number of integer values being read.
   *
   * @throws ModelException with error code TAB_PAR_FILE_PREMATURE_END if the end
   * of the file is reached at any point.
   */

  protected void ReadParam(FileReader in, Integer[] p_iVarName,
                           int iNumParams) throws
      ModelException {

    try {

      Vector p_oLine = ModelFileFunctions.ReadLine(in);
      String sData;
      int i;
      //Make sure there were enough values read
      if (p_oLine.size() < iNumParams + 1) { //+1 for comment
        ModelException exp = new ModelException();
        exp.iErrorCode = ErrorGUI.TAB_FILE_PREMATURE_END;
        throw (exp);
      }

      for (i = 1; i <= iNumParams; i++) {
        sData = (String) p_oLine.elementAt(i);
        p_iVarName[i - 1] = new Integer(sData);
      }

    }
    catch (java.io.IOException e) {
      ModelException exp = new ModelException();
      exp.iErrorCode = ErrorGUI.TAB_FILE_PREMATURE_END;
      throw (exp);
    }
  }

  /**
   * Reads character values from a tab-delimited file into a String array.  The
   * values are expected to be on one line separated by tabs.  Once the specified
   * number of values is read, the file stream is advanced to the beginning of
   * the next line, even if there are more values present.
   * <p>
   * There is no provision made for making sure that the array is large enough
   * to hold the number of parameters being read into it.
   *
   * @param in The file stream to read from.
   * @param p_sVarName A String array in which to read the values from file.
   * @param iNumParams The number of String values being read.
   *
   * @throws ModelException with error code TAB_PAR_FILE_PREMATURE_END if the
   * end of the file is reached at any point.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */

  protected void ReadParam(FileReader in, String[] p_sVarName, int iNumParams) throws
      ModelException {

    try {

      Vector p_oLine = ModelFileFunctions.ReadLine(in);
      int i;
      //Make sure there were enough values read
      if (p_oLine.size() < iNumParams + 1) { //+1 for comment
        ModelException exp = new ModelException();
        exp.iErrorCode = ErrorGUI.TAB_FILE_PREMATURE_END;
        throw (exp);
      }

      for (i = 1; i <= iNumParams; i++) {
        p_sVarName[i - 1] = (String) p_oLine.elementAt(i);
      }

    }
    catch (java.io.IOException e) {
      ModelException exp = new ModelException();
      exp.iErrorCode = ErrorGUI.TAB_FILE_PREMATURE_END;
      throw (exp);
    }
  }

  /**
   * Checks to see if the currently loaded dataset is adequate for creating a
   * run and internally valid.  The dataset is valid if there is at least
   * one behavior enabled, and each managed object successfully validates its
   * internal data.
   * @throws ModelException if the dataset is not valid.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  protected void ValidateDataSet() throws ModelException {

    TreePopulation oPop = GetTreePopulation();
    int i;
    boolean bAnyEnabled = false;

    //Make sure there are enabled behaviors
    BehaviorTypeBase[] p_oAllBehaviors = GetAllBehaviorGroups();
    for (i = 0; i < p_oAllBehaviors.length; i++) {
      if (p_oAllBehaviors[i].AnyBehaviorsEnabled()) {
        bAnyEnabled = true;
      }
    }

    if (!bAnyEnabled) {
      ModelException oErr = new ModelException();
      oErr.sMessage = "The current run has no behaviors.  Please use Edit->Model Flow to define behaviors for this run.";
      oErr.sCallingFunction = "JAVA";
      throw (oErr);
    }

    //Have each object validate its data
    for (i = 0; i < mp_oManagedObjects.length; i++) {
      mp_oManagedObjects[i].ValidateData(oPop);
    }
  }

  /**
   * Writes out the XML parameter file with the current settings.  Note that
   * it is not required that the data set be adequate or valid; this way
   * work can be saved in progress.  However, if the dataset is invalid,
   * the user will be warned of that fact.
   * @param sFileName  Filename of new file.
   * @return Whether or not a file was saved.
   * @throws ModelException If there is a problem validating the data or
   * writing the file.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public boolean WriteParameterFile(String sFileName) throws ModelException {
    java.io.FileOutputStream jOutputStream = null;
    java.io.OutputStreamWriter jOutStreamWriter = null;
    java.io.BufferedWriter jOut = null;
    try {
      TreePopulation oPop = GetTreePopulation();

      //See if the dataset is valid; if it's not, warn the user
      try {
        ValidateDataSet();
      }
      catch (ModelException oErr) {

        //Invalid - warn user
        int iReturnValue = new JOptionPane().showConfirmDialog(m_oMainWindow,
            "The data is not valid for a run.  Message:  " + oErr.sMessage +
            "  Would you like to save this file anyway?", "Continue with save?",
            JOptionPane.YES_NO_OPTION, JOptionPane.ERROR_MESSAGE);
        if (iReturnValue == JOptionPane.NO_OPTION) {
          return false;
        }
      }
      int i;

      jOutputStream = new java.io.FileOutputStream(sFileName);
      //Write the file in UTF-8 characters because that's what the XML parser
      //wants
      jOutStreamWriter = new java.io.OutputStreamWriter(jOutputStream, "UTF-8");
      jOut = new java.io.BufferedWriter(jOutStreamWriter);

      //Write the opening tag
      jOut.write(
          "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\" ?>");
      jOut.write("<paramFile fileCode=\"06050101\">");

      //Write the plot info
      GetPlot().WriteXML(jOut, oPop);

      //Write the tree info
      oPop.WriteXML(jOut, oPop);

      //Write any grid maps
      for (i = 0; i < mp_oManagedObjects.length; i++) {
        if (mp_oManagedObjects[i] instanceof BehaviorTypeBase) {
          BehaviorTypeBase oBehavior = (BehaviorTypeBase) mp_oManagedObjects[i];
          oBehavior.WriteGridXML(jOut);
        }
      }

      //Write the behavior list
      jOut.write("<behaviorList>");

      //Write out all behaviors
      for (i = 0; i < mp_oManagedObjects.length; i++) {
        if (mp_oManagedObjects[i] instanceof BehaviorTypeBase) {
          BehaviorTypeBase oBehavior = (BehaviorTypeBase) mp_oManagedObjects[i];
          oBehavior.WriteBehaviorNodes(jOut, oPop);
        }
      }

      jOut.write("</behaviorList>");

      //Write all other data objects
      for (i = 0; i < mp_oManagedObjects.length; i++) {
        if (! (mp_oManagedObjects[i] instanceof TreePopulation) &&
            ! (mp_oManagedObjects[i] instanceof Plot)) {
          mp_oManagedObjects[i].WriteXML(jOut, oPop);
        }
      }

      //Write the closing tag
      jOut.write("</paramFile>");
      jOut.flush();

      //Set the newly written file as the current parameter file
      m_sParameterFile = sFileName;
      //Set the modified flag to false
      m_bWasParameterFileModified = false;
      return true;
    }
    catch (java.io.IOException e) {
      ModelException oErr = new ModelException();
      oErr.sCallingFunction = "JAVA";
      oErr.sMessage = "There was a problem writing the parameter file.";
      oErr.iErrorCode = ErrorGUI.BAD_FILE;
      throw (oErr);
    }
    finally {
      try {
        if (jOut != null) {
          jOut.close();
        }
      }
      catch (java.io.IOException oErr) {
        ; //do nothing
      }
      try {
        if (jOutStreamWriter != null) {
          jOutStreamWriter.close();
        }
      }
      catch (java.io.IOException oErr) {
        ; //do nothing
      }
      try {
        if (jOutputStream != null) {
          jOutputStream.close();
        }
      }
      catch (java.io.IOException oErr) {
        ; //do nothing
      }
    }
  }

  /**
   * Gets the MainWindow object.
   * @return MainWindow MainWindow object.
   */
  public MainWindow GetMainWindow() {
    return m_oMainWindow;
  }

  /**
   * Gets the descriptive names list of behaviors for a tree
   * species/type combo.
   * @param iSpecies Species
   * @param iType Type
   * @return String array of behavior names.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public String[] GetBehaviorList(int iSpecies, int iType) {
    try {
      SpeciesTypeCombo oCombo;
      Vector oList = new Vector(0);
      int iNumBehaviors, i, j, k;

      for (i = 0; i < mp_oManagedObjects.length; i++) {
        if (mp_oManagedObjects[i] instanceof BehaviorTypeBase) {
          BehaviorTypeBase oBase = (BehaviorTypeBase) mp_oManagedObjects[i];
          if (oBase.mp_oChildBehaviors != null) {
            iNumBehaviors = oBase.mp_oChildBehaviors.length;
            for (j = 0; j < iNumBehaviors; j++) {
              for (k = 0; k < oBase.mp_oChildBehaviors[j].GetNumberOfCombos();
                   k++) {
                oCombo = oBase.mp_oChildBehaviors[j].GetSpeciesTypeCombo(k);
                if (oCombo.GetSpecies() == iSpecies &&
                    oCombo.GetType() == iType) {
                  oList.add(oBase.mp_oChildBehaviors[j].GetDescriptor());
                  break;
                }
              }
            }
          }
        }
      }

      String[] sList = new String[oList.size()];
      for (i = 0; i < oList.size(); i++) {
        sList[i] = (String) oList.elementAt(i);
      }
      return sList;
    }
    catch (ModelException oErr) {
      return null;
    }
  }

  /**
   * Sets the list of behaviors for a tree species/type combo.  This replaces
   * any existing assignments.
   * @param oToSet Species/type combo to set behaviors for
   * @param p_sBehaviors String array of behavior DISPLAY names.
   * @throws ModelException if there is an unrecognized behavior present.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void SetBehaviorList(SpeciesTypeCombo oToSet, String[] p_sBehaviors) throws
      ModelException {

    BehaviorTypeBase[] p_oBehaviorGroups = GetAllBehaviorGroups();
    SpeciesTypeCombo oCombo;
    Behavior oBehavior;
    int iNumBehaviors, i, j, k;
    boolean bFound;

    //First remove all existing assignments for this type/species
    for (i = 0; i < p_oBehaviorGroups.length; i++) {
      if (p_oBehaviorGroups[i].mp_oChildBehaviors != null) {
        iNumBehaviors = p_oBehaviorGroups[i].mp_oChildBehaviors.length;
        for (j = 0; j < iNumBehaviors; j++) {
          for (k = 0;
               k < p_oBehaviorGroups[i].mp_oChildBehaviors[j].GetNumberOfCombos();
               k++) {
            oCombo = p_oBehaviorGroups[i].mp_oChildBehaviors[j].
                GetSpeciesTypeCombo(k);
            if (oCombo.equals(oToSet)) {
              p_oBehaviorGroups[i].mp_oChildBehaviors[j].DeleteSpeciesTypeCombo(
                  k);
            }
          }
        }
      }
    }

    //Now match all behaviors and assign them the species/type combo
    if (p_sBehaviors == null || p_sBehaviors.length == 0) {
      return;
    }
    for (i = 0; i < p_sBehaviors.length; i++) {

      //Find this behavior
      bFound = false;
      for (j = 0; j < p_oBehaviorGroups.length; j++) {
        oBehavior = p_oBehaviorGroups[j].GetBehaviorByDisplayName(p_sBehaviors[
            i]);
        if (oBehavior != null) {
          oBehavior.AddSpeciesTypeCombo(oToSet);
          bFound = true;
          break;
        }
      }

      if (!bFound) {
        //Error - unrecognized behavior
        ModelException oErr = new ModelException();
        oErr.iErrorCode = ErrorGUI.BAD_ARGUMENT;
        oErr.sMessage = "There was an attempt to apply the tree type \"" +
            oToSet.GetType() + "\" species \"" + oToSet.GetSpecies() +
            "\" to an unrecognized behavior \"" + p_sBehaviors[i] + "\".";
        oErr.sCallingFunction = "JAVA";
        throw (oErr);
      }
    }
  }

  /**
   * Takes all loaded data and prepares a parameter file to pass to the core to
   * run, returning the filename of the parameter file.
   * @return The filename of the parameter file which can be passed to the core
   * code.
   * @throws ModelException if there are any data validation problems.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public String PrepToRun() throws ModelException {

    //What was input?  If there's no parameter file name, throw an exception
    if (m_sParameterFile.length() == 0) {
      ModelException oErr = new ModelException();
      oErr.iErrorCode = ErrorGUI.MODEL_NOT_READY;
      oErr.sMessage = "Please enter a parameter file.";
      oErr.sCallingFunction = "JAVA";
    }

    //Validate the dataset
    ValidateDataSet();

    //If our file was not modified, pass it through
    if (!m_bWasParameterFileModified) {
      return m_sParameterFile;
    }

    //Ask if they want to save this file
    int iReturnVal = new JOptionPane().showConfirmDialog(m_oMainWindow,
        "Do you wish to save a copy of this file now?", "Model",
        JOptionPane.YES_NO_OPTION);

    if (iReturnVal == JOptionPane.YES_OPTION) {
      m_oMainWindow.DoFileSave();
      if (!m_bWasParameterFileModified) { //they may have cancelled out
        return m_sParameterFile;
      }
    }

    //The file was modified but the user doesn't want to save; so we'll have
    //to write a temp file.
    //This is just a temporary filename I've made up.  We'll always need one
    //if parameter file data has been modified but not saved.

    String sFilename = "sortie_temp.xml";
    WriteParameterFile(sFilename);

    return sFilename;
  }

  /**
   * Performs the actions necessary to start a new parameter file.
   * @throws ModelException if there is a problem with setup.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void CreateNewParameterFile() throws ModelException {

    //Trigger a save if an existing file was being edited
    if (m_bWasParameterFileModified) {
      int iReturnVal = new JOptionPane().showConfirmDialog(m_oMainWindow,
          "Do you wish to save a copy of the current file?", "Model",
          JOptionPane.YES_NO_OPTION);

      if (iReturnVal == JOptionPane.YES_OPTION) {
        m_oMainWindow.DoFileSave();
        if (m_bWasParameterFileModified) { //they may have cancelled out
          return;
        }
      }
    }

    ClearCurrentData();
    m_oMainWindow.SetModelState(MainWindowStateSetter.NO_PAR_FILE);

    //Open the tree setup window
    DisplayTreeSetupWindow();

    //If they successfully completed it, ask them to open the model
    //flow window
    if (m_sParameterFile.equals("<New>")) {
      DisplayModelFlowWindow();
    }
  }

  /**
   * This clears old settings in order to accept a new parameter file.  If
   * there's no current data, this will make sure everything is set up to
   * accept new data.
   * <p>
   * The parameter file string is set to empty, and the parameter file modified
   * flag is set to false.
   * @throws ModelException if there are any problems.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  protected void ClearCurrentData() throws ModelException {
    m_sParameterFile = "";
    m_bWasParameterFileModified = false;

    if (mp_oManagedObjects != null) {

      //Garbage-collect everything currently in there
      int i;

      for (i = 0; i < mp_oManagedObjects.length; i++) {
        mp_oManagedObjects[i] = null;
      }

      mp_oManagedObjects = null;
    }

    mp_oManagedObjects = new WorkerBase[15];
    mp_oManagedObjects[0] = new Plot(this);
    mp_oManagedObjects[1] = new Allometry(this);
    mp_oManagedObjects[2] = new TreePopulation(this);
    mp_oManagedObjects[3] = new DisturbanceBehaviors(this);
    mp_oManagedObjects[4] = new LightBehaviors(this);
    mp_oManagedObjects[5] = new GrowthBehaviors(this);
    mp_oManagedObjects[6] = new MortalityBehaviors(this);
    mp_oManagedObjects[7] = new SubstrateBehaviors(this);
    mp_oManagedObjects[8] = new RemoveDeadBehaviors(this);
    mp_oManagedObjects[9] = new DisperseBehaviors(this);
    mp_oManagedObjects[10] = new SeedPredationBehaviors(this);
    mp_oManagedObjects[11] = new EstablishmentBehaviors(this);
    mp_oManagedObjects[12] = new PlantingBehaviors(this);
    mp_oManagedObjects[13] = new AnalysisBehaviors(this);
    mp_oManagedObjects[14] = new OutputBehaviors(this);

  }

  /**
   * Gets the OutputBehaviors object, or null if none exists.
   * @return OutputBehaviors object.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public OutputBehaviors GetOutputBehaviors() {

    for (int i = 0; i < mp_oManagedObjects.length; i++) {
      if (mp_oManagedObjects[i] instanceof OutputBehaviors) {
        return (OutputBehaviors) mp_oManagedObjects[i];
      }
    }

    return null;
  }

  /**
   * Gets the DisperseBehaviors object, or null if none exists.
   * @return DisperseBehaviors object.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public DisperseBehaviors GetDisperseBehaviors() {

    for (int i = 0; i < mp_oManagedObjects.length; i++) {
      if (mp_oManagedObjects[i] instanceof DisperseBehaviors) {
        return (DisperseBehaviors) mp_oManagedObjects[i];
      }
    }

    return null;
  }

  /**
   * Gets the AnalysisBehaviors object, or null if none exists.
   * @return AnalysisBehaviors object.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>October 27, 2004: Created in beta version (LEM)
   */
  public AnalysisBehaviors GetAnalysisBehaviors() {

    for (int i = 0; i < mp_oManagedObjects.length; i++) {
      if (mp_oManagedObjects[i] instanceof AnalysisBehaviors) {
        return (AnalysisBehaviors) mp_oManagedObjects[i];
      }
    }

    return null;
  }

  /**
   * Gets the EstablishmentBehaviors object, or null if none exists.
   * @return EstablishmentBehaviors object.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public EstablishmentBehaviors GetEstablishmentBehaviors() {

    for (int i = 0; i < mp_oManagedObjects.length; i++) {
      if (mp_oManagedObjects[i] instanceof EstablishmentBehaviors) {
        return (EstablishmentBehaviors) mp_oManagedObjects[i];
      }
    }

    return null;
  }

  /**
   * Gets the Substrate object.
   * @return Substrate object, or null if none exists.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public SubstrateBehaviors GetSubstrateBehaviors() {

    for (int i = 0; i < mp_oManagedObjects.length; i++) {
      if (mp_oManagedObjects[i] instanceof SubstrateBehaviors) {
        return (SubstrateBehaviors) mp_oManagedObjects[i];
      }
    }

    return null;
  }

  /**
   * Gets the MortalityBehaviors object.
   * @return MortalityBehaviors object, or null if none exists.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public MortalityBehaviors GetMortalityBehaviors() {

    for (int i = 0; i < mp_oManagedObjects.length; i++) {
      if (mp_oManagedObjects[i] instanceof MortalityBehaviors) {
        return (MortalityBehaviors) mp_oManagedObjects[i];
      }
    }

    return null;
  }

  /**
   * Gets the GrowthBehaviors object.
   * @return GrowthBehaviors object, or null if none exists.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public GrowthBehaviors GetGrowthBehaviors() {

    for (int i = 0; i < mp_oManagedObjects.length; i++) {
      if (mp_oManagedObjects[i] instanceof GrowthBehaviors) {
        return (GrowthBehaviors) mp_oManagedObjects[i];
      }
    }

    return null;
  }

  /**
   * Gets the Plot object.
   * @return Plot object, or null if none exists.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public Plot GetPlot() {

    for (int i = 0; i < mp_oManagedObjects.length; i++) {
      if (mp_oManagedObjects[i] instanceof Plot) {
        return (Plot) mp_oManagedObjects[i];
      }
    }

    return null;
  }

  /**
   * Gets the tree population object.
   * @return TreePopulation object, or null if none exists.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public TreePopulation GetTreePopulation() {

    for (int i = 0; i < mp_oManagedObjects.length; i++) {
      if (mp_oManagedObjects[i] instanceof TreePopulation) {
        return (TreePopulation) mp_oManagedObjects[i];
      }
    }
    return null;
  }

  /**
   * Gets the allometry object.
   * @return Allometry object, or null if none exists.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 25, 2005: Created (LEM)
   */
  public Allometry GetAllometry() {

    for (int i = 0; i < mp_oManagedObjects.length; i++) {
      if (mp_oManagedObjects[i] instanceof Allometry) {
        return (Allometry) mp_oManagedObjects[i];
      }
    }
    return null;
  }

  /**
   * Gets the disturbance behaviors object, or null if none exists.
   * @return DisturbanceBehaviors object, or null if none exists.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public DisturbanceBehaviors GetDisturbanceBehaviors() {
    for (int i = 0; i < mp_oManagedObjects.length; i++) {
      if (mp_oManagedObjects[i] instanceof DisturbanceBehaviors) {
        return (DisturbanceBehaviors) mp_oManagedObjects[i];
      }
    }
    return null;
  }

  /**
   * Gets the planting behaviors object, or null if none exists.
   * @return PlantingBehaviors object, or null if none exists.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public PlantingBehaviors GetPlantingBehaviors() {
    for (int i = 0; i < mp_oManagedObjects.length; i++) {
      if (mp_oManagedObjects[i] instanceof PlantingBehaviors) {
        return (PlantingBehaviors) mp_oManagedObjects[i];
      }
    }
    return null;
  }

  /**
   * Gets the light behaviors object.
   * @return LightBehaviors object, or null if none exists.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public LightBehaviors GetLightBehaviors() {
    for (int i = 0; i < mp_oManagedObjects.length; i++) {
      if (mp_oManagedObjects[i] instanceof LightBehaviors) {
        return (LightBehaviors) mp_oManagedObjects[i];
      }
    }
    return null;
  }

  public RemoveDeadBehaviors GetRemoveDeadBehaviors() {
    for (int i = 0; i < mp_oManagedObjects.length; i++) {
      if (mp_oManagedObjects[i] instanceof RemoveDeadBehaviors) {
        return (RemoveDeadBehaviors) mp_oManagedObjects[i];
      }
    }
    return null;
  }

  /**
   * Reads in a harvest regime.  Passes it off to the harvest and planting
   * behaviors to do with it what they will.  Currently this is not supported
   * if the currently loaded parameter file is XML.
   * @param sFileName Harvest regime file name.
   * @throws ModelException If there is a problem reading the file.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void InputHarvestRegime(String sFileName) throws ModelException {

    GetDisturbanceBehaviors().ReadOldHarvestRegimeFile(sFileName, GetPlot(),
        GetTreePopulation());
    GetDisturbanceBehaviors().DoSetup(GetTreePopulation());
    GetPlantingBehaviors().ReadOldHarvestRegimeFile(sFileName, GetPlot(),
        GetTreePopulation());
    GetPlantingBehaviors().DoSetup(GetTreePopulation());
    m_bWasParameterFileModified = true;
  }

  /**
   * Gets all of the objects descended from BehaviorTypeBase.
   * @return An array of behaviors.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public BehaviorTypeBase[] GetAllBehaviorGroups() {
    Vector oBehaviors = new Vector(0);
    int i;

    if (mp_oManagedObjects == null) {
      return null;
    }

    for (i = 0; i < mp_oManagedObjects.length; i++) {
      if (mp_oManagedObjects[i] instanceof BehaviorTypeBase) {
        oBehaviors.add(mp_oManagedObjects[i]);
      }
    }

    if (oBehaviors.size() == 0) {
      return null;
    }
    BehaviorTypeBase[] oToReturn = new BehaviorTypeBase[oBehaviors.size()];
    for (i = 0; i < oBehaviors.size(); i++) {
      oToReturn[i] = (BehaviorTypeBase) oBehaviors.elementAt(i);
    }
    return oToReturn;
  }

  /**
   * Gets the array of managed objects.
   * @return The array of managed objects.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public WorkerBase[] GetAllObjects() {
    return mp_oManagedObjects;
  }

  /**
   * Allows the user to edit parameters.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void DisplayRunParameters() {
    ParameterEdit oEditWindow = new ParameterEdit(m_oMainWindow, this);
    oEditWindow.pack();
    oEditWindow.setLocationRelativeTo(null);
    oEditWindow.setVisible(true);
    m_bWasParameterFileModified = true;
  }

  /**
   * Allows the user to edit simulation flow and behavior order.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void DisplayModelFlowWindow() {
    ModelFlowSetup oFlowWindow = new ModelFlowSetup(m_oMainWindow, this);
    oFlowWindow.pack();
    oFlowWindow.setLocationRelativeTo(null);
    oFlowWindow.setVisible(true);
    m_bWasParameterFileModified = true;
  }

  /**
   * Allows the user to edit episodic event data.  This specifically doesn't
   * actually display the window.  The MainWindow object that called this
   * gets the option to resize it because it's so big.  If I had lots of time,
   * I'd do this for all windows.
   * @throws ModelException passing through from called methods.
   * @return The new window.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>May 4, 2004: Created (LEM)
   */
  public JDialog DisplayEpisodicEventsWindow() throws ModelException {
    EpisodicEventsWindow oWindow = new EpisodicEventsWindow(m_oMainWindow,
        GetDisturbanceBehaviors(), GetPlantingBehaviors());
    m_bWasParameterFileModified = true;
    return oWindow;
  }

  /**
   * Extracts the data from the parameter window and passes it to objects.
   * @param jDataWindow Window holding parameters.
   * @throws ModelException if the data extracted is invalid.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void ExtractParameters(JDialog jDataWindow) throws ModelException {

    TreePopulation oPop = GetTreePopulation();
    JPanel jMainPanel = null;
    JScrollPane jScroller = null;
    Vector p_oData;
    String sObjectName = null, sComponentName;
    int i, j, k;

    //Get the main data panel
    for (i = 0; i < jDataWindow.getContentPane().getComponentCount(); i++) {
      sComponentName = jDataWindow.getContentPane().getComponent(i).getName();
      if (sComponentName != null && sComponentName.equals("Scroller")) {
        jScroller = (JScrollPane) jDataWindow.getContentPane().getComponent(i);
        break;
      }
    }

    if (jScroller == null) {
      //Panic
      ModelException oErr = new ModelException();
      oErr.sMessage = "Can't find data in the parameter window.";
      oErr.sCallingFunction = "JAVA";
      throw (oErr);
    }

    for (i = 0; i < jScroller.getViewport().getComponentCount(); i++) {
      sComponentName = jScroller.getViewport().getComponent(i).getName();
      if (sComponentName != null && sComponentName.equals("Main Panel")) {
        jMainPanel = (JPanel) jScroller.getViewport().getComponent(i);
        break;
      }
    }

    if (jMainPanel == null) {
      //Panic
      ModelException oErr = new ModelException();
      oErr.sMessage = "Can't find data in the parameter window.";
      oErr.sCallingFunction = "JAVA";
      throw (oErr);
    }

    //Each set of tables is isolated in its own panel.  Find all those named
    //"Object Panel".
    for (i = 0; i < jMainPanel.getComponentCount(); i++) {
      sComponentName = jMainPanel.getComponent(i).getName();
      if (sComponentName != null && sComponentName.equals("Object Panel")) {
        JPanel jPanel = (JPanel) jMainPanel.getComponent(i);

        p_oData = null;
        p_oData = new Vector();

        //Find the label with the object's name
        for (j = 0; j < jPanel.getComponentCount(); j++) {
          sComponentName = jPanel.getComponent(j).getName();
          if (sComponentName != null && sComponentName.equals("Object Name")) {
            JLabel jLabel = (JLabel) jPanel.getComponent(j);
            sObjectName = jLabel.getText();
          }
        }

        //Now format the tables into a vector
        for (j = 0; j < jPanel.getComponentCount(); j++) {
          sComponentName = jPanel.getComponent(j).getName();
          if (sComponentName != null && sComponentName.equals("Table Panel")) {
            JPanel jTablePanel = (JPanel) jPanel.getComponent(j);
            for (k = 0; k < jTablePanel.getComponentCount(); k++) {
              if (jTablePanel.getComponent(k) instanceof EnhancedTable) {
                EnhancedTable oTable = (EnhancedTable) jTablePanel.
                    getComponent(k);
                TableModel jModel = oTable.getModel();
                Object[][] p_oTableData = new Object[jModel.getRowCount()][
                    jModel.getColumnCount()];
                int iRow, iCol;
                for (iRow = 0; iRow < jModel.getRowCount(); iRow++) {
                  for (iCol = 0; iCol < jModel.getColumnCount(); iCol++) {
                    p_oTableData[iRow][iCol] = jModel.getValueAt(iRow, iCol);
                  }
                }
                Object[] p_oHeader = new Object[oTable.getColumnCount()];
                for (iCol = 0; iCol < p_oHeader.length; iCol++) {
                  p_oHeader[iCol] = oTable.getColumnModel().getColumn(iCol).
                      getHeaderValue();
                }

                p_oData.add(WorkerBase.MakeTableVector(p_oHeader,
                    p_oTableData));
              }
            }
          }
        }

        //Now get the managed object that matches the label
        for (j = 0; j < mp_oManagedObjects.length; j++) {
          if (mp_oManagedObjects[j].GetName().equals(sObjectName)) {
            mp_oManagedObjects[j].ReadDataFromDisplay(p_oData, oPop);
            mp_oManagedObjects[j].ValidateData(oPop);
          }
        }
      }
    }
  }

  /**
   * Triggers the DoSetup() methods of all BehaviorTypeBase objects under
   * management.
   * @throws ModelException passed through from the BehaviorTypeBase objects.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void DoSetup() throws ModelException {
    TreePopulation oPop = GetTreePopulation();
    if (oPop == null) {
      return;
    }
    int i;
    for (i = 0; i < mp_oManagedObjects.length; i++) {
      mp_oManagedObjects[i].DoSetup(oPop);
    }
  }
}
