package javawrapper;

/**
* Controls the data relating to growth behaviors.
* <p>Copyright: Copyright (c) 2003 Charles D. Canham</p>
* <p>Company: Institute of Ecosystem Studies</p>
* @author Lora E. Murphy
* @version 1.0
*
* <br>Edit history:
* <br>------------------
* <br>April 28, 2004: Submitted in beta version (LEM)
* <br>August 27, 2004: Added PR growth (LEM)
* <br>October 12, 2004:  Changed lambda format for NCI (LEM)
* <br>November 30, 2004:  Fixed a bug for when you remove a species - it
* resets the neighbor lambdas (LEM)
* <br>January 18, 2005:  Put in the consolidated NCI behavior (LEM)
* <br>February 8, 2005: Put in the double resource growth behavior (LEM)
* <br>February 10, 2005:  Put in the stochastic gap growth behavior (LEM)
* <br>March 16, 2005:  Upgraded NCI growth (LEM)
* <br>May 5, 2005:  Added linear bi-level growth (LEM)
* <br>May 9, 2005:  Added lognormal bi-level height growth (LEM)
* <br>May 12, 2005:  Added new flag for NCI (LEM)
*/

public class GrowthBehaviors
    extends BehaviorTypeBase {

  /**Asymptotic diameter growth for each species*/
  protected ModelVector mp_fAsymptoticDiameterGrowth = new ModelVector(
      "Asymptotic Diameter Growth (A)",
      "gr_asympDiameterGrowth", "gr_adgVal", 0, ModelVector.FLOAT);

  /**Slope of growth response for each species*/
  protected ModelVector mp_fSlopeOfGrowthResponse = new ModelVector(
      "Slope of Growth Response (S)",
      "gr_slopeGrowthResponse",
      "gr_sgrVal", 0,
      ModelVector.FLOAT);

  /**Adult constant radial growth for each species, in mm/yr*/
  protected ModelVector mp_fAdultConstantRadialGrowth = new ModelVector(
      "Adult Constant Radial Growth in mm/yr", "gr_adultConstRadialInc",
      "gr_acriVal", 0, ModelVector.FLOAT);

  /**Adult constant area increment, in cm2/yr*/
  protected ModelVector mp_fAdultConstantAreaInc = new ModelVector(
      "Adult Constant Area Growth in sq. cm/yr", "gr_adultConstAreaInc",
      "gr_acaiVal", 0, ModelVector.FLOAT);

  /**NCI max potential growth for each species*/
  protected ModelVector mp_fNCIMaxPotentialGrowth = new ModelVector(
      "NCI Maximum Potential Growth, cm/yr",
      "gr_nciMaxPotentialGrowth",
      "gr_nmpgVal", 0,
      ModelVector.FLOAT);

  /**NCI maximum crowding distance, in m, for each species*/
  protected ModelVector mp_fNCIMaxCrowdingRadius = new ModelVector(
      "NCI Maximum Crowding Distance, in meters", "gr_nciMaxCrowdingRadius",
      "gr_nmcrVal", 0, ModelVector.FLOAT);

  /**NCI minimum DBH for crowding neighbors, for each species; all species
   * required*/
  protected ModelVector mp_fNCIMinNeighborDBH = new ModelVector(
      "NCI Minimum Neighbor DBH, in cm", "gr_nciMinNeighborDBH",
      "gr_nmndVal", 0, ModelVector.FLOAT, true);

  /**NCI alpha for each species*/
  protected ModelVector mp_fNCIAlpha = new ModelVector("NCI Alpha",
      "gr_nciAlpha", "gr_naVal", 0,
      ModelVector.FLOAT);

  /**NCI beta for each species*/
  protected ModelVector mp_fNCIBeta = new ModelVector("NCI Beta", "gr_nciBeta",
      "gr_nbVal", 0,
      ModelVector.FLOAT);

  /**NCI growth - Size sensitivity to NCI parameter (gamma)*/
  protected ModelVector mp_fNCISizeSensToNCI = new ModelVector(
      "NCI Size Sensitivity to NCI (gamma)", "gr_nciSizeSensNCI",
      "gr_nssnVal", 0, ModelVector.FLOAT);

  /**NCI growth - Crowding Effect Slope (C)*/
  protected ModelVector mp_fNCICrowdingEffectSlope = new ModelVector(
      "NCI Crowding Effect Slope (C)", "gr_nciCrowdingSlope", "gr_ncslVal", 0,
      ModelVector.FLOAT);

  /**NCI growth - Crowding Effect Steepness (D)*/
  protected ModelVector mp_fNCICrowdingEffectSteepness = new ModelVector(
      "NCI Crowding Effect Steepness (D)", "gr_nciCrowdingSteepness",
      "gr_ncstVal", 0, ModelVector.FLOAT);

  /**NCI growth - Neighbor storm effect - medium damage*/
  protected ModelVector mp_fNCINeighStormEffMed = new ModelVector(
      "NCI Neighbor Storm Damage (eta) - Medium (0-1)",
      "gr_nciNeighStormEffMedDmg", "gr_nnsemdVal", 0, ModelVector.FLOAT, true);

  /**NCI growth - Neighbor storm effect - full damage*/
  protected ModelVector mp_fNCINeighStormEffFull = new ModelVector(
      "NCI Neighbor Storm Damage (eta) - Complete (0-1)",
      "gr_nciNeighStormEffFullDmg", "gr_nnsefdVal", 0, ModelVector.FLOAT, true);

  /**NCI growth - Size effect mode (X0)*/
  protected ModelVector mp_fNCISizeEffectMode = new ModelVector(
      "NCI Size Effect Mode, in cm (X0)", "gr_nciSizeEffectMode",
      "gr_nsemVal", 0, ModelVector.FLOAT);

  /**NCI growth - Size effect variance (Xb)*/
  protected ModelVector mp_fNCISizeEffectVariance = new ModelVector(
      "NCI Size Effect Variance, in cm (Xb)", "gr_nciSizeEffectVariance",
      "gr_nsevVal", 0, ModelVector.FLOAT);

  /**NCI growth - Shading coefficient (m)*/
  protected ModelVector mp_fNCIShadingEffectCoefficient = new ModelVector(
      "NCI Shading Effect Coefficient (m)", "gr_nciShadingCoefficient",
      "gr_nscVal", 0, ModelVector.FLOAT);

  /**NCI growth - Shading exponent (n)*/
  protected ModelVector mp_fNCIShadingEffectExponent = new ModelVector(
      "NCI Shading Effect Exponent (n)", "gr_nciShadingExponent",
      "gr_nseVal", 0, ModelVector.FLOAT);

  /**NCI growth - Storm effect - medium damage*/
  protected ModelVector mp_fNCIStormEffectMed = new ModelVector(
      "NCI Damage Effect - Medium Storm Damage (0-1)", "gr_nciStormEffMedDmg",
      "gr_nsemdVal", 0, ModelVector.FLOAT);

  /**NCI growth - Storm effect - full damage*/
  protected ModelVector mp_fNCIStormEffectFull = new ModelVector(
      "NCI Damage Effect - Complete Storm Damage (0-1)",
      "gr_nciStormEffFullDmg", "gr_nsefdVal", 0, ModelVector.FLOAT);

  /**Length of last suppression factor for each species*/
  protected ModelVector mp_fLengthOfLastSuppressionFactor = new ModelVector(
      "Length of Last Suppression Factor", "gr_lengthLastSuppFactor",
      "gr_llsfVal", 0, ModelVector.FLOAT);

  /**Length of current release factor for each species*/
  protected ModelVector mp_fLengthOfCurrentReleaseFactor = new ModelVector(
      "Length of Current Release Factor", "gr_lengthCurrReleaseFactor",
      "gr_lcrfVal", 0, ModelVector.FLOAT);

  /**Logistic diameter growth - asymptotic growth at full light*/
  protected ModelVector mp_fLogisticDiamAsympGrowthFullLight = new ModelVector(
      "Logistic - Asymptotic Diam Growth - Full Light in mm/yr (a)",
      "gr_diamAsympGrowthAtFullLight", "gr_dagaflVal", 0, ModelVector.FLOAT);

  /**Logistic height growth - asymptotic growth at full light*/
  protected ModelVector mp_fLogisticHeightAsympGrowthFullLight = new
      ModelVector(
      "Logistic - Asymptotic Height Growth - Full Light in cm/yr (a)",
      "gr_heightAsympGrowthAtFullLight", "gr_hagaflVal", 0, ModelVector.FLOAT);

  /**Logistic diameter growth - shape parameter 1 - b*/
  protected ModelVector mp_fLogisticDiamShape1b = new ModelVector(
      "Logistic - Diam Shape Param 1 (b)", "gr_diamShape1b", "gr_ds1bVal",
      0, ModelVector.FLOAT);

  /**Logistic height growth - shape parameter 1 - b*/
  protected ModelVector mp_fLogisticHeightShape1b = new ModelVector(
      "Logistic - Height Shape Param 1 (b)", "gr_heightShape1b",
      "gr_hs1bVal", 0, ModelVector.FLOAT);

  /**Logistic diameter growth - shape parameter 2 - c*/
  protected ModelVector mp_fLogisticDiamShape2c = new ModelVector(
      "Logistic - Diam Shape Param 2 (c)", "gr_diamShape2c", "gr_ds2cVal",
      0, ModelVector.FLOAT);

  /**Logistic height growth - shape parameter 2 - c*/
  protected ModelVector mp_fLogisticHeightShape2c = new ModelVector(
      "Logistic - Height Shape Param 2 (c)", "gr_heightShape2c", "gr_hs2cVal",
      0, ModelVector.FLOAT);

  /**Simple linear diameter growth - intercept*/
  protected ModelVector mp_fSimpLinDiamIntercept = new ModelVector(
      "Simple Linear - Diam Intercept in mm/yr (a)",
      "gr_diamSimpleLinearIntercept",
      "gr_dsliVal", 0, ModelVector.FLOAT);

  /**Simple linear height growth - intercept*/
  protected ModelVector mp_fSimpLinHeightIntercept = new ModelVector(
      "Simple Linear - Height Intercept in cm/yr (a)",
      "gr_heightSimpleLinearIntercept",
      "gr_hsliVal", 0, ModelVector.FLOAT);

  /**Simple linear diameter growth - slope*/
  protected ModelVector mp_fSimpLinDiamSlope = new ModelVector(
      "Simple Linear - Diam Slope (b)", "gr_diamSimpleLinearSlope",
      "gr_dslsVal", 0, ModelVector.FLOAT);

  /**Simple linear height growth - slope*/
  protected ModelVector mp_fSimpLinHeightSlope = new ModelVector(
      "Simple Linear - Height Slope (b)", "gr_heightSimpleLinearSlope",
      "gr_hslsVal", 0, ModelVector.FLOAT);

  /** Linear growth with exponential reduction for shade diameter growth - intercept*/
  protected ModelVector mp_fLinShadeDiamIntercept = new ModelVector(
      "Shaded Linear - Diam Intercept in mm/yr (a)",
      "gr_diamShadedLinearIntercept",
      "gr_dshliVal", 0, ModelVector.FLOAT);

  /** Linear growth with exponential reduction for shade height growth - intercept*/
  protected ModelVector mp_fLinShadeHeightIntercept = new ModelVector(
      "Shaded Linear - Height Intercept in cm/yr (a)",
      "gr_heightShadedLinearIntercept",
      "gr_hshliVal", 0, ModelVector.FLOAT);

  /** Linear growth with exponential reduction for shade diameter growth - slope*/
  protected ModelVector mp_fLinShadeDiamSlope = new ModelVector(
      "Shaded Linear - Diam Slope (b)", "gr_diamShadedLinearSlope",
      "gr_dshlsVal", 0, ModelVector.FLOAT);

  /** Linear growth with exponential reduction for shade height growth - slope*/
  protected ModelVector mp_fLinShadeHeightSlope = new ModelVector(
      "Shaded Linear - Height Slope (b)", "gr_heightShadedLinearSlope",
      "gr_hshlsVal", 0, ModelVector.FLOAT);

  /** Linear growth with exponential reduction for shade diameter growth - shade exponent*/
  protected ModelVector mp_fLinShadeDiamShadeExp = new ModelVector(
      "Shaded Linear - Diam Shade Exponent (c)", "gr_diamShadedLinearShadeExp",
      "gr_dshlseVal", 0, ModelVector.FLOAT);

  /** Linear growth with exponential reduction for shade height growth - shade exponent*/
  protected ModelVector mp_fLinShadeHeightShadeExp = new ModelVector(
      "Shaded Linear - Height Shade Exponent (c)",
      "gr_heightShadedLinearShadeExp", "gr_hshlseVal", 0, ModelVector.FLOAT);

  /** Logistic growth with size dependent asymptote - diameter intercept*/
  protected ModelVector mp_fSizeDepLogisticDiamIntercept = new ModelVector(
      "Size Dep. Logistic - Diam Intercept (a)",
      "gr_diamSizeDepLogisticIntercept", "gr_dsdliVal", 0, ModelVector.FLOAT);

  /** Logistic growth with size dependent asymptote - height intercept*/
  protected ModelVector mp_fSizeDepLogisticHeightIntercept = new ModelVector(
      "Size Dep. Logistic - Height Intercept (a)",
      "gr_heightSizeDepLogisticIntercept", "gr_hsdliVal", 0, ModelVector.FLOAT);

  /** Logistic growth with size dependent asymptote - diameter slope*/
  protected ModelVector mp_fSizeDepLogisticDiamSlope = new ModelVector(
      "Size Dep. Logistic - Diam Slope (b)", "gr_diamSizeDepLogisticSlope",
      "gr_dsdlsVal", 0, ModelVector.FLOAT);

  /** Logistic growth with size dependent asymptote - height slope*/
  protected ModelVector mp_fSizeDepLogisticHeightSlope = new ModelVector(
      "Size Dep. Logistic - Height Slope (b)", "gr_heightSizeDepLogisticSlope",
      "gr_hsdlsVal", 0, ModelVector.FLOAT);

  /** Logistic growth with size dependent asymptote - diameter shape parameter 1 - c*/
  protected ModelVector mp_fSizeDepLogisticDiamShape1c = new ModelVector(
      "Size Dep. Logistic - Diam Shape Param 1 (c)",
      "gr_diamSizeDepLogisticShape1c", "gr_dsdls1cVal", 0, ModelVector.FLOAT);

  /** Logistic growth with size dependent asymptote - height shape parameter 1 - c*/
  protected ModelVector mp_fSizeDepLogisticHeightShape1c = new ModelVector(
      "Size Dep. Logistic - Height Shape Param 1 (c)",
      "gr_heightSizeDepLogisticShape1c", "gr_hsdls1cVal", 0,
      ModelVector.FLOAT);

  /** Logistic growth with size dependent asymptote - diameter shape parameter 2 - d*/
  protected ModelVector mp_fSizeDepLogisticDiamShape2d = new ModelVector(
      "Size Dep. Logistic - Diam Shape Param 2 (d)",
      "gr_diamSizeDepLogisticShape2d", "gr_dsdls2dVal", 0, ModelVector.FLOAT);

  /** Logistic growth with size dependent asymptote - height shape parameter 2 - d*/
  protected ModelVector mp_fSizeDepLogisticHeightShape2d = new ModelVector(
      "Size Dep. Logistic - Height Shape Param 2 (d)",
      "gr_heightSizeDepLogisticShape2d", "gr_hsdls2dVal", 0, ModelVector.FLOAT);

  /** Lognormal diameter growth - growth increment at diameter 36*/
  protected ModelVector mp_fLognormalDiamIncAtDiam36 = new ModelVector(
      "Lognormal - Diam Growth Increment at Diam 36, in mm/yr (a)",
      "gr_diamLognormalIncAtDiam36", "gr_dliad36Val", 0, ModelVector.FLOAT);

  /** Lognormal height growth - height increment at diameter 36*/
  protected ModelVector mp_fLognormalHeightIncAtDiam36 = new ModelVector(
      "Lognormal - Height Growth Increment at Diam 36, in cm/yr (a)",
      "gr_heightLognormalIncAtDiam36", "gr_hliad36Val", 0, ModelVector.FLOAT);

  /** Lognormal diameter growth - shape parameter*/
  protected ModelVector mp_fLognormalDiamShapeParam = new ModelVector(
      "Lognormal - Diam Shape Parameter (b)", "gr_diamLognormalShapeParam",
      "gr_dlspVal", 0, ModelVector.FLOAT);

  /** Lognormal height growth - shape parameter*/
  protected ModelVector mp_fLognormalHeightShapeParam = new ModelVector(
      "Lognormal - Height Shape Parameter (b)", "gr_heightLognormalShapeParam",
      "gr_hlspVal", 0, ModelVector.FLOAT);

  /** Lognormal diameter growth - effect of shade*/
  protected ModelVector mp_fLognormalDiamEffectOfShade = new ModelVector(
      "Lognormal - Diam Effect of Shade (c)", "gr_diamLognormalEffectOfShade",
      "gr_dleosVal", 0, ModelVector.FLOAT);

  /** Lognormal height growth - effect of shade*/
  protected ModelVector mp_fLognormalHeightEffectOfShade = new ModelVector(
      "Lognormal - Height Effect of Shade (c)",
      "gr_heightLognormalEffectOfShade", "gr_hleosVal", 0, ModelVector.FLOAT);

  /** Double resource relative growth - influence of resource*/
  protected ModelVector mp_fDoubleResourceInfluence = new ModelVector(
      "Double resource - Influence of Resource (C)",
      "gr_diamDoubleMMResourceInfluence", "gr_ddmmriVal", 0, ModelVector.FLOAT);

  /** Linear bi-level growth - threshold for high-light growth*/
  protected ModelVector mp_fBiLevelHighLightThreshold = new ModelVector(
      "Linear Bi-Level - Threshold for High-Light Growth (0 - 100)",
      "gr_linearBilevHiLiteThreshold", "gr_lbhltVal", 0, ModelVector.FLOAT);

  /** Linear bi-level growth - slope for high-light growth*/
  protected ModelVector mp_fBiLevelHighLightSlope = new ModelVector(
      "Linear Bi-Level - Slope for High-Light Growth (b)",
      "gr_linearBilevHiLiteSlope", "gr_lbhlsVal", 0, ModelVector.FLOAT);

  /** Linear bi-level growth - intercept for high-light growth*/
  protected ModelVector mp_fBiLevelHighLightIntercept = new ModelVector(
      "Linear Bi-Level - Intercept for High-Light Growth (a)",
      "gr_linearBilevHiLiteIntercept", "gr_lbhliVal", 0, ModelVector.FLOAT);

  /** Linear bi-level growth - slope for low-light growth*/
  protected ModelVector mp_fBiLevelLowLightSlope = new ModelVector(
      "Linear Bi-Level - Slope for Low-Light Growth (b)",
      "gr_linearBilevLoLiteSlope", "gr_lbllsVal", 0, ModelVector.FLOAT);

  /** Linear bi-level growth - intercept for low-light growth*/
  protected ModelVector mp_fBiLevelLowLightIntercept = new ModelVector(
      "Linear Bi-Level - Intercept for Low-Light Growth (a)",
      "gr_linearBilevLoLiteIntercept", "gr_lblliVal", 0, ModelVector.FLOAT);

  /** Lognormal bi-level growth - X0 for low-light growth*/
  protected ModelVector mp_fLogBiLevelLoLiteX0 = new ModelVector(
      "Lognormal Bi-Level - X0 for Low-Light Growth",
      "gr_lognormalBilevLoLiteX0", "gr_lbllx0Val", 0, ModelVector.FLOAT);

  /** Lognormal bi-level growth - Xb for low-light growth*/
  protected ModelVector mp_fLogBiLevelLoLiteXb = new ModelVector(
      "Lognormal Bi-Level - Xb for Low-Light Growth",
      "gr_lognormalBilevLoLiteXb", "gr_lbllxbVal", 0, ModelVector.FLOAT);

  /** Lognormal bi-level growth - max growth for low-light growth*/
  protected ModelVector mp_fLogBiLevelLoLiteMaxGrwth = new ModelVector(
      "Lognormal Bi-Level - Max Growth in Low Light (m)",
      "gr_lognormalBilevLoLiteMaxGrowth", "gr_lbllmgVal", 0, ModelVector.FLOAT);

  /** Lognormal bi-level growth - X0 for high-light growth*/
  protected ModelVector mp_fLogBiLevelHiLiteX0 = new ModelVector(
      "Lognormal Bi-Level - X0 for High-Light Growth",
      "gr_lognormalBilevHiLiteX0", "gr_lbhlx0Val", 0, ModelVector.FLOAT);

  /** Lognormal bi-level growth - Xb for high-light growth*/
  protected ModelVector mp_fLogBiLevelHiLiteXb = new ModelVector(
      "Lognormal Bi-Level - Xb for High-Light Growth",
      "gr_lognormalBilevHiLiteXb", "gr_lbhlxbVal", 0, ModelVector.FLOAT);

  /** Lognormal bi-level growth - max growth for high-light growth*/
  protected ModelVector mp_fLogBiLevelHiLiteMaxGrwth = new ModelVector(
      "Lognormal Bi-Level - Max Growth in High Light (m)",
      "gr_lognormalBilevHiLiteMaxGrowth", "gr_lbhlmgVal", 0, ModelVector.FLOAT);

  /** Lognormal bi-level growth - threshold for high-light growth*/
  protected ModelVector mp_fLogBiLevelHighLightThreshold = new ModelVector(
      "Lognormal Bi-Level - Threshold for High-Light Growth (0 - 100)",
      "gr_lognormalBilevHiLiteThreshold", "gr_lobhltVal", 0, ModelVector.FLOAT);

  /** PR semi-stochastic growth - height threshold for stochastic growth*/
  protected ModelVector mp_fPRHiteThreshold = new ModelVector(
      "PR - Height Threshold for Stochastic Growth (m)",
      "gr_prStochHiteThreshold", "gr_pshtVal", 0, ModelVector.FLOAT);

  /** PR semi-stochastic growth - "a" for deterministic growth*/
  protected ModelVector mp_fPRDetermA = new ModelVector(
      "PR - \"a\" Parameter for Deterministic Growth",
      "gr_prStochDetermA", "gr_psdaVal", 0, ModelVector.FLOAT);

  /** PR semi-stochastic growth - "b" for deterministic growth*/
  protected ModelVector mp_fPRDetermB = new ModelVector(
      "PR - \"b\" Parameter for Deterministic Growth",
      "gr_prStochDetermB", "gr_psdbVal", 0, ModelVector.FLOAT);

  /** PR semi-stochastic growth - mean DBH for stochastic growth*/
  protected ModelVector mp_fPRMeanDBH = new ModelVector(
      "PR - Mean DBH (cm) for Stochastic Growth",
      "gr_prStochMeanDBH", "gr_psmdVal", 0, ModelVector.FLOAT);

  /** PR semi-stochastic growth - DBH standard deviation for stochastic growth*/
  protected ModelVector mp_fPRDBHStdDev = new ModelVector(
      "PR - DBH Standard Deviation for Stochastic Growth",
      "gr_prStochStdDev", "gr_pssdVal", 0, ModelVector.FLOAT);

  /**Morality rate at suppression*/
  protected ModelFloat m_fMortalityRateAtSuppression = new ModelFloat(0,
      "Mortality Threshold for Suppression", "gr_mortRateAtSuppression");

  /**NCI - NCI DBH divisor*/
  protected ModelFloat m_fNCIDbhDivisor = new ModelFloat(1,
      "NCI DBH Divisor (q)", "gr_nciDbhDivisor");

  /**Years exceeding threshold before a tree is suppressed*/
  protected ModelInt m_iYrsExceedingThresholdBeforeSuppressed = new ModelInt(0,
      "Years Exceeding Threshold Before a Tree is Suppressed",
      "gr_yrsExceedThresholdBeforeSupp");

  /**NCI - Whether or not to include snags in NCI calculations - LEM 05/12/05*/
  protected ModelEnum m_iIncludeSnagsInNCI =
      new ModelEnum(new int[] {0, 1},
                    new String[] {"false", "true"},
                    "Include Snags in NCI Calculations",
                    "gr_nciIncludeSnagsInNCI");


  /**NCI - Whether or not to relativize NCI to NCImax
   * @deprecated - will be removed in version 7 (?)*/
  protected ModelEnum m_iRelativizeNCI =
      new ModelEnum(new int[] {0, 1},
                    new String[] {"false", "true"},
                    "Relativize NCI with NCI Max",
                    "gr_nciRelativizeWithNCIMax");

  /**
   * Constructor.
   * @param oManager GUIManager object.
   * @throws ModelException Passed through from called functions.  Shouldn't
   * ever be thrown.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   * <br>January 18, 2005:  Replaced all NCI behaviors with generic (LEM)
   */
  public GrowthBehaviors(GUIManager oManager) throws ModelException {
    super(oManager, "Growth", "growth");

    m_sHelpID = "data.growth_behaviors";

    //Set up growth behaviors
    mp_oChildBehaviors = new Behavior[41];
    int iIndex = -1;

    //Default the snags in NCI flag to false
    m_iIncludeSnagsInNCI.SetValue("false");

    //*****************
     //Absolute growth limited by basal area - auto-height
     //*****************
    iIndex++;
    mp_oChildBehaviors[iIndex] = new Behavior("absba",
                                         "Absolute growth limited to basal area increment - diam with auto height",
                                         "absbagrowth",
                                         (float) 1);
    //Data
    mp_oChildBehaviors[iIndex].AddRequiredData(m_fMortalityRateAtSuppression);
    mp_oChildBehaviors[iIndex].AddRequiredData(
        m_iYrsExceedingThresholdBeforeSuppressed);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fAdultConstantAreaInc);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fAsymptoticDiameterGrowth);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fSlopeOfGrowthResponse);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fLengthOfLastSuppressionFactor);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fLengthOfCurrentReleaseFactor);
    //Data members
    mp_oChildBehaviors[iIndex].mp_oNewTreeDataMembers.add(new DataMember(
        "Years suppressed", "yls", DataMember.INTEGER));
    mp_oChildBehaviors[iIndex].mp_oNewTreeDataMembers.add(new DataMember(
        "Years released", "ylr", DataMember.INTEGER));
    mp_oChildBehaviors[iIndex].mp_oNewTreeDataMembers.add(new DataMember("Diameter growth", "Growth", DataMember.FLOAT));

    //*****************
     //Absolute growth limited by radial increment - auto-height
     //*****************
      iIndex++;
    mp_oChildBehaviors[iIndex] = new Behavior("absradial",
                                         "Absolute growth limited to radial increment - diam with auto height",
                                         "absradialgrowth",
                                         (float) 1);
    //Data
    mp_oChildBehaviors[iIndex].AddRequiredData(m_fMortalityRateAtSuppression);
    mp_oChildBehaviors[iIndex].AddRequiredData(
        m_iYrsExceedingThresholdBeforeSuppressed);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fAdultConstantRadialGrowth);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fAsymptoticDiameterGrowth);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fSlopeOfGrowthResponse);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fLengthOfLastSuppressionFactor);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fLengthOfCurrentReleaseFactor);
    //Data members
    mp_oChildBehaviors[iIndex].mp_oNewTreeDataMembers.add(new DataMember(
        "Years suppressed", "yls", DataMember.INTEGER));
    mp_oChildBehaviors[iIndex].mp_oNewTreeDataMembers.add(new DataMember(
        "Years released", "ylr", DataMember.INTEGER));
    mp_oChildBehaviors[iIndex].mp_oNewTreeDataMembers.add(new DataMember("Diameter growth", "Growth", DataMember.FLOAT));

    //*****************
     //Constant basal area increment - auto-height
     //*****************
      iIndex++;
    mp_oChildBehaviors[iIndex] = new Behavior("constba",
                                         "Constant basal area growth - diam with auto height",
                                         "constbagrowth", (float) 1);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fAdultConstantAreaInc);
    mp_oChildBehaviors[iIndex].mp_oNewTreeDataMembers.add(new DataMember("Diameter growth", "Growth", DataMember.FLOAT));

    //*****************
     //Constant radial increment - auto-height
     //*****************
      iIndex++;
    mp_oChildBehaviors[iIndex] = new Behavior("constradial",
                                         "Constant radial growth - diam with auto height",
                                         "constradialgrowth", (float) 1);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fAdultConstantRadialGrowth);
    mp_oChildBehaviors[iIndex].mp_oNewTreeDataMembers.add(new DataMember("Diameter growth", "Growth", DataMember.FLOAT));

    //*****************
     //Simple linear regression - auto-height
     //*****************
      iIndex++;
    mp_oChildBehaviors[iIndex] = new Behavior("linear",
                                         "Linear growth - diam with auto height",
                                         "simple linear growth",
                                         (float) 1);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fSimpLinDiamIntercept);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fSimpLinDiamSlope);
    mp_oChildBehaviors[iIndex].mp_oNewTreeDataMembers.add(new DataMember("Diameter growth", "Growth", DataMember.FLOAT));

    //*****************
     //Linear growth with exponential reduction for shade - auto-height
     //*****************
      iIndex++;
    mp_oChildBehaviors[iIndex] = new Behavior("shaded linear",
                                         "Linear growth w/ exponential shade reduction - diam with auto height",
                                         "shaded linear growth", (float) 1);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fLinShadeDiamIntercept);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fLinShadeDiamSlope);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fLinShadeDiamShadeExp);
    mp_oChildBehaviors[iIndex].mp_oNewTreeDataMembers.add(new DataMember("Diameter growth", "Growth", DataMember.FLOAT));

    //*****************
     //Linear bi-level growth - auto-height
     //*****************
      iIndex++;
    mp_oChildBehaviors[iIndex] = new Behavior("linear bi-level",
                                         "Linear bi-level growth - diam with auto height",
                                         "linear bilevel growth", (float) 1);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fBiLevelHighLightThreshold);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fBiLevelLowLightSlope);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fBiLevelLowLightIntercept);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fBiLevelHighLightSlope);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fBiLevelHighLightIntercept);
    mp_oChildBehaviors[iIndex].mp_oNewTreeDataMembers.add(new DataMember("Diameter growth", "Growth", DataMember.FLOAT));

    //*****************
     //Logistic growth - auto-height
     //*****************
      iIndex++;
    mp_oChildBehaviors[iIndex] = new Behavior("logistic growth",
                                         "Logistic growth - diam with auto height",
                                         "logistic growth", (float) 1);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fLogisticDiamAsympGrowthFullLight);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fLogisticDiamShape1b);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fLogisticDiamShape2c);
    mp_oChildBehaviors[iIndex].mp_oNewTreeDataMembers.add(new DataMember("Diameter growth", "Growth", DataMember.FLOAT));

    //*****************
     //Logistic growth with size dependent asymptote - auto-height
     //*****************
      iIndex++;
    mp_oChildBehaviors[iIndex] = new Behavior("size dependent logistic growth",
                                         "Logistic growth w/ size dependent asymptote - diam with auto height",
                                         "size dependent logistic growth",
                                         (float) 1);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fSizeDepLogisticDiamIntercept);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fSizeDepLogisticDiamSlope);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fSizeDepLogisticDiamShape1c);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fSizeDepLogisticDiamShape2d);
    mp_oChildBehaviors[iIndex].mp_oNewTreeDataMembers.add(new DataMember("Diameter growth", "Growth", DataMember.FLOAT));

    //*****************
     //Lognormal growth with exponential reduction for shade - auto-height
     //*****************
      iIndex++;
    mp_oChildBehaviors[iIndex] = new Behavior("lognormal growth",
                                         "Lognormal with exponential shade reduction - diam with auto height",
                                         "lognormal growth",
                                         (float) 1);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fLognormalDiamIncAtDiam36);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fLognormalDiamShapeParam);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fLognormalDiamEffectOfShade);
    mp_oChildBehaviors[iIndex].mp_oNewTreeDataMembers.add(new DataMember("Diameter growth", "Growth", DataMember.FLOAT));

    //*****************
     //NCI growth - auto-height
     //*****************
      iIndex++;
    mp_oChildBehaviors[iIndex] = new Behavior("nci",
                                         "NCI growth - diam with auto height",
                                         "ncigrowth",
                                         (float) 2.1);

    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fNCIMaxPotentialGrowth);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fNCIMaxCrowdingRadius);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fNCIAlpha);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fNCIBeta);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fNCISizeSensToNCI);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fNCICrowdingEffectSlope);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fNCICrowdingEffectSteepness);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fNCINeighStormEffMed);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fNCINeighStormEffFull);
    mp_oChildBehaviors[iIndex].AddRequiredData(m_fNCIDbhDivisor);
    mp_oChildBehaviors[iIndex].AddRequiredData(m_iIncludeSnagsInNCI);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fNCIMinNeighborDBH);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fNCISizeEffectMode);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fNCISizeEffectVariance);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fNCIShadingEffectCoefficient);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fNCIShadingEffectExponent);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fNCIStormEffectMed);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fNCIStormEffectFull);
    //Disallow for seedlings
    mp_oChildBehaviors[iIndex].SetCanApplyTo(TreePopulation.SEEDLING, false);
    mp_oChildBehaviors[iIndex].mp_oNewTreeDataMembers.add(new DataMember("Diameter growth", "Growth", DataMember.FLOAT));

    //*****************
     //Relative growth limited by basal area - auto-height
     //*****************
      iIndex++;
    mp_oChildBehaviors[iIndex] = new Behavior("relba",
                                          "Relative growth limited to basal area increment - diam with auto height",
                                          "relbagrowth",
                                          (float) 1);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fAdultConstantAreaInc);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fAsymptoticDiameterGrowth);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fSlopeOfGrowthResponse);
    mp_oChildBehaviors[iIndex].mp_oNewTreeDataMembers.add(new DataMember("Diameter growth", "Growth", DataMember.FLOAT));

    //*****************
     //Relative growth limited by radial increment - auto-height
     //*****************
      iIndex++;
    mp_oChildBehaviors[iIndex] = new Behavior("relradial",
                                          "Relative growth limited to radial increment - diam with auto height",
                                          "relradialgrowth",
                                          (float) 1);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fAdultConstantRadialGrowth);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fAsymptoticDiameterGrowth);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fSlopeOfGrowthResponse);
    mp_oChildBehaviors[iIndex].mp_oNewTreeDataMembers.add(new DataMember("Diameter growth", "Growth", DataMember.FLOAT));

    //*****************
     //Unlimited absolute growth - auto-height
     //*****************
      iIndex++;
    mp_oChildBehaviors[iIndex] = new Behavior("absunlim",
                                          "Unlimited absolute growth - diam with auto height",
                                          "absunlimgrowth", (float) 1);
    //Data
    mp_oChildBehaviors[iIndex].AddRequiredData(m_fMortalityRateAtSuppression);
    mp_oChildBehaviors[iIndex].AddRequiredData(
        m_iYrsExceedingThresholdBeforeSuppressed);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fAsymptoticDiameterGrowth);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fSlopeOfGrowthResponse);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fLengthOfLastSuppressionFactor);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fLengthOfCurrentReleaseFactor);
    //Data members
    mp_oChildBehaviors[iIndex].mp_oNewTreeDataMembers.add(new DataMember(
        "Years suppressed", "yls", DataMember.INTEGER));
    mp_oChildBehaviors[iIndex].mp_oNewTreeDataMembers.add(new DataMember(
        "Years released", "ylr", DataMember.INTEGER));
    mp_oChildBehaviors[iIndex].mp_oNewTreeDataMembers.add(new DataMember("Diameter growth", "Growth", DataMember.FLOAT));

    //*****************
     //Unlimited relative growth - auto-height
     //*****************
      iIndex++;
    mp_oChildBehaviors[iIndex] = new Behavior("relunlim",
                                          "Unlimited relative growth - diam with auto height",
                                          "relunlimgrowth", (float) 1);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fAsymptoticDiameterGrowth);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fSlopeOfGrowthResponse);
    mp_oChildBehaviors[iIndex].mp_oNewTreeDataMembers.add(new DataMember("Diameter growth", "Growth", DataMember.FLOAT));

    //*****************
     //Double resource Michaelis-Menton relative growth - auto-height
     //*****************
      iIndex++;
    mp_oChildBehaviors[iIndex] = new Behavior("double resource",
                                          "Double resource relative growth - diam with auto height",
                                          "Double resource relative diam with auto height", (float) 1);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fAsymptoticDiameterGrowth);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fSlopeOfGrowthResponse);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fDoubleResourceInfluence);
    mp_oChildBehaviors[iIndex].mp_oNewTreeDataMembers.add(new DataMember("Diameter growth", "Growth", DataMember.FLOAT));

    //*****************
     //Stochastic gap growth behavior
     //*****************
      iIndex++;
    mp_oChildBehaviors[iIndex] = new Behavior("stochastic gap",
                                          "Stochastic Gap Growth",
                                          "Stochastic Gap Growth", (float) 1);
    mp_oChildBehaviors[iIndex].mp_oNewTreeDataMembers.add(new DataMember("Diameter growth", "Growth", DataMember.FLOAT));

    //*****************
     //Absolute growth limited by basal area - diam only
     //*****************
      iIndex++;
    mp_oChildBehaviors[iIndex] = new Behavior("absba diam only",
                                          "Absolute growth limited to basal area increment - diam only",
                                          "absbagrowth diam only",
                                          (float) 1);
    //Data
    mp_oChildBehaviors[iIndex].AddRequiredData(m_fMortalityRateAtSuppression);
    mp_oChildBehaviors[iIndex].AddRequiredData(
        m_iYrsExceedingThresholdBeforeSuppressed);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fAdultConstantAreaInc);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fAsymptoticDiameterGrowth);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fSlopeOfGrowthResponse);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fLengthOfLastSuppressionFactor);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fLengthOfCurrentReleaseFactor);
    //Data members
    mp_oChildBehaviors[iIndex].mp_oNewTreeDataMembers.add(new DataMember(
        "Years suppressed", "yls", DataMember.INTEGER));
    mp_oChildBehaviors[iIndex].mp_oNewTreeDataMembers.add(new DataMember(
        "Years released", "ylr", DataMember.INTEGER));
    mp_oChildBehaviors[iIndex].mp_oNewTreeDataMembers.add(new DataMember("Diameter growth", "Growth", DataMember.FLOAT));

    //*****************
     //Absolute growth limited by radial increment - diam only
     //*****************
      iIndex++;
    mp_oChildBehaviors[iIndex] = new Behavior("absradial diam only",
                                          "Absolute growth limited to radial increment - diam only",
                                          "absradialgrowth diam only",
                                          (float) 1);
    //Data
    mp_oChildBehaviors[iIndex].AddRequiredData(m_fMortalityRateAtSuppression);
    mp_oChildBehaviors[iIndex].AddRequiredData(
        m_iYrsExceedingThresholdBeforeSuppressed);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fAdultConstantRadialGrowth);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fAsymptoticDiameterGrowth);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fSlopeOfGrowthResponse);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fLengthOfLastSuppressionFactor);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fLengthOfCurrentReleaseFactor);
    //Data members
    mp_oChildBehaviors[iIndex].mp_oNewTreeDataMembers.add(new DataMember(
        "Years suppressed", "yls", DataMember.INTEGER));
    mp_oChildBehaviors[iIndex].mp_oNewTreeDataMembers.add(new DataMember(
        "Years released", "ylr", DataMember.INTEGER));
    mp_oChildBehaviors[iIndex].mp_oNewTreeDataMembers.add(new DataMember("Diameter growth", "Growth", DataMember.FLOAT));

    //*****************
     //Constant basal area increment - diam only
     //*****************
      iIndex++;
    mp_oChildBehaviors[iIndex] = new Behavior("constba diam only",
                                          "Constant basal area growth - diam only",
                                          "constbagrowth diam only", (float) 1);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fAdultConstantAreaInc);
    mp_oChildBehaviors[iIndex].mp_oNewTreeDataMembers.add(new DataMember("Diameter growth", "Growth", DataMember.FLOAT));

    //*****************
     //Constant radial increment - diam only
     //*****************
      iIndex++;
    mp_oChildBehaviors[iIndex] = new Behavior("constradial diam only",
                                          "Constant radial growth - diam only",
                                          "constradialgrowth diam only",
                                          (float) 1);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fAdultConstantRadialGrowth);
    mp_oChildBehaviors[iIndex].mp_oNewTreeDataMembers.add(new DataMember("Diameter growth", "Growth", DataMember.FLOAT));

    //*****************
     //Simple linear regression - diam only
     //*****************
      iIndex++;
    mp_oChildBehaviors[iIndex] = new Behavior("linear diam only",
                                          "Linear growth - diam only",
                                          "simple linear growth diam only",
                                          (float) 1);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fSimpLinDiamIntercept);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fSimpLinDiamSlope);
    mp_oChildBehaviors[iIndex].mp_oNewTreeDataMembers.add(new DataMember("Diameter growth", "Growth", DataMember.FLOAT));

    //*****************
     //Linear bi-level growth - diam only
     //*****************
      iIndex++;
    mp_oChildBehaviors[iIndex] = new Behavior("linear bi-level diam only",
                                         "Linear bi-level growth - diam only",
                                         "linear bilevel growth diam only", (float) 1);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fBiLevelHighLightThreshold);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fBiLevelLowLightSlope);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fBiLevelLowLightIntercept);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fBiLevelHighLightSlope);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fBiLevelHighLightIntercept);
    mp_oChildBehaviors[iIndex].mp_oNewTreeDataMembers.add(new DataMember("Diameter growth", "Growth", DataMember.FLOAT));

    //*****************
     //Linear growth with exponential reduction for shade - diam only
     //*****************
      iIndex++;
    mp_oChildBehaviors[iIndex] = new Behavior("shaded linear diam only",
                                          "Linear growth w/ exponential shade reduction - diam only",
                                          "shaded linear growth diam only",
                                          (float) 1);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fLinShadeDiamIntercept);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fLinShadeDiamSlope);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fLinShadeDiamShadeExp);
    mp_oChildBehaviors[iIndex].mp_oNewTreeDataMembers.add(new DataMember("Diameter growth", "Growth", DataMember.FLOAT));

    //*****************
     //Logistic growth - diam only
     //*****************
      iIndex++;
    mp_oChildBehaviors[iIndex] = new Behavior("logistic growth diam only",
                                          "Logistic growth - diam only",
                                          "logistic growth diam only",
                                          (float) 1);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fLogisticDiamAsympGrowthFullLight);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fLogisticDiamShape1b);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fLogisticDiamShape2c);
    mp_oChildBehaviors[iIndex].mp_oNewTreeDataMembers.add(new DataMember("Diameter growth", "Growth", DataMember.FLOAT));

    //*****************
     //Logistic growth with size dependent asymptote - diam only
     //*****************
      iIndex++;
    mp_oChildBehaviors[iIndex] = new Behavior(
        "size dependent logistic growth diam only",
        "Logistic growth w/ size dependent asymptote - diam only",
        "size dependent logistic growth diam only",
        (float) 1);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fSizeDepLogisticDiamIntercept);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fSizeDepLogisticDiamSlope);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fSizeDepLogisticDiamShape1c);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fSizeDepLogisticDiamShape2d);
    mp_oChildBehaviors[iIndex].mp_oNewTreeDataMembers.add(new DataMember("Diameter growth", "Growth", DataMember.FLOAT));

    //*****************
     //Lognormal growth with exponential reduction for shade - diam only
     //*****************
      iIndex++;
    mp_oChildBehaviors[iIndex] = new Behavior(
        "lognormal growth diam only",
        "Lognormal with exponential shade reduction - diam only",
        "lognormal growth diam only",
        (float) 1);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fLognormalDiamIncAtDiam36);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fLognormalDiamShapeParam);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fLognormalDiamEffectOfShade);
    mp_oChildBehaviors[iIndex].mp_oNewTreeDataMembers.add(new DataMember("Diameter growth", "Growth", DataMember.FLOAT));

    //*****************
     //NCI growth - diam only
     //*****************
      iIndex++;
    mp_oChildBehaviors[iIndex] = new Behavior("nci diam only",
                                          "NCI growth - diam only",
                                          "ncigrowth diam only",
                                          (float) 2.1);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fNCIMaxPotentialGrowth);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fNCIMaxCrowdingRadius);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fNCIAlpha);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fNCIBeta);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fNCISizeSensToNCI);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fNCICrowdingEffectSlope);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fNCICrowdingEffectSteepness);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fNCINeighStormEffMed);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fNCINeighStormEffFull);
    mp_oChildBehaviors[iIndex].AddRequiredData(m_fNCIDbhDivisor);
    mp_oChildBehaviors[iIndex].AddRequiredData(m_iIncludeSnagsInNCI);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fNCIMinNeighborDBH);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fNCISizeEffectMode);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fNCISizeEffectVariance);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fNCIShadingEffectCoefficient);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fNCIShadingEffectExponent);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fNCIStormEffectMed);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fNCIStormEffectFull);
    //Disallow for seedlings
    mp_oChildBehaviors[iIndex].SetCanApplyTo(TreePopulation.SEEDLING, false);
    mp_oChildBehaviors[iIndex].mp_oNewTreeDataMembers.add(new DataMember("Diameter growth", "Growth", DataMember.FLOAT));

    //*****************
     //Relative growth limited by basal area - diam only
     //*****************
      iIndex++;
    mp_oChildBehaviors[iIndex] = new Behavior("relba diam only",
                                          "Relative growth limited to basal area increment - diam only",
                                          "relbagrowth diam only",
                                          (float) 1);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fAdultConstantAreaInc);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fAsymptoticDiameterGrowth);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fSlopeOfGrowthResponse);
    mp_oChildBehaviors[iIndex].mp_oNewTreeDataMembers.add(new DataMember("Diameter growth", "Growth", DataMember.FLOAT));

    //*****************
     //Relative growth limited by radial increment - diam only
     //*****************
      iIndex++;
    mp_oChildBehaviors[iIndex] = new Behavior("relradial diam only",
                                          "Relative growth limited to radial increment - diam only",
                                          "relradialgrowth diam only",
                                          (float) 1);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fAdultConstantRadialGrowth);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fAsymptoticDiameterGrowth);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fSlopeOfGrowthResponse);
    mp_oChildBehaviors[iIndex].mp_oNewTreeDataMembers.add(new DataMember("Diameter growth", "Growth", DataMember.FLOAT));

    //*****************
     //Unlimited absolute growth - diam only
     //*****************
      iIndex++;
    mp_oChildBehaviors[iIndex] = new Behavior("absunlim diam only",
                                          "Unlimited absolute growth - diam only",
                                          "absunlimgrowth diam only", (float) 1);
    //Data
    mp_oChildBehaviors[iIndex].AddRequiredData(m_fMortalityRateAtSuppression);
    mp_oChildBehaviors[iIndex].AddRequiredData(
        m_iYrsExceedingThresholdBeforeSuppressed);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fAsymptoticDiameterGrowth);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fSlopeOfGrowthResponse);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fLengthOfLastSuppressionFactor);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fLengthOfCurrentReleaseFactor);
    //Data members
    mp_oChildBehaviors[iIndex].mp_oNewTreeDataMembers.add(new DataMember(
        "Years suppressed", "yls", DataMember.INTEGER));
    mp_oChildBehaviors[iIndex].mp_oNewTreeDataMembers.add(new DataMember(
        "Years released", "ylr", DataMember.INTEGER));
    mp_oChildBehaviors[iIndex].mp_oNewTreeDataMembers.add(new DataMember("Diameter growth", "Growth", DataMember.FLOAT));

    //*****************
     //Unlimited relative growth - diam only
     //*****************
      iIndex++;
    mp_oChildBehaviors[iIndex] = new Behavior("relunlim diam only",
                                          "Unlimited relative growth - diam only",
                                          "relunlimgrowth diam only", (float) 1);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fAsymptoticDiameterGrowth);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fSlopeOfGrowthResponse);
    mp_oChildBehaviors[iIndex].mp_oNewTreeDataMembers.add(new DataMember("Diameter growth", "Growth", DataMember.FLOAT));

    //*****************
     //Puerto Rico semi-stochastic growth - diam only
     //*****************
      iIndex++;
    mp_oChildBehaviors[iIndex] = new Behavior("pr semi-stochastic diam only",
                                          "Puerto Rico semi-stochastic - diam only",
                                          "PR semi-stochastic diam only", (float) 1);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fPRHiteThreshold);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fPRDetermA);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fPRDetermB);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fPRMeanDBH);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fPRDBHStdDev);
    mp_oChildBehaviors[iIndex].mp_oNewTreeDataMembers.add(new DataMember("Diameter growth", "Growth", DataMember.FLOAT));

    //*****************
     //Double resource Michaelis-Menton relative growth - auto-height
     //*****************
      iIndex++;
    mp_oChildBehaviors[iIndex] = new Behavior("double resource diam only",
                                          "Double resource relative growth - diam only",
                                          "Double resource relative diam only", (float) 1);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fAsymptoticDiameterGrowth);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fSlopeOfGrowthResponse);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fDoubleResourceInfluence);
    mp_oChildBehaviors[iIndex].mp_oNewTreeDataMembers.add(new DataMember("Diameter growth", "Growth", DataMember.FLOAT));

    //*****************
     //Height incrementer - height only
     //*****************
      iIndex++;
    mp_oChildBehaviors[iIndex] = new Behavior("height incrementer",
                                          "Allometric height incrementer",
                                          "height incrementer",
                                          (float) 1);

    //*****************
     //Simple linear regression - height only
     //*****************
      iIndex++;
    mp_oChildBehaviors[iIndex] = new Behavior("linear height only",
                                          "Linear growth - height only",
                                          "simple linear growth height only",
                                          (float) 1);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fSimpLinHeightIntercept);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fSimpLinHeightSlope);

    //*****************
     //Linear growth with exponential reduction for shade - height only
     //*****************
      iIndex++;
    mp_oChildBehaviors[iIndex] = new Behavior("shaded linear height only",
                                          "Linear growth w/ exponential shade reduction - height only",
                                          "shaded linear growth height only",
                                          (float) 1);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fLinShadeHeightIntercept);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fLinShadeHeightSlope);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fLinShadeHeightShadeExp);

    //*****************
     //Logistic growth - height only
     //*****************
      iIndex++;
    mp_oChildBehaviors[iIndex] = new Behavior("logistic growth height only",
                                          "Logistic growth - height only",
                                          "logistic growth height only",
                                          (float) 1);
    mp_oChildBehaviors[iIndex].AddRequiredData(
        mp_fLogisticHeightAsympGrowthFullLight);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fLogisticHeightShape1b);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fLogisticHeightShape2c);

    //*****************
     //Logistic growth with size dependent asymptote - height only
     //*****************
      iIndex++;
    mp_oChildBehaviors[iIndex] = new Behavior(
        "size dependent logistic growth height only",
        "Logistic growth w/ size dependent asymptote - height only",
        "size dependent logistic growth height only",
        (float) 1);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fSizeDepLogisticHeightIntercept);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fSizeDepLogisticHeightSlope);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fSizeDepLogisticHeightShape1c);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fSizeDepLogisticHeightShape2d);

    //*****************
     //Lognormal growth with exponential reduction for shade - height only
     //*****************
      iIndex++;
    mp_oChildBehaviors[iIndex] = new Behavior(
        "lognormal growth height only",
        "Lognormal with exponential shade reduction - height only",
        "lognormal growth height only",
        (float) 1);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fLognormalHeightIncAtDiam36);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fLognormalHeightShapeParam);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fLognormalHeightEffectOfShade);

    //*****************
     //Lognormal bi-level growth - height only
     //*****************
      iIndex++;
    mp_oChildBehaviors[iIndex] = new Behavior(
        "lognormal bi-level growth height only",
        "Lognormal bi-level - height only",
        "log bilevel growth height only",
        (float) 1);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fLogBiLevelHighLightThreshold);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fLogBiLevelLoLiteMaxGrwth);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fLogBiLevelLoLiteX0);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fLogBiLevelLoLiteXb);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fLogBiLevelHiLiteMaxGrwth);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fLogBiLevelHiLiteX0);
    mp_oChildBehaviors[iIndex].AddRequiredData(mp_fLogBiLevelHiLiteXb);



    mp_oAllData.add(mp_fNCIMaxPotentialGrowth);
    mp_oAllData.add(mp_fNCIMaxCrowdingRadius);
    mp_oAllData.add(mp_fNCIAlpha);
    mp_oAllData.add(mp_fNCIBeta);
    mp_oAllData.add(mp_fNCISizeSensToNCI);
    mp_oAllData.add(mp_fNCICrowdingEffectSlope);
    mp_oAllData.add(mp_fNCICrowdingEffectSteepness);
    mp_oAllData.add(mp_fNCINeighStormEffMed);
    mp_oAllData.add(mp_fNCINeighStormEffFull);
    mp_oAllData.add(m_fNCIDbhDivisor);
    mp_oAllData.add(m_iIncludeSnagsInNCI);
    mp_oAllData.add(mp_fNCIMinNeighborDBH);
    mp_oAllData.add(mp_fNCISizeEffectMode);
    mp_oAllData.add(mp_fNCISizeEffectVariance);
    mp_oAllData.add(mp_fNCIShadingEffectCoefficient);
    mp_oAllData.add(mp_fNCIShadingEffectExponent);
    mp_oAllData.add(mp_fNCIStormEffectMed);
    mp_oAllData.add(mp_fNCIStormEffectFull);
    mp_oAllData.add(m_iRelativizeNCI);
    mp_oAllData.add(mp_fAdultConstantRadialGrowth);
    mp_oAllData.add(mp_fAdultConstantAreaInc);
    mp_oAllData.add(mp_fAsymptoticDiameterGrowth);
    mp_oAllData.add(mp_fSlopeOfGrowthResponse);
    mp_oAllData.add(m_fMortalityRateAtSuppression);
    mp_oAllData.add(m_iYrsExceedingThresholdBeforeSuppressed);
    mp_oAllData.add(mp_fLengthOfLastSuppressionFactor);
    mp_oAllData.add(mp_fLengthOfCurrentReleaseFactor);
    mp_oAllData.add(mp_fLognormalDiamIncAtDiam36);
    mp_oAllData.add(mp_fLognormalDiamShapeParam);
    mp_oAllData.add(mp_fLognormalDiamEffectOfShade);
    mp_oAllData.add(mp_fLogisticDiamAsympGrowthFullLight);
    mp_oAllData.add(mp_fLogisticDiamShape1b);
    mp_oAllData.add(mp_fLogisticDiamShape2c);
    mp_oAllData.add(mp_fSizeDepLogisticDiamIntercept);
    mp_oAllData.add(mp_fSizeDepLogisticDiamSlope);
    mp_oAllData.add(mp_fSizeDepLogisticDiamShape1c);
    mp_oAllData.add(mp_fSizeDepLogisticDiamShape2d);
    mp_oAllData.add(mp_fLinShadeDiamIntercept);
    mp_oAllData.add(mp_fLinShadeDiamSlope);
    mp_oAllData.add(mp_fLinShadeDiamShadeExp);
    mp_oAllData.add(mp_fSimpLinDiamIntercept);
    mp_oAllData.add(mp_fSimpLinDiamSlope);
    mp_oAllData.add(mp_fLognormalHeightIncAtDiam36);
    mp_oAllData.add(mp_fLognormalHeightShapeParam);
    mp_oAllData.add(mp_fLognormalHeightEffectOfShade);
    mp_oAllData.add(mp_fLogisticHeightAsympGrowthFullLight);
    mp_oAllData.add(mp_fLogisticHeightShape1b);
    mp_oAllData.add(mp_fLogisticHeightShape2c);
    mp_oAllData.add(mp_fSizeDepLogisticHeightIntercept);
    mp_oAllData.add(mp_fSizeDepLogisticHeightSlope);
    mp_oAllData.add(mp_fSizeDepLogisticHeightShape1c);
    mp_oAllData.add(mp_fSizeDepLogisticHeightShape2d);
    mp_oAllData.add(mp_fLinShadeHeightIntercept);
    mp_oAllData.add(mp_fLinShadeHeightSlope);
    mp_oAllData.add(mp_fLinShadeHeightShadeExp);
    mp_oAllData.add(mp_fSimpLinHeightIntercept);
    mp_oAllData.add(mp_fSimpLinHeightSlope);
    mp_oAllData.add(mp_fDoubleResourceInfluence);
    mp_oAllData.add(mp_fBiLevelHighLightThreshold);
    mp_oAllData.add(mp_fBiLevelLowLightSlope);
    mp_oAllData.add(mp_fBiLevelLowLightIntercept);
    mp_oAllData.add(mp_fBiLevelHighLightSlope);
    mp_oAllData.add(mp_fBiLevelHighLightIntercept);
    mp_oAllData.add(mp_fLogBiLevelHighLightThreshold);
    mp_oAllData.add(mp_fLogBiLevelLoLiteMaxGrwth);
    mp_oAllData.add(mp_fLogBiLevelLoLiteX0);
    mp_oAllData.add(mp_fLogBiLevelLoLiteXb);
    mp_oAllData.add(mp_fLogBiLevelHiLiteMaxGrwth);
    mp_oAllData.add(mp_fLogBiLevelHiLiteX0);
    mp_oAllData.add(mp_fLogBiLevelHiLiteXb);
    mp_oAllData.add(mp_fPRHiteThreshold);
    mp_oAllData.add(mp_fPRDetermA);
    mp_oAllData.add(mp_fPRDetermB);
    mp_oAllData.add(mp_fPRMeanDBH);
    mp_oAllData.add(mp_fPRDBHStdDev);

    //Grids
    mp_oGrids = new Grid[1];
    //Resource grid
    String sGridName = "Resource";

    //Create the data member
    DataMember[] p_oDataMembers = new DataMember[1];

    p_oDataMembers[0] = new DataMember("Resource", "Resource", DataMember.FLOAT);
    mp_oGrids[0] = new Grid(sGridName, p_oDataMembers, null, 8, 8);

    //Assign to double-resource relative growth
    GetBehaviorByKey("double resource").AddGrid(mp_oGrids[0]);
    GetBehaviorByKey("double resource diam only").AddGrid(mp_oGrids[0]);
  }

  /**
   * Performs setup tasks.  This method adds the NCI lambdas as required data,
   * and causes all storm damage parameters to default to 1.
   * @param oPop Tree population
   * @throws ModelException Won't throw.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   * <br>October 12, 2004:  Changed lambda format (LEM)
   * <br>November 30, 2004:  Made sure the lambdas wouldn't be added again
   * and old lambdas would be cleaned out (LEM)
   * <br>January 18, 2005:  Replaced all NCI lambdas with generic behavior
   * lambdas (LEM)
   * <br>March 16, 2005:  Initialized minimum neighbor DBH for NCI to allow
   * for easy backwards compatibility (LEM)
   */
  public void DoSetup(TreePopulation oPop) throws ModelException {
    Behavior oBeh1 = GetBehaviorByKey("nci"),
        oBeh2 = GetBehaviorByKey("nci diam only");
    int iNumSpecies = oPop.GetNumberOfSpecies(), i, j;
    boolean bAlreadyLambdas = false;

    //See if we already had lambdas
    for (i = 0; i < mp_oAllData.size(); i++) {
      ModelData oData = (ModelData) mp_oAllData.elementAt(i);
      String sKey = oData.GetDescriptor().toLowerCase();
      if (sKey.indexOf("lambda neighbor") > -1) {
        bAlreadyLambdas = true;
        break;
      }
    }

    if (bAlreadyLambdas == false) {

      for (i = 0; i < iNumSpecies; i++) {
        String sSpName = oPop.GetSpeciesNameFromCode(i);
        ModelVector oVector = new ModelVector(
            sSpName.replace('_', ' ') + " NCI Lambda Neighbors",
            "gr_nci" + sSpName + "NeighborLambda", "gr_nlVal",
            0, ModelVector.FLOAT);

        mp_oAllData.add(oVector);
        oBeh1.AddRequiredData(oVector);
        oBeh2.AddRequiredData(oVector);
      }
    }
    else {
      //We already had lambdas.  What we want to do is preserve any for species
      //that didn't change, because they may have values.

      //Create an array of the lambda arrays, placed in the new species
      //order
      ModelData[] p_oNCI = new ModelData[iNumSpecies];
      int iCode;

      for (i = 0; i < iNumSpecies; i++) {
        p_oNCI[i] = null;
      }

      //Get the existing lambdas and put them where they go in our arrays -
      //any for species that were removed will be left behind
      for (i = 0; i < mp_oAllData.size(); i++) {
        ModelData oData = (ModelData) mp_oAllData.elementAt(i);
        String sKey = oData.GetDescriptor().toLowerCase();
        if (sKey.indexOf("lambda neighbor") > -1) {
          //get the species
          String sSpecies;
          sKey = oData.GetDescriptor();
          sSpecies = sKey.substring(0, sKey.indexOf(" NCI")).replace(' ', '_');
          iCode = oPop.GetSpeciesCodeFromName(sSpecies);
          if (iCode > -1) {
            p_oNCI[iCode] = oData;
          }
        }
      }

      //Now remove all lambdas from the required data and from all behaviors
      for (i = 0; i < mp_oAllData.size(); i++) {
        ModelData oData = (ModelData) mp_oAllData.elementAt(i);
        String sKey = oData.GetDescriptor().toLowerCase();
        if (sKey.indexOf("lambda neighbor") > -1) {
          mp_oAllData.remove(i);
          i--;
        }
      }
      for (i = 0; i < mp_oChildBehaviors.length; i++) {
        for (j = 0; j < mp_oChildBehaviors[i].mp_oRequiredData.size(); j++) {
          ModelData oData = (ModelData) mp_oChildBehaviors[i].
              mp_oRequiredData.elementAt(j);
          String sKey = oData.GetDescriptor().toLowerCase();
          if (sKey.indexOf("lambda neighbor") > -1) {
            mp_oChildBehaviors[i].mp_oRequiredData.remove(j);
            j--;
          }
        }
      }

      //Now add back the lambdas, creating new ones for any new species
      for (i = 0; i < p_oNCI.length; i++) {
        if (p_oNCI[i] == null) {
          String sSpName = oPop.GetSpeciesNameFromCode(i);
          p_oNCI[i] = new ModelVector(
              sSpName.replace('_', ' ') + " NCI Lambda Neighbors",
              "gr_nci" + sSpName + "NeighborLambda", "gr_nlVal",
              0, ModelVector.FLOAT);
        }
        mp_oAllData.add(p_oNCI[i]);
        oBeh1.AddRequiredData(p_oNCI[i]);
        oBeh2.AddRequiredData(p_oNCI[i]);
      }
    }

    //If the damage parameters have not already been set up, set all their
    //values to a default of 1.  Otherwise leave them alone.
    if (mp_fNCINeighStormEffMed.GetValue().size() == 0) {
      for (i = 0; i < iNumSpecies; i++) {
        mp_fNCINeighStormEffMed.GetValue().add(new Float(1.0));
      }
    }

    if (mp_fNCINeighStormEffFull.GetValue().size() == 0) {
      for (i = 0; i < iNumSpecies; i++) {
        mp_fNCINeighStormEffFull.GetValue().add(new Float(1.0));
      }
    }

    if (mp_fNCIStormEffectMed.GetValue().size() == 0) {
      for (i = 0; i < iNumSpecies; i++) {
        mp_fNCIStormEffectMed.GetValue().add(new Float(1.0));
      }
    }

    if (mp_fNCIStormEffectFull.GetValue().size() == 0) {
      for (i = 0; i < iNumSpecies; i++) {
        mp_fNCIStormEffectFull.GetValue().add(new Float(1.0));
      }
    }

    //Now set minimum adult DBH to 0 to allow easy backwards compatibility -
    //remove this section when you remove the deprecated m_iRelativizeNCI
    //LEM 03-16-05
    if (mp_fNCIMinNeighborDBH.GetValue().size() == 0) {
      for (i = 0; i < iNumSpecies; i++) {
        mp_fNCIMinNeighborDBH.GetValue().add(new Float(0.0));
      }
    }
  }

  /**
   * Sets the values for asymptotic diameter growth.  One per species is
   * normal.
   * @param p_fVals Array of Floats which contain the values to be read, in
   * species order.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void SetAsymptoticDiameterGrowth(Float[] p_fVals) {
    SetVectorValues(mp_fAsymptoticDiameterGrowth, p_fVals);
  }

  /**
   * Sets the values for slope of growth response.  One per species is normal.
   * @param p_fVals Array of Floats which contain the values to be read, in
   * species order.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void SetSlopeOfGrowthResponse(Float[] p_fVals) {
    SetVectorValues(mp_fSlopeOfGrowthResponse, p_fVals);
  }

  /**
   * Sets the values for adult constant radial growth.  One per species is
   * normal.
   * @param p_fVals Array of Floats which contain the values to be read, in
   * species order.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void SetAdultConstantRadialGrowth(Float[] p_fVals) {
    SetVectorValues(mp_fAdultConstantRadialGrowth, p_fVals);
  }

  /**
   * Sets the values for adult constant area growth.  One per species is
   * normal.
   * @param p_fVals Array of Floats which contain the values to be read, in
   * species order.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void SetAdultConstantAreaGrowth(Float[] p_fVals) {
    SetVectorValues(mp_fAdultConstantAreaInc, p_fVals);
  }

  /**
   * Sets the values for length of last suppression factor.  One per species is
   * normal.
   * @param p_fVals Array of Floats which contain the values to be read, in
   * species order.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void SetLengthOfLastSuppressionFactor(Float[] p_fVals) {
    SetVectorValues(mp_fLengthOfLastSuppressionFactor, p_fVals);
  }

  /**
   * Sets the values for length of current release factor.  One per species is
   * normal.
   * @param p_fVals Array of Floats which contain the values to be read, in
   * species order.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void SetLengthOfCurrentReleaseFactor(Float[] p_fVals) {
    SetVectorValues(mp_fLengthOfCurrentReleaseFactor, p_fVals);
  }

  /**
   * Sets the mortality rate at suppression.
   * @param f Proportion of trees that die if suppressed, as a value between
   * 0 and 1.
   * @throws ModelException if the value is not between 0 and 1.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void SetMortalityRateAtSuppression(float f) throws ModelException {
    ModelFloat fTest = new ModelFloat(f,
                                      m_fMortalityRateAtSuppression.
                                      GetDescriptor(),
                                      "");
    MakeSureIsProportion(fTest);
    m_fMortalityRateAtSuppression.SetValue(f);
  }

  /**
   * Sets the value for years exceeding suppression threshold before a tree is
   * suppressed.
   * @param i Value, in years.
   * @throws ModelException if the value is less than 0.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void SetYearsExceedingThresholdBeforeSuppressed(int i) throws
      ModelException {
    ModelInt iTest = new ModelInt(i,
                                  m_iYrsExceedingThresholdBeforeSuppressed.
                                  GetDescriptor(),
                                  "");
    MakeSurePositive(iTest);
    m_iYrsExceedingThresholdBeforeSuppressed.SetValue(i);
  }

  /**
   * Sets which species for a give type use the absolute unlimited growth
   * behavior.
   * <p>
   * This method would normally not be used, as this object manages its own
   * data and would make the decision of which behaviors are enabled and which
   * are not.  However, in the case of input of a tab-delimited parameter file,
   * the information must be fed to this class.
   * @param p_iVals Array of use flags, one per species.  A 0 value means that
   * the given species doesn't use the behavior, a non-zero value means
   * they do.
   * @param iType Tree type to which the species list applies.
   * @param oPop Tree population object for species/type validation.
   * @throws ModelException if the type number is invalid, or if not all
   * species are represented.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void SetUsesAbsoluteUnlimitedGrowth(Integer[] p_iVals, int iType,
                                             TreePopulation oPop) throws
      ModelException {

    SetUsesBehavior(p_iVals, iType, oPop, "absunlim");
  }

  /**
   * Sets which species for a give type use the constant basal area increment
   * growth behavior.
   * <p>
   * This method would normally not be used, as this object manages its own
   * data and would make the decision of which behaviors are enabled and which
   * are not.  However, in the case of input of a tab-delimited parameter file,
   * the information must be fed to this class.
   * @param p_iVals Array of use flags, one per species.  A 0 value means that
   * the given species doesn't use the behavior, a non-zero value means
   * they do.
   * @param iType Tree type to which the species list applies.
   * @param oPop Tree population object for species/type validation.
   * @throws ModelException if the type number is invalid, or if not all
   * species are represented.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */

  public void SetUsesConstantBasalAreaIncrement(Integer[] p_iVals, int iType,
                                                TreePopulation oPop) throws
      ModelException {
    SetUsesBehavior(p_iVals, iType, oPop, "constba");
  }

  /**
   * Sets which species for a give type use the absolute basal area growth behavior.
   * <p>
   * This method would normally not be used, as this object manages its own
   * data and would make the decision of which behaviors are enabled and which
   * are not.  However, in the case of input of a tab-delimited parameter file,
   * the information must be fed to this class.
   * @param p_iVals Array of use flags, one per species.  A 0 value means that
   * the given species doesn't use the behavior, a non-zero value means
   * they do.
   * @param iType Tree type to which the species list applies.
   * @param oPop Tree population object for species/type validation.
   * @throws ModelException if the type number is invalid, or if not all
   * species are represented.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */

  public void SetUsesAbsoluteBasalAreaGrowth(Integer[] p_iVals, int iType,
                                             TreePopulation oPop) throws
      ModelException {
    SetUsesBehavior(p_iVals, iType, oPop, "absba");
  }

  /**
   * Sets which species for a give type use the unlimited relative growth
   * behavior.
   * <p>
   * This method would normally not be used, as this object manages its own
   * data and would make the decision of which behaviors are enabled and which
   * are not.  However, in the case of input of a tab-delimited parameter file,
   * the information must be fed to this class.
   * @param p_iVals Array of use flags, one per species.  A 0 value means that
   * the given species doesn't use the behavior, a non-zero value means
   * they do.
   * @param iType Tree type to which the species list applies.
   * @param oPop Tree population object for species/type validation.
   * @throws ModelException if the type number is invalid, or if not all
   * species are represented.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */

  public void SetUsesUnlimitedRelativeGrowth(Integer[] p_iVals, int iType,
                                             TreePopulation oPop) throws
      ModelException {
    SetUsesBehavior(p_iVals, iType, oPop, "relunlim");
  }

  /**
   * Sets which species for a give type use the constant radial increment
   * growth behavior.
   * <p>
   * This method would normally not be used, as this object manages its own
   * data and would make the decision of which behaviors are enabled and which
   * are not.  However, in the case of input of a tab-delimited parameter file,
   * the information must be fed to this class.
   * @param p_iVals Array of use flags, one per species.  A 0 value means that
   * the given species doesn't use the behavior, a non-zero value means
   * they do.
   * @param iType Tree type to which the species list applies.
   * @param oPop Tree population object for species/type validation.
   * @throws ModelException if the type number is invalid, or if not all
   * species
   * are represented.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */

  public void SetUsesConstantRadialIncrement(Integer[] p_iVals, int iType,
                                             TreePopulation oPop) throws
      ModelException {
    SetUsesBehavior(p_iVals, iType, oPop, "constradial");
  }

  /**
   * Sets which species for a give type use the relative basal area
   * growth behavior.
   * <p>
   * This method would normally not be used, as this object manages its own
   * data and would make the decision of which behaviors are enabled and which
   * are not.  However, in the case of input of a tab-delimited parameter file,
   * the information must be fed to this class.
   * @param p_iVals Array of use flags, one per species.  A 0 value means that
   * the given species doesn't use the behavior, a non-zero value means
   * they do.
   * @param iType Tree type to which the species list applies.
   * @param oPop Tree population object for species/type validation.
   * @throws ModelException if the type number is invalid, or if not all
   * species are represented.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */

  public void SetUsesRelativeBasalArea(Integer[] p_iVals, int iType,
                                       TreePopulation oPop) throws
      ModelException {
    SetUsesBehavior(p_iVals, iType, oPop, "relba");
  }

  /**
   * Validates the data.
   * @param oPop TreePopulation object.
   * @throws ModelException if:
   * <ul>
   * <li>NCI Max. radius of neighbor effects is <= 0 for any species</li>
   * <li>NCI Max growth is <= 0 for any species</li>
   * <li>NCI size effect mode for any species is <= 0</li>
   * <li>NCI Size effect variance = 0 for any species</li>
   * <li>NCI Eta for any species for any damage category is not between 0 and
   * 1</li>
   * <li>NCI Storm Effect parameters for any species for any damage category
   * are not between 0 and 1</li>
   * <li>NCI DBH divisor is <= 0</li>
   * <li>Minimun neighbor DBH value is < 0</li>
   * <li>Linear bi-level growth is enabled and the light threshold is not
   * between 0 and 100</li>
   * <li>Lognormal bi-level growth is enabled and:
   * <ul>
   * <li>Max growth for high or low light is negative</li>
   * <li>X0 for high or low light is zero</li>
   * <li>Xb for high or low light is zero</li>
   * <li>Light threshold is not between 0 and 100</li>
   * </ul></li>
   * </ul>
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   * <br>January 18, 2005:  Updated NCI validation (LEM)
   * <br>May 5, 2005:  Added validation for linear bi-level growth
   */
  public void ValidateData(TreePopulation oPop) throws ModelException {
    ModelVector oVectorTest;
    Behavior oNCI;
    boolean[] p_bAppliesTo;
    int iNumSpecies = oPop.GetNumberOfSpecies(),
        i;

    //If no behaviors are enabled, exit
    if (!AnyBehaviorsEnabled()) {
      return;
    }

    p_bAppliesTo = new boolean[iNumSpecies];
    for (i = 0; i < p_bAppliesTo.length; i++) {
      p_bAppliesTo[i] = false;
    }

    if (IsThisDataUsed(mp_fAdultConstantRadialGrowth)) {
      MakeSureAllPositive(mp_fAdultConstantRadialGrowth, p_bAppliesTo);
    }

    //If an absolute growth behavior is enabled, validate suppression/release
    for (i = 0; i < mp_oChildBehaviors.length; i++) {
      if (mp_oChildBehaviors[i].GetKey().indexOf("abs") >= 0 &&
          mp_oChildBehaviors[i].m_bIsEnabled) {
        MakeSureRightSize(mp_fLengthOfLastSuppressionFactor, iNumSpecies);
        MakeSureRightSize(mp_fLengthOfCurrentReleaseFactor, iNumSpecies);
        MakeSureIsProportion(m_fMortalityRateAtSuppression);
        MakeSurePositive(m_fMortalityRateAtSuppression);
        MakeSurePositive(m_iYrsExceedingThresholdBeforeSuppressed);

        //Make sure that BC mortality is enabled
        Behavior oBeh = m_oManager.GetMortalityBehaviors().GetBehaviorByKey(
            "bcmort");
        if (oBeh == null) {
          throw (new ModelException(ErrorGUI.DATA_MISSING, "JAVA", "The " +
                                    "behavior \"" +
                                    mp_oChildBehaviors[i].GetDescriptor() +
                                    "\" requires the BC Mortality behavior."));
        }
        else if (oBeh.m_bIsEnabled == false) {
          throw (new ModelException(ErrorGUI.DATA_MISSING, "JAVA", "The " +
                                    "behavior \"" +
                                    mp_oChildBehaviors[i].GetDescriptor() +
                                    "\" requires the behavior \"" +
                                    oBeh.GetDescriptor() + "\"."));
        }
        break;
      }
    }

    //Validate that the NCI vectors are the right size if NCI growth is being
    //used
    if (GetBehaviorByKey("nci").m_bIsEnabled ||
        GetBehaviorByKey("nci diam only").m_bIsEnabled) {

      oNCI = GetBehaviorByKey("nci");
      p_bAppliesTo = oNCI.GetWhichSpeciesUsed(oPop);
      oNCI = GetBehaviorByKey("nci diam only");
      p_bAppliesTo = MergeBooleans(p_bAppliesTo, oNCI.GetWhichSpeciesUsed(oPop));

      //Count the number of lambdas
      int iNumLambdas = 0;
      for (i = 0; i < oNCI.GetNumberOfRequiredDataObjects(); i++) {
        ModelData oData = (ModelData) oNCI.GetRequiredData(i);
        String sDescriptor = oData.GetDescriptor().toLowerCase();
        if (sDescriptor.indexOf("lambda") > -1 &&
            sDescriptor.indexOf("pr nci") == -1) {
          oVectorTest = (ModelVector) oData;
          MakeSureAllAreProportions(oVectorTest, p_bAppliesTo);
          iNumLambdas++;
        }
      }

      if (iNumSpecies > iNumLambdas) {
        ModelException oErr = new ModelException();
        oErr.iErrorCode = ErrorGUI.BAD_DATA;
        oErr.sCallingFunction = "JAVA";
        oErr.sMessage = "There is the wrong number of NCI lambda neighbors.";
      }
      MakeSureAllPositive(mp_fNCIMaxCrowdingRadius, p_bAppliesTo);
      MakeSureAllPositive(mp_fNCIMaxPotentialGrowth, p_bAppliesTo);
      MakeSureAllPositive(mp_fNCISizeEffectMode, p_bAppliesTo);
      MakeSureAllNonZero(mp_fNCISizeEffectVariance, p_bAppliesTo);
      MakeSureAllAreProportions(mp_fNCINeighStormEffFull, p_bAppliesTo);
      MakeSureAllAreProportions(mp_fNCINeighStormEffMed, p_bAppliesTo);
      MakeSureAllAreProportions(mp_fNCIStormEffectFull, p_bAppliesTo);
      MakeSureAllAreProportions(mp_fNCIStormEffectMed, p_bAppliesTo);
      MakeSurePositive(m_fNCIDbhDivisor);
      MakeSureAllNonNegative(mp_fNCIMinNeighborDBH); //this is for all species
    }

    //LEM 05-05-05
    if (GetBehaviorByKey("linear bi-level").m_bIsEnabled ||
        GetBehaviorByKey("linear bi-level diam only").m_bIsEnabled) {
      oNCI = GetBehaviorByKey("linear bi-level");
      p_bAppliesTo = oNCI.GetWhichSpeciesUsed(oPop);
      oNCI = GetBehaviorByKey("linear bi-level diam only");
      p_bAppliesTo = MergeBooleans(p_bAppliesTo, oNCI.GetWhichSpeciesUsed(oPop));
      MakeSureAllAreBounded(mp_fBiLevelHighLightThreshold, p_bAppliesTo, 0, 100);
    }

    //LEM 05-09-05
    if (GetBehaviorByKey("lognormal bi-level growth height only").m_bIsEnabled) {
      p_bAppliesTo = GetBehaviorByKey("lognormal bi-level growth height only").GetWhichSpeciesUsed(oPop);
      MakeSureAllAreBounded(mp_fLogBiLevelHighLightThreshold, p_bAppliesTo, 0, 100);
      MakeSureAllNonZero(mp_fLogBiLevelHiLiteX0, p_bAppliesTo);
      MakeSureAllNonZero(mp_fLogBiLevelLoLiteX0, p_bAppliesTo);
      MakeSureAllNonZero(mp_fLogBiLevelHiLiteXb, p_bAppliesTo);
      MakeSureAllNonZero(mp_fLogBiLevelLoLiteXb, p_bAppliesTo);
      MakeSureAllNonNegative(mp_fLogBiLevelHiLiteMaxGrwth, p_bAppliesTo);
      MakeSureAllNonNegative(mp_fLogBiLevelLoLiteMaxGrwth, p_bAppliesTo);
    }
  }

  /**
   * Before letting the base class write the XML, this verifies that there is
   * Resource grid information entered if the double resource relative behavior
   * is used.  I expect to remove this when there is some formal behavior for
   * taking care of the resources.
   * @param jOut FileWriter
   * @param oPop TreePopulation
   * @throws ModelException
   */
  public void WriteXML(java.io.BufferedWriter jOut, TreePopulation oPop) throws ModelException {
    if (GetBehaviorByKey("double resource").m_bIsEnabled ||
        GetBehaviorByKey("double resource diam only").m_bIsEnabled) {
      if (mp_oGrids[0].mp_oGridVals == null || mp_oGrids[0].mp_oGridVals.size() == 0) {
        throw new ModelException(ErrorGUI.DATA_MISSING, "JAVA", "To use either the behavior \"" +
                                 GetBehaviorByKey("double resource").GetDescriptor() +
                                 "\"\n or the behavior \"" + GetBehaviorByKey("double resource diam only").GetDescriptor() +
                                 "\"\n you must populate grid values for the grid \"" +
                                 mp_oGrids[0].GetName() + "\".");
      }
    }
    super.WriteXML(jOut, oPop);
  }
}
