package javawrapper;

import java.io.BufferedWriter;
import java.util.Vector;
import org.xml.sax.Attributes;

/**
 * Manages output behaviors and data
 * <p>Copyright: Copyright (c) Charles D. Canham 2003</p>
 * <p>Company: Institute of Ecosystem Studies</p>
 * @author Lora E. Murphy
 * @version 1.0
 *
 * <br>Edit history:
 * <br>------------------
 * <br>April 28, 2004: Submitted in beta version (LEM)
 */

public class OutputBehaviors
    extends BehaviorTypeBase {

  /**For short output - collection of
       TreeOutputSaveInfo objects - vector is sized number of types*/
  protected Vector mp_oShortTreeSaveSettings;

  /**For detailed output - collection of DetailedTreeSettings objects*/
  protected Vector mp_oDetailedTreeSaveSettings;

  /**For detailed output - collection of DetailedGridSettings objects*/
  protected Vector mp_oDetailedGridSaveSettings;

  /**For short output - the list of Subplot objects*/
  protected Vector mp_oSubplots;

  /**Tree population*/
  protected TreePopulation m_oPop;

  /**File name for short output*/
  protected ModelString m_sShortOutputFilename,
      /**File name for detailed output*/
      m_sDetailedOutputFilename;

  /**When parsing an XML file, this is the setting currently receiving data*/
  protected DetailedOutputSettings m_oCurrentDetailedSettings;

  /**When parsing an XML file, this is the subplot currently receiving data*/
  protected Subplot m_oCurrentSubplot;

  /**When parsing an XML file, this is the current short output type*/
  protected int m_iCurrentType;

  public final static String DETAILED_OUTPUT_EXTENSION = ".gz.tar";
  public final static String SUMMARY_EXTENSION = ".out";

  /**
   * Constructor.
   * @param oManager GUIManager object.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public OutputBehaviors(GUIManager oManager) {
    super(oManager, "Output", "output");
    m_bAutoDisplayParameters = false;

    m_iCurrentType = -1;
    mp_oShortTreeSaveSettings = new Vector(0);
    m_sShortOutputFilename = new ModelString("", "Filename for summary output",
                                             "so_filename");

    mp_oDetailedTreeSaveSettings = new Vector(0);
    mp_oDetailedGridSaveSettings = new Vector(0);
    m_sDetailedOutputFilename = new ModelString("", "Detailed output filename",
                                         "ou_filename");

    mp_oSubplots = new Vector(0);

    //Make our child behaviors
    mp_oChildBehaviors = new Behavior[2];
    mp_oChildBehaviors[0] = new Behavior("short output", "Summary Output",
                                         "short output", (float)1);
    mp_oChildBehaviors[0].m_bMustHaveTrees = false;
    mp_oChildBehaviors[0].m_bCanEditAutomatically = false;
    mp_oChildBehaviors[0].AddRequiredData(m_sShortOutputFilename);

    mp_oChildBehaviors[1] = new Behavior("detailed", "Detailed Output", "output",
                                         (float)1);
    mp_oChildBehaviors[1].m_bMustHaveTrees = false;
    mp_oChildBehaviors[1].m_bCanEditAutomatically = false;
    mp_oChildBehaviors[1].AddRequiredData(m_sDetailedOutputFilename);

    mp_oAllData.add(m_sDetailedOutputFilename);
    mp_oAllData.add(m_sShortOutputFilename);
  }

  /**
   * Creates a subplot.
   * @param sName Name of the subplot.
   * @return The new Subplot object.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public Subplot AddSubplot(String sName) {
    Subplot oSubplot = new Subplot(sName);
    mp_oSubplots.add(oSubplot);
    return oSubplot;
  }

  /**
   * Sets the short output filename.
   * @param sFilename Filename.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   * <br>November 18, 2004:  Added extension enforcement (LEM)
   * <br>December 3, 2004:  Fixed a bug when the filename was empty (LEM)
   */
  public void SetShortOutputFileName(String sFilename) {
    sFilename = sFilename.trim();
    //Make sure the file extension's on it
    if (sFilename.length() > 0 &&
        sFilename.toLowerCase().endsWith(SUMMARY_EXTENSION) == false) {
      sFilename += SUMMARY_EXTENSION;
    }

    m_sShortOutputFilename.SetValue(sFilename);
  }

  /**
   * Gets the short output filename.
   * @return Filename.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public String GetShortOutputFileName() {
    return m_sShortOutputFilename.GetValue();
  }

  /**
   * Sets the detailed output filename.
   * @param sFilename Filename.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   * <br>November 18, 2004:  Added extension enforcement (LEM)
   * <br>December 3, 2004:  Fixed a bug when the filename was empty (LEM)
   */
  public void SetDetailedOutputFileName(String sFilename) {
    sFilename = sFilename.trim();
    //Make sure the file extension's on it
    if (sFilename.length() > 0 &&
        sFilename.toLowerCase().endsWith(DETAILED_OUTPUT_EXTENSION) == false) {
      sFilename += DETAILED_OUTPUT_EXTENSION;
    }

    m_sDetailedOutputFilename.SetValue(sFilename);
  }

  /**
   * Gets the detailed output filename.
   * @return Filename.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public String GetDetailedOutputFileName() {
    return m_sDetailedOutputFilename.GetValue();
  }

  /**
   * Gets the number of tree settings for which there is detailed output info
   * (each setting is one tree species/type combo).
   * @return Number of settings.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public int GetNumberOfDetailedTreeSettings() {
    return mp_oDetailedTreeSaveSettings.size();
  }

  /**
   * Gets the number of grid settings for which there is detailed output info.
   * @return Number of settings.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public int GetNumberOfDetailedGridSettings() {
    return mp_oDetailedGridSaveSettings.size();
  }

  /**
   * Gets whether or not short output is being saved.
   * @return Save status.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public boolean SavingShortOutputData() {
    Behavior oBeh = GetBehaviorByKey("short output");
    return oBeh.m_bIsEnabled;
  }

  /**
   * Returns the detailed output tree setting at a specified index.
   * @param iIndex The index.
   * @return The DetailedTreeSettings object, or null if the index is invalid.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public DetailedTreeSettings GetDetailedTreeSetting(int iIndex) {
    if (iIndex < 0 || iIndex > mp_oDetailedTreeSaveSettings.size()) {
      return null;
    }
    else {
      return (DetailedTreeSettings) mp_oDetailedTreeSaveSettings.elementAt(iIndex);
    }
  }

  /**
   * Returns the detailed output grid setting at a specified index.
   * @param iIndex The index.
   * @return The DetailedGridSettings object, or null if the index is invalid.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public DetailedGridSettings GetDetailedGridSetting(int iIndex) {
    if (iIndex < 0 || iIndex > mp_oDetailedGridSaveSettings.size()) {
      return null;
    }
    else {
      return (DetailedGridSettings) mp_oDetailedGridSaveSettings.elementAt(iIndex);
    }
  }

  /**
   * Keeps the behavior enabled flag current.  Any change of save flags has
   * the potential to enable the behavior (by making the first save flag true)
   * or unenabling (by making the last save flag false).
   * @param oBeh Behavior to check.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  protected void UpdateIsEnabled(Behavior oBeh) {
    if (oBeh != GetBehaviorByKey("short output")) {
      return;
    }

    TreeOutputSaveInfo oData;
    int i;
    boolean bAnySaveFlags = false;
    for (i = 0; i < mp_oShortTreeSaveSettings.size(); i++) {
      oData = (TreeOutputSaveInfo) mp_oShortTreeSaveSettings.elementAt(i);
      if (oData.bSaveABA || oData.bSaveADN || oData.bSaveRBA || oData.bSaveRDN) {
        bAnySaveFlags = true;
        break;
      }
    }

    oBeh.m_bIsEnabled = bAnySaveFlags;
  }

  /**
   * Sets the save absolute density attribute for a type.
   * @param iType Tree type.
   * @param bUse Whether or not to use this.
   * @throws ModelException If the type is not valid.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void SetSaveAbsoluteDensity(int iType, boolean bUse) throws
      ModelException {
    if (iType < 0 || iType >= m_oPop.GetNumberOfTypes()) {
      ModelException oErr = new ModelException();
      oErr.iErrorCode = ErrorGUI.ILLEGAL_OP;
      oErr.sCallingFunction = "JAVA";
      oErr.sMessage = "Invalid type passed to output";
      throw oErr;
    }

    TreeOutputSaveInfo oInfo = (TreeOutputSaveInfo) mp_oShortTreeSaveSettings.
        elementAt(iType);
    oInfo.bSaveADN = bUse;

    UpdateIsEnabled(GetBehaviorByKey("short output"));
  }

  /**
   * Sets the save relative density attribute for a type.
   * @param iType Tree type.
   * @param bUse Whether or not to use this.
   * @throws ModelException If the type is not valid.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void SetSaveRelativeDensity(int iType, boolean bUse) throws
      ModelException {
    if (iType < 0 || iType >= m_oPop.GetNumberOfTypes()) {
      ModelException oErr = new ModelException();
      oErr.iErrorCode = ErrorGUI.ILLEGAL_OP;
      oErr.sCallingFunction = "JAVA";
      oErr.sMessage = "Invalid type passed to output";
      throw oErr;
    }

    TreeOutputSaveInfo oInfo = (TreeOutputSaveInfo) mp_oShortTreeSaveSettings.
        elementAt(iType);
    oInfo.bSaveRDN = bUse;

    UpdateIsEnabled(GetBehaviorByKey("short output"));
  }

  /**
   * Sets the save absolute basal area attribute for a type.  This is ignored
   * if it's not sapling or adult.
   * @param iType Tree type.
   * @param bUse Whether or not to use this.
   * @throws ModelException If the type is not valid.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void SetSaveAbsoluteBasalArea(int iType, boolean bUse) throws
      ModelException {
    if (iType < 0 || iType >= m_oPop.GetNumberOfTypes()) {
      ModelException oErr = new ModelException();
      oErr.iErrorCode = ErrorGUI.ILLEGAL_OP;
      oErr.sCallingFunction = "JAVA";
      oErr.sMessage = "Invalid type passed to output";
      throw oErr;
    }

    if (iType == TreePopulation.SEEDLING) {
      return;
    }

    TreeOutputSaveInfo oInfo = (TreeOutputSaveInfo) mp_oShortTreeSaveSettings.
        elementAt(iType);
    oInfo.bSaveABA = bUse;

    UpdateIsEnabled(GetBehaviorByKey("short output"));
  }

  /**
   * Sets the save relative basal area attribute for a type.  This is ignored
   * if it's not sapling or adult.
   * @param iType Tree type.
   * @param bUse Whether or not to use this.
   * @throws ModelException If the type is not valid.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void SetSaveRelativeBasalArea(int iType, boolean bUse) throws
      ModelException {
    if (iType < 0 || iType >= m_oPop.GetNumberOfTypes()) {
      ModelException oErr = new ModelException();
      oErr.iErrorCode = ErrorGUI.ILLEGAL_OP;
      oErr.sCallingFunction = "JAVA";
      oErr.sMessage = "Invalid type passed to output";
      throw oErr;
    }

    if (iType == TreePopulation.SEEDLING) {
      return;
    }

    TreeOutputSaveInfo oInfo = (TreeOutputSaveInfo) mp_oShortTreeSaveSettings.
        elementAt(iType);
    oInfo.bSaveRBA = bUse;

    UpdateIsEnabled(GetBehaviorByKey("short output"));
  }

  /**
   * Gets whether or not absolute basal area is saved for a tree type.
   * @param iType Tree type.
   * @return Save flag.
   * @throws ModelException if the type is not valid.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public boolean GetSaveAbsoluteBasalArea(int iType) throws ModelException {
    if (iType < 0 || iType >= m_oPop.GetNumberOfTypes()) {
      ModelException oErr = new ModelException();
      oErr.iErrorCode = ErrorGUI.ILLEGAL_OP;
      oErr.sCallingFunction = "JAVA";
      oErr.sMessage = "Invalid type passed to output";
      throw oErr;
    }

    TreeOutputSaveInfo oInfo = (TreeOutputSaveInfo) mp_oShortTreeSaveSettings.
        elementAt(iType);
    return (oInfo.bSaveABA);
  }

  /**
   * Gets whether or not relative basal area is saved for a tree type.
   * @param iType Tree type.
   * @return Save flag.
   * @throws ModelException if the type is not valid.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public boolean GetSaveRelativeBasalArea(int iType) throws ModelException {
    if (iType < 0 || iType >= m_oPop.GetNumberOfTypes()) {
      ModelException oErr = new ModelException();
      oErr.iErrorCode = ErrorGUI.ILLEGAL_OP;
      oErr.sCallingFunction = "JAVA";
      oErr.sMessage = "Invalid type passed to output";
      throw oErr;
    }

    TreeOutputSaveInfo oInfo = (TreeOutputSaveInfo) mp_oShortTreeSaveSettings.
        elementAt(iType);
    return (oInfo.bSaveRBA);
  }

  /**
   * Gets whether or not absolute density is saved for a tree type.
   * @param iType Tree type.
   * @return Save flag.
   * @throws ModelException if the type is not valid.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public boolean GetSaveAbsoluteDensity(int iType) throws ModelException {
    if (iType < 0 || iType >= m_oPop.GetNumberOfTypes()) {
      ModelException oErr = new ModelException();
      oErr.iErrorCode = ErrorGUI.ILLEGAL_OP;
      oErr.sCallingFunction = "JAVA";
      oErr.sMessage = "Invalid type passed to output";
      throw oErr;
    }

    TreeOutputSaveInfo oInfo = (TreeOutputSaveInfo) mp_oShortTreeSaveSettings.
        elementAt(iType);
    return (oInfo.bSaveADN);
  }

  /**
   * Gets whether or not relative density is saved for a tree type.
   * @param iType Tree type.
   * @return Save flag.
   * @throws ModelException if the type is not valid.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public boolean GetSaveRelativeDensity(int iType) throws ModelException {
    if (iType < 0 || iType >= m_oPop.GetNumberOfTypes()) {
      ModelException oErr = new ModelException();
      oErr.iErrorCode = ErrorGUI.ILLEGAL_OP;
      oErr.sCallingFunction = "JAVA";
      oErr.sMessage = "Invalid type passed to output";
      throw oErr;
    }

    TreeOutputSaveInfo oInfo = (TreeOutputSaveInfo) mp_oShortTreeSaveSettings.
        elementAt(iType);
    return (oInfo.bSaveRDN);
  }

  /**
   * Overridden from base because output parameters are special.
   * @param jOut File to write to.
   * @param oPop TreePopulation object.
   * @throws ModelException if there is a problem writing the file.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void WriteXML(BufferedWriter jOut, TreePopulation oPop) throws
      ModelException {

    try {
      WriteShortOutputXML(jOut);

      WriteDetailedOutputXML(jOut, oPop);
    }
    catch (java.io.IOException e) {
      ModelException oErr = new ModelException();
      oErr.iErrorCode = ErrorGUI.BAD_FILE;
      oErr.sCallingFunction = "JAVA";
      oErr.sMessage = "There was a problem writing the parameter file.";
      throw (oErr);
    }
  }

  /**
   * Writes the parameter file data for short output.  Does nothing if there
   * are no short output save settings.
   * @param jOut File to write
   * @throws java.io.IOException passed through from file writing
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  protected void WriteShortOutputXML(BufferedWriter jOut) throws java.io.
      IOException {
    //Make sure we have anything to save
    TreeOutputSaveInfo oInfo;
    Subplot oSubplot;
    Cell oCell;
    int i, j;
    boolean bAnyOutput = false;
    for (i = 0; i < mp_oShortTreeSaveSettings.size(); i++) {
      oInfo = (TreeOutputSaveInfo) mp_oShortTreeSaveSettings.elementAt(i);

      if (oInfo.bSaveABA || oInfo.bSaveADN || oInfo.bSaveRBA ||
          oInfo.bSaveRDN) {
        bAnyOutput = true;
        break;
      }
    }

    if (!bAnyOutput) {
      return;
    }

    jOut.write("<shortoutput>");

    //Write file name
    WriteData(jOut, m_sShortOutputFilename);

    //Write the tree types
    TreePopulation oPop = m_oManager.GetTreePopulation();
    for (i = 0; i < mp_oShortTreeSaveSettings.size(); i++) {
      oInfo = (TreeOutputSaveInfo) mp_oShortTreeSaveSettings.elementAt(i);

      if (oInfo.bSaveABA || oInfo.bSaveADN || oInfo.bSaveRBA ||
          oInfo.bSaveRDN) {
        jOut.write("<so_treeTypeInfo type=\"" + oPop.GetTypeNameFromCode(i) +
                   "\">");

        //Save relative basal area
        if (i > TreePopulation.SEEDLING) {
          jOut.write("<so_saveRBA save=\"");
          if (oInfo.bSaveRBA) {
            jOut.write("true");
          }
          else {
            jOut.write("false");
          }
          jOut.write("\"/>");
        }

        //Save absolute basal area
        if (i > TreePopulation.SEEDLING) {
          jOut.write("<so_saveABA save=\"");
          if (oInfo.bSaveABA) {
            jOut.write("true");
          }
          else {
            jOut.write("false");
          }
          jOut.write("\"/>");
        }

        //Save relative density
        jOut.write("<so_saveRDN save=\"");
        if (oInfo.bSaveRDN) {
          jOut.write("true");
        }
        else {
          jOut.write("false");
        }
        jOut.write("\"/>");

        //Save absolute density
        jOut.write("<so_saveADN save=\"");
        if (oInfo.bSaveADN) {
          jOut.write("true");
        }
        else {
          jOut.write("false");
        }
        jOut.write("\"/>");

        jOut.write("</so_treeTypeInfo>");
      }
    }

    //Now write subplots
    for (i = 0; i < mp_oSubplots.size(); i++) {
      oSubplot = (Subplot) mp_oSubplots.elementAt(i);
      if (oSubplot.GetNumberOfCells() > 0) {
        jOut.write("<so_subplot>");
        jOut.write("<so_subplotName>" + oSubplot.GetSubplotName() +
                   "</so_subplotName>");
        jOut.write("<pointSet>");
        for (j = 0; j < oSubplot.GetNumberOfCells(); j++) {

          oCell = oSubplot.GetCell(j);
          jOut.write("<po_point x=\"" + oCell.GetX() + "\" y=\"" + oCell.GetY() +
                     "\"/>");
        }
        jOut.write("</pointSet>");

        jOut.write("</so_subplot>");
      }
    }

    jOut.write("</shortoutput>");

  }

  /**
   * Writes the parameter file data for detailed output.  Does nothing if there
   * are no detailed output save settings.
   * @param jOut File to write
   * @param oPop TreePopulation object
   * @throws java.io.IOException passed through from file writing
   * @throws ModelException passed through from tree population
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  protected void WriteDetailedOutputXML(BufferedWriter jOut, TreePopulation oPop) throws
      java.io.
      IOException, ModelException {

    Behavior oBeh = GetBehaviorByKey("detailed");
    if (!oBeh.m_bIsEnabled) {
      return;
    }

    //Just an extra double-check
    if (mp_oDetailedTreeSaveSettings.size() == 0 &&
        mp_oDetailedGridSaveSettings.size() == 0) {
      return;
    }

    DetailedTreeSettings oTreeSettings;
    DetailedGridSettings oGridSettings;
    int i, j;

    jOut.write("<output>");

    //Write filename
    WriteData(jOut, m_sDetailedOutputFilename);

    //Write tree settings
    for (i = 0; i < mp_oDetailedTreeSaveSettings.size(); i++) {
      oTreeSettings = (DetailedTreeSettings) mp_oDetailedTreeSaveSettings.
          elementAt(i);

      jOut.write("<ou_treeInfo species=\"" +
                 oPop.GetSpeciesNameFromCode(oTreeSettings.GetSpecies()) +
                 "\" type=\"" + oPop.GetTypeNameFromCode(oTreeSettings.GetType()) +
                 "\">");

      //Save frequency
      jOut.write("<ou_saveFreq>" +
                 String.valueOf(oTreeSettings.GetSaveFrequency()) +
                 "</ou_saveFreq>");

      //Ints
      for (j = 0; j < oTreeSettings.GetNumberOfInts(); j++) {
        jOut.write("<ou_int>" + oTreeSettings.GetInt(j).GetCodeName() +
                   "</ou_int>");
      }

      //Floats
      for (j = 0; j < oTreeSettings.GetNumberOfFloats(); j++) {
        jOut.write("<ou_float>" + oTreeSettings.GetFloat(j).GetCodeName() +
                   "</ou_float>");
      }

      //Chars
      for (j = 0; j < oTreeSettings.GetNumberOfChars(); j++) {
        jOut.write("<ou_char>" + oTreeSettings.GetChar(j).GetCodeName() +
                   "</ou_char>");
      }

      //Bools
      for (j = 0; j < oTreeSettings.GetNumberOfBools(); j++) {
        jOut.write("<ou_bool>" + oTreeSettings.GetBool(j).GetCodeName() +
                   "</ou_bool>");
      }

      jOut.write("</ou_treeInfo>");
    }

    //Write grids stuff
    for (i = 0; i < mp_oDetailedGridSaveSettings.size(); i++) {
      oGridSettings = (DetailedGridSettings) mp_oDetailedGridSaveSettings.
          elementAt(i);

      jOut.write("<ou_gridInfo gridName=\"" + oGridSettings.GetName() + "\">");

      //Save frequency
      jOut.write("<ou_saveFreq>" +
                 String.valueOf(oGridSettings.GetSaveFrequency()) +
                 "</ou_saveFreq>");

      //Ints
      for (j = 0; j < oGridSettings.GetNumberOfInts(); j++) {
        jOut.write("<ou_int>" + oGridSettings.GetInt(j).GetCodeName() +
                   "</ou_int>");
      }

      //Floats
      for (j = 0; j < oGridSettings.GetNumberOfFloats(); j++) {
        jOut.write("<ou_float>" + oGridSettings.GetFloat(j).GetCodeName() +
                   "</ou_float>");
      }

      //Chars
      for (j = 0; j < oGridSettings.GetNumberOfChars(); j++) {
        jOut.write("<ou_char>" + oGridSettings.GetChar(j).GetCodeName() +
                   "</ou_char>");
      }

      //Bools
      for (j = 0; j < oGridSettings.GetNumberOfBools(); j++) {
        jOut.write("<ou_bool>" + oGridSettings.GetBool(j).GetCodeName() +
                   "</ou_bool>");
      }

      //Package ints
      for (j = 0; j < oGridSettings.GetNumberOfPackageInts(); j++) {
        jOut.write("<ou_packageInt>" +
                   oGridSettings.GetPackageInt(j).GetCodeName() +
                   "</ou_packageInt>");
      }

      //Package floats
      for (j = 0; j < oGridSettings.GetNumberOfPackageFloats(); j++) {
        jOut.write("<ou_packageFloat>" +
                   oGridSettings.GetPackageFloat(j).GetCodeName() +
                   "</ou_packageFloat>");
      }

      //Package chars
      for (j = 0; j < oGridSettings.GetNumberOfPackageChars(); j++) {
        jOut.write("<ou_packageChar>" +
                   oGridSettings.GetPackageChar(j).GetCodeName() +
                   "</ou_packageChar>");
      }

      //Package bools
      for (j = 0; j < oGridSettings.GetNumberOfPackageBools(); j++) {
        jOut.write("<ou_packageBool>" +
                   oGridSettings.GetPackageBool(j).GetCodeName() +
                   "</ou_packageBool>");
      }

      jOut.write("</ou_gridInfo>");
    }

    jOut.write("</output>");
  }

  /**
   * Validates the data before writing to a parameter file.
   * @throws ModelException if the filename is empty but there are save
   * settings.
   * @param oPop Not used.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void ValidateData(TreePopulation oPop) throws ModelException {
    Behavior oBeh;
    int i;

    //If no behaviors are enabled, exit
    boolean bAnybodyEnabled = false;
    for (i = 0; i < mp_oChildBehaviors.length; i++) {
      if (mp_oChildBehaviors[i].m_bIsEnabled) {
        bAnybodyEnabled = true;
      }
    }

    if (!bAnybodyEnabled) {
      return;
    }

    // Short output
    oBeh = GetBehaviorByKey("short output");

    if (oBeh.m_bIsEnabled && m_sShortOutputFilename.GetValue().length() == 0) {
      ModelException oErr = new ModelException();
      oErr.iErrorCode = ErrorGUI.BAD_DATA;
      oErr.sCallingFunction = "JAVA";
      oErr.sMessage = "An output file name is required.";
      throw (oErr);
    }

    //Detailed output

    //Make sure all detailed tree settings actually have settings in them -
    //if not, delete them
    DetailedTreeSettings oTreeSettings;
    for (i = 0; i < mp_oDetailedTreeSaveSettings.size(); i++) {
      oTreeSettings = (DetailedTreeSettings) mp_oDetailedTreeSaveSettings.
          elementAt(i);
      if (oTreeSettings.GetNumberOfFloats() == 0 &&
          oTreeSettings.GetNumberOfInts() == 0 &&
          oTreeSettings.GetNumberOfBools() == 0 &&
          oTreeSettings.GetNumberOfChars() == 0) {
        mp_oDetailedTreeSaveSettings.remove(i);
        i--;
      }
    }

    DetailedGridSettings oGridSettings;
    for (i = 0; i < mp_oDetailedGridSaveSettings.size(); i++) {
      oGridSettings = (DetailedGridSettings) mp_oDetailedGridSaveSettings.
          elementAt(i);
      if (oGridSettings.GetNumberOfFloats() == 0 &&
          oGridSettings.GetNumberOfInts() == 0 &&
          oGridSettings.GetNumberOfBools() == 0 &&
          oGridSettings.GetNumberOfChars() == 0 &&
          oGridSettings.GetNumberOfPackageFloats() == 0 &&
          oGridSettings.GetNumberOfPackageInts() == 0 &&
          oGridSettings.GetNumberOfPackageBools() == 0 &&
          oGridSettings.GetNumberOfPackageChars() == 0) {
        mp_oDetailedGridSaveSettings.remove(i);
        i--;
      }
    }

    oBeh = GetBehaviorByKey("detailed");
    if (mp_oDetailedTreeSaveSettings.size() == 0 &&
        mp_oDetailedGridSaveSettings.size() == 0) {
      oBeh.m_bIsEnabled = false;
    }
    //   else {
    //     oBeh.m_bIsEnabled = true;
    //   }

    //Detailed output
    if (oBeh.m_bIsEnabled && m_sDetailedOutputFilename.GetValue().length() == 0) {
      ModelException oErr = new ModelException();
      oErr.iErrorCode = ErrorGUI.BAD_DATA;
      oErr.sCallingFunction = "JAVA";
      oErr.sMessage = "A detailed output file name is required.";
      throw (oErr);
    }
  }

  /**
   * Performs setup.
   * @param oPop TreePopulation object.
   * @throws javawrapper.ModelException Doesn't throw this exception.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void DoSetup(TreePopulation oPop) throws javawrapper.ModelException {
    m_oPop = oPop;
    int i;

    //Make a tree settings object for each tree type
    if (mp_oShortTreeSaveSettings.size() == 0) {
      for (i = 0; i < m_oPop.GetNumberOfTypes(); i++) {
        mp_oShortTreeSaveSettings.add(i, new TreeOutputSaveInfo(i));
      }
    }
  }

  /**
   * Displays the output setup options.
   * @param oWindow Main application window.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  void DisplayWindow(MainWindow oWindow) {
    OutputSetup oOutputFrame = new OutputSetup(this, oWindow);
    oOutputFrame.pack();
    oOutputFrame.setLocationRelativeTo(oWindow);
    oOutputFrame.setVisible(true);
  }

  /**
   * This has its own display window; it won't contribute data to the display.
   * @param oPop TreePopulation object.
   * @return null.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public Vector FormatDataForDisplay(TreePopulation oPop) {
    return null;
  }

  /**
   * Gets the tree population.
   * @return Tree population.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public TreePopulation GetTreePopulation() {
    return m_oPop;
  }

  /**
   * Deletes all detailed output tree settings.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void ClearDetailedTreeSettings() {
    mp_oDetailedTreeSaveSettings.clear();

    //Make sure the detailed output behavior is not enabled if there is also
    //no grids
    Behavior oBeh = GetBehaviorByKey("detailed");
    if (mp_oDetailedGridSaveSettings.size() == 0) {
      oBeh.m_bIsEnabled = false;
    }

    //Clear the behavior's apply to settings
    oBeh.ClearSpeciesTypeCombos();
  }

  /**
   * Deletes all detailed output grid settings.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void ClearDetailedGridSettings() {
    mp_oDetailedGridSaveSettings.clear();

    //Make sure the detailed output behavior is not enabled if there is also no grids
    Behavior oBeh = GetBehaviorByKey("detailed");
    if (mp_oDetailedTreeSaveSettings.size() == 0) {
      oBeh.m_bIsEnabled = false;
    }

  }

  /**
   * Adds a new detailed output tree settings object.  This will overwrite any
   * existing settings for that species and type.
   * @param oTreeSettings The new settings object.
   * @throws ModelException if either the type or the species name is invalid.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void AddDetailedTreeSettings(DetailedTreeSettings oTreeSettings) throws
      ModelException {

    if (null == oTreeSettings) {
      return;
    }

    DetailedTreeSettings oData;
    int i,
        iSpecies = oTreeSettings.GetSpecies(),
        iType = oTreeSettings.GetType();

    //Validate the type
    if (iType < 0 || iType >= m_oPop.GetNumberOfTypes()) {
      ModelException oErr = new ModelException();
      oErr.iErrorCode = ErrorGUI.BAD_DATA;
      oErr.sCallingFunction = "JAVA";
      oErr.sMessage = "Invalid type passed to output  - " +
          String.valueOf(iType);
      throw (oErr);
    }

    //Validate the species
    if (iSpecies < 0 || iSpecies >= m_oPop.GetNumberOfSpecies()) {
      ModelException oErr = new ModelException();
      oErr.iErrorCode = ErrorGUI.BAD_DATA;
      oErr.sCallingFunction = "JAVA";
      oErr.sMessage = "Invalid species passed to output - " +
          String.valueOf(iSpecies);
      throw (oErr);
    }

    //Make sure there are some settings in here
    /*    if (oTreeSettings.GetNumberOfBools() == 0 &&
            oTreeSettings.GetNumberOfChars() == 0 &&
            oTreeSettings.GetNumberOfFloats() == 0 &&
            oTreeSettings.GetNumberOfInts() == 0) {
          return;
        } */

    //Check to see if this type/species combo exists already
    for (i = 0; i < mp_oDetailedTreeSaveSettings.size(); i++) {
      oData = (DetailedTreeSettings) mp_oDetailedTreeSaveSettings.elementAt(i);
      if (iSpecies == oData.GetSpecies() && iType == oData.GetType()) {
        //Remove this one
        mp_oDetailedTreeSaveSettings.remove(i);
        i--;
      }
    }

    mp_oDetailedTreeSaveSettings.add(oTreeSettings);

    //Make sure the detailed output behavior is enabled
    Behavior oBeh = GetBehaviorByKey("detailed");
    oBeh.m_bIsEnabled = true;

    //Add this species/type combo
    oBeh.AddSpeciesTypeCombo(new SpeciesTypeCombo(oTreeSettings.GetSpecies(),
                                                  oTreeSettings.GetType(),
                                                  m_oPop));
  }

  /**
   * Adds a new detailed output grid settings object.  This will overwrite any
   * existing settings for that grid.
   * @param oTreeSettings The new settings object.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void AddDetailedGridSettings(DetailedGridSettings oTreeSettings) {
    DetailedGridSettings oData;
    int i;

    if (null == oTreeSettings) {
      return;
    }

    //Make sure there are some settings in here
    /*   if (oTreeSettings.GetNumberOfBools() == 0 &&
           oTreeSettings.GetNumberOfChars() == 0 &&
           oTreeSettings.GetNumberOfFloats() == 0 &&
           oTreeSettings.GetNumberOfInts() == 0) {
         return;
       } */

    //Check to see if this type/species combo exists already
    for (i = 0; i < mp_oDetailedGridSaveSettings.size(); i++) {
      oData = (DetailedGridSettings) mp_oDetailedGridSaveSettings.elementAt(i);
      if (oTreeSettings.GetName() == oData.GetName()) {
        //Remove this one
        mp_oDetailedGridSaveSettings.remove(i);
        i--;
      }
    }

    mp_oDetailedGridSaveSettings.add(oTreeSettings);

    //Make sure the detailed output behavior is enabled
    Behavior oBeh = GetBehaviorByKey("detailed");
    oBeh.m_bIsEnabled = true;
  }

  /**
   * Sets a data object's value.  Overriden from the base class to capture
   * output settings.
   * @param sXMLTag XML tag of data object whose value is to be set.
   * @param sXMLParentTag The immediate parent tag that sXMLTag is within.
   * @param oAttributes Attributes of XML tag
   * @param oData Data value appropriate to the data type
   * @return true if the value was set successfully; false if the value could
   * not be found.  (This would not be an error, because I need a way to
   * cycle through the objects until one of the objects comes up with a
   * match.)
   * @throws ModelException if the value could not be assigned to the data
   * object.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public boolean SetSingleValueByXMLTag(String sXMLTag, String sXMLParentTag,
                                        Attributes oAttributes,
                                        Object oData) throws
      ModelException {

    if (sXMLTag.equals("ou_float")) {
      //Detailed output float value - tree or grid
      m_oCurrentDetailedSettings.AddFloat(String.valueOf(oData));
      return true;
    }
    else if (sXMLTag.equals("ou_int")) {
      //Detailed output int value - tree or grid
      m_oCurrentDetailedSettings.AddInt(String.valueOf(oData));
      return true;
    }
    else if (sXMLTag.equals("ou_saveFreq")) {
      //Detailed output save frequency - tree or grid
      m_oCurrentDetailedSettings.SetSaveFrequency(new Integer(String.valueOf(
          oData)).
                                                 intValue());
      return true;
    }
    else if (sXMLTag.equals("ou_bool")) {
      //Detailed output bool value - tree or grid
      m_oCurrentDetailedSettings.AddBool(String.valueOf(oData));
      return true;
    }
    else if (sXMLTag.equals("ou_char")) {
      //Detailed output char value - tree or grid
      m_oCurrentDetailedSettings.AddChar(String.valueOf(oData));
      return true;
    }
    else if (sXMLTag.equals("ou_packageFloat")) {
      //Detailed output package float value for grid
      DetailedGridSettings oGrid = (DetailedGridSettings)
          m_oCurrentDetailedSettings;
      oGrid.AddPackageFloat(String.valueOf(oData));
      return true;
    }
    else if (sXMLTag.equals("ou_packageInt")) {
      //Detailed output package int value for grid
      DetailedGridSettings oGrid = (DetailedGridSettings)
          m_oCurrentDetailedSettings;
      oGrid.AddPackageInt(String.valueOf(oData));
      return true;
    }
    else if (sXMLTag.equals("ou_packageBool")) {
      //Detailed output package bool value for grid
      DetailedGridSettings oGrid = (DetailedGridSettings)
          m_oCurrentDetailedSettings;
      oGrid.AddPackageBool(String.valueOf(oData));
      return true;
    }
    else if (sXMLTag.equals("ou_packageChar")) {
      //Detailed output package char value for grid
      DetailedGridSettings oGrid = (DetailedGridSettings)
          m_oCurrentDetailedSettings;
      oGrid.AddPackageChar(String.valueOf(oData));
      return true;
    }

    else if (sXMLTag.equals("so_subplotName")) {
      //Subplot name - create the subplot
      m_oCurrentSubplot = AddSubplot( (String) oData);
      return true;
    }

    return super.SetSingleValueByXMLTag(sXMLTag, sXMLTag, oAttributes, oData);
  }

  /**
   * Sets a data vector's value.  Overridden from the base class.  Due to
   * the vagaries of parameter file parsing, our values may come in through
   * here instead of in single values because there could be repeats.
   * @param sXMLTag Parent XML tag of data vector whose value is to be set.
   * @param sXMLParentTag The immediate parent tag that sXMLTag is within.
   * @param p_oData Vector of data values appropriate to the data type
   * @param p_sChildXMLTags The XML tags of the child elements
   * @param p_bAppliesTo Array of booleans saying which of the vector values
   * should be set.  This is important in the case of species-specifics - the
   * vector index is the species number but not all species are set.
   * @param oParentAttributes Attributes of parent tag.  May be useful when
   * overridding this for unusual tags.
   * @param p_oAttributes Attributes passed from parser.  This may be needed
   * when overriding this function.  Basic species-specific values are
   * already handled by this function.
   * @return true if the value was set successfully; false if the value could
   * not be found.  (This would not be an error, because I need a way to
   * cycle through the objects until one of the objects comes up with a
   * match.)  If a match to a data object is made via XML tag, but the found
   * object is not a ModelVector, this returns false.
   * @throws ModelException if the value could not be assigned to the data
   * object.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public boolean SetVectorValueByXMLTag(String sXMLTag, String sXMLParentTag,
                                        Vector p_oData,
                                        String[] p_sChildXMLTags,
                                        boolean[] p_bAppliesTo,
                                        Attributes oParentAttributes,
                                        Attributes[] p_oAttributes) throws
      ModelException {

    int i;
    boolean bReturn = false;
    for (i = 0; i < p_sChildXMLTags.length; i++) {
      if (p_sChildXMLTags[i] != null) {
        if (SetSingleValueByXMLTag(p_sChildXMLTags[i], sXMLParentTag,
                                   p_oAttributes[i], p_oData.elementAt(i))) {
          bReturn = true;
        }
      }
    }

    return bReturn;
  }

  /**
   * Accepts an XML parent tag (empty, no data) from the parser.  This function
   * watches for the following output tags:
   * <ul>
   * <li>ou_treeInfo
   * <li>ou_gridInfo
   * <li>so_treeTypeInfo
   * <li>so_saveRBA
   * <li>so_saveABA
   * <li>so_saveRDN
   * <li>so_saveADN
   * <li>po_point
   * </ul>
   * @param sXMLTag The XML tag.
   * @param oAttributes The attributes of this object.
   * @throws ModelException if data is missing or invalid.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void ReadXMLParentTag(String sXMLTag, Attributes oAttributes) throws
      ModelException {

    if (sXMLTag.equals("ou_treeInfo")) {

      //This is detailed output tree info

      //Extract species and type
      int iSpecies = m_oPop.GetSpeciesCodeFromName(oAttributes.getValue(
          "species")),
          iType = m_oPop.GetTypeCodeFromName(oAttributes.getValue("type"));

      if (iSpecies == -1 || iType == -1) {
        ModelException oErr = new ModelException();
        oErr.sMessage =
            "Attributes missing from tag \"ou_treeInfo\" - \"species\" and/or \"type\"";
        throw (oErr);
      }

      m_oCurrentDetailedSettings = new DetailedTreeSettings(iType, iSpecies);
      AddDetailedTreeSettings( (DetailedTreeSettings) m_oCurrentDetailedSettings);

    }
    else if (sXMLTag.equals("ou_gridInfo")) {

      //This is detailed output grid info

      //Extract the grid name
      String sGridName = oAttributes.getValue("gridName");
      if (sGridName == null) {
        ModelException oErr = new ModelException();
        oErr.sMessage =
            "Attribute missing from tag \"ou_gridInfo\" - \"gridName\"";
        throw (oErr);
      }

      m_oCurrentDetailedSettings = new DetailedGridSettings(sGridName);
      AddDetailedGridSettings( (DetailedGridSettings) m_oCurrentDetailedSettings);
    }
    else if (sXMLTag.equals("so_treeTypeInfo")) {

      //This is short output tree info

      //Extract type
      m_iCurrentType = m_oPop.GetTypeCodeFromName(oAttributes.getValue("type"));

      if (m_iCurrentType == -1) {
        ModelException oErr = new ModelException();
        oErr.sMessage =
            "Attribute missing from tag \"so_treeTypeInfo\" - \"type\"";
        throw (oErr);
      }
    }
    else if (sXMLTag.equals("so_saveRBA")) {
      //Short output - save relative basal area
      //Get whether or not to save
      String sSave = oAttributes.getValue("save");
      boolean bSave;
      if (sSave == null) {
        ModelException oErr = new ModelException();
        oErr.sMessage = "Output - attribute \"save\" missing from XML tag " +
            "\"so_saveRBA\" in input file.";
        throw (oErr);
      }
      if (sSave.equals("true")) {
        bSave = true;
      }
      else if (sSave.equals("false")) {
        bSave = false;
      }
      else {
        ModelException oErr = new ModelException();
        oErr.sMessage = "Output - attribute \"save\" of XML tag " +
            "\"so_saveRBA\" in input file has unrecognized value \"" + sSave +
            "\".";
        throw (oErr);
      }
      SetSaveRelativeBasalArea(m_iCurrentType, bSave);
    }
    else if (sXMLTag.equals("so_saveABA")) {
      //Short output - save absolute basal area
      //Get whether or not to save
      String sSave = oAttributes.getValue("save");
      boolean bSave;
      if (sSave == null) {
        ModelException oErr = new ModelException();
        oErr.sMessage = "Output - attribute \"save\" missing from XML tag " +
            "\"so_saveABA\" in input file.";
        throw (oErr);
      }
      if (sSave.equals("true")) {
        bSave = true;
      }
      else if (sSave.equals("false")) {
        bSave = false;
      }
      else {
        ModelException oErr = new ModelException();
        oErr.sMessage = "Output - attribute \"save\" of XML tag " +
            "\"so_saveABA\" in input file has unrecognized value \"" + sSave +
            "\".";
        throw (oErr);
      }
      SetSaveAbsoluteBasalArea(m_iCurrentType, bSave);
    }
    else if (sXMLTag.equals("so_saveRDN")) {
      //Short output - save relative density
      //Get whether or not to save
      String sSave = oAttributes.getValue("save");
      boolean bSave;
      if (sSave == null) {
        ModelException oErr = new ModelException();
        oErr.sMessage = "Output - attribute \"save\" missing from XML tag " +
            "\"so_saveRDN\" in input file.";
        throw (oErr);
      }
      if (sSave.equals("true")) {
        bSave = true;
      }
      else if (sSave.equals("false")) {
        bSave = false;
      }
      else {
        ModelException oErr = new ModelException();
        oErr.sMessage = "Output - attribute \"save\" of XML tag " +
            "\"so_saveRDN\" in input file has unrecognized value \"" + sSave +
            "\".";
        throw (oErr);
      }
      SetSaveRelativeDensity(m_iCurrentType, bSave);
    }
    else if (sXMLTag.equals("so_saveADN")) {
      //Short output - save absolute density
      //Get whether or not to save
      String sSave = oAttributes.getValue("save");
      boolean bSave;
      if (sSave == null) {
        ModelException oErr = new ModelException();
        oErr.sMessage = "Output - attribute \"save\" missing from XML tag " +
            "\"so_saveADN\" in input file.";
        throw (oErr);
      }
      if (sSave.equals("true")) {
        bSave = true;
      }
      else if (sSave.equals("false")) {
        bSave = false;
      }
      else {
        ModelException oErr = new ModelException();
        oErr.sMessage = "Output - attribute \"save\" of XML tag " +
            "\"so_saveADN\" in input file has unrecognized value \"" + sSave +
            "\".";
        throw (oErr);
      }
      SetSaveAbsoluteDensity(m_iCurrentType, bSave);
    }
    else if (sXMLTag.equals("po_point") && m_oCurrentSubplot != null) {
      Plot oPlot = m_oManager.GetPlot();

      int iX = new Integer( (String) oAttributes.getValue("x")).
          intValue(),
          iY = new Integer( (String) oAttributes.getValue("y")).
          intValue();
      m_oCurrentSubplot.AddCell(iX, iY, oPlot);
    }
    super.ReadXMLParentTag(sXMLTag, oAttributes);
  }
}
