package javawrapper;

import junit.framework.*;
import java.io.FileWriter;
import java.io.File;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.XMLReaderFactory;
import org.xml.sax.helpers.DefaultHandler;
import org.xml.sax.InputSource;

/**
 * Tests the OutputBehaviors class.
 * <p>Copyright: Copyright (c) Charles D. Canham 2003</p>
 * <p>Company: Institute of Ecosystem Studies</p>
 * @author Lora E. Murphy
 * @version 1.0
 *
 * <br>Edit history:
 * <br>------------------
 * <br>April 28, 2004: Submitted in beta version (LEM)

 */
public class TestOutputBehaviors
    extends TestCase {
  private OutputBehaviors p_oOutputBehaviors = null;

  protected void setUp() throws Exception {
    super.setUp();

    try {
      TreePopulation oPop = new TreePopulation(null);
      oPop.SetSpeciesNames(new String[] {"ACSA", "ACRU", "QURU"});
      p_oOutputBehaviors = new OutputBehaviors(null);
      p_oOutputBehaviors.DoSetup(oPop);
    }
    catch (ModelException oErr) {
      throw new Exception(oErr.sMessage);
    }
    assertEquals(0, p_oOutputBehaviors.GetNumberOfDetailedTreeSettings());
    assertEquals(0, p_oOutputBehaviors.GetNumberOfDetailedGridSettings());
  }

  protected void tearDown() throws Exception {
    p_oOutputBehaviors = null;
    super.tearDown();
  }

  public void testAddDetailedGridSettings() {
    DetailedGridSettings oSettings = null, oTest;
    p_oOutputBehaviors.AddDetailedGridSettings(oSettings);

    //Normal processing

    //First added
    oSettings = new DetailedGridSettings("Grid 1");
    oSettings.AddChar("Char1");
    oSettings.AddInt("Integer 1");
    oSettings.SetSaveFrequency(2);
    p_oOutputBehaviors.AddDetailedGridSettings(oSettings);
    assertEquals(1, p_oOutputBehaviors.GetNumberOfDetailedGridSettings());
    oTest = p_oOutputBehaviors.GetDetailedGridSetting(0);
    assertTrue(testEquals(oSettings, oTest));

    //Second added
    oSettings = null;
    oSettings = new DetailedGridSettings("Grid 2");
    oSettings.AddChar("Char1");
    oSettings.AddInt("Integer 1");
    oSettings.AddBool("OneBool");
    oSettings.AddFloat("OneFloat");
    oSettings.AddFloat("OtherFloat");
    oSettings.AddFloat("TwoFloat");
    oSettings.AddChar("OneChar");
    oSettings.SetSaveFrequency(2);
    p_oOutputBehaviors.AddDetailedGridSettings(oSettings);
    assertEquals(2, p_oOutputBehaviors.GetNumberOfDetailedGridSettings());
    oTest = p_oOutputBehaviors.GetDetailedGridSetting(1);
    assertTrue(testEquals(oSettings, oTest));
    assertFalse(testEquals(oSettings,
                           p_oOutputBehaviors.GetDetailedGridSetting(0)));

    //No settings - should not be added
    /*    oSettings = null;
        oSettings = new DetailedGridSettings("Grid 3");
        oSettings.SetSaveFrequency(0);
        p_oOutputBehaviors.AddDetailedGridSettings(oSettings);
        assertEquals(2, p_oOutputBehaviors.GetNumberOfDetailedGridSettings());*/

    //Same type and species as first - should replace
    oSettings = null;
    oSettings = new DetailedGridSettings("Grid 1");
    oSettings.AddFloat("OneFloat");
    oSettings.AddFloat("OtherFloat");
    oSettings.SetSaveFrequency(0);
    p_oOutputBehaviors.AddDetailedGridSettings(oSettings);
    assertEquals(2, p_oOutputBehaviors.GetNumberOfDetailedGridSettings());
    oTest = p_oOutputBehaviors.GetDetailedGridSetting(1);
    assertTrue(testEquals(oSettings, oTest));

    System.out.println(
        "Normal processing succeeded for AddDetailedGridSettings.");
  }

  public void testAddDetailedTreeSettings() {
    DetailedTreeSettings oSettings = null,
        oTest;

    try {
      p_oOutputBehaviors.AddDetailedTreeSettings(oSettings);

      //Normal processing

      //First added
      oSettings = new DetailedTreeSettings(TreePopulation.ADULT, 0);
      oSettings.AddBool("OneBool");
      oSettings.AddFloat("OneFloat");
      oSettings.SetSaveFrequency(2);
      p_oOutputBehaviors.AddDetailedTreeSettings(oSettings);
      assertEquals(1, p_oOutputBehaviors.GetNumberOfDetailedTreeSettings());
      oTest = p_oOutputBehaviors.GetDetailedTreeSetting(0);
      assertTrue(testEquals(oSettings, oTest));

      //No settings - should not be added
      /*      oSettings = null;
            oSettings = new DetailedTreeSettings(TreePopulation.SEEDLING, 1);
            oSettings.SetSaveFrequency(1);
            p_oOutputBehaviors.AddDetailedTreeSettings(oSettings);
            p_oOutputBehaviors.AddDetailedTreeSettings(oSettings);
       assertEquals(1, p_oOutputBehaviors.GetNumberOfDetailedTreeSettings());
            oSettings = p_oOutputBehaviors.GetDetailedTreeSetting(0);
            assertTrue(testEquals(oSettings, oTest));*/

      //Second added
      oSettings = null;
      oSettings = new DetailedTreeSettings(TreePopulation.SEEDLING, 1);
      oSettings.AddBool("OneBool");
      oSettings.AddFloat("OneFloat");
      oSettings.AddFloat("OneFloat");
      oSettings.AddFloat("TwoFloat");
      oSettings.AddChar("OneChar");
      oSettings.SetSaveFrequency(0);
      p_oOutputBehaviors.AddDetailedTreeSettings(oSettings);
      assertEquals(2, p_oOutputBehaviors.GetNumberOfDetailedTreeSettings());
      oTest = p_oOutputBehaviors.GetDetailedTreeSetting(1);
      assertTrue(testEquals(oSettings, oTest));

      //Same type and species as second - should replace
      oSettings = null;
      oSettings = new DetailedTreeSettings(TreePopulation.SEEDLING, 1);
      oSettings.AddBool("OneBool");
      oSettings.AddFloat("OneFloat");
      oSettings.AddFloat("OneFloat");
      oSettings.AddFloat("DifferentFloat");
      oSettings.AddChar("OneChar");
      p_oOutputBehaviors.AddDetailedTreeSettings(oSettings);
      assertEquals(2, p_oOutputBehaviors.GetNumberOfDetailedTreeSettings());
      oSettings = p_oOutputBehaviors.GetDetailedTreeSetting(1);
      assertFalse(testEquals(oSettings, oTest)); //oTest is still from before

      //Same type but not same species - should add
      oSettings = null;
      oSettings = new DetailedTreeSettings(TreePopulation.SEEDLING, 0);
      oSettings.AddBool("OneBool");
      oSettings.AddFloat("OneFloat");
      oSettings.AddFloat("OneFloat");
      oSettings.AddFloat("DifferentFloat");
      oSettings.AddChar("OneChar");
      p_oOutputBehaviors.AddDetailedTreeSettings(oSettings);
      assertEquals(3, p_oOutputBehaviors.GetNumberOfDetailedTreeSettings());

      //Same species but not same type - should add
      oSettings = null;
      oSettings = new DetailedTreeSettings(TreePopulation.SAPLING, 1);
      oSettings.AddBool("OneBool");
      oSettings.AddFloat("OneFloat");
      oSettings.AddFloat("OneFloat");
      oSettings.AddFloat("DifferentFloat");
      oSettings.AddChar("OneChar");
      p_oOutputBehaviors.AddDetailedTreeSettings(oSettings);
      assertEquals(4, p_oOutputBehaviors.GetNumberOfDetailedTreeSettings());

    }
    catch (ModelException oErr) {
      fail("Normal processing failed for AddDetailedTreeSettings with message " +
           oErr.sMessage);
    }

    System.out.println(
        "Normal processing succeeded for AddDetailedTreeSettings.");

    //Exception processing

    //Invalid species
    try {
      oSettings = null;
      oSettings = new DetailedTreeSettings(TreePopulation.SEEDLING, 5);
      oSettings.AddBool("OneBool");
      oSettings.AddFloat("OneFloat");
      oSettings.AddFloat("OneFloat");
      oSettings.AddFloat("DifferentFloat");
      oSettings.AddChar("OneChar");
      p_oOutputBehaviors.AddDetailedTreeSettings(oSettings);
      fail("AddDetailedTreeSettings failed to catch invalid species");
    }
    catch (ModelException oErr) {
      ;
    }

    //Invalid type
    try {
      oSettings = null;
      oSettings = new DetailedTreeSettings(17, 0);
      oSettings.AddBool("OneBool");
      oSettings.AddFloat("OneFloat");
      oSettings.AddFloat("OneFloat");
      oSettings.AddFloat("DifferentFloat");
      oSettings.AddChar("OneChar");
      p_oOutputBehaviors.AddDetailedTreeSettings(oSettings);
      fail("AddDetailedTreeSettings failed to catch invalid type");
    }
    catch (ModelException oErr) {
      ;
    }
    System.out.println(
        "Exception processing succeeded for AddDetailedTreeSettings.");
  }

  public void testClearDetailedGridSettings() {
    try {
      p_oOutputBehaviors.ClearDetailedGridSettings();
      assertEquals(0, p_oOutputBehaviors.GetNumberOfDetailedGridSettings());
      p_oOutputBehaviors.ClearDetailedGridSettings();

      //Make sure only grids are cleared, not trees
      DetailedTreeSettings oTreeSettings = new DetailedTreeSettings(
          TreePopulation.ADULT, 0);
      oTreeSettings.AddBool("OneBool");
      oTreeSettings.AddFloat("OneFloat");
      oTreeSettings.SetSaveFrequency(2);
      p_oOutputBehaviors.AddDetailedTreeSettings(oTreeSettings);

      DetailedGridSettings oGridSettings = new DetailedGridSettings("Grid 1");
      oGridSettings.AddBool("OneBool");
      oGridSettings.AddFloat("OneFloat");
      oGridSettings.SetSaveFrequency(2);
      p_oOutputBehaviors.AddDetailedGridSettings(oGridSettings);

      p_oOutputBehaviors.ClearDetailedGridSettings();
      assertEquals(0, p_oOutputBehaviors.GetNumberOfDetailedGridSettings());
      assertTrue(p_oOutputBehaviors.GetNumberOfDetailedTreeSettings() > 0);
    }
    catch (ModelException oErr) {
      fail("ClearDetailedGridSettings failed with message " + oErr.sMessage);
    }
  }

  public void testClearDetailedTreeSettings() {
    try {
      p_oOutputBehaviors.ClearDetailedTreeSettings();
      assertEquals(0, p_oOutputBehaviors.GetNumberOfDetailedTreeSettings());
      p_oOutputBehaviors.ClearDetailedTreeSettings();

      //Make sure only trees are cleared, not grids
      DetailedTreeSettings oTreeSettings = new DetailedTreeSettings(
          TreePopulation.ADULT, 0);
      oTreeSettings.AddBool("OneBool");
      oTreeSettings.AddFloat("OneFloat");
      oTreeSettings.SetSaveFrequency(2);
      p_oOutputBehaviors.AddDetailedTreeSettings(oTreeSettings);

      DetailedGridSettings oGridSettings = new DetailedGridSettings("Grid 1");
      oGridSettings.AddBool("OneBool");
      oGridSettings.AddFloat("OneFloat");
      oGridSettings.SetSaveFrequency(2);
      p_oOutputBehaviors.AddDetailedGridSettings(oGridSettings);

      p_oOutputBehaviors.ClearDetailedTreeSettings();
      assertEquals(0, p_oOutputBehaviors.GetNumberOfDetailedTreeSettings());
      assertTrue(p_oOutputBehaviors.GetNumberOfDetailedGridSettings() > 0);
    }
    catch (ModelException oErr) {
      fail("ClearDetailedGridSettings failed with message " + oErr.sMessage);
    }

  }

  /**
   * Tests reading of parameter file XML rundata settings.
   */
  public void testReadXMLRundataSettings() {
    GUIManager oManager = null;
    try {
      oManager = new GUIManager(null);
      oManager.ClearCurrentData();

      //Set up and parse our test file
      //Create our parser
      XMLReader oParser = XMLReaderFactory.createXMLReader(
          "org.apache.xerces.parsers.SAXParser");
      //Create the handler that looks for setup data and register it
      DefaultHandler oHandler = new ParameterFileParser(oManager);
      oParser.setContentHandler(oHandler);
      //Parse the file
      String sFileName = WriteXMLFile1();
      InputSource oToParse = new InputSource(sFileName);
      oParser.parse(oToParse);

      OutputBehaviors oOutput = oManager.GetOutputBehaviors();
      assertEquals(10, oOutput.GetNumberOfDetailedGridSettings());
      assertEquals(6, oOutput.GetNumberOfDetailedTreeSettings());

      int i;
      for (i = 0; i < oOutput.GetNumberOfDetailedTreeSettings(); i++) {
        DetailedTreeSettings oSettings = (DetailedTreeSettings) oOutput.
            GetDetailedTreeSetting(i);
        if (0 == oSettings.GetSpecies()) {
          assertEquals(1, oSettings.GetSaveFrequency());
          assertEquals(1, oSettings.GetNumberOfBools());
          assertEquals(1, oSettings.GetNumberOfChars());
          assertEquals(1, oSettings.GetNumberOfFloats());
          assertEquals(1, oSettings.GetNumberOfInts());
          assertEquals("species 1 test float 1",
                       oSettings.GetFloat(0).GetCodeName());
          assertEquals("species 1 test char 1",
                       oSettings.GetChar(0).GetCodeName());
          assertEquals("species 1 test int 1",
                       oSettings.GetInt(0).GetCodeName());
          assertEquals("species 1 test bool 1",
                       oSettings.GetBool(0).GetCodeName());

        }
        else if (1 == oSettings.GetSpecies()) {

          assertEquals(2, oSettings.GetSaveFrequency());
          assertEquals(2, oSettings.GetNumberOfBools());
          assertEquals(2, oSettings.GetNumberOfChars());
          assertEquals(2, oSettings.GetNumberOfFloats());
          assertEquals(2, oSettings.GetNumberOfInts());
          assertEquals("species 2 test float 1",
                       oSettings.GetFloat(0).GetCodeName());
          assertEquals("species 2 test char 1",
                       oSettings.GetChar(0).GetCodeName());
          assertEquals("species 2 test int 1",
                       oSettings.GetInt(0).GetCodeName());
          assertEquals("species 2 test bool 1",
                       oSettings.GetBool(0).GetCodeName());
          assertEquals("species 2 test float 2",
                       oSettings.GetFloat(1).GetCodeName());
          assertEquals("species 2 test char 2",
                       oSettings.GetChar(1).GetCodeName());
          assertEquals("species 2 test int 2",
                       oSettings.GetInt(1).GetCodeName());
          assertEquals("species 2 test bool 2",
                       oSettings.GetBool(1).GetCodeName());

        }
        else if (2 == oSettings.GetSpecies()) {

          assertEquals(3, oSettings.GetSaveFrequency());
          assertEquals(0, oSettings.GetNumberOfBools());
          assertEquals(1, oSettings.GetNumberOfChars());
          assertEquals(0, oSettings.GetNumberOfFloats());
          assertEquals(0, oSettings.GetNumberOfInts());
          assertEquals("species 3 test char 1",
                       oSettings.GetChar(0).GetCodeName());

        }
        else if (3 == oSettings.GetSpecies()) {

          assertEquals(4, oSettings.GetSaveFrequency());
          assertEquals(0, oSettings.GetNumberOfBools());
          assertEquals(0, oSettings.GetNumberOfChars());
          assertEquals(1, oSettings.GetNumberOfFloats());
          assertEquals(0, oSettings.GetNumberOfInts());
          assertEquals("species 4 test float 1",
                       oSettings.GetFloat(0).GetCodeName());

        }
        else if (4 == oSettings.GetSpecies()) {

          assertEquals(5, oSettings.GetSaveFrequency());
          assertEquals(1, oSettings.GetNumberOfBools());
          assertEquals(0, oSettings.GetNumberOfChars());
          assertEquals(0, oSettings.GetNumberOfFloats());
          assertEquals(0, oSettings.GetNumberOfInts());
          assertEquals("species 5 test bool 1",
                       oSettings.GetBool(0).GetCodeName());

        }
        else if (5 == oSettings.GetSpecies()) {

          assertEquals(6, oSettings.GetSaveFrequency());
          assertEquals(0, oSettings.GetNumberOfBools());
          assertEquals(0, oSettings.GetNumberOfChars());
          assertEquals(0, oSettings.GetNumberOfFloats());
          assertEquals(1, oSettings.GetNumberOfInts());
          assertEquals("species 6 test int 1",
                       oSettings.GetInt(0).GetCodeName());

        }
        else {
          ModelException oErr = new ModelException();
          oErr.sMessage = "Unrecognized species in tree.";
          throw (oErr);
        }
      }

      for (i = 0; i < oOutput.GetNumberOfDetailedGridSettings(); i++) {
        DetailedGridSettings oSettings = (DetailedGridSettings) oOutput.
            GetDetailedGridSetting(i);
        if (oSettings.GetName().equals("grid 1")) {
          assertEquals(1, oSettings.GetSaveFrequency());
          assertEquals(1, oSettings.GetNumberOfBools());
          assertEquals(1, oSettings.GetNumberOfChars());
          assertEquals(1, oSettings.GetNumberOfFloats());
          assertEquals(1, oSettings.GetNumberOfInts());
          assertEquals(0, oSettings.GetNumberOfPackageBools());
          assertEquals(0, oSettings.GetNumberOfPackageChars());
          assertEquals(0, oSettings.GetNumberOfPackageFloats());
          assertEquals(0, oSettings.GetNumberOfPackageInts());
          assertEquals("grid 1 test float 1",
                       oSettings.GetFloat(0).GetCodeName());
          assertEquals("grid 1 test char 1",
                       oSettings.GetChar(0).GetCodeName());
          assertEquals("grid 1 test int 1",
                       oSettings.GetInt(0).GetCodeName());
          assertEquals("grid 1 test bool 1",
                       oSettings.GetBool(0).GetCodeName());

        }
        else if (oSettings.GetName().equals("grid 2")) {

          assertEquals(2, oSettings.GetSaveFrequency());
          assertEquals(2, oSettings.GetNumberOfBools());
          assertEquals(2, oSettings.GetNumberOfChars());
          assertEquals(2, oSettings.GetNumberOfFloats());
          assertEquals(2, oSettings.GetNumberOfInts());
          assertEquals(0, oSettings.GetNumberOfPackageBools());
          assertEquals(0, oSettings.GetNumberOfPackageChars());
          assertEquals(0, oSettings.GetNumberOfPackageFloats());
          assertEquals(0, oSettings.GetNumberOfPackageInts());
          assertEquals("grid 2 test float 1",
                       oSettings.GetFloat(0).GetCodeName());
          assertEquals("grid 2 test char 1",
                       oSettings.GetChar(0).GetCodeName());
          assertEquals("grid 2 test int 1",
                       oSettings.GetInt(0).GetCodeName());
          assertEquals("grid 2 test bool 1",
                       oSettings.GetBool(0).GetCodeName());
          assertEquals("grid 2 test float 2",
                       oSettings.GetFloat(1).GetCodeName());
          assertEquals("grid 2 test char 2",
                       oSettings.GetChar(1).GetCodeName());
          assertEquals("grid 2 test int 2",
                       oSettings.GetInt(1).GetCodeName());
          assertEquals("grid 2 test bool 2",
                       oSettings.GetBool(1).GetCodeName());

        }
        else if (oSettings.GetName().equals("grid 3")) {

          assertEquals(3, oSettings.GetSaveFrequency());
          assertEquals(0, oSettings.GetNumberOfBools());
          assertEquals(1, oSettings.GetNumberOfChars());
          assertEquals(0, oSettings.GetNumberOfFloats());
          assertEquals(0, oSettings.GetNumberOfInts());
          assertEquals(0, oSettings.GetNumberOfPackageBools());
          assertEquals(0, oSettings.GetNumberOfPackageChars());
          assertEquals(0, oSettings.GetNumberOfPackageFloats());
          assertEquals(0, oSettings.GetNumberOfPackageInts());
          assertEquals("grid 3 test char 1",
                       oSettings.GetChar(0).GetCodeName());

        }
        else if (oSettings.GetName().equals("grid 4")) {

          assertEquals(4, oSettings.GetSaveFrequency());
          assertEquals(0, oSettings.GetNumberOfBools());
          assertEquals(0, oSettings.GetNumberOfChars());
          assertEquals(1, oSettings.GetNumberOfFloats());
          assertEquals(0, oSettings.GetNumberOfInts());
          assertEquals(0, oSettings.GetNumberOfPackageBools());
          assertEquals(0, oSettings.GetNumberOfPackageChars());
          assertEquals(0, oSettings.GetNumberOfPackageFloats());
          assertEquals(0, oSettings.GetNumberOfPackageInts());
          assertEquals("grid 4 test float 1",
                       oSettings.GetFloat(0).GetCodeName());

        }
        else if (oSettings.GetName().equals("grid 5")) {

          assertEquals(5, oSettings.GetSaveFrequency());
          assertEquals(1, oSettings.GetNumberOfBools());
          assertEquals(0, oSettings.GetNumberOfChars());
          assertEquals(0, oSettings.GetNumberOfFloats());
          assertEquals(0, oSettings.GetNumberOfInts());
          assertEquals(0, oSettings.GetNumberOfPackageBools());
          assertEquals(0, oSettings.GetNumberOfPackageChars());
          assertEquals(0, oSettings.GetNumberOfPackageFloats());
          assertEquals(0, oSettings.GetNumberOfPackageInts());
          assertEquals("grid 5 test bool 1",
                       oSettings.GetBool(0).GetCodeName());

        }
        else if (oSettings.GetName().equals("grid 6")) {

          assertEquals(6, oSettings.GetSaveFrequency());
          assertEquals(0, oSettings.GetNumberOfBools());
          assertEquals(0, oSettings.GetNumberOfChars());
          assertEquals(0, oSettings.GetNumberOfFloats());
          assertEquals(1, oSettings.GetNumberOfInts());
          assertEquals(0, oSettings.GetNumberOfPackageBools());
          assertEquals(0, oSettings.GetNumberOfPackageChars());
          assertEquals(0, oSettings.GetNumberOfPackageFloats());
          assertEquals(0, oSettings.GetNumberOfPackageInts());
          assertEquals("grid 6 test int 1",
                       oSettings.GetInt(0).GetCodeName());

        }
        else if (oSettings.GetName().equals("grid 7")) {

          assertEquals(7, oSettings.GetSaveFrequency());
          assertEquals(1, oSettings.GetNumberOfBools());
          assertEquals(1, oSettings.GetNumberOfChars());
          assertEquals(1, oSettings.GetNumberOfFloats());
          assertEquals(1, oSettings.GetNumberOfInts());
          assertEquals(1, oSettings.GetNumberOfPackageBools());
          assertEquals(1, oSettings.GetNumberOfPackageChars());
          assertEquals(1, oSettings.GetNumberOfPackageFloats());
          assertEquals(1, oSettings.GetNumberOfPackageInts());
          assertEquals("grid 7 test int 1",
                       oSettings.GetInt(0).GetCodeName());
          assertEquals("grid 7 test float 1",
                       oSettings.GetFloat(0).GetCodeName());
          assertEquals("grid 7 test char 1",
                       oSettings.GetChar(0).GetCodeName());
          assertEquals("grid 7 test bool 1",
                       oSettings.GetBool(0).GetCodeName());
          assertEquals("grid 7 test package int 1",
                       oSettings.GetPackageInt(0).GetCodeName());
          assertEquals("grid 7 test package float 1",
                       oSettings.GetPackageFloat(0).GetCodeName());
          assertEquals("grid 7 test package char 1",
                       oSettings.GetPackageChar(0).GetCodeName());
          assertEquals("grid 7 test package bool 1",
                       oSettings.GetPackageBool(0).GetCodeName());
        }
        else if (oSettings.GetName().equals("grid 8")) {

          assertEquals(8, oSettings.GetSaveFrequency());
          assertEquals(2, oSettings.GetNumberOfBools());
          assertEquals(2, oSettings.GetNumberOfChars());
          assertEquals(2, oSettings.GetNumberOfFloats());
          assertEquals(2, oSettings.GetNumberOfInts());
          assertEquals(5, oSettings.GetNumberOfPackageBools());
          assertEquals(4, oSettings.GetNumberOfPackageChars());
          assertEquals(3, oSettings.GetNumberOfPackageFloats());
          assertEquals(2, oSettings.GetNumberOfPackageInts());
          assertEquals("grid 8 test float 1",
                       oSettings.GetFloat(0).GetCodeName());
          assertEquals("grid 8 test char 1",
                       oSettings.GetChar(0).GetCodeName());
          assertEquals("grid 8 test int 1",
                       oSettings.GetInt(0).GetCodeName());
          assertEquals("grid 8 test bool 1",
                       oSettings.GetBool(0).GetCodeName());
          assertEquals("grid 8 test float 2",
                       oSettings.GetFloat(1).GetCodeName());
          assertEquals("grid 8 test char 2",
                       oSettings.GetChar(1).GetCodeName());
          assertEquals("grid 8 test int 2",
                       oSettings.GetInt(1).GetCodeName());
          assertEquals("grid 8 test bool 2",
                       oSettings.GetBool(1).GetCodeName());
          assertEquals("grid 8 test package int 1",
                       oSettings.GetPackageInt(0).GetCodeName());
          assertEquals("grid 8 test package int 2",
                       oSettings.GetPackageInt(1).GetCodeName());
          assertEquals("grid 8 test package float 1",
                       oSettings.GetPackageFloat(0).GetCodeName());
          assertEquals("grid 8 test package float 2",
                       oSettings.GetPackageFloat(1).GetCodeName());
          assertEquals("grid 8 test package float 3",
                       oSettings.GetPackageFloat(2).GetCodeName());
          assertEquals("grid 8 test package char 1",
                       oSettings.GetPackageChar(0).GetCodeName());
          assertEquals("grid 8 test package char 2",
                       oSettings.GetPackageChar(1).GetCodeName());
          assertEquals("grid 8 test package char 3",
                       oSettings.GetPackageChar(2).GetCodeName());
          assertEquals("grid 8 test package char 4",
                       oSettings.GetPackageChar(3).GetCodeName());
          assertEquals("grid 8 test package bool 1",
                       oSettings.GetPackageBool(0).GetCodeName());
          assertEquals("grid 8 test package bool 2",
                       oSettings.GetPackageBool(1).GetCodeName());
          assertEquals("grid 8 test package bool 3",
                       oSettings.GetPackageBool(2).GetCodeName());
          assertEquals("grid 8 test package bool 4",
                       oSettings.GetPackageBool(3).GetCodeName());
          assertEquals("grid 8 test package bool 5",
                       oSettings.GetPackageBool(4).GetCodeName());
        }
        else if (oSettings.GetName().equals("grid 9")) {

          assertEquals(9, oSettings.GetSaveFrequency());
          assertEquals(0, oSettings.GetNumberOfBools());
          assertEquals(0, oSettings.GetNumberOfChars());
          assertEquals(0, oSettings.GetNumberOfFloats());
          assertEquals(0, oSettings.GetNumberOfInts());
          assertEquals(1, oSettings.GetNumberOfPackageBools());
          assertEquals(1, oSettings.GetNumberOfPackageChars());
          assertEquals(1, oSettings.GetNumberOfPackageFloats());
          assertEquals(1, oSettings.GetNumberOfPackageInts());
          assertEquals("grid 9 test package int 1",
                       oSettings.GetPackageInt(0).GetCodeName());
          assertEquals("grid 9 test package float 1",
                       oSettings.GetPackageFloat(0).GetCodeName());
          assertEquals("grid 9 test package char 1",
                       oSettings.GetPackageChar(0).GetCodeName());
          assertEquals("grid 9 test package bool 1",
                       oSettings.GetPackageBool(0).GetCodeName());
        }
        else if (oSettings.GetName().equals("grid 10")) {

          assertEquals(10, oSettings.GetSaveFrequency());
          assertEquals(0, oSettings.GetNumberOfBools());
          assertEquals(0, oSettings.GetNumberOfChars());
          assertEquals(0, oSettings.GetNumberOfFloats());
          assertEquals(0, oSettings.GetNumberOfInts());
          assertEquals(5, oSettings.GetNumberOfPackageBools());
          assertEquals(4, oSettings.GetNumberOfPackageChars());
          assertEquals(3, oSettings.GetNumberOfPackageFloats());
          assertEquals(2, oSettings.GetNumberOfPackageInts());
          assertEquals("grid 10 test package int 1",
                       oSettings.GetPackageInt(0).GetCodeName());
          assertEquals("grid 10 test package int 2",
                       oSettings.GetPackageInt(1).GetCodeName());
          assertEquals("grid 10 test package float 1",
                       oSettings.GetPackageFloat(0).GetCodeName());
          assertEquals("grid 10 test package float 2",
                       oSettings.GetPackageFloat(1).GetCodeName());
          assertEquals("grid 10 test package float 3",
                       oSettings.GetPackageFloat(2).GetCodeName());
          assertEquals("grid 10 test package char 1",
                       oSettings.GetPackageChar(0).GetCodeName());
          assertEquals("grid 10 test package char 2",
                       oSettings.GetPackageChar(1).GetCodeName());
          assertEquals("grid 10 test package char 3",
                       oSettings.GetPackageChar(2).GetCodeName());
          assertEquals("grid 10 test package char 4",
                       oSettings.GetPackageChar(3).GetCodeName());
          assertEquals("grid 10 test package bool 1",
                       oSettings.GetPackageBool(0).GetCodeName());
          assertEquals("grid 10 test package bool 2",
                       oSettings.GetPackageBool(1).GetCodeName());
          assertEquals("grid 10 test package bool 3",
                       oSettings.GetPackageBool(2).GetCodeName());
          assertEquals("grid 10 test package bool 4",
                       oSettings.GetPackageBool(3).GetCodeName());
          assertEquals("grid 10 test package bool 5",
                       oSettings.GetPackageBool(4).GetCodeName());
        }
        else {
          ModelException oErr = new ModelException();
          oErr.sMessage = "Unrecognized grid settings.";
          throw (oErr);
        }
      }

      new File(sFileName).delete();
      System.out.println("XML rundata reading test succeeded.");
    }
    catch (ModelException oErr) {
      fail("XML rundata reading test failed with message " + oErr.sMessage);
    }
    catch (org.xml.sax.SAXException oE) {
      fail("Parse test could not parse parameter file - message: " +
           oE.getMessage());
    }
    catch (java.io.IOException oE) {
      fail("Caught IOException.  Message: " + oE.getMessage());
    }

  }

  /**
   * Tests reading of parameter file XML rundata settings.
   */
  public void testReadXMLShortSettings1() {
    GUIManager oManager = null;
    try {
      oManager = new GUIManager(null);
      oManager.ClearCurrentData();

      //Set up and parse our test file
      //Create our parser
      XMLReader oParser = XMLReaderFactory.createXMLReader(
          "org.apache.xerces.parsers.SAXParser");
      //Create the handler that looks for setup data and register it
      DefaultHandler oHandler = new ParameterFileParser(oManager);
      oParser.setContentHandler(oHandler);

      //Parse the file
      String sFileName = WriteXMLFile2();
      InputSource oToParse = new InputSource(sFileName);
      oParser.parse(oToParse);

      OutputBehaviors oOutput = oManager.GetOutputBehaviors();

      assertTrue(oOutput.GetSaveAbsoluteDensity(TreePopulation.SEEDLING));
      assertTrue(oOutput.GetSaveRelativeDensity(TreePopulation.SEEDLING));
      assertFalse(oOutput.GetSaveAbsoluteBasalArea(TreePopulation.SEEDLING));
      assertFalse(oOutput.GetSaveRelativeBasalArea(TreePopulation.SEEDLING));

      assertTrue(oOutput.GetSaveAbsoluteDensity(TreePopulation.SAPLING));
      assertTrue(oOutput.GetSaveRelativeDensity(TreePopulation.SAPLING));
      assertTrue(oOutput.GetSaveAbsoluteBasalArea(TreePopulation.SAPLING));
      assertTrue(oOutput.GetSaveRelativeBasalArea(TreePopulation.SAPLING));

      assertTrue(oOutput.GetSaveAbsoluteDensity(TreePopulation.ADULT));
      assertTrue(oOutput.GetSaveRelativeDensity(TreePopulation.ADULT));
      assertTrue(oOutput.GetSaveAbsoluteBasalArea(TreePopulation.ADULT));
      assertTrue(oOutput.GetSaveRelativeBasalArea(TreePopulation.ADULT));

      //Subplots
      assertEquals(2, oOutput.mp_oSubplots.size());

      Subplot oSubplot = (Subplot) oOutput.mp_oSubplots.elementAt(0);
      assertEquals("test 1", oSubplot.GetSubplotName());
      assertEquals(1, oSubplot.GetNumberOfCells());
      assertEquals(0, oSubplot.GetCell(0).GetX());
      assertEquals(12, oSubplot.GetCell(0).GetY());

      oSubplot = (Subplot) oOutput.mp_oSubplots.elementAt(1);
      assertEquals("test 2", oSubplot.GetSubplotName());
      assertEquals(5, oSubplot.GetNumberOfCells());
      assertEquals(6, oSubplot.GetCell(0).GetX());
      assertEquals(14, oSubplot.GetCell(0).GetY());
      assertEquals(7, oSubplot.GetCell(1).GetX());
      assertEquals(15, oSubplot.GetCell(1).GetY());
      assertEquals(8, oSubplot.GetCell(2).GetX());
      assertEquals(16, oSubplot.GetCell(2).GetY());
      assertEquals(9, oSubplot.GetCell(3).GetX());
      assertEquals(17, oSubplot.GetCell(3).GetY());
      assertEquals(10, oSubplot.GetCell(4).GetX());
      assertEquals(18, oSubplot.GetCell(4).GetY());

      new File(sFileName).delete();
      System.out.println("XML short output reading test succeeded.");
    }
    catch (ModelException oErr) {
      fail("XML short output reading test failed with message " + oErr.sMessage);
    }
    catch (org.xml.sax.SAXException oE) {
      fail("Parse test could not parse parameter file - message: " +
           oE.getMessage());
    }
    catch (java.io.IOException oE) {
      fail("Caught IOException.  Message: " + oE.getMessage());
    }
  }

  /**
   * Tests reading of parameter file XML rundata settings.
   */
  public void testReadXMLShortSettings2() {
    GUIManager oManager = null;
    try {
      oManager = new GUIManager(null);
      oManager.ClearCurrentData();

      //Set up and parse our test file
      //Create our parser
      XMLReader oParser = XMLReaderFactory.createXMLReader(
          "org.apache.xerces.parsers.SAXParser");
      //Create the handler that looks for setup data and register it
      DefaultHandler oHandler = new ParameterFileParser(oManager);
      oParser.setContentHandler(oHandler);

      //Parse the file
      String sFileName = WriteXMLFile3();
      InputSource oToParse = new InputSource(sFileName);
      oParser.parse(oToParse);

      OutputBehaviors oOutput = oManager.GetOutputBehaviors();

      assertFalse(oOutput.GetSaveAbsoluteDensity(TreePopulation.SEEDLING));
      assertFalse(oOutput.GetSaveRelativeDensity(TreePopulation.SEEDLING));
      assertFalse(oOutput.GetSaveAbsoluteBasalArea(TreePopulation.SEEDLING));
      assertFalse(oOutput.GetSaveRelativeBasalArea(TreePopulation.SEEDLING));

      assertFalse(oOutput.GetSaveAbsoluteDensity(TreePopulation.SAPLING));
      assertTrue(oOutput.GetSaveRelativeDensity(TreePopulation.SAPLING));
      assertFalse(oOutput.GetSaveAbsoluteBasalArea(TreePopulation.SAPLING));
      assertTrue(oOutput.GetSaveRelativeBasalArea(TreePopulation.SAPLING));

      assertTrue(oOutput.GetSaveAbsoluteDensity(TreePopulation.ADULT));
      assertFalse(oOutput.GetSaveRelativeDensity(TreePopulation.ADULT));
      assertFalse(oOutput.GetSaveAbsoluteBasalArea(TreePopulation.ADULT));
      assertFalse(oOutput.GetSaveRelativeBasalArea(TreePopulation.ADULT));

      new File(sFileName).delete();
      System.out.println("XML short output reading test succeeded.");
    }
    catch (ModelException oErr) {
      fail("XML short output reading test failed with message " + oErr.sMessage);
    }
    catch (org.xml.sax.SAXException oE) {
      fail("Parse test could not parse parameter file - message: " +
           oE.getMessage());
    }
    catch (java.io.IOException oE) {
      fail("Caught IOException.  Message: " + oE.getMessage());
    }
  }

  /**
   * This writes an XML file to test value setting.  This file contains rundata
   * only.
   * @return Name of file.
   * @throws ModelException if there is an IO problem.
   */
  protected String WriteXMLFile1() throws ModelException {
    try {
      String sFileName = "\\loratestxml1.xml";
      FileWriter oOut = new FileWriter(sFileName);

      oOut.write(
          "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\" ?>");
      oOut.write("<paramFile fileCode=\"06010101\">");
      oOut.write("<trees>");
      oOut.write("<tr_speciesList>");
      oOut.write("<tr_species speciesName=\"Western_Hemlock\"/>");
      oOut.write("<tr_species speciesName=\"Western_Redcedar\"/>");
      oOut.write("<tr_species speciesName=\"Amabilis_Fir\"/>");
      oOut.write("<tr_species speciesName=\"Subalpine_Fir\"/>");
      oOut.write("<tr_species speciesName=\"Hybrid_Spruce\"/>");
      oOut.write("<tr_species speciesName=\"Lodgepole_Pine\"/>");
      oOut.write("<tr_species speciesName=\"Trembling_Aspen\"/>");
      oOut.write("<tr_species speciesName=\"Black_Cottonwood\"/>");
      oOut.write("<tr_species speciesName=\"Paper_Birch\"/>");
      oOut.write("</tr_speciesList>");
      oOut.write("</trees>");

      //Include an empty behavior list - this triggers setup
      oOut.write("<behaviorList>");
      oOut.write("</behaviorList>");

      oOut.write("<output>");

      //Straight-up - one per data type
      oOut.write("<ou_filename>OutputSizeTest</ou_filename>");
      oOut.write("<ou_treeInfo species=\"Western_Hemlock\" type=\"sapling\">");
      oOut.write("<ou_saveFreq>1</ou_saveFreq>");
      oOut.write("<ou_float>species 1 test float 1</ou_float>");
      oOut.write("<ou_char>species 1 test char 1</ou_char>");
      oOut.write("<ou_int>species 1 test int 1</ou_int>");
      oOut.write("<ou_bool>species 1 test bool 1</ou_bool>");
      oOut.write("</ou_treeInfo>");

      //Multiples of each - not grouped together by type
      oOut.write("<ou_treeInfo species=\"Western_Redcedar\" type=\"sapling\">");
      oOut.write("<ou_saveFreq>2</ou_saveFreq>");
      oOut.write("<ou_float>species 2 test float 1</ou_float>");
      oOut.write("<ou_char>species 2 test char 1</ou_char>");
      oOut.write("<ou_int>species 2 test int 1</ou_int>");
      oOut.write("<ou_bool>species 2 test bool 1</ou_bool>");
      oOut.write("<ou_float>species 2 test float 2</ou_float>");
      oOut.write("<ou_char>species 2 test char 2</ou_char>");
      oOut.write("<ou_int>species 2 test int 2</ou_int>");
      oOut.write("<ou_bool>species 2 test bool 2</ou_bool>");
      oOut.write("</ou_treeInfo>");

      //One char only
      oOut.write("<ou_treeInfo species=\"Amabilis_Fir\" type=\"sapling\">");
      oOut.write("<ou_saveFreq>3</ou_saveFreq>");
      oOut.write("<ou_char>species 3 test char 1</ou_char>");
      oOut.write("</ou_treeInfo>");

      //One float only
      oOut.write("<ou_treeInfo species=\"Subalpine_Fir\" type=\"sapling\">");
      oOut.write("<ou_saveFreq>4</ou_saveFreq>");
      oOut.write("<ou_float>species 4 test float 1</ou_float>");
      oOut.write("</ou_treeInfo>");

      //One bool only
      oOut.write("<ou_treeInfo species=\"Hybrid_Spruce\" type=\"sapling\">");
      oOut.write("<ou_saveFreq>5</ou_saveFreq>");
      oOut.write("<ou_bool>species 5 test bool 1</ou_bool>");
      oOut.write("</ou_treeInfo>");

      //One int only
      oOut.write("<ou_treeInfo species=\"Lodgepole_Pine\" type=\"sapling\">");
      oOut.write("<ou_saveFreq>6</ou_saveFreq>");
      oOut.write("<ou_int>species 6 test int 1</ou_int>");
      oOut.write("</ou_treeInfo>");

      //Now grids - same for all
      //Straight-up - one per data type - no package
      oOut.write("<ou_gridInfo gridName=\"grid 1\">");
      oOut.write("<ou_saveFreq>1</ou_saveFreq>");
      oOut.write("<ou_float>grid 1 test float 1</ou_float>");
      oOut.write("<ou_char>grid 1 test char 1</ou_char>");
      oOut.write("<ou_int>grid 1 test int 1</ou_int>");
      oOut.write("<ou_bool>grid 1 test bool 1</ou_bool>");
      oOut.write("</ou_gridInfo>");

      //Multiples of each - not grouped together by type
      oOut.write("<ou_gridInfo gridName=\"grid 2\">");
      oOut.write("<ou_saveFreq>2</ou_saveFreq>");
      oOut.write("<ou_float>grid 2 test float 1</ou_float>");
      oOut.write("<ou_char>grid 2 test char 1</ou_char>");
      oOut.write("<ou_int>grid 2 test int 1</ou_int>");
      oOut.write("<ou_bool>grid 2 test bool 1</ou_bool>");
      oOut.write("<ou_float>grid 2 test float 2</ou_float>");
      oOut.write("<ou_char>grid 2 test char 2</ou_char>");
      oOut.write("<ou_int>grid 2 test int 2</ou_int>");
      oOut.write("<ou_bool>grid 2 test bool 2</ou_bool>");
      oOut.write("</ou_gridInfo>");

      //One char only
      oOut.write("<ou_gridInfo gridName=\"grid 3\">");
      oOut.write("<ou_saveFreq>3</ou_saveFreq>");
      oOut.write("<ou_char>grid 3 test char 1</ou_char>");
      oOut.write("</ou_gridInfo>");

      //One float only
      oOut.write("<ou_gridInfo gridName=\"grid 4\">");
      oOut.write("<ou_saveFreq>4</ou_saveFreq>");
      oOut.write("<ou_float>grid 4 test float 1</ou_float>");
      oOut.write("</ou_gridInfo>");

      //One bool only
      oOut.write("<ou_gridInfo gridName=\"grid 5\">");
      oOut.write("<ou_saveFreq>5</ou_saveFreq>");
      oOut.write("<ou_bool>grid 5 test bool 1</ou_bool>");
      oOut.write("</ou_gridInfo>");

      //One int only
      oOut.write("<ou_gridInfo gridName=\"grid 6\">");
      oOut.write("<ou_saveFreq>6</ou_saveFreq>");
      oOut.write("<ou_int>grid 6 test int 1</ou_int>");
      oOut.write("</ou_gridInfo>");

      //One per data type, both package and regular
      oOut.write("<ou_gridInfo gridName=\"grid 7\">");
      oOut.write("<ou_saveFreq>7</ou_saveFreq>");
      oOut.write("<ou_float>grid 7 test float 1</ou_float>");
      oOut.write("<ou_char>grid 7 test char 1</ou_char>");
      oOut.write("<ou_int>grid 7 test int 1</ou_int>");
      oOut.write("<ou_bool>grid 7 test bool 1</ou_bool>");
      oOut.write("<ou_packageInt>grid 7 test package int 1</ou_packageInt>");
      oOut.write(
          "<ou_packageFloat>grid 7 test package float 1</ou_packageFloat>");
      oOut.write("<ou_packageChar>grid 7 test package char 1</ou_packageChar>");
      oOut.write("<ou_packageBool>grid 7 test package bool 1</ou_packageBool>");
      oOut.write("</ou_gridInfo>");

      //Multiples of each, both package and regular
      oOut.write("<ou_gridInfo gridName=\"grid 8\">");
      oOut.write("<ou_saveFreq>8</ou_saveFreq>");
      oOut.write("<ou_float>grid 8 test float 1</ou_float>");
      oOut.write("<ou_char>grid 8 test char 1</ou_char>");
      oOut.write("<ou_int>grid 8 test int 1</ou_int>");
      oOut.write("<ou_bool>grid 8 test bool 1</ou_bool>");
      oOut.write("<ou_float>grid 8 test float 2</ou_float>");
      oOut.write("<ou_char>grid 8 test char 2</ou_char>");
      oOut.write("<ou_int>grid 8 test int 2</ou_int>");
      oOut.write("<ou_bool>grid 8 test bool 2</ou_bool>");
      oOut.write("<ou_packageInt>grid 8 test package int 1</ou_packageInt>");
      oOut.write("<ou_packageInt>grid 8 test package int 2</ou_packageInt>");
      oOut.write(
          "<ou_packageFloat>grid 8 test package float 1</ou_packageFloat>");
      oOut.write(
          "<ou_packageFloat>grid 8 test package float 2</ou_packageFloat>");
      oOut.write(
          "<ou_packageFloat>grid 8 test package float 3</ou_packageFloat>");
      oOut.write("<ou_packageChar>grid 8 test package char 1</ou_packageChar>");
      oOut.write("<ou_packageChar>grid 8 test package char 2</ou_packageChar>");
      oOut.write("<ou_packageChar>grid 8 test package char 3</ou_packageChar>");
      oOut.write("<ou_packageChar>grid 8 test package char 4</ou_packageChar>");
      oOut.write("<ou_packageBool>grid 8 test package bool 1</ou_packageBool>");
      oOut.write("<ou_packageBool>grid 8 test package bool 2</ou_packageBool>");
      oOut.write("<ou_packageBool>grid 8 test package bool 3</ou_packageBool>");
      oOut.write("<ou_packageBool>grid 8 test package bool 4</ou_packageBool>");
      oOut.write("<ou_packageBool>grid 8 test package bool 5</ou_packageBool>");
      oOut.write("</ou_gridInfo>");

      //One per data type, package only
      oOut.write("<ou_gridInfo gridName=\"grid 9\">");
      oOut.write("<ou_saveFreq>9</ou_saveFreq>");
      oOut.write("<ou_packageInt>grid 9 test package int 1</ou_packageInt>");
      oOut.write(
          "<ou_packageFloat>grid 9 test package float 1</ou_packageFloat>");
      oOut.write("<ou_packageChar>grid 9 test package char 1</ou_packageChar>");
      oOut.write("<ou_packageBool>grid 9 test package bool 1</ou_packageBool>");
      oOut.write("</ou_gridInfo>");

      //Multiples of each, package only
      oOut.write("<ou_gridInfo gridName=\"grid 10\">");
      oOut.write("<ou_saveFreq>10</ou_saveFreq>");
      oOut.write("<ou_packageInt>grid 10 test package int 1</ou_packageInt>");
      oOut.write("<ou_packageInt>grid 10 test package int 2</ou_packageInt>");
      oOut.write(
          "<ou_packageFloat>grid 10 test package float 1</ou_packageFloat>");
      oOut.write(
          "<ou_packageFloat>grid 10 test package float 2</ou_packageFloat>");
      oOut.write(
          "<ou_packageFloat>grid 10 test package float 3</ou_packageFloat>");
      oOut.write("<ou_packageChar>grid 10 test package char 1</ou_packageChar>");
      oOut.write("<ou_packageChar>grid 10 test package char 2</ou_packageChar>");
      oOut.write("<ou_packageChar>grid 10 test package char 3</ou_packageChar>");
      oOut.write("<ou_packageChar>grid 10 test package char 4</ou_packageChar>");
      oOut.write("<ou_packageBool>grid 10 test package bool 1</ou_packageBool>");
      oOut.write("<ou_packageBool>grid 10 test package bool 2</ou_packageBool>");
      oOut.write("<ou_packageBool>grid 10 test package bool 3</ou_packageBool>");
      oOut.write("<ou_packageBool>grid 10 test package bool 4</ou_packageBool>");
      oOut.write("<ou_packageBool>grid 10 test package bool 5</ou_packageBool>");
      oOut.write("</ou_gridInfo>");

      oOut.write("</output>");
      oOut.write("</paramFile>");

      oOut.close();
      return sFileName;
    }
    catch (java.io.IOException oE) {
      ModelException oErr = new ModelException();
      oErr.sMessage = oE.getMessage();
      throw (oErr);
    }
  }

  /**
   * This writes an XML file to test value setting.  This file contains short
   * output settings only, for all (normal) types with everything set to true.
   * It also contains subplots.
   * @return Name of file.
   * @throws ModelException if there is an IO problem.
   */
  protected String WriteXMLFile2() throws ModelException {
    try {
      String sFileName = "\\loratestxml2.xml";
      FileWriter oOut = new FileWriter(sFileName);

      oOut.write(
          "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\" ?>");
      oOut.write("<paramFile fileCode=\"06010101\">");

      //Plot - so the subplot cells can be set up
      oOut.write("<plot>");
      oOut.write("<plot_lenX>200.0</plot_lenX>");
      oOut.write("<plot_lenY>200.0</plot_lenY>");
      oOut.write("<plot_latitude>55.37</plot_latitude>");
      oOut.write("</plot>");

      oOut.write("<trees>");
      oOut.write("<tr_speciesList>");
      oOut.write("<tr_species speciesName=\"Western_Hemlock\"/>");
      oOut.write("<tr_species speciesName=\"Western_Redcedar\"/>");
      oOut.write("<tr_species speciesName=\"Amabilis_Fir\"/>");
      oOut.write("<tr_species speciesName=\"Subalpine_Fir\"/>");
      oOut.write("<tr_species speciesName=\"Hybrid_Spruce\"/>");
      oOut.write("<tr_species speciesName=\"Lodgepole_Pine\"/>");
      oOut.write("<tr_species speciesName=\"Trembling_Aspen\"/>");
      oOut.write("<tr_species speciesName=\"Black_Cottonwood\"/>");
      oOut.write("<tr_species speciesName=\"Paper_Birch\"/>");
      oOut.write("</tr_speciesList>");
      oOut.write("</trees>");

      //Include an empty behavior list - this triggers setup
      oOut.write("<behaviorList>");
      oOut.write("</behaviorList>");

      oOut.write("<shortoutput>");

      //Set all save options to true for all recognized types
      oOut.write("<so_filename>testfile</so_filename>");

      oOut.write("<so_treeTypeInfo type=\"seedling\">");
      oOut.write("<so_saveRBA save=\"true\"/>"); //this shouldn't happen
      oOut.write("<so_saveABA save=\"true\"/>"); //this shouldn't happen
      oOut.write("<so_saveRDN save=\"true\"/>");
      oOut.write("<so_saveADN save=\"true\"/>");
      oOut.write("</so_treeTypeInfo>");

      oOut.write("<so_treeTypeInfo type=\"sapling\">");
      oOut.write("<so_saveRBA save=\"true\"/>");
      oOut.write("<so_saveABA save=\"true\"/>");
      oOut.write("<so_saveRDN save=\"true\"/>");
      oOut.write("<so_saveADN save=\"true\"/>");
      oOut.write("</so_treeTypeInfo>");

      oOut.write("<so_treeTypeInfo type=\"adult\">");
      oOut.write("<so_saveRBA save=\"true\"/>");
      oOut.write("<so_saveABA save=\"true\"/>");
      oOut.write("<so_saveRDN save=\"true\"/>");
      oOut.write("<so_saveADN save=\"true\"/>");
      oOut.write("</so_treeTypeInfo>");

      //Now subplots
      oOut.write("<so_subplot>");
      oOut.write("<so_subplotName>test 1</so_subplotName>");
      //One point
      oOut.write("<pointSet>");
      oOut.write("<po_point x=\"0\" y=\"12\"/>");
      oOut.write("</pointSet>");
      oOut.write("</so_subplot>");

      //5 points
      oOut.write("<so_subplot>");
      oOut.write("<so_subplotName>test 2</so_subplotName>");
      oOut.write("<pointSet>");
      oOut.write("<po_point x=\"6\" y=\"14\"/>");
      oOut.write("<po_point x=\"7\" y=\"15\"/>");
      oOut.write("<po_point x=\"8\" y=\"16\"/>");
      oOut.write("<po_point x=\"9\" y=\"17\"/>");
      oOut.write("<po_point x=\"10\" y=\"18\"/>");
      oOut.write("</pointSet>");
      oOut.write("</so_subplot>");

      oOut.write("</shortoutput>");
      oOut.write("</paramFile>");

      oOut.close();
      return sFileName;
    }
    catch (java.io.IOException oE) {
      ModelException oErr = new ModelException();
      oErr.sMessage = oE.getMessage();
      throw (oErr);
    }
  }

  /**
   * This writes an XML file to test value setting.  This file contains short
   * output settings only, for all (normal) types with a mix of true, false,
   * and absent.
   * @return Name of file.
   * @throws ModelException if there is an IO problem.
   */
  protected String WriteXMLFile3() throws ModelException {
    try {
      String sFileName = "\\loratestxml3.xml";
      FileWriter oOut = new FileWriter(sFileName);

      oOut.write(
          "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\" ?>");
      oOut.write("<paramFile fileCode=\"06010101\">");
      oOut.write("<trees>");
      oOut.write("<tr_speciesList>");
      oOut.write("<tr_species speciesName=\"Western_Hemlock\"/>");
      oOut.write("<tr_species speciesName=\"Western_Redcedar\"/>");
      oOut.write("<tr_species speciesName=\"Amabilis_Fir\"/>");
      oOut.write("<tr_species speciesName=\"Subalpine_Fir\"/>");
      oOut.write("<tr_species speciesName=\"Hybrid_Spruce\"/>");
      oOut.write("<tr_species speciesName=\"Lodgepole_Pine\"/>");
      oOut.write("<tr_species speciesName=\"Trembling_Aspen\"/>");
      oOut.write("<tr_species speciesName=\"Black_Cottonwood\"/>");
      oOut.write("<tr_species speciesName=\"Paper_Birch\"/>");
      oOut.write("</tr_speciesList>");
      oOut.write("</trees>");

      //Include an empty behavior list - this triggers setup
      oOut.write("<behaviorList>");
      oOut.write("</behaviorList>");

      oOut.write("<shortoutput>");

      //Set all save options to true for all recognized types
      oOut.write("<so_filename>testfile</so_filename>");

      oOut.write("<so_treeTypeInfo type=\"sapling\">");
      oOut.write("<so_saveRBA save=\"true\"/>");
      oOut.write("<so_saveABA save=\"false\"/>");
      oOut.write("<so_saveRDN save=\"true\"/>");
      oOut.write("</so_treeTypeInfo>");

      oOut.write("<so_treeTypeInfo type=\"adult\">");
      oOut.write("<so_saveRBA save=\"false\"/>");
      oOut.write("<so_saveADN save=\"true\"/>");
      oOut.write("</so_treeTypeInfo>");

      oOut.write("</shortoutput>");
      oOut.write("</paramFile>");

      oOut.close();
      return sFileName;
    }
    catch (java.io.IOException oE) {
      ModelException oErr = new ModelException();
      oErr.sMessage = oE.getMessage();
      throw (oErr);
    }
  }

  private boolean testEquals(DetailedTreeSettings o1, DetailedTreeSettings o2) {
    if (o1.GetNumberOfBools() != o2.GetNumberOfBools() ||
        o1.GetNumberOfFloats() != o2.GetNumberOfFloats() ||
        o1.GetNumberOfChars() != o2.GetNumberOfChars() ||
        o1.GetNumberOfInts() != o2.GetNumberOfInts() ||
        o1.GetSaveFrequency() != o2.GetSaveFrequency() ||
        o1.GetSpecies() != o2.GetSpecies() ||
        o1.GetType() != o2.GetType()) {
      return false;
    }

    int i;
    for (i = 0; i < o1.GetNumberOfBools(); i++) {
      if (o1.GetBool(i) != o2.GetBool(i)) {
        return false;
      }
    }

    for (i = 0; i < o1.GetNumberOfChars(); i++) {
      if (o1.GetChar(i) != o2.GetChar(i)) {
        return false;
      }
    }

    for (i = 0; i < o1.GetNumberOfFloats(); i++) {
      if (o1.GetFloat(i) != o2.GetFloat(i)) {
        return false;
      }
    }

    for (i = 0; i < o1.GetNumberOfInts(); i++) {
      if (o1.GetInt(i) != o2.GetInt(i)) {
        return false;
      }
    }

    return true;
  }

  /**
   * Test two rundata settings for equality.
   * @param o1 First settings to test.
   * @param o2 Second settings to test.
   * @return True if all settings are the same, false if not.
   */
  private boolean testEquals(DetailedGridSettings o1, DetailedGridSettings o2) {
    if (o1.GetNumberOfBools() != o2.GetNumberOfBools() ||
        o1.GetNumberOfFloats() != o2.GetNumberOfFloats() ||
        o1.GetNumberOfChars() != o2.GetNumberOfChars() ||
        o1.GetNumberOfInts() != o2.GetNumberOfInts() ||
        o1.GetSaveFrequency() != o2.GetSaveFrequency() ||
        o1.GetName() != o2.GetName()) {
      return false;
    }

    int i;
    for (i = 0; i < o1.GetNumberOfBools(); i++) {
      if (o1.GetBool(i) != o2.GetBool(i)) {
        return false;
      }
    }

    for (i = 0; i < o1.GetNumberOfChars(); i++) {
      if (o1.GetChar(i) != o2.GetChar(i)) {
        return false;
      }
    }

    for (i = 0; i < o1.GetNumberOfFloats(); i++) {
      if (o1.GetFloat(i) != o2.GetFloat(i)) {
        return false;
      }
    }

    for (i = 0; i < o1.GetNumberOfInts(); i++) {
      if (o1.GetInt(i) != o2.GetInt(i)) {
        return false;
      }
    }

    return true;
  }



}
